/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.classify;

import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.distribution.Distribution;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.event.EventCollector;

/**
 * Classifies file system usage into a distribution.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class FileSystemClassifier extends ClassifierBase implements IClassifier<FileSystemData> {

    // TODO Make these numbers generic & look in more detail at this!
    // Calibrated using iStat Menus on the Mac Min throughput per process = 35 MB/sec Max throughput per process = 35 MB/sec.
    private static final double WRITE_MAX_BYTES_PER_SECOND = 35000000;
    private static final double READ_MAX_BYTES_PER_SECOND = 35000000;

    private static final double[] READ_THRESHOLDS = new double[]{READ_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_1, READ_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_2, READ_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_3, READ_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_4};
    private static final double[] WRITE_THRESHOLDS = new double[]{WRITE_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_1, WRITE_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_2, WRITE_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_3, WRITE_MAX_BYTES_PER_SECOND * THRESHOLD_FACTOR_4};

    private final EventCollector event_collector;
    private final String machine_id;
    private final String router_mac;

    private FileSystemData last = null; // Remembers last measurement taken.

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a collector.
     *
     * @param event_collector a collector that collects events
     * @param machine_id the ID of the calling machine.
     */
    public FileSystemClassifier(final EventCollector event_collector, final String machine_id, final String router_mac) {

        super();
        this.event_collector = event_collector;
        this.machine_id = machine_id;
        this.router_mac = router_mac;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Distribution[] classify(final FileSystemData data) {

        if (last == null) { // don't classify if we have no previous reading to compare to
            last = data;
            return null;
        }

        if (!last.getFileSystemName().equals(data.getFileSystemName())) {
            event_collector.addEvent(new Event(Event.FILESYSTEMS_CHANGED, machine_id, null, router_mac));
        }

        final long milliseconds_elapsed = data.getDate().getTime() - last.getDate().getTime();
        long seconds_elapsed = milliseconds_elapsed / MILLI_SECONDS_PER_SECOND;

        // avoid arithmetic exception
        if (seconds_elapsed == 0) {
            seconds_elapsed = 1;
        }

        final long reads_bytes_per_second = (data.getBytesRead() - last.getBytesRead()) / seconds_elapsed;
        final long writes_bytes_per_second = (data.getBytesWritten() - last.getBytesWritten()) / seconds_elapsed;

        final String attr = data.getFileSystemName();

        final Distribution[] result = new Distribution[]{classify("fs_read", attr, data.getDate(), reads_bytes_per_second, READ_THRESHOLDS), classify("fs_write", attr, data.getDate(), writes_bytes_per_second, WRITE_THRESHOLDS)};

        last = data; // Remember this reading.
        return result;
    }

    @Override
    public FileSystemData getContentType() {

        return new FileSystemData().getNewInstance();
    }
}
