/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.classify;

import java.util.ArrayList;
import java.util.List;

import uk.ac.standrews.cs.numonic.data.NetworkData;
import uk.ac.standrews.cs.numonic.distribution.Distribution;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.event.EventCollector;

/**
 * Classifier for network data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class NetworkClassifier extends ClassifierBase implements IClassifier<NetworkData> {

    // TODO Make these numbers generic & look in more detail at this!

    // from http://ndt.iupui.donar.measurement-lab.org:7123/
    // On Node IO - Macbook Pro with
    // running 10s outbound test (client-to-server [C2S]) . . . . . 37.13Mb/s = 38,933,626 bytes/sec
    // running 10s inbound test (server-to-client [S2C]) . . . . . . 91.25Mb/s = 915,682,560 bytes/sec
    // The slowest link in the end-to-end path is a 100 Mbps Full duplex Fast Ethernet subnet
    // On EIGG - Windows 7 wins...
    // running 10s outbound test (client-to-server [C2S]) . . . . . 93.90Mb/s
    // running 10s inbound test (server-to-client [S2C]) . . . . . . 93.95Mb/s

    private static final double NETWORK_TX_MAX = 98461286.4;
    private static final double NETWORK_RX_MAX = 98513715.2;

    private static final double[] TX_THRESHOLDS = new double[]{NETWORK_TX_MAX * THRESHOLD_FACTOR_1, NETWORK_TX_MAX * THRESHOLD_FACTOR_2, NETWORK_TX_MAX * THRESHOLD_FACTOR_3, NETWORK_TX_MAX * THRESHOLD_FACTOR_4};
    private static final double[] RX_THRESHOLDS = new double[]{NETWORK_RX_MAX * THRESHOLD_FACTOR_1, NETWORK_RX_MAX * THRESHOLD_FACTOR_2, NETWORK_RX_MAX * THRESHOLD_FACTOR_3, NETWORK_RX_MAX * THRESHOLD_FACTOR_4};

    private final EventCollector event_collector;
    private final String machine_id;
    private final String router_mac;

    private NetworkData last = null; // Remembers last measurement taken.

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a collector.
     *
     * @param event_collector a collector that collects events
     * @param machine_id the ID of the calling machine.
     */
    public NetworkClassifier(final EventCollector event_collector, final String machine_id, final String router_mac) {

        super();
        this.event_collector = event_collector;
        this.machine_id = machine_id;
        this.router_mac = router_mac;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Distribution[] classify(final NetworkData data) {

        if (last == null) { // don't classify if we have no previous reading to compare to
            last = data;
            return null;
        }

        if (!last.getDeviceName().equals(data.getDeviceName())) {

            final List<String> list = new ArrayList<String>();
            list.add(last.getDeviceName());
            list.add(data.getDeviceName());
            event_collector.addEvent(new Event(Event.IP_ADDRESS_CHANGE, machine_id, list, router_mac));
        }

        final long milliseconds_elapsed = data.getDate().getTime() - last.getDate().getTime();
        final long seconds_elapsed = milliseconds_elapsed / MILLI_SECONDS_PER_SECOND;
        final long rx_bytes_per_second = (data.getRXBytes() - last.getRXBytes()) / seconds_elapsed;
        final long tx_bytes_per_second = (data.getTXBytes() - last.getTXBytes()) / seconds_elapsed;

        final String attr = data.getDeviceName() + "/" + data.getIPAddress();

        final Distribution[] result = new Distribution[]{classify("net_tx", attr, data.getDate(), tx_bytes_per_second, TX_THRESHOLDS), classify("net_rx", attr, data.getDate(), rx_bytes_per_second, RX_THRESHOLDS)};

        last = data; // remember this reading.
        return result;
    }

    @Override
    public NetworkData getContentType() {

        return new NetworkData().getNewInstance();
    }
}
