/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.data;

/**
 * Used to store file system usage data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class FileSystemData extends Data implements IData<FileSystemData> {

    public String machine_id; // Only used when getting data from the database; can be null when data is added to the database.

    public String file_system_location;
    public String file_system_type;
    public String file_system_name;

    public long fs_space_used; // Amount of space used in the file system (in Kilobytes)
    public long fs_space_free; // Amount of space free in the file system (in Kilobytes)
    public long fs_size; // Total size of file system (in Kilobytes)
    public long fs_number_of_files; // Total number of files in the file system
    public long fs_disk_reads; // Total number of disk reads
    public long fs_bytes_read; // Total number of bytes read
    public long fs_disk_writes; // Total number of disk writes
    public long fs_bytes_written; // Total number of bytes

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a data object.
     */
    public FileSystemData() {

        super();
    }

    /**
     * Initialises a data object.
     * @param date the date at which the measurement was made
     */
    public FileSystemData(final java.sql.Date date) {

        super(date);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();

        builder.append("FileSystemResults [file_System_Location=");
        builder.append(file_system_location);
        builder.append(", file_System_Type=");
        builder.append(file_system_type);
        builder.append(", file_System_Name=");
        builder.append(file_system_name);
        builder.append(", Fs_used=");
        builder.append(fs_space_used);
        builder.append(", Fs_free=");
        builder.append(fs_space_free);
        builder.append(", Fs_size=");
        builder.append(fs_size);
        builder.append(", Fs_disk_reads=");
        builder.append(fs_disk_reads);
        builder.append(", Fs_files=");
        builder.append(fs_number_of_files);
        builder.append(", Fs_disk_read_bytes=");
        builder.append(fs_bytes_read);
        builder.append(", Fs_disk_writes=");
        builder.append(fs_disk_reads);
        builder.append(", Fs_disk_write_bytes=");
        builder.append(fs_bytes_written);
        builder.append("]");

        return builder.toString();
    }

    /**
     * Returns a new object in which each numerical field contains the minimum value from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the minimum values
     */
    @Override
    public FileSystemData min(final FileSystemData other) {

        final FileSystemData min = new FileSystemData();

        min.file_system_location = other.file_system_location;
        min.file_system_type = other.file_system_type;
        min.file_system_name = other.file_system_name;

        min.fs_size = fs_size;

        min.fs_space_used = min(fs_space_used, other.fs_space_used);
        min.fs_space_free = min(fs_space_free, other.fs_space_free);
        min.fs_disk_reads = min(fs_disk_reads, other.fs_disk_reads);
        min.fs_number_of_files = min(fs_number_of_files, other.fs_number_of_files);
        min.fs_bytes_read = min(fs_bytes_read, other.fs_bytes_read);
        min.fs_disk_reads = min(fs_disk_reads, other.fs_disk_reads);
        min.fs_bytes_written = min(fs_bytes_written, other.fs_bytes_written);

        return min;
    }

    /**
     * Returns a new object in which each numerical field contains the maximum value from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the maximum values
     */
    @Override
    public FileSystemData max(final FileSystemData other) {

        final FileSystemData max = new FileSystemData();

        max.file_system_location = other.file_system_location;
        max.file_system_type = other.file_system_type;
        max.file_system_name = other.file_system_name;

        max.fs_size = fs_size;

        max.fs_space_used = max(fs_space_used, other.fs_space_used);
        max.fs_space_free = max(fs_space_free, other.fs_space_free);
        max.fs_disk_reads = max(fs_disk_reads, other.fs_disk_reads);
        max.fs_number_of_files = max(fs_number_of_files, other.fs_number_of_files);
        max.fs_bytes_read = max(fs_bytes_read, other.fs_bytes_read);
        max.fs_disk_reads = max(fs_disk_reads, other.fs_disk_reads);
        max.fs_bytes_written = max(fs_bytes_written, other.fs_bytes_written);

        return max;
    }

    /**
     * Returns a new object in which each numerical field contains the sum of the values from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the summed values
     */
    @Override
    public FileSystemData sum(final FileSystemData other) {

        final FileSystemData total = new FileSystemData();

        total.file_system_location = other.file_system_location;
        total.file_system_type = other.file_system_type;
        total.file_system_name = other.file_system_name;

        total.fs_size = fs_size;

        total.fs_space_used = fs_space_used + other.fs_space_used;
        total.fs_space_free = fs_space_free + other.fs_space_free;
        total.fs_disk_reads = fs_disk_reads + other.fs_disk_reads;
        total.fs_number_of_files = fs_number_of_files + other.fs_number_of_files;
        total.fs_bytes_read = fs_bytes_read + other.fs_bytes_read;
        total.fs_disk_reads = fs_disk_reads + other.fs_disk_reads;
        total.fs_bytes_written = fs_bytes_written + other.fs_bytes_written;

        return total;
    }

    /**
     * Returns a new object in which each numerical field contains the value from this object divided by the given number of measurements.
     * Other fields are taken from the other object.
     *
     * @param numberOfMeasurements the divisor
     * @return an object containing the converted values
     */
    @Override
    public FileSystemData convertToAverage(final int numberOfMeasurements) {

        final FileSystemData average = new FileSystemData();

        average.file_system_location = file_system_location;
        average.file_system_type = file_system_type;
        average.file_system_name = file_system_name;

        average.fs_size = fs_size;

        average.fs_space_used = fs_space_used / numberOfMeasurements;
        average.fs_space_free = fs_space_free / numberOfMeasurements;
        average.fs_disk_reads = fs_disk_reads / numberOfMeasurements;
        average.fs_number_of_files = fs_number_of_files / numberOfMeasurements;
        average.fs_bytes_read = fs_bytes_read / numberOfMeasurements;
        average.fs_disk_reads = fs_disk_reads / numberOfMeasurements;
        average.fs_bytes_written = fs_bytes_written / numberOfMeasurements;

        return average;
    }

    /**
     * Returns a new object in which each numerical field contains the value from another given object subtracted from the value from this object.
     * Other fields are taken from the other object.
     *
     * @param base the other object
     * @return an object containing the adjusted values
     */
    @Override
    public FileSystemData adjustForBase(final FileSystemData base) {

        final FileSystemData adjusted = new FileSystemData();

        // Not adjusted.
        adjusted.file_system_location = file_system_location;
        adjusted.file_system_type = file_system_type;
        adjusted.file_system_name = file_system_name;
        adjusted.fs_size = fs_size;
        adjusted.fs_space_used = fs_space_used;
        adjusted.fs_space_free = fs_space_free;

        // Adjusted.
        adjusted.fs_number_of_files = fs_number_of_files - base.fs_number_of_files;

        adjusted.fs_disk_reads = fs_disk_reads - base.fs_disk_reads;

        adjusted.fs_bytes_read = fs_bytes_read - base.fs_bytes_read;
        adjusted.fs_disk_reads = fs_disk_reads - base.fs_disk_reads;
        adjusted.fs_bytes_written = fs_bytes_written - base.fs_bytes_written;

        return adjusted;
    }

    /**
     * Whether this dataset has any data in it that needs to be adjusted for the base - whether the {@link IData#adjustForBase(Object)}
     * method needs to be called.
     *
     * @return true if the {@link IData#adjustForBase(Object)} method does any work.
     */
    @Override
    public boolean needsAdjustedForBase() {

        return true;
    }

    /**
     * Returns a new instance.
     * @return a new instance
     */
    @Override
    public FileSystemData getNewInstance() {

        return new FileSystemData();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public int hashCode() {

        final int prime = 31;
        int result = 1;
        result = prime * result + (file_system_location == null ? 0 : file_system_location.hashCode());
        result = prime * result + (file_system_name == null ? 0 : file_system_name.hashCode());
        result = prime * result + (file_system_type == null ? 0 : file_system_type.hashCode());
        return result;
    }

    @Override
    public boolean equals(final Object obj) {

        if (this == obj) { return true; }
        if (obj == null) { return false; }
        if (getClass() != obj.getClass()) { return false; }
        final FileSystemData other = (FileSystemData) obj;
        if (file_system_location == null) {
            if (other.file_system_location != null) { return false; }
        }
        else if (!file_system_location.equals(other.file_system_location)) { return false; }
        if (file_system_name == null) {
            if (other.file_system_name != null) { return false; }
        }
        else if (!file_system_name.equals(other.file_system_name)) { return false; }
        if (file_system_type == null) {
            if (other.file_system_type != null) { return false; }
        }
        else if (!file_system_type.equals(other.file_system_type)) { return false; }
        return true;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the machine ID.
     * @return the machine ID
     */
    public String getMachineId() {

        return machine_id;
    }

    /**
     * Sets the machine ID.
     * @param machine_id the machine ID
     */
    public void setMachineId(final String machine_id) {

        this.machine_id = machine_id;
    }

    /**
     * Returns the file system location.
     * @return the file system location
     */
    public String getFileSystemLocation() {

        return file_system_location;
    }

    /**
     * Sets the file system location.
     * @param file_system_location the file system location
     */
    public void setFileSystemLocation(final String file_system_location) {

        this.file_system_location = file_system_location;
    }

    /**
     * Returns the file system type.
     * @return the file system type
     */
    public String getFileSystemType() {

        return file_system_type;
    }

    /**
     * Sets the file system type.
     * @param file_system_type the file system type
     */
    public void setFileSystemType(final String file_system_type) {

        this.file_system_type = file_system_type;
    }

    /**
     * Returns the file system name.
     * @return the file system name
     */
    public String getFileSystemName() {

        return file_system_name;
    }

    /**
     * Sets the file system name.
     * @param file_system_name the file system name
     */
    public void setFileSystemName(final String file_system_name) {

        this.file_system_name = file_system_name;
    }

    /**
     * Returns the space used in the file system.
     * @return the space used in the file system in KB
     */
    public long getSpaceUsed() {

        return fs_space_used;
    }

    /**
     * Sets the space used in the file system.
     * @param space_used the space used in the file system in KB
     */
    public void setSpaceUsed(final long space_used) {

        this.fs_space_used = space_used;
    }

    /**
     * Returns the space free in the file system.
     * @return the space free in the file system in KB
     */
    public long getSpaceFree() {

        return fs_space_free;
    }

    /**
     * Sets the space free in the file system.
     * @param space_free the space free in the file system in KB
     */
    public void setSpaceFree(final long space_free) {

        this.fs_space_free = space_free;
    }

    /**
     * Returns the total size of the file system.
     * @return the total size of the file system in KB
     */
    public long getSize() {

        return fs_size;
    }

    /**
     * Sets the total size of the file system.
     * @param size the total size of the file system in KB
     */
    public void setSize(final long size) {

        this.fs_size = size;
    }

    /**
     * Returns the number of files in the file system.
     * @return the number of files in the file system
     */
    public long getNumberOfFiles() {

        return fs_number_of_files;
    }

    /**
     * Sets the number of files in the file system.
     * @param number_of_files the number of files in the file system
     */
    public void setNumberOfFiles(final long number_of_files) {

        this.fs_number_of_files = number_of_files;
    }

    /**
     * Returns the number of disk reads.
     * @return the number of disk reads
     */
    public long getNumberOfDiskReads() {

        return fs_disk_reads;
    }

    /**
     * Sets the number of disk reads.
     * @param number_of_disk_reads the number of disk reads
     */
    public void setNumberOfDiskReads(final long number_of_disk_reads) {

        this.fs_disk_reads = number_of_disk_reads;
    }

    /**
     * Returns the number of bytes read.
     * @return the number of bytes read
     */
    public long getBytesRead() {

        return fs_bytes_read;
    }

    /**
     * Sets the number of bytes read.
     * @param bytes_read the number of bytes read
     */
    public void setBytesRead(final long bytes_read) {

        this.fs_bytes_read = bytes_read;
    }

    /**
     * Returns the number of disk writes.
     * @return the number of disk writes
     */
    public long getNumberOfDiskWrites() {

        return fs_disk_writes;
    }

    /**
     * Sets the number of disk writes.
     * @param number_of_disk_writes the number of disk writes
     */
    public void setNumberOfDiskWrites(final long number_of_disk_writes) {

        this.fs_disk_writes = number_of_disk_writes;
    }

    /**
     * Returns the number of bytes written.
     * @return the number of bytes written
     */
    public long getBytesWritten() {

        return fs_bytes_written;
    }

    /**
     * Sets the number of bytes written.
     * @param bytes_written the number of bytes written
     */
    public void setBytesWritten(final long bytes_written) {

        this.fs_bytes_written = bytes_written;
    }
}
