/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.data;

import java.util.Date;

/**
 * Implemented by value classes containing the results of resource monitoring.
 * @param <T> The class implementing this interface.
 * @author angus
 */
public interface IData<T> {

    /**
     * Creates a new instance of the class implementing this interface.
     * @return a new instance of the class implementing this interface
     */
    T getNewInstance();

    /**
     * Compares this instance against another of the same type and returns the minimum values for each element.
     * @param other - The instance this is being compared against.
     * @return New Instance of T containing the smallest values at each element.
     */
    T min(T other);

    /**
     * Compare this instance against another of the same type and return the maxmimum values for each element.
     * @param other The instance this is being compared against.
     * @return New Instance of T containing the largest values at each element.
     */
    T max(T other);

    /**
     * Add each element in this instance against the corresponding element in another instance of the same type.
     * @param other The instance this is being compared against.
     * @return New Instance of T containing the sum of each element.
     */
    T sum(T other);

    /**
     * Get the average values for this monitoring set by dividing the total by the number of measurements.
     * @param numberOfMeasurements Number of measurements taken in this summary set.
     * @return the average values for this monitoring set
     */
    T convertToAverage(int numberOfMeasurements);

    /**
     * Some measurements are constantly incrementing and only reset after the system has been restarted. To obtain something meaningful from
     * these measurements this method takes them and subtracts the previous value to produce the delta from the previous measurement.
     * This method performs this operation on only the methods where it is necessary.
     * @param previous The previous value for each measurement. This is subtracted from the currently measured value.
     * @return the delta from the previous measurement
     */
    T adjustForBase(T previous);

    /**
     * Whether this dataset has any data in it that needs to be adjusted for the base - whether the {@link IData#adjustForBase(Object)}
     * method needs to be called.
     *
     * @return true if the {@link IData#adjustForBase(Object)} method does any work.
     */
    boolean needsAdjustedForBase();

    /**
     * Returns the Date when this object was created.
     * @return the Date when this object was created
     */
    Date getDate();
}
