/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.data;

import java.util.Date;

/**
 * Used to store latency and bandwidth data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class LatencyAndBandwidthData {

    public final String local_machine_ID;
    public final String local_network_adapter_name;
    public final String local_gateway;
    public final String remote_machine_ID;
    public final String remote_network_adapter_name;
    public final String remote_gateway;
    public final long latency;
    public final long bandwidth;
    public final Date date;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a latency/bandwidth record.
     *
     * @param local_machine_ID the local machine ID
     * @param local_network_adapter_name the local NIC name
     * @param local_gateway the local gateway
     * @param remote_machine_ID the remote machine ID
     * @param remote_network_adapter_name the remote NIC name
     * @param remote_gateway the remote gateway
     * @param latency the latency in ms
     * @param bandwidth the bandwidth in KB/s
     * @param date the time of recording
     */
    public LatencyAndBandwidthData(final String local_machine_ID, final String local_network_adapter_name, final String local_gateway, final String remote_machine_ID, final String remote_network_adapter_name, final String remote_gateway, final long latency, final long bandwidth, final Date date) {

        this.local_machine_ID = local_machine_ID;
        this.local_network_adapter_name = local_network_adapter_name;
        this.local_gateway = local_gateway;
        this.remote_machine_ID = remote_machine_ID;
        this.remote_network_adapter_name = remote_network_adapter_name;
        this.remote_gateway = remote_gateway;
        this.latency = latency;
        this.bandwidth = bandwidth;
        this.date = (Date) date.clone();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();
        builder.append("LatencyAndBandwidthData [localMachineID=");
        builder.append(local_machine_ID);
        builder.append(", localNetworkAdapterName=");
        builder.append(local_network_adapter_name);
        builder.append(", localGateway=");
        builder.append(local_gateway);
        builder.append(", remoteMachineID=");
        builder.append(remote_machine_ID);
        builder.append(", remoteNetworkAdapterName=");
        builder.append(remote_network_adapter_name);
        builder.append(", remoteGateway=");
        builder.append(remote_gateway);
        builder.append(", latency=");
        builder.append(latency);
        builder.append(", bandwidth=");
        builder.append(bandwidth);
        builder.append(", time=");
        builder.append(date);
        builder.append("]");
        return builder.toString();
    }

    /**
     * Returns the local machine ID.
     * @return the local machine ID
     */
    public String getLocalMachineID() {

        return local_machine_ID;
    }

    /**
     * Returns the local NIC name.
     * @return the local NIC name
     */
    public String getLocalNetworkAdapterName() {

        return local_network_adapter_name;
    }

    /**
     * Returns the local gateway.
     * @return the local gateway
     */
    public String getLocalGateway() {

        return local_gateway;
    }

    /**
     * Returns the remote machine ID.
     * @return the remote machine ID
     */
    public String getRemoteMachineID() {

        return remote_machine_ID;
    }

    /**
     * Returns the remote NIC name.
     * @return the remote NIC name
     */
    public String getRemoteNetworkAdapterName() {

        return remote_network_adapter_name;
    }

    /**
     * Returns the remote gateway.
     * @return the remote gateway
     */
    public String getRemoteGateway() {

        return remote_gateway;
    }

    /**
     * Returns the latency in ms.
     * @return the latency in ms
     */
    public long getLatency() {

        return latency;
    }

    /**
     * Returns the bandwidth in KB/s.
     * @return the bandwidth in KB/s
     */
    public long getBandwidth() {

        return bandwidth;
    }

    /**
     * Returns the time of recording.
     * @return the time of recording
     */
    public Date getDate() {

        return (Date) date.clone();
    }
}
