/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.data;

/**
 * Used to store process data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class ProcessData extends Data implements IData<ProcessData> {

    public String process_name; // The name of the process
    public double process_cpu_percent; // Percentage CPU used by this process
    public long process_start_time; // When this process was started stored as a Unix timestamp (no. of seconds since Jan 1 1970).
    public long process_mem_size; // Virtual memory usage of entire process. vm_lib + vm_exe + vm_data + vm_stack (in kiloBytes).
    public long process_resident; // Resident set size, the non-swapped physical memory that a task has used (in kiloBytes).

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();

        builder.append("ProcessMonResults [Process_name=");
        builder.append(process_name);
        builder.append(" ,Process_Cpu_Percent=");
        builder.append(process_cpu_percent);
        builder.append(", Process_start_time=");
        builder.append(process_start_time);
        builder.append(", Process_mem_size=");
        builder.append(process_mem_size);
        builder.append(", Process_resident=");
        builder.append(process_resident);
        builder.append("]");

        return builder.toString();
    }

    /**
     * Returns a new object in which each numerical field contains the minimum value from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the minimum values
     */
    @Override
    public ProcessData min(final ProcessData other) {

        final ProcessData min = new ProcessData();

        min.process_name = other.process_name;
        min.process_start_time = process_start_time;

        min.process_cpu_percent = process_cpu_percent < other.process_cpu_percent ? process_cpu_percent : other.process_cpu_percent;

        min.process_mem_size = process_mem_size < other.process_mem_size ? process_mem_size : other.process_mem_size;
        min.process_resident = process_resident < other.process_resident ? process_resident : other.process_resident;

        return min;
    }

    /**
     * Returns a new object in which each numerical field contains the maximum value from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the maximum values
     */
    @Override
    public ProcessData max(final ProcessData other) {

        final ProcessData max = new ProcessData();

        max.process_name = other.process_name;
        max.process_start_time = process_start_time;

        max.process_cpu_percent = process_cpu_percent > other.process_cpu_percent ? process_cpu_percent : other.process_cpu_percent;

        max.process_mem_size = process_mem_size > other.process_mem_size ? process_mem_size : other.process_mem_size;
        max.process_resident = process_resident > other.process_resident ? process_resident : other.process_resident;

        return max;
    }

    /**
     * Returns a new object in which each numerical field contains the sum of the values from this object and the other given object.
     * Other fields are taken from the other object.
     *
     * @param other the other object
     * @return an object containing the summed values
     */
    @Override
    public ProcessData sum(final ProcessData other) {

        final ProcessData total = new ProcessData();

        total.process_name = other.process_name;
        total.process_start_time = process_start_time;

        total.process_cpu_percent = process_cpu_percent + other.process_cpu_percent;
        total.process_mem_size = process_mem_size + other.process_mem_size;
        total.process_resident = process_resident + other.process_resident;

        return total;
    }

    /**
     * Returns a new object in which each numerical field contains the value from this object divided by the given number of measurements.
     * Other fields are taken from the other object.
     *
     * @param numberOfMeasurements the divisor
     * @return an object containing the converted values
     */
    @Override
    public ProcessData convertToAverage(final int numberOfMeasurements) {

        final ProcessData average = new ProcessData();

        average.process_name = process_name;
        average.process_start_time = process_start_time;

        average.process_cpu_percent = process_cpu_percent / numberOfMeasurements;
        average.process_mem_size = process_mem_size / numberOfMeasurements;
        average.process_resident = process_resident / numberOfMeasurements;

        return average;
    }

    /**
     * Placeholder method, since nothing needs to be adjusted.
     *
     * @param base the other object
     * @return this object
     */
    @Override
    public ProcessData adjustForBase(final ProcessData base) {

        return this;
    }

    /**
     * Whether this dataset has any data in it that needs to be adjusted for the base - whether the {@link IData#adjustForBase(Object)}
     * method needs to be called.
     *
     * @return true if the {@link IData#adjustForBase(Object)} method does any work.
     */
    @Override
    public boolean needsAdjustedForBase() {

        return false;
    }

    /**
     * Returns a new instance.
     * @return a new instance
     */
    @Override
    public ProcessData getNewInstance() {

        return new ProcessData();
    }

    @Override
    public int hashCode() {

        final int prime = 31;
        int result = 1;
        result = prime * result + (process_name == null ? 0 : process_name.hashCode());
        return result;
    }

    @Override
    public boolean equals(final Object obj) {

        if (this == obj) { return true; }
        if (obj == null) { return false; }
        if (getClass() != obj.getClass()) { return false; }
        final ProcessData other = (ProcessData) obj;
        if (process_name == null) {
            if (other.process_name != null) { return false; }
        }
        else if (!process_name.equals(other.process_name)) { return false; }
        return true;
    }

    /**
     * Returns the process name.
     * @return the process name
     */
    public String getProcessName() {

        return process_name;
    }

    /**
     * Sets the process name.
     * @param process_name the process name
     */
    public void setProcessName(final String process_name) {

        this.process_name = process_name;
    }

    /**
     * Returns the cpu utilization.
     * @return the cpu utilization
     */
    public double getProcessCpuPercent() {

        return process_cpu_percent;
    }

    /**
     * Sets the process CPU percentage.
     * @param process_cpu_percent the CPU percentage
     */
    public void setProcessCpuPercent(final double process_cpu_percent) {

        this.process_cpu_percent = process_cpu_percent;
    }

    /**
     * Returns the process start time.
     * @return the process start time in Unix time
     */
    public long getProcessStartTime() {

        return process_start_time;
    }

    /**
     * Sets the process start time.
     * @param process_start_time the process start time in Unix time
     */
    public void setProcessStartTime(final long process_start_time) {

        this.process_start_time = process_start_time;
    }

    /**
     * Returns the virtual memory usage.
     * @return the virtual memory usage in KB
     */
    public long getProcessMemSize() {

        return process_mem_size;
    }

    /**
     * Sets the virtual memory usage.
     * @param process_mem_size the virtual memory usage in KB
     */
    public void setProcessMemSize(final long process_mem_size) {

        this.process_mem_size = process_mem_size;
    }

    /**
     * Returns the resident set size.
     * @return the resident set size in KB
     */
    public long getProcessResident() {

        return process_resident;
    }

    /**
     * Sets the resident set size.
     * @param process_resident the resident set size in KB
     */
    public void setProcessResident(final long process_resident) {

        this.process_resident = process_resident;
    }
}
