/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.db;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.LatencyAndBandwidthData;
import uk.ac.standrews.cs.numonic.data.MachineUtilisationData;
import uk.ac.standrews.cs.numonic.data.NetworkData;
import uk.ac.standrews.cs.numonic.data.ProcessData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.db.objects.DistributionObject;
import uk.ac.standrews.cs.numonic.db.objects.FileSystemSummary;
import uk.ac.standrews.cs.numonic.db.objects.MachineUtilSummary;
import uk.ac.standrews.cs.numonic.db.objects.NetworkDataSummary;
import uk.ac.standrews.cs.numonic.db.objects.ProcessDataSummary;
import uk.ac.standrews.cs.numonic.distribution.Distribution;
import uk.ac.standrews.cs.numonic.distribution.DistributionCollector;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.main.NumonicLogger;
import uk.ac.standrews.cs.numonic.reporting.IReporting;
import uk.ac.standrews.cs.numonic.sort.data.DistributionData;
import uk.ac.standrews.cs.numonic.summary.MultipleSummary;
import uk.ac.standrews.cs.numonic.summary.SingleSummary;

import com.mysql.jdbc.exceptions.jdbc4.CommunicationsException;

/**
 * Database-specific reporting interface.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class DatabaseConnection implements IReporting {

    private Connection connection;
    private String machineID = null;

    private QueryBuilder distribution_data_prepared_statement = new QueryBuilder();
    private QueryBuilder event_prepared_statement = new QueryBuilder();
    private QueryBuilder fs_prepared_statement = new QueryBuilder();
    private QueryBuilder util_prepared_statement = new QueryBuilder();
    private QueryBuilder net_prepared_statement = new QueryBuilder();
    private QueryBuilder process_prepared_statement = new QueryBuilder();
    private QueryBuilder sys_info_prepared_statement = new QueryBuilder();
    private QueryBuilder latency_bandwith_prepared_statement = new QueryBuilder();

    private final String database_url;
    private final String database_user_name;
    private final String database_password;
    /**
     * The date format used to add dates to the database.
     */
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss"); // 2010-08-04 13:51:14.539
    public final static Logger LOGGER = NumonicLogger.LOGGER;

    Queue<Object> workQueue = new LinkedList<Object>();
    int counter = 0;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the connection.
     *
     * @param database_url the database URL
     * @param database_user_name the user name
     * @param database_password the password
     * @param drop_database_tables true if all tables should be dropped
     */
    public DatabaseConnection(final String database_url, final String database_user_name, final String database_password, final boolean drop_database_tables) {

        this.database_url = database_url;
        this.database_user_name = database_user_name;
        this.database_password = database_password;

        boolean isMySQL = false;

        try {
            String databaseDriverClassName = null;

            if (database_url.startsWith("jdbc:mysql")) {
                databaseDriverClassName = "com.mysql.jdbc.Driver";
                isMySQL = true;
            }
            else {
                databaseDriverClassName = "org.h2.Driver";
            }

            Class.forName(databaseDriverClassName).newInstance();
            connection = DriverManager.getConnection(database_url, database_user_name, database_password);

        }
        catch (final CommunicationsException e) {
            LOGGER.log(Level.SEVERE, "Couldnt initialise a DB connection -  quitting the program");
            System.exit(1);

        }
        catch (final Exception e) {
            ErrorHandling.hardError(e, "Quitting: couldn't initialise database connection to: " + database_url);
        }

        try {
            if (drop_database_tables) {
                dropAllTables();
            }
        }
        catch (final SQLException e) {
            ErrorHandling.hardError(e, "Quitting: couldn't drop all tables from database running at: " + database_url);
        }

        // Create tables.
        try {
            if (!isMySQL) {
                executeSQLScript("sql-schema" + File.separator + "monitorSchema.sql");
                executeSQLScript("sql-schema" + File.separator + "views.sql");
            }
        }
        catch (final Exception e) {
            ErrorHandling.hardError(e, "Quitting: couldn't configure tables in the database running at: " + database_url);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Drops all existing tables.
     * @throws SQLException if an error occurs
     */
    public void dropAllTables() throws SQLException {

        Statement statement = null;

        try {
            statement = connection.createStatement();
            statement.executeUpdate("DROP ALL OBJECTS;");
        }
        finally {
            if (statement != null) {
                statement.close();
            }
        }
    }

    private void executeSQLScript(final String filename) throws Exception {

        InputStream resource_stream = null;
        Statement statement = null;

        try {
            // First try to read query from the jar file, in which case need absolute path.
            resource_stream = getClass().getResourceAsStream(File.separator + filename);

            // If that doesn't work, try from local file system with relative path.
            if (resource_stream == null) {
                resource_stream = new FileInputStream(new File(filename));
            }

            final StringBuilder query = new StringBuilder();

            // Read query.
            int ch;
            while ((ch = resource_stream.read()) != -1) {
                query.append((char) ch);
            }

            // Execute query.
            statement = connection.createStatement();
            statement.executeUpdate(query.toString());

        }
        finally {
            try {
                if (resource_stream != null) {
                    resource_stream.close();
                }
            }
            finally {
                if (statement != null) {
                    statement.close();
                }
            }
        }
    }

    private boolean isLockException(final SQLException e) {

        return e.getMessage().contains("Couldn't obtain locks");
    }

    /**
     * executes a prepared statement
     * <p> 
     * tests if the connection is valid and only then executes the statement
     * 
     * @param prepared_statement to be executed
     * @return boolean true if the statement was executed, false when an exception was thrown
     */
    private boolean executeStatement(final PreparedStatement prepared_statement) {

        LOGGER.log(Level.INFO, "XXX - execute Statement " + prepared_statement);
        try {

            if (connection.isValid(5)) {
                prepared_statement.execute();
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.INFO, "CAUGHT EXCEPTION", e);
            return false;
        }
        catch (final Exception e) {
            LOGGER.log(Level.INFO, "CAUGHT EXCEPTION2", e);
            return false;
        }
        return true;
    }

    /**
     * tests if the queue is empty and calls <code>sendQueuedItems</code> if necessary
     * <p>
     * if the queue is empty the methods returns true. If the queue is not empty <code>sendQueuedItems</code>
     * will try to send off all items that are in the queue and returns true if successful in which case 
     * true is also returned by the method.
     * 
     * @return true if the queue is empty and all elements were successfully sent to the database
     */
    private boolean emptyQueue() {

        if (!workQueue.isEmpty() && !sendQueuedItems()) { return false; }
        return true;
    }

    /**
     * tries to send queued data to the database by casting the objects in the queue and calling 
     * an according method on the object to create a statement to be executed 
     * 
     * @return true if queue was emptied successfully
     */
    private boolean sendQueuedItems() {

        LOGGER.log(Level.INFO, "2a - try to send items in queue - no of items: " + workQueue.size());

        boolean connected = true;
        try {
            while (!workQueue.isEmpty() && connected) {

                LOGGER.log(Level.INFO, "item: " + workQueue.peek().toString());
                final Object object = workQueue.peek();
                if (object instanceof DistributionObject) {
                    connected = sendDistribution((DistributionObject) object);
                }
                else if (object instanceof FileSystemSummary) {
                    connected = sendFileSystemData((FileSystemSummary) object);
                }
                else if (object instanceof NetworkDataSummary) {
                    connected = sendNetworkData((NetworkDataSummary) object);
                }
                else if (object instanceof MachineUtilSummary) {
                    connected = sendMachineUtilData((MachineUtilSummary) object);
                }
                else if (object instanceof ProcessDataSummary) {
                    connected = sendProcessData((ProcessDataSummary) object);
                }
                else if (object instanceof LatencyAndBandwidthData) { //Latency Bandwidth
                    connected = sendLatencyAndBandwidth((LatencyAndBandwidthData) object);

                }
                else if (object instanceof Event) { //must be events
                    connected = insertEvent((Event) object);
                }
                else if (object instanceof SystemInfoData) {
                    connected = reportSystemInfo((SystemInfoData) object);
                }
                if (connected) {
                    workQueue.poll(); // remove top item if it was sent to the database
                }
            }

            if (workQueue.isEmpty()) {
                return true;
            }
            else {
                return false;
            }
        }
        catch (final Exception e) {
            LOGGER.log(Level.WARNING, "caught exception while trying to empty queeu", e);
            return false;
        }
    }

    @Override
    public synchronized boolean reportDistributionData(final DistributionCollector<?> distribution) throws SQLException, TimeoutException {

        //        LOGGER.log(Level.INFO, "1 - Report distr data");

        final String machine_id = distribution.getMachineId();
        final Distribution[] dataString = distribution.getValues(); // character representation
        final int num_measurements = distribution.getNumberOfMeasurements();
        final Date start_time = distribution.getStartTime();
        final Date end_time = distribution.getEndTime();

        boolean connected = true;
        DistributionObject obj;

        if (!emptyQueue()) {
            connected = false;
        }

        for (final Distribution data : dataString) {
            obj = new DistributionObject(machine_id, data, num_measurements, start_time, end_time);
            if (connected) {
                if (!sendDistribution(obj)) {
                    connected = false;
                    workQueue.add(obj);
                }
            }
            else {
                workQueue.add(obj);
            }
        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }

    }

    private boolean sendDistribution(final DistributionObject distribution) {

        try {
            if (connection.isValid(5)) {
                final boolean sent = executeStatement(distribution_data_prepared_statement.doInsertQuery("RESOURCE_MONITORING.DISTRIBUTION", connection, distribution.getMachine_id(), DATE_FORMAT.format(distribution.getStart_time()), DATE_FORMAT.format(distribution.getEnd_time()),
                                distribution.getNum_measurements(), distribution.getDataString().getInstrumentName(), distribution.getDataString().getAttributes(), distribution.getDataString().getCharChain()));
                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in sendDistribution", e);
            return false;
        }

        return true;
    }

    @Override
    public synchronized boolean reportLatencyAndBandwidthData(final LatencyAndBandwidthData data) throws SQLException, TimeoutException {

        LOGGER.log(Level.INFO, "1 - Report l_andB data");
        if (emptyQueue() && sendLatencyAndBandwidth(data)) {
            LOGGER.log(Level.INFO, "4 - success");
            return true;
        }
        workQueue.add(data);
        LOGGER.log(Level.INFO, "5 - no success - try to reset the connection");
        resetConnection();
        return false;
    }

    private boolean sendLatencyAndBandwidth(final LatencyAndBandwidthData data) {

        try {
            if (connection.isValid(5)) {
                final boolean sent = executeStatement(latency_bandwith_prepared_statement.doInsertQuery("RESOURCE_MONITORING.LATENCY_AND_BANDWIDTH", connection, data.getLocalMachineID(), DATE_FORMAT.format(data.getDate()), data.getLocalNetworkAdapterName(), data.getLocalGateway(),
                                data.getRemoteMachineID(), data.getRemoteNetworkAdapterName(), data.getRemoteGateway(), data.getLatency(), data.getBandwidth()));
                if (!sent) { return false; }
            }
            else {
                return false;
            }
        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException", e);
            return false;
        }
        return true;
    }

    @Override
    public synchronized boolean reportFileSystemData(final MultipleSummary<FileSystemData> fileSystemSummary) throws SQLException, TimeoutException {

        boolean connected = true;
        FileSystemSummary obj;

        if (!emptyQueue()) {
            connected = false;
        }
        for (final SingleSummary<FileSystemData> summary : fileSystemSummary.getSummaries()) {
            obj = new FileSystemSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());
            if (connected) {
                if (!sendFileSystemData(obj)) {
                    connected = false;
                    workQueue.add(obj);
                }
            }
            else {
                workQueue.add(obj);
            }

        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }

    }

    private synchronized boolean sendFileSystemData(final FileSystemSummary fileSystemSummary) {

        LOGGER.log(Level.INFO, "3 try to send to database");

        try {
            if (connection.isValid(5)) {
                final boolean sent = executeStatement(fs_prepared_statement.doInsertQuery("RESOURCE_MONITORING.FILE_SYSTEM", connection, fileSystemSummary.getMachine_id(), DATE_FORMAT.format(fileSystemSummary.getStart_time()), DATE_FORMAT.format(fileSystemSummary.getEnd_time()), fileSystemSummary
                                .getMeasurements_before_summary(), fileSystemSummary.getAvg().getFileSystemName(), fileSystemSummary.getAvg().getFileSystemLocation(), fileSystemSummary.getAvg().getFileSystemType(), fileSystemSummary.getAvg().getSize(), fileSystemSummary.getAvg().getSpaceFree(),
                                fileSystemSummary.getAvg().getSpaceUsed(), fileSystemSummary.getAvg().getNumberOfFiles(), fileSystemSummary.getMin().getNumberOfFiles(), fileSystemSummary.getMax().getNumberOfFiles(), fileSystemSummary.getAvg().getNumberOfDiskReads(), fileSystemSummary.getMin()
                                                .getNumberOfDiskReads(), fileSystemSummary.getMax().getNumberOfDiskReads(), fileSystemSummary.getAvg().getBytesRead(), fileSystemSummary.getMin().getBytesRead(), fileSystemSummary.getMax().getBytesRead(), fileSystemSummary.getAvg()
                                                .getNumberOfDiskWrites(), fileSystemSummary.getMin().getNumberOfDiskWrites(), fileSystemSummary.getMax().getNumberOfDiskWrites(), fileSystemSummary.getAvg().getBytesWritten(), fileSystemSummary.getMin().getBytesWritten(), fileSystemSummary.getMax()
                                                .getBytesWritten()));

                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in sendFileSystemSummary", e);
            return false;
        }
        return true;
    }

    @Override
    public synchronized boolean reportMachineUtilData(final SingleSummary<MachineUtilisationData> summary) throws SQLException, TimeoutException {

        boolean connected = true;
        final MachineUtilSummary obj = new MachineUtilSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());

        if (!emptyQueue()) {
            connected = false;
        }
        if (connected) {
            if (!sendMachineUtilData(obj)) {
                connected = false;
                workQueue.add(obj);
            }
        }
        else {
            workQueue.add(obj);
        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }

    }

    private boolean sendMachineUtilData(final MachineUtilSummary summary) {

        try {
            if (connection.isValid(5)) {
                final boolean sent = executeStatement(util_prepared_statement.doInsertQuery("RESOURCE_MONITORING.MACHINE_UTIL", connection, summary.getMachine_id(), DATE_FORMAT.format(summary.getStart_time()), DATE_FORMAT.format(summary.getEnd_time()), summary.getMeasurements_before_summary(),
                                summary.getAvg().getCpuUserTotal(), summary.getMin().getCpuUserTotal(), summary.getMax().getCpuUserTotal(), summary.getAvg().getCpuSysTotal(), summary.getMin().getCpuSysTotal(), summary.getMax().getCpuSysTotal(), summary.getAvg().getCpuNiceTotal(), summary.getMin()
                                                .getCpuNiceTotal(), summary.getMax().getCpuNiceTotal(), summary.getAvg().getCpuIdleTotal(), summary.getMin().getCpuIdleTotal(), summary.getMax().getCpuIdleTotal(), summary.getAvg().getCpuWaitTotal(), summary.getMin().getCpuWaitTotal(), summary
                                                .getMax().getCpuWaitTotal(), summary.getAvg().getMemoryUsed(), summary.getMin().getMemoryUsed(), summary.getMax().getMemoryUsed(), summary.getAvg().getMemoryFree(), summary.getMin().getMemoryFree(), summary.getMax().getMemoryFree(), summary.getAvg()
                                                .getSwapUsed(), summary.getMin().getSwapUsed(), summary.getMax().getSwapUsed(), summary.getAvg().getSwapFree(), summary.getMin().getSwapFree(), summary.getMax().getSwapFree()));
                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in sendMachineUtilSummary", e);
            return false;
        }
        return true;

    }

    @Override
    public synchronized boolean reportNetworkData(final SingleSummary<NetworkData> summary) throws SQLException, TimeoutException {

        boolean connected = true;
        final NetworkDataSummary obj = new NetworkDataSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());

        if (!emptyQueue()) {
            connected = false;
        }
        if (connected) {
            if (!sendNetworkData(obj)) {
                connected = false;
                workQueue.add(obj);
            }
        }
        else {
            workQueue.add(obj);
        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }

    }

    private boolean sendNetworkData(final NetworkDataSummary summary) {

        try {

            if (connection.isValid(5)) {

                final boolean sent = executeStatement(net_prepared_statement.doInsertQuery("RESOURCE_MONITORING.NETWORK", connection, summary.getMachine_id(), DATE_FORMAT.format(summary.getStart_time()), DATE_FORMAT.format(summary.getEnd_time()), summary.getMeasurements_before_summary(), summary
                                .getAvg().getDeviceName(), summary.getAvg().getIPAddress(), summary.getAvg().getRXBytes(), summary.getMin().getRXBytes(), summary.getMax().getRXBytes(), summary.getAvg().getRXPackets(), summary.getMin().getRXPackets(), summary.getMax().getRXPackets(), summary
                                .getAvg().getRXErrors(), summary.getMin().getRXErrors(), summary.getMax().getRXErrors(), summary.getAvg().getRXDropped(), summary.getMin().getRXDropped(), summary.getMax().getRXDropped(), summary.getAvg().getRXOverruns(), summary.getMin().getRXOverruns(), summary
                                .getMax().getRXOverruns(), summary.getAvg().getTXBytes(), summary.getMin().getTXBytes(), summary.getMax().getTXBytes(), summary.getAvg().getTXPackets(), summary.getMin().getTXPackets(), summary.getMax().getTXPackets(), summary.getAvg().getTXErrors(), summary.getMin()
                                .getTXErrors(), summary.getMax().getTXErrors(), summary.getAvg().getTXDropped(), summary.getMin().getTXDropped(), summary.getMax().getTXDropped(), summary.getAvg().getTXOverruns(), summary.getMin().getTXOverruns(), summary.getMax().getTXOverruns()));
                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in sendNetworkSystemSummary", e);
            return false;
        }

        return true;
    }

    @Override
    public synchronized boolean reportProcessData(final SingleSummary<ProcessData> summary) throws SQLException, TimeoutException {

        boolean connected = true;
        final ProcessDataSummary obj = new ProcessDataSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());

        if (!emptyQueue()) {
            connected = false;
        }
        if (connected) {
            if (!sendProcessData(obj)) {
                connected = false;
                workQueue.add(obj);
            }
        }
        else {
            workQueue.add(obj);
        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }

    }

    public boolean sendProcessData(final ProcessDataSummary summary) {

        try {
            if (connection.isValid(5)) {
                final boolean sent = executeStatement(process_prepared_statement.doInsertQuery("RESOURCE_MONITORING.PROCESS", connection, summary.getMachine_id(), DATE_FORMAT.format(summary.getStart_time()), DATE_FORMAT.format(summary.getEnd_time()), summary.getMeasurements_before_summary(),
                                summary.getAvg().getProcessName(), summary.getAvg().getProcessStartTime(), summary.getAvg().getProcessCpuPercent(), summary.getMin().getProcessCpuPercent(), summary.getMax().getProcessCpuPercent(), summary.getAvg().getProcessMemSize(), summary.getMin()
                                                .getProcessMemSize(), summary.getMax().getProcessMemSize(), summary.getAvg().getProcessResident(), summary.getMin().getProcessResident(), summary.getMax().getProcessResident()));

                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in sendProcessSummary", e);
            return false;
        }

        return true;
    }

    private boolean insertEvent(final Event event) {

        try {
            // create a comma separated list of attributes
            if (connection.isValid(5)) {
                final StringBuffer attributes = new StringBuffer();
                boolean more = false;
                if (event.getAttributes() != null) {

                    for (final String attribute : event.getAttributes()) {
                        if (more) {
                            attributes.append(",");
                        }
                        attributes.append(attribute);
                        more = true;
                    }
                }

                final boolean sent = executeStatement(event_prepared_statement.doInsertQuery("RESOURCE_MONITORING.EVENTS", connection, event.getMachineId(), event.getLabel(), DATE_FORMAT.format(event.getTime()), attributes.toString(), event.getMacAddress()));

                if (!sent) { return false; }
            }
            else {
                return false;
            }

        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLEXception", e);
            return false;

        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "EXception", e);
            return false;
        }
        return true;
    }

    @Override
    public synchronized boolean reportEventData(final List<Event> events) {

        //        LOGGER.log(Level.INFO, "1 - Report event data");
        boolean connected = true;
        for (final Event event : events) {
            if (connected && emptyQueue()) {
                if (!insertEvent(event)) {
                    connected = false;
                    workQueue.add(event);
                }
            }
            else {
                workQueue.add(event);
            }
        }
        if (connected) {
            return true;
        }
        else {
            //        LOGGER.log(Level.INFO, "5 - no success - try to reset the connection");
            LOGGER.log(Level.INFO, "QUEUE: added  " + events.toString());
            resetConnection();
            return false;
        }

    }

    @Override
    public synchronized boolean reportAllNetworkData(final MultipleSummary<NetworkData> summaries) {

        boolean connected = true;
        NetworkDataSummary obj;

        if (!emptyQueue()) {
            connected = false;
        }
        for (final SingleSummary<NetworkData> summary : summaries.getSummaries()) {
            obj = new NetworkDataSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());
            if (connected) {
                if (!sendNetworkData(obj)) {
                    connected = false;
                    workQueue.add(obj);
                }
            }
            else {
                workQueue.add(obj);
            }

        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }
    }

    @Override
    public synchronized boolean reportAllProcessData(final MultipleSummary<ProcessData> summaries) {

        boolean connected = true;
        ProcessDataSummary obj;

        if (!emptyQueue()) {
            connected = false;
        }
        for (final SingleSummary<ProcessData> summary : summaries.getSummaries()) {
            obj = new ProcessDataSummary(summary.getAverage(), summary.getMin(), summary.getMax(), summary.getStartTime(), summary.getEndTime(), summary.getMachineID(), summary.getNumberOfMeasurements());
            if (connected) {
                if (!sendProcessData(obj)) {
                    connected = false;
                    workQueue.add(obj);
                }
            }
            else {
                workQueue.add(obj);
            }

        }

        if (connected) {
            return true;
        }
        else {
            resetConnection();
            return false;
        }
    }

    @Override
    public synchronized boolean reportSystemInfo(final SystemInfoData data) {

        LOGGER.log(Level.INFO, "1 - Report ALLSystm data");
        if (entryExist(data.getMachineId())) {
            updateSystemInfo(data);
        }
        else {
            try {

                if (connection.isValid(5)) {
                    final boolean sent = executeStatement(sys_info_prepared_statement.doInsertQuery("RESOURCE_MONITORING.SYS_INFO", connection, data.getMachineId(), DATE_FORMAT.format(new Date()), data.getHostname(), data.getPrimaryIp(), data.getCpuVendor(), data.getCpuModel(),
                                    data.getNumberOfCores(), data.getNumberOfCpus(), data.getCpuClockSpeed(), data.getCpuCacheSize(), data.getOsName(), data.getOsVersion(), data.getDefaultGatewayIp(), data.getDefaultGatewayMAC(), data.getActive(), data.getMemoryTotal(), data.getSwapTotal()));
                    if (!sent) { return false; }
                }
                else {
                    workQueue.add(data);
                    return false;
                }

            }
            catch (final SQLException e) {
                workQueue.add(data);
                LOGGER.log(Level.SEVERE, "SQLException in reportAllNetworkData", e);
                return false;
            }
        }
        return true;
    }

    private boolean updateSystemInfo(final SystemInfoData data) {

        LOGGER.log(Level.INFO, "1a - UPDATE SYS_INFO data");
        try {
            final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO WHERE machine_id='" + data.getMachineId() + "';";
            Statement stat = null;
            if (connection.isValid(5)) {
                stat = connection.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_UPDATABLE);
                final ResultSet rs = stat.executeQuery(query);

                rs.first();
                rs.updateString("hostname", data.getHostname());
                rs.updateString("primary_ip", data.getPrimaryIp());
                rs.updateString("default_gateway_ip", data.getDefaultGatewayIp());
                rs.updateString("default_gateway_mac", data.getDefaultGatewayMAC());
                rs.updateString("active", "true");
                rs.updateRow();
                rs.close();
                stat.close();
            }
            else {
                LOGGER.log(Level.WARNING, "Connection is not valid anymore");
                workQueue.add(data);
                return false;
            }
        }
        catch (final SQLException e) {
            workQueue.add(data);
            LOGGER.log(Level.SEVERE, "SQLException in updateSystemInfo", e);
            return false;
        }
        return true;

    }

    @Override
    public boolean setHostActive() {

        LOGGER.log(Level.INFO, "SET HOST ACTIVE");
        try {
            final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO WHERE machine_id='" + machineID + "';";
            Statement stat = null;
            if (connection.isValid(5)) {
                stat = connection.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_UPDATABLE);
                final ResultSet rs = stat.executeQuery(query);

                rs.first();
                rs.updateString("active", "true");
                rs.updateRow();
                rs.close();
                stat.close();
            }
            else {
                LOGGER.log(Level.WARNING, "Connection is not valid anymore");
                return false;
            }
        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "SQLException in updateSystemInfo", e);
            return false;
        }
        return true;
    }

    private boolean entryExist(final String machine_id) {

        final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO WHERE machine_id='" + machine_id + "';";
        Statement stat = null;
        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement();
                final ResultSet rs = stat.executeQuery(query);
                int rowCount = 0;
                while (rs.next()) {
                    rowCount++;
                }
                if (rowCount > 0) {

                    return true;
                }
                else {
                    return false;
                }
            }

            stat.close();
        }
        catch (final SQLException e) {
            LOGGER.log(Level.SEVERE, "couldnt check if entry exists", e);
            return false;
        }
        return true;
    }

    /**
     * Returns system information.
     * @return system information
     * @throws SQLException if an error occurs
     */
    @Override
    public List<SystemInfoData> getSystemInfo() throws SQLException {

        final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO;";

        final List<SystemInfoData> machinesData = new LinkedList<SystemInfoData>();

        Statement stat = null;

        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement();
                final ResultSet rs = stat.executeQuery(query);

                while (rs.next()) {
                    final SystemInfoData data = new SystemInfoData(rs.getString("machine_id"), rs.getString("os_name"), rs.getString("os_version"), rs.getString("hostname"), rs.getString("primary_ip"), rs.getString("default_gateway_ip"), rs.getString("default_gateway_mac"),
                                    rs.getString("cpu_vendor"), rs.getString("cpu_model"), rs.getInt("num_cores"), rs.getInt("num_cpus"), rs.getInt("cpu_mhz"), rs.getLong("cpu_cache_size"), rs.getLong("memory_total"), rs.getLong("swap_total"), rs.getString("active"));

                    machinesData.add(data);
                }
            }
            else {
                return null;
            }
        }
        finally {
            stat.close();
        }
        return machinesData;
    }

    /**
     * Returns the known hosts.
     * @return the known hosts
     * @throws SQLException if an error occurs
     */
    @Override
    public synchronized List<String> getActiveKnownHosts() {

        final String query = "SELECT primary_ip FROM RESOURCE_MONITORING.SYS_INFO WHERE active = 'true';";
        final List<String> ipNames = new LinkedList<String>();
        Statement stat = null;

        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement();
                final ResultSet rs = stat.executeQuery(query);

                while (rs.next()) {

                    ipNames.add(rs.getString("primary_ip"));
                }

                stat.close();
            }
            else {
                return null;
            }
        }
        catch (final SQLException e) {
            LOGGER.log(Level.WARNING, "cant get active hosts", e);
            return null;
        }

        return ipNames;
    }

    @Override
    public synchronized void setHostPassive(final String primary_ip) throws SQLException {

        final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO WHERE primary_ip='" + primary_ip + "';";
        Statement stat = null;

        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_UPDATABLE);
                final ResultSet rs = stat.executeQuery(query);
                rs.first();
                rs.updateString("active", "false");
                rs.updateRow();
                rs.close();
            }
            else {
                LOGGER.log(Level.WARNING, "couldnt set host passive ...");
            }

        }
        finally {

            stat.close();
        }
    }

    @Override
    public List<FileSystemData> getFileSystemData(final int last_n_minutes) throws SQLException {

        final String query = "SELECT " + getBasicMonitoringInfo() + getFSInfo() + ofLastNTimePeriod(last_n_minutes, "MINUTE") + groupBy("machine_id", "fs_name");
        final List<FileSystemData> fsData = new LinkedList<FileSystemData>();
        Statement stat = null;

        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement();
                final ResultSet rs = stat.executeQuery(query);

                while (rs.next()) {
                    final FileSystemData data = new FileSystemData(rs.getDate("start_ts"));
                    data.setMachineId(rs.getString("machine_id"));
                    data.setFileSystemLocation(rs.getString("fs_location"));
                    data.setFileSystemName(rs.getString("fs_name"));
                    data.setFileSystemType(rs.getString("fs_type"));

                    data.setSpaceFree(rs.getLong("fs_free"));
                    data.setSpaceUsed(rs.getLong("fs_used"));
                    data.setSize(rs.getLong("fs_size"));

                    data.setBytesRead(rs.getLong("fs_disk_read_bytes_avg"));
                    data.setNumberOfDiskReads(rs.getLong("fs_disk_reads_avg"));
                    data.setBytesWritten(rs.getLong("fs_disk_write_bytes_avg"));
                    data.setNumberOfDiskWrites(rs.getLong("fs_disk_writes_avg"));
                    data.setNumberOfFiles(rs.getLong("fs_files_avg"));

                    fsData.add(data);
                }
            }
            else {
                throw new SQLException();
            }
        }
        finally {
            stat.close();
        }

        return fsData;
    }

    @Override
    public Collection<Event> getEvents(final int last_n_minutes) throws SQLException {

        //        final String query = "SELECT machine_id, label, time FROM RESOURCE_MONITORING.EVENTS" + ofLastNTimePeriod(last_n_minutes, "MINUTE");
        final List<Event> data = new LinkedList<Event>();
        //        Statement stat = null;
        //
        //        try {
        //            stat = connection.createStatement();
        //            final ResultSet rs = stat.executeQuery(query);
        //
        //            while (rs.next()) {
        //                final String machine_id = rs.getString("machine_id");
        //                final String label = rs.getString("label");
        //
        //                final Event data = new Event();
        //
        //                data.add(data);
        //            }
        //        }
        //        finally {
        //            stat.close();
        //        }

        //TODO AL IS HERE

        return data;
    }

    private String groupBy(final String... columns) {

        final StringBuilder s = new StringBuilder();
        s.append(" GROUP BY ");

        for (int i = 0; i < columns.length; i++) {
            s.append(columns[i]);

            if (i < columns.length - 1) {
                s.append(", ");
            }
        }
        return s.toString();
    }

    private String getBasicMonitoringInfo() {

        final StringBuilder builder = new StringBuilder();
        builder.append("machine_id,");
        builder.append("min(start_ts) as start_ts,");
        builder.append("sum(measurements) as measurements,");
        return builder.toString();
    }

    private String getFSInfo() {

        final StringBuilder builder = new StringBuilder();
        builder.append("fs_name, fs_type,fs_location,max(fs_size) as fs_size,min(fs_free) as fs_free, max(fs_used) as fs_used, avg(fs_files_avg) as fs_files_avg, min(fs_files_min) as fs_files_min, max(fs_files_max) as fs_files_max, avg(fs_disk_reads_avg) as fs_disk_reads_avg, min(fs_disk_reads_min) as fs_disk_reads_min, max(fs_disk_reads_max) as fs_disk_reads_max, ");
        builder.append("avg(fs_disk_read_bytes_avg) as fs_disk_read_bytes_avg, min(fs_disk_read_bytes_min) as fs_disk_read_bytes_min, max(fs_disk_read_bytes_max) as fs_disk_read_bytes_max, avg(fs_disk_writes_avg) as fs_disk_writes_avg, min(fs_disk_writes_min) as fs_disk_writes_min, max(fs_disk_writes_max) as fs_disk_writes_max, ");
        builder.append("avg(fs_disk_write_bytes_avg) as fs_disk_write_bytes_avg, min(fs_disk_write_bytes_min) as fs_disk_write_bytes_min, max(fs_disk_write_bytes_max) as fs_disk_write_bytes_max FROM resource_monitoring.file_system ");
        return builder.toString();
    }

    private String getDistributionData() {

        return "MACHINE_ID, min(START_TS) as start_ts, max(END_TS) as end_ts, sum(MEASUREMENTS) as measurements, " + "INSTRUMENT_NAME, ATTRIBUTES, avg(HIGH) as HIGH, avg(MEDIUM_HIGH) as MEDIUM_HIGH, avg(MEDIUM) as MEDIUM, " + "avg(MEDIUM_LOW) as medium_low, avg(LOW) as LOW "
                        + "FROM RESOURCE_MONITORING.DISTRIBUTION";
    }

    private String ofLastNTimePeriod(final int n, final String timePeriod) {

        return " WHERE start_ts >= DATEADD('" + timePeriod + "', -" + n + ", NOW())";
    }

    @Override
    public Collection<DistributionData> getDistributionData(final int time) throws SQLException {

        final Map<String, DistributionData> newData = new HashMap<String, DistributionData>();
        final String query = "SELECT " + getDistributionData() + ofLastNTimePeriod(time, "MINUTE") + groupBy("instrument_name", "attributes", "machine_id");
        Statement stat = null;

        try {
            if (connection.isValid(5)) {
                stat = connection.createStatement();
                final ResultSet rs = stat.executeQuery(query);

                while (rs.next()) {

                    final String machine_id = rs.getString("machine_id");

                    final String instrument_name = rs.getString("instrument_name");
                    final String attributes = rs.getString("attributes");
                    final Date date = rs.getDate("start_ts");
                    final float high = rs.getFloat("HIGH");
                    final float medium_high = rs.getFloat("MEDIUM_HIGH");
                    final float medium = rs.getFloat("MEDIUM");
                    final float medium_low = rs.getFloat("medium_low");
                    final float low = rs.getFloat("LOW");

                    final Distribution dist = new Distribution(instrument_name, attributes, date, high, medium_high, medium, medium_low, low);

                    DistributionData dd = newData.get(machine_id);

                    if (dd == null) {
                        dd = new DistributionData(machine_id);
                    }

                    dd.add(dist);

                    newData.put(machine_id, dd);
                }
            }
            else {
                throw new SQLException();
            }
        }
        finally {
            stat.close();
        }

        return newData.values();
    }

    private boolean closeConnection() {

        try {
            if (!connection.isClosed()) {
                connection.close();
                LOGGER.log(Level.INFO, "closed the connection");
            }
        }
        catch (final CommunicationsException e) {
            LOGGER.log(Level.SEVERE, "Couldn't close database connection");
        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "Exception when trying to close connection", e);
            return false;
        }
        return true;
    }

    /**
     * try to reset the connection 
     * @return successful reset if true
     */
    private boolean resetConnection() {

        LOGGER.log(Level.INFO, " RESET 0 - try to reset connection");
        try {
            closeConnection();
            {
                //                LOGGER.log(Level.INFO, "RESET 1 - closed existing connection ");
                String databaseDriverClassName = null;

                if (database_url.startsWith("jdbc:mysql")) {
                    databaseDriverClassName = "com.mysql.jdbc.Driver";
                }
                else {
                    databaseDriverClassName = "org.h2.Driver";
                }

                Class.forName(databaseDriverClassName).newInstance();
                //                LOGGER.log(Level.INFO, "RESET 2 - created new Driver instance");
                DriverManager.setLoginTimeout(2000);
                connection = DriverManager.getConnection(database_url, database_user_name, database_password);
                LOGGER.log(Level.INFO, "RESET 3 - Connected to database");
                resetQueryBuilders();
            }
        }
        catch (final CommunicationsException e) {
            LOGGER.log(Level.WARNING, " RESET Couldn't connect to remote database");
            return false;

        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "other exception ...");
            return false;

        }
        return true;
    }

    public void setMachineID(final String machineID) {

        this.machineID = machineID;
    }

    /**
     * reset the queryBuilders
     */
    private void resetQueryBuilders() {

        distribution_data_prepared_statement = new QueryBuilder();
        event_prepared_statement = new QueryBuilder();
        fs_prepared_statement = new QueryBuilder();
        util_prepared_statement = new QueryBuilder();
        net_prepared_statement = new QueryBuilder();
        process_prepared_statement = new QueryBuilder();
        sys_info_prepared_statement = new QueryBuilder();
        latency_bandwith_prepared_statement = new QueryBuilder();

    }
}
