/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.distribution;

import java.util.Date;

/**
 * Models data distributions.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class Distribution {

    /**
     * Possible intervals.
     */
    public static enum Interval {

        /**
         * High.
         */
        HIGH,

        /**
         * Medium high.
         */
        MEDIUM_HIGH,

        /**
         * Medium.
         */
        MEDIUM,

        /**
         * Medium low.
         */
        MEDIUM_LOW,

        /**
         * Low.
         */
        LOW
    };

    // -------------------------------------------------------------------------------------------------------

    private int number_added = 0; // The number of distributions that have been added to this one

    private final String instrument_name;
    private final String attributes;
    private final Date date;

    private float probability_high;
    private float probability_medium_high;
    private float probability_medium;
    private float probability_medium_low;
    private float probability_low;

    private String char_chain;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the distribution.
     *
     * @param instrument_name the instrument name
     * @param attributes the attributes
     * @param date the time of creation
     * @param probability_high the high probability
     * @param probability_medium_high the medium high probability
     * @param probability_medium the medium probability
     * @param probability_medium_low the medium low probability
     * @param probability_low the low probability
     */
    public Distribution(final String instrument_name, final String attributes, final Date date, final float probability_high, final float probability_medium_high, final float probability_medium, final float probability_medium_low, final float probability_low) {

        this.instrument_name = instrument_name;
        this.attributes = attributes;
        this.date = (Date) date.clone();

        this.probability_high = probability_high;
        this.probability_medium_high = probability_medium_high;
        this.probability_medium = probability_medium;
        this.probability_medium_low = probability_medium_low;
        this.probability_low = probability_low;
    }

    public Distribution(final String instrument_name, final String attributes, final Date date, final String char_chain) {

        this.instrument_name = instrument_name;
        this.attributes = attributes;
        this.date = (Date) date.clone();
        this.char_chain = char_chain;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the number of measurements.
     * @return the number of measurements
     */
    public int numberOfMeasurements() {

        return number_added;
    }

    /**
     * Adds values from another instance.
     * @param other the other instance
     */
    public void add(final Distribution other) {

        probability_high += other.probability_high;
        probability_medium_high += other.probability_medium_high;
        probability_medium += other.probability_medium;
        probability_medium_low += other.probability_medium_low;
        probability_low += other.probability_low;
        number_added++;
    }

    /**
     * Converts values to the average of all added distributions.
     */
    public void convertToAverage() {

        if (number_added != 0) {
            probability_high = probability_high / number_added;
            probability_medium_high = probability_medium_high / number_added;
            probability_medium = probability_medium / number_added;
            probability_medium_low = probability_medium_low / number_added;
            probability_low = probability_low / number_added;
        }
    }

    /**
     * Get the value of this probability according to a metric which defines the value of this resource in general, and the amount of
     * information needed.
     *
     * @param upTo the number of intervals used to obtain the value. For example, if 'MEDIUM' is specified here the value will be calculated
     *            based on the probability that something is between MEDIUM and LOW utilization
     * @return the value of this resource
     */
    public double getValue(final Interval upTo) {

        float toMeasure = 0;

        switch (upTo) {
            case HIGH:
                toMeasure += probability_high;
                break;
            case MEDIUM_HIGH:
                toMeasure += probability_medium_high;
                break;
            case MEDIUM:
                toMeasure += probability_medium;
                break;
            case MEDIUM_LOW:
                toMeasure += probability_medium_low;
                break;
            case LOW:
                toMeasure += probability_low;
                break;
            default:
                break;
        }

        return probability_low;
    }

    public char getCharRepresenation() {

        if (probability_high > 0) {
            return '5';
        }
        else if (probability_medium_high > 0) {
            return '4';
        }
        else if (probability_medium > 0) {
            return '3';
        }
        else if (probability_medium_low > 0) {
            return '2';
        }
        else if (probability_low > 0) {
            return '1';
        }
        else {
            return '6';
        }
    }

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();
        builder.append(instrument_name);
        builder.append(" attr: [");
        builder.append(attributes);
        builder.append("]");
        builder.append("distribution<");
        builder.append(probability_high);
        builder.append(",");
        builder.append(probability_medium_high);
        builder.append(",");
        builder.append(probability_medium);
        builder.append(",");
        builder.append(probability_medium_low);
        builder.append(",");
        builder.append(probability_low);
        builder.append(",");
        builder.append(">");
        return builder.toString();
    }

    /**
     * Returns the instrument name.
     * @return the instrument name
     */
    public String getInstrumentName() {

        return instrument_name;
    }

    /**
     * Returns the attributes.
     * @return the attributes
     */
    public String getAttributes() {

        return attributes;
    }

    /**
     * Returns the time of creation.
     * @return the time of creation
     */
    public Date getDate() {

        return (Date) date.clone();
    }

    /**
     * Returns the high probability.
     * @return the high probability
     */
    public float getProbabilityHigh() {

        return probability_high;
    }

    /**
     * Returns the medium high probability.
     * @return the medium high probability
     */
    public float getProbabilityMediumHigh() {

        return probability_medium_high;
    }

    /**
     * Returns the medium probability.
     * @return the medium probability
     */
    public float getProbabilityMedium() {

        return probability_medium;
    }

    /**
     * Returns the medium low probability.
     * @return the medium low probability
     */
    public float getProbabilityMediumLow() {

        return probability_medium_low;
    }

    /**
     * Returns the low probability.
     * @return the low probability
     */
    public float getProbabilityLow() {

        return probability_low;
    }

    public String getCharChain() {

        return char_chain;
    }

}
