/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.distribution;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.PrettyPrinter;
import uk.ac.standrews.cs.numonic.classify.IClassifier;

/**
 * Collects classifications of values of a given type.
 *
 * @param <T> the value type
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class DistributionCollector<T> {

    // List of probability distributions for a collection of instruments.
    // for example LIST< (cpu - (0.2,0.8),memory - (03,0.7),(cpu- (0.2,0.8),memory - (03,0.7)),(cpu - (0.2,0.8),memory - (03,0.7)) >
    private final List<Distribution[]> classifications;

    private final IClassifier<T> classifier;
    private final int measurements_before_summarization;
    private final String machineID;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a collector.
     *
     * @param classifier a classifier for the values
     * @param machineID the machine ID
     * @param measurementsBeforeSummarization the number of classifications to be accumulated before they are summarized
     */
    public DistributionCollector(final IClassifier<T> classifier, final String machineID, final int measurementsBeforeSummarization) {

        this.classifier = classifier;
        this.machineID = machineID;
        this.measurements_before_summarization = measurementsBeforeSummarization;

        classifications = new ArrayList<Distribution[]>();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests whether the collector is full.
     * @return true if the number of accumulated classifications has reached the threshold
     */
    public boolean isFull() {

        return classifications.size() >= measurements_before_summarization;
    }

    /**
     * Clears the accumulated classifications.
     */
    public void reset() {

        classifications.clear();
    }

    /**
     * Classifies a new measurement and stores the resulting distribution.
     * @param info the measurement
     */
    public void addNewMeasurement(final T info) {

        final Distribution[] distribution = classifier.classify(info);
        if (distribution != null) {
            classifications.add(distribution);
        }
    }

    /**
     * returns a characterRepresentation of the values measured
     * i.e. cpu - [1,1,2,1,1] where the following mapping is used: 1 - low; 2 - medLow, 
     * 3 - medium, 4 - medHigh, 5 - high
     */
    public Distribution[] getValues() {

        final Distribution[] charChain;
        try {
            final int num_instruments = classifications.get(0).length;
            charChain = new Distribution[num_instruments];
            final String[] charString = new String[num_instruments];

            for (int i = 0; i < charChain.length; i++) {
                charString[i] = "";
            }

            for (final Distribution[] array : classifications) {
                for (int i = 0; i < array.length; i++) {
                    charString[i] += array[i].getCharRepresenation();
                }
            }

            for (int i = 0; i < charChain.length; i++) {
                charChain[i] = new Distribution(classifications.get(0)[i].getInstrumentName(), classifications.get(0)[i].getAttributes(), classifications.get(0)[i].getDate(), charString[i]);

            }
        }
        catch (final NullPointerException e) {
            return null;
        }

        return charChain;
    }

    /**
     * Returns an array of probability distributions which is the aggregate of the originals.
     * For example: [cpu - (0.2,0.8),memory - (03,0.7)]
     *
     * @return returns an array of probability distributions which is the aggregate of the originals
     */
    public Distribution[] getProbabilities() {

        final int num_instruments = classifications.get(0).length; // The number of different instruments for which we have probability distributions.

        final Distribution[] probabilities = new Distribution[num_instruments];

        // Initialise the arrays.
        for (int i = 0; i < num_instruments; i++) {
            probabilities[i] = new Distribution(classifications.get(0)[i].getInstrumentName(), classifications.get(0)[i].getAttributes(), classifications.get(0)[i].getDate(), 0.0f, 0.0f, 0.0f, 0.0f, 0.0f);
        }

        for (final Distribution[] array : classifications) {
            for (int i = 0; i < array.length; i++) {
                probabilities[i].add(array[i]);
            }
        }

        for (int i = 0; i < num_instruments; i++) {
            probabilities[i].convertToAverage();
        }

        // DEBUG
        if (Diagnostic.getLevel() == DiagnosticLevel.FULL) {
            for (int i = 0; i < num_instruments; i++) {
                Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Prob[" + i + "]= " + probabilities[i]);
            }
        }

        return probabilities;
    }

    /**
     * Returns the number of accumulated measurements.
     * @return the number of accumulated measurements
     */
    public int getNumberOfMeasurements() {

        return classifications.size();
    }

    /**
     * Returns the machine ID.
     * @return the machine ID
     */
    public String getMachineId() {

        return machineID;
    }

    /**
     * Returns the time of the first stored classification.
     * @return the time of the first stored classification
     */
    public Date getStartTime() {

        return classifications.get(0)[0].getDate();
    }

    /**
     * Returns the time of the last stored classification.
     * @return the time of the last stored classification
     */
    public Date getEndTime() {

        return classifications.get(classifications.size() - 1)[0].getDate();
    }

    @Override
    public String toString() {

        return "DistributionCollector [classifications=" + PrettyPrinter.toString(classifications) + ", measurements_before_summarization=" + measurements_before_summarization + ", machineID=" + machineID + "]";
    }

    public T getContentType() {

        return classifier.getContentType();
    }

}
