/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.event;

import java.util.Date;
import java.util.List;

/**
 * Representation of modelling event.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class Event {

    /**
     * Start monitoring event.
     */
    public static final String START_MONITORING = "START_MONITORING";

    /**
     * Stop monitoring event.
     */
    public static final String STOPPED_MONITORING = "STOPPED_MONITORING";

    /**
     * File systems changed event.
     */
    public static final String FILESYSTEMS_CHANGED = "FILESYSTEMS_CHANGED";

    /**
     * Primary network changed event.
     */
    public static final String IP_ADDRESS_CHANGE = "IP_ADDRESS_CHANGE";

    /**
     * Suspected system sleep event.
     */
    public static final String SUSPECTED_SYSTEM_SLEEP = "SUSPECTED_SYSTEM_SLEEP";

    /**
     * Suspected system wake event.
     */
    public static final String SUSPECTED_SYSTEM_WAKE = "SUSPECTED_SYSTEM_WAKE";

    private final Date time;
    private final String label;
    private final List<String> attributes;
    private final String machine_id;
    private final String mac_Address;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the event.
     *
     * @param label the event label
     * @param machine_id the machine ID
     * @param date the time
     * @param attributes the event attributes
     */
    public Event(final String label, final String machine_id, final Date date, final List<String> attributes, final String mac_Address) {

        time = (Date) date.clone();
        this.machine_id = machine_id;
        this.label = label;
        this.attributes = attributes;
        this.mac_Address = mac_Address;
    }

    /**
     * Initializes the event with the current time.
     *
     * @param label the event label
     * @param machine_id the machine ID
     * @param attributes the event attributes
     */
    public Event(final String label, final String machine_id, final List<String> attributes, final String mac_Address) {

        this(label, machine_id, new Date(), attributes, mac_Address);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the time of recording.
     * @return the time of recording
     */
    public Date getTime() {

        return (Date) time.clone();
    }

    /**
     * Returns the event label.
     * @return the event label
     */
    public String getLabel() {

        return label;
    }

    /**
     * Returns the event attributes.
     * @return the event attributes
     */
    public List<String> getAttributes() {

        return attributes;
    }

    /**
     * Returns the machine ID.
     * @return the machine ID
     */
    public String getMachineId() {

        return machine_id;
    }

    public String getMacAddress() {

        return mac_Address;
    }

    @Override
    public String toString() {

        return "Event [time=" + time + ", label=" + label + ", attributes=" + attributes + ", machine_id=" + machine_id + "]";
    }

}
