/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.main;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.UnknownHostException;
import java.nio.channels.FileLock;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.hyperic.sigar.NetInterfaceConfig;
import org.hyperic.sigar.Sigar;
import org.hyperic.sigar.SigarException;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;
import uk.ac.standrews.cs.nds.p2p.util.SHA1KeyFactory;
import uk.ac.standrews.cs.nds.util.CommandLineArgs;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.Duration;
import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.nds.util.NetworkUtil;
import uk.ac.standrews.cs.numonic.classify.FileSystemClassifier;
import uk.ac.standrews.cs.numonic.classify.MachineUtilisationClassifier;
import uk.ac.standrews.cs.numonic.classify.NetworkClassifier;
import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.MachineUtilisationData;
import uk.ac.standrews.cs.numonic.data.NetworkData;
import uk.ac.standrews.cs.numonic.data.ProcessData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.distribution.DistributionCollector;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.event.EventCollector;
import uk.ac.standrews.cs.numonic.monitor.FileSystemMonitor;
import uk.ac.standrews.cs.numonic.monitor.H2OProcessMonitor;
import uk.ac.standrews.cs.numonic.monitor.MachineUtilisationMonitor;
import uk.ac.standrews.cs.numonic.monitor.NetworkMonitor;
import uk.ac.standrews.cs.numonic.monitor.ProcessMonitor;
import uk.ac.standrews.cs.numonic.monitor.SystemMonitor;
import uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency.BandwidthAndLatencyServer;
import uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency.SimpleBandwidthAndLatencyClient;
import uk.ac.standrews.cs.numonic.reporting.IRemoteHosts;
import uk.ac.standrews.cs.numonic.reporting.IReporting;
import uk.ac.standrews.cs.numonic.reportingFactories.IRemoteHostsFactory;
import uk.ac.standrews.cs.numonic.reportingFactories.IReportingFactory;
import uk.ac.standrews.cs.numonic.reportingFactories.ReportingInitialisationException;
import uk.ac.standrews.cs.numonic.summary.MultipleSummary;
import uk.ac.standrews.cs.numonic.summary.SingleSummary;
import uk.ac.standrews.cs.numonic.util.IPropertiesWrapper;
import uk.ac.standrews.cs.numonic.util.PropertiesWrapper;

import com.mysql.jdbc.exceptions.jdbc4.CommunicationsException;

/**
 * Monitors the processes and the availability of resources on the local machine and reports it to a reporting class specified in a
 * properties file..
 */
public final class Numonic extends Thread {

    private static final int SLEEP_TIME_THRESHOLD_FACTOR = 300000; // five minute
    private static final String TEMP_DIR_PROPERTY_NAME = "java.io.tmpdir";
    private static final String NUMONIC_LOCK_FILE_NAME = "numonic_lock.txt";
    private static final String HEARTBEAT_FILE_NAME = "numonic_heartbeart.txt";

    private static FileLock numonic_lock;
    private static FileOutputStream lock_output_stream;

    // Names of properties
    private static final String PROPERTY_COMMENT = "Numonic properties file";
    private static final String MACHINEID = "MACHINEID";
    private static final String H2O_PROCESS_SUMMARY_ENABLED = "H2O_process_summary_enabled";
    private static final String MACHINE_UTILISATION_SUMMARY_ENABLED = "machine_utilisation_summary_enabled";
    private static final String MACHINE_UTILISATION_DISTRIBUTION_ENABLED = "machine_utilisation_distribution_enabled";
    private static final String ALL_PROCESS_SUMMARY_ENABLED = "all_process_summary_enabled";
    private static final String ALL_NETWORK_SUMMARY_ENABLED = "all_network_summary_enabled";
    private static final String PRIMARY_NETWORK_SUMMARY_ENABLED = "primary_network_summary_enabled";
    private static final String FILE_SYSTEM_SUMMARY_ENABLED = "file_system_summary_enabled";
    private static final String EVENT_COLLECTION_ENABLED = "event_collection_enabled";
    private static final String ALL_NETWORK_DISTRIBUTION_ENABLED = "all_network_distribution_enabled";
    private static final String PRIMARY_NETWORK_DISTRIBUTION_ENABLED = "primary_network_distribution_enabled";
    private static final String FILE_SYSTEM_DISTRIBUTION_ENABLED = "file_system_distribution_enabled";
    private static final String BANDWIDTH_AND_LATENCY_ENABLED = "bandwidth_and_latency_enabled";
    private static final String TIME_BETWEEN_MEASUREMENTS = "time_between_measurements";
    private static final String PROCESS_MEASUREMENTS_BEFORE_SUMMARIZATION = "processMeasurementsBeforeSummarization";
    private static final String MACHINE_UTILISATION_MEASUREMENTS_BEFORE_SUMMARIZATION = "machineUtilisationMeasurementsBeforeSummarization";
    private static final String NETWORK_MEASUREMENTS_BEFORE_SUMMARIZATION = "networkMeasurementsBeforeSummarization";
    private static final String FILESYSTEM_MEASUREMENTS_BEFORE_SUMMARIZATION = "filesystemMeasurementsBeforeSummarization";
    private static final String EVENTS_BEFORE_SUMMARIZATION = "eventsBeforeSummarization";
    private static final String MACHINE_UTILISATION_MEASUREMENTS_BEFORE_DISTRIBUTION = "machineUtilisationMeasurementsBeforeDistribution";
    private static final String NETWORK_MEASUREMENTS_BEFORE_DISTRIBUTION = "networkMeasurementsBeforeDistribution";
    private static final String REPORTING_FACTORY_CLASS = "reportingFactoryClass";
    private static final String REMOTE_HOSTS_FACTORY_CLASS = "remoteHostsFactoryClass";

    // Defaults
    private static final long DEFAULT_SLEEP_TIME = 1000; // 1000 ms = 1 second
    private static final int DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION = 5;
    private static final String MACHINE_ID_PROPERTIES_FILE_LOCATION = "numonic-machineid.properties";

    private static final File TEMP_DIR = new File(System.getProperty(TEMP_DIR_PROPERTY_NAME));

    // Property fields
    private boolean h2o_process_summary_enabled = false;
    private boolean machine_utilisation_summary_enabled = true;
    private boolean machine_utilisation_distribution_enabled = true;
    private boolean all_process_summary_enabled = false;
    private boolean primary_network_summary_enabled = true;
    private boolean all_network_summary_enabled = false;
    private boolean file_system_summary_enabled = true;
    private boolean event_collection_enabled = true;
    private boolean primary_network_distribution_enabled = true;
    private boolean all_network_distribution_enabled = false;
    private boolean file_system_distribution_enabled = true;
    private boolean bandwidth_and_latency_enabled = false;

    // Summary and distribution settings
    private int processMeasurementsBeforeSummarization = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int machineUtilisationMeasurementsBeforeSummarization = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int networkMeasurementsBeforeSummarization = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int filesystemMeasurementsBeforeSummarization = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int eventsBeforeSummarization = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int machineUtilisationMeasurementsBeforeDistribution = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;
    private int networkMeasurementsBeforeDistribution = DEFAULT_MEASUREMENTS_BEFORE_SUMMARIZATION;

    private boolean doing_machine_utilisation_monitoring = false;
    private boolean doing_fs_monitoring = false;
    private boolean doing_all_network_monitoring = false;
    private boolean doing_primary_network_monitoring = false;

    // Monitoring classes
    private H2OProcessMonitor h2OProcessMonitor;
    private ProcessMonitor processMonitor;
    private MachineUtilisationMonitor utilization;
    private FileSystemMonitor fileSysMonitor;
    private NetworkMonitor networkMonitor;

    // Summary Instances.
    private SingleSummary<ProcessData> h2oProcessSummary;
    private MultipleSummary<ProcessData> allProcessSummary;
    private SingleSummary<MachineUtilisationData> machineSummary;
    private SingleSummary<NetworkData> primaryNetworkSummary;
    private MultipleSummary<NetworkData> allNetworkSummary;
    private MultipleSummary<FileSystemData> fileSystemSummary;

    // Distribution collection Instances.
    private DistributionCollector<MachineUtilisationData> machineUtilisationDistribution;
    private DistributionCollector<NetworkData> primaryNetworkDistribution;
    private Map<String, DistributionCollector<NetworkData>> allNetworkDistribution;
    private EventCollector eventCollector;
    private Map<String, DistributionCollector<FileSystemData>> fileSystemDistributionCollectors;

    // General variables.
    private final IPropertiesWrapper properties;
    private long sleepTime = DEFAULT_SLEEP_TIME;
    private boolean running = true;
    private IReporting reporter;
    private IRemoteHosts remoteHosts;
    private String machine_id;
    private IPropertiesWrapper localNumonicProperties;

    private String primaryIPAddress = null;
    private final Sigar sigar;

    static String propertiesFileLocation;
    private SimpleBandwidthAndLatencyClient b_and_lclient;
    private BandwidthAndLatencyServer r;
    public static Logger LOGGER = NumonicLogger.LOGGER;
    static String logFileName;
    private SystemInfoData sysInfoResults;
    private SystemMonitor sysInfo;
    private int counter;

    // -------------------------------------------------------------------------------------------------------

    public Numonic(final IPropertiesWrapper properties) throws UnknownHostException {

        LOGGER.log(Level.INFO, "Start Numonic");
        setName("numonic-monitoring-thread");
        this.properties = properties;
        sigar = new Sigar();
        processProperties();
        newSystemInfo();

        getLastHeartBeatUpdate(); // record when we were last up
        touchHeartBeatFile(); // record that we are running again

        // Conditionally start up the monitors

        if (all_process_summary_enabled) {
            processMonitor = new ProcessMonitor();
        }
        if (h2o_process_summary_enabled) {
            h2OProcessMonitor = new H2OProcessMonitor();
        }
        if (doing_machine_utilisation_monitoring) {
            utilization = new MachineUtilisationMonitor();
        }
        if (doing_fs_monitoring) {
            fileSysMonitor = new FileSystemMonitor();
        }
        if (doing_primary_network_monitoring || doing_all_network_monitoring) {
            networkMonitor = new NetworkMonitor();
        }

        // Instantiate Summary Instances.

        if (h2o_process_summary_enabled) {
            h2oProcessSummary = new SingleSummary<ProcessData>(new ProcessData(), machine_id, processMeasurementsBeforeSummarization);
        }
        if (machine_utilisation_summary_enabled) {
            machineSummary = new SingleSummary<MachineUtilisationData>(new MachineUtilisationData(), machine_id, machineUtilisationMeasurementsBeforeSummarization);
        }
        if (all_process_summary_enabled) {
            allProcessSummary = new MultipleSummary<ProcessData>(processMeasurementsBeforeSummarization, machine_id);
        }
        if (primary_network_summary_enabled) {
            primaryNetworkSummary = new SingleSummary<NetworkData>(new NetworkData(), machine_id, networkMeasurementsBeforeSummarization);
        }
        if (all_network_summary_enabled) {
            allNetworkSummary = new MultipleSummary<NetworkData>(networkMeasurementsBeforeSummarization, machine_id);
        }
        if (file_system_summary_enabled || file_system_distribution_enabled) {
            fileSystemSummary = new MultipleSummary<FileSystemData>(filesystemMeasurementsBeforeSummarization, machine_id);
        }
        if (event_collection_enabled) {
            eventCollector = new EventCollector(eventsBeforeSummarization);
            final List<String> attributes = new ArrayList<String>();
            attributes.add(sysInfoResults.getPrimaryIp());

            eventCollector.addEvent(new Event(Event.START_MONITORING, machine_id, attributes, sysInfoResults.getDefaultGatewayMAC())); // Record start of Monitoring event *********
        }

        // Instantiate Distribution Instances.

        if (machine_utilisation_distribution_enabled) {
            machineUtilisationDistribution = new DistributionCollector<MachineUtilisationData>(new MachineUtilisationClassifier(), machine_id, machineUtilisationMeasurementsBeforeDistribution);
        }
        if (primary_network_distribution_enabled) {
            primaryNetworkDistribution = new DistributionCollector<NetworkData>(new NetworkClassifier(eventCollector, machine_id, sysInfoResults.getDefaultGatewayMAC()), machine_id, networkMeasurementsBeforeDistribution);
        }
        if (all_network_distribution_enabled) {
            allNetworkDistribution = new HashMap<String, DistributionCollector<NetworkData>>();
        }
        if (file_system_distribution_enabled) {
            fileSystemDistributionCollectors = new HashMap<String, DistributionCollector<FileSystemData>>();
        }

        // Start up the inter-node Bandwidth & Latency Monitoring

        if (bandwidth_and_latency_enabled) {
            counter = 0;

            r = new BandwidthAndLatencyServer();
            r.startServices();

            b_and_lclient = new SimpleBandwidthAndLatencyClient(remoteHosts, reporter);
            b_and_lclient.start();
        }

        reportMonitoringStatus();
    }

    /**
     * Initialises the monitoring service.
     * 
     * @param propertiesFileLocation
     *            location of the properties file as a local path or URL
     * @throws IOException
     */
    public Numonic(final String propertiesFileLocation) throws IOException {

        this(loadPropertiesFile(propertiesFileLocation));
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Starts a Numonic instance.
     * 
     * @param args
     *            command line arguments including property file location specified with '-p'
     * @throws IOException
     */
    public static void main(final String[] args) {

        if (canLockExclusively()) {

            Diagnostic.setLevel(DiagnosticLevel.RUN);

            Diagnostic.setTimestampFlag(true);
            Diagnostic.setTimestampFormat(new SimpleDateFormat("HH:mm:ss:SSS "));
            Diagnostic.setTimestampDelimiterFlag(false);

            if (!CommandLineArgs.containsArg(args, "-p")) {
                ErrorHandling.hardError("No properties file or log name specified: use -p to specify numonic properties - use -l to specify logfile name");
            }

            final Map<String, String> arguments = CommandLineArgs.parseCommandLineArgs(args);
            propertiesFileLocation = arguments.get("-p");

            logFileName = arguments.get("-l");
            if (logFileName == null) {
                logFileName = new SimpleDateFormat("yyyy-MM-dd-HH-mm-ss").format(new Date()) + ".log";
            }

            Numonic systemMonitorBase;
            try {
                systemMonitorBase = new Numonic(propertiesFileLocation);
                systemMonitorBase.start();

            }
            catch (final IOException e) {
                LOGGER.log(Level.SEVERE, "IO Exception in main method", e);
            }

        }
        else {
            Diagnostic.trace(DiagnosticLevel.FINAL, "detected another numonic instance, quitting");
        }
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void run() {

        try {
            if (!reportStaticSystemInfo()) {
                LOGGER.log(Level.WARNING, "lost connection - quit program");
                System.exit(1);
            }

            while (isRunning()) {

                try {
                    Thread.sleep(2000);
                    takeMeasurements();
                }
                catch (final InterruptedException e) {
                    LOGGER.log(Level.WARNING, "Interrupted exception", e);
                }
            }

            releaseLock();
        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "Exception in RUN", e);
        }
    }

    /**
     * Returns true if the monitoring service is running.
     * 
     * @return true if the monitoring service is running
     */
    public synchronized boolean isRunning() {

        return running;
    }

    /**
     * Controls the monitoring service.
     * 
     * @param running
     *            true if the monitoring service should run
     */
    public synchronized void setRunning(final boolean running) {

        this.running = running;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Used to avoid running multiple Numonic instances on the same machine concurrently.
     * 
     * @return true if an exclusive lock has been obtained
     */
    private static synchronized boolean canLockExclusively() {

        try {
            final File lock_file = new File(TEMP_DIR, NUMONIC_LOCK_FILE_NAME);

            lock_output_stream = new FileOutputStream(lock_file);
            numonic_lock = lock_output_stream.getChannel().tryLock();
            return numonic_lock != null;
        }
        catch (final FileNotFoundException e) {
            return false;
        }
        catch (final IOException e) {
            return false;
        }
    }

    private static synchronized void releaseLock() {

        try {
            numonic_lock.release();
            lock_output_stream.close();
        }
        catch (final Exception e) {
            // Ignore.
        }
    }

    private void processProperties() {

        getMachineIDFromLocalPropertiesFile();

        try {

            getDataFromPropertiesFile();
        }
        catch (final Exception e) {
            ErrorHandling.hardExceptionError(e, "Quitting: cannot load properties.");
        }
    }

    public static IPropertiesWrapper loadPropertiesFile(final String propertiesFileLocation) throws IOException {

        final IPropertiesWrapper properties = PropertiesWrapper.getWrapper(propertiesFileLocation);
        properties.setPropertyComment(PROPERTY_COMMENT);

        properties.loadProperties();

        return properties;
    }

    public void getDataFromPropertiesFile() {

        // remoteHosts class
        final String remoteHostsclassname = properties.getProperty(REMOTE_HOSTS_FACTORY_CLASS);
        if (remoteHostsclassname == null) {
            Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "No Reporting Class specified (property " + REMOTE_HOSTS_FACTORY_CLASS + "). A reporting class MUST be specified by an application calling Numonic using Numonic.setReporter(IReporting).");
        }
        else {
            setRemoteHosts(remoteHostsclassname);
        }

        // reporting class
        final String reportingclassname = properties.getProperty(REPORTING_FACTORY_CLASS);
        if (reportingclassname == null) {
            Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "No Reporting Class specified (property " + REPORTING_FACTORY_CLASS + "). A reporting class MUST be specified by an application calling Numonic using Numonic.setReporter(IReporting).");
        }
        else {
            setReporter(reportingclassname);
        }

        // Process the rest of the properties

        h2o_process_summary_enabled = properties.isEnabled(H2O_PROCESS_SUMMARY_ENABLED);
        machine_utilisation_summary_enabled = properties.isEnabled(MACHINE_UTILISATION_SUMMARY_ENABLED);
        machine_utilisation_distribution_enabled = properties.isEnabled(MACHINE_UTILISATION_DISTRIBUTION_ENABLED);

        all_process_summary_enabled = properties.isEnabled(ALL_PROCESS_SUMMARY_ENABLED);
        primary_network_summary_enabled = properties.isEnabled(PRIMARY_NETWORK_SUMMARY_ENABLED);
        all_network_summary_enabled = properties.isEnabled(ALL_NETWORK_SUMMARY_ENABLED);
        file_system_summary_enabled = properties.isEnabled(FILE_SYSTEM_SUMMARY_ENABLED);
        event_collection_enabled = properties.isEnabled(EVENT_COLLECTION_ENABLED);
        bandwidth_and_latency_enabled = properties.isEnabled(BANDWIDTH_AND_LATENCY_ENABLED);
        primary_network_distribution_enabled = properties.isEnabled(PRIMARY_NETWORK_DISTRIBUTION_ENABLED);
        all_network_distribution_enabled = properties.isEnabled(ALL_NETWORK_DISTRIBUTION_ENABLED);
        file_system_distribution_enabled = properties.isEnabled(FILE_SYSTEM_DISTRIBUTION_ENABLED);

        sleepTime = getIntProperty(TIME_BETWEEN_MEASUREMENTS);
        processMeasurementsBeforeSummarization = getIntProperty(PROCESS_MEASUREMENTS_BEFORE_SUMMARIZATION);
        machineUtilisationMeasurementsBeforeSummarization = getIntProperty(MACHINE_UTILISATION_MEASUREMENTS_BEFORE_SUMMARIZATION);
        networkMeasurementsBeforeSummarization = getIntProperty(NETWORK_MEASUREMENTS_BEFORE_SUMMARIZATION);
        filesystemMeasurementsBeforeSummarization = getIntProperty(FILESYSTEM_MEASUREMENTS_BEFORE_SUMMARIZATION);
        eventsBeforeSummarization = getIntProperty(EVENTS_BEFORE_SUMMARIZATION);
        machineUtilisationMeasurementsBeforeDistribution = getIntProperty(MACHINE_UTILISATION_MEASUREMENTS_BEFORE_DISTRIBUTION);
        networkMeasurementsBeforeDistribution = getIntProperty(NETWORK_MEASUREMENTS_BEFORE_DISTRIBUTION);

        doing_machine_utilisation_monitoring = machine_utilisation_summary_enabled || machine_utilisation_distribution_enabled;
        doing_fs_monitoring = file_system_summary_enabled || file_system_distribution_enabled;

        doing_all_network_monitoring = all_network_summary_enabled || all_network_distribution_enabled;
        doing_primary_network_monitoring = primary_network_summary_enabled || primary_network_distribution_enabled;
    }

    public void getMachineIDFromLocalPropertiesFile() {

        // Get the Machine ID
        localNumonicProperties = PropertiesWrapper.getWrapper(MACHINE_ID_PROPERTIES_FILE_LOCATION);
        try {
            localNumonicProperties.loadProperties();
        }
        catch (final IOException e) {
            try {
                localNumonicProperties.createNewFile();
            }
            catch (final IOException e1) {
                ErrorHandling.hardError("Cannot create properties file" + MACHINE_ID_PROPERTIES_FILE_LOCATION);
            }
        }

        machine_id = localNumonicProperties.getProperty(MACHINEID);

        if (machine_id == null || machine_id.equals("")) {
            //            final IKey my_key = new SHA1KeyFactory().generateKey(new Date().toString().getBytes()); // use the time to seed a unique id for
            final String network_ifs = generateMachineID();
            final IKey my_key = new SHA1KeyFactory().generateKey(network_ifs); // this node.
            machine_id = my_key.toString(Key.DEFAULT_RADIX);
            localNumonicProperties.setProperty(MACHINEID, machine_id);
            try {
                localNumonicProperties.saveAndClose();
            }
            catch (final IOException e) {
                ErrorHandling.hardError("Cannot save changes to properties file" + MACHINE_ID_PROPERTIES_FILE_LOCATION);
            }
        }
    }

    private String generateMachineID() {

        String id = "";
        NetInterfaceConfig config = null;
        try {
            final String[] names = sigar.getNetInterfaceList();
            for (final String name : names) {
                config = sigar.getNetInterfaceConfig(name);
                if (config.getHwaddr() != null && name.compareTo("lo0") != 0 && name.compareTo("lo") != 0) {
                    id += config.getHwaddr();
                }
                //                System.out.println("HardwareAddr on " + name + ": " + config.getHwaddr());
            }
        }
        catch (final SigarException e) {
            LOGGER.log(Level.SEVERE, "couldnt get the machine id", e);
        }

        return id.replaceAll(":", "");
    }

    private int getIntProperty(final String key) {

        try {
            return Integer.valueOf(properties.getProperty(key));
        }
        catch (final NumberFormatException e) {
            throw new NumberFormatException("invalid integer format for: " + key);
        }
    }

    /**
     * Set the reporter class by passing an existing instance of the class to numonic.
     * 
     * @param reportingInstance
     *            Reporter class to be used.
     */
    public void setReporter(final IReporting reportingInstance) {

        reporter = reportingInstance;
        reporter.setMachineID(machine_id);
    }

    /**
     * Set the reporter class by specifying where the FQ-name of the reporter classes factory. This method instantiates a new reporting
     * class using the factory.
     * 
     * @param reportingClassName
     *            The factory class used to create the reporter class. This class must implement {@link IReportingFactory}.
     */
    public void setReporter(final String reportingClassName) {

        try {
            final Class<?> clazz = Class.forName(reportingClassName);
            final IReportingFactory factory = (IReportingFactory) clazz.newInstance();
            factory.init(properties);
            setReporter(factory.create());
        }
        catch (final ClassNotFoundException e) {
            ErrorHandling.hardError("Cannot load classname " + reportingClassName + " (property " + REPORTING_FACTORY_CLASS + ")");
        }
        catch (final InstantiationException e) {
            ErrorHandling.hardError("Classname " + reportingClassName + "cannot be instantiated");
        }
        catch (final IllegalAccessException e) {
            ErrorHandling.hardError("Classname " + reportingClassName + "cannot be instantiated");
        }
        catch (final ReportingInitialisationException e) {
            ErrorHandling.hardError("Classname " + reportingClassName + "cannot be instantiated");
        }
    }

    /**
     * Return a reference to the reporting instance being used by numonic.
     * 
     * @return instance that monitoring results are being sent to.
     */
    public IReporting getReporter() {

        return reporter;
    }

    private void setRemoteHosts(final String remoteHostsClassName) {

        try {
            final Class<?> clazz = Class.forName(remoteHostsClassName);
            final IRemoteHostsFactory factory = (IRemoteHostsFactory) clazz.newInstance();
            factory.init(properties);
            remoteHosts = factory.create();

        }
        catch (final ClassNotFoundException e) {
            ErrorHandling.hardError("Cannot load classname " + remoteHostsClassName + " (property " + REPORTING_FACTORY_CLASS + ")");
        }
        catch (final InstantiationException e) {
            ErrorHandling.hardError("Classname " + remoteHostsClassName + "cannot be instantiated");
        }
        catch (final IllegalAccessException e) {
            ErrorHandling.hardError("Classname " + remoteHostsClassName + "cannot be instantiated");
        }
        catch (final ReportingInitialisationException e) {
            ErrorHandling.hardError("Classname " + remoteHostsClassName + "cannot be instantiated");
        }
    }

    // Record Distributions of data

    private void recordPrimaryNetworkDistribution(final NetworkData nmd) throws Exception {

        primaryNetworkDistribution.addNewMeasurement(nmd);

        if (primaryNetworkDistribution.isFull()) {
            reporter.reportDistributionData(primaryNetworkDistribution);
            primaryNetworkDistribution.reset();
        }
    }

    private void recordAllNetworkDistribution(final List<NetworkData> all_net_data) throws Exception {

        for (final NetworkData data : all_net_data) {
            final String nic_name = data.getDeviceName();
            DistributionCollector<NetworkData> netDataDistribution = allNetworkDistribution.get(nic_name);

            if (netDataDistribution == null) { // need to create a new entry for a new nic
                netDataDistribution = new DistributionCollector<NetworkData>(new NetworkClassifier(eventCollector, machine_id, sysInfoResults.getDefaultGatewayMAC()), machine_id, networkMeasurementsBeforeSummarization);
                allNetworkDistribution.put(nic_name, netDataDistribution);
            }
            netDataDistribution.addNewMeasurement(data);

            if (netDataDistribution.isFull()) {
                reporter.reportDistributionData(netDataDistribution);
                netDataDistribution.reset();
            }
        }
    }

    private void recordFileSystemUtilizationDistribution(final List<FileSystemData> lfsd) throws Exception {

        for (final FileSystemData fsd : lfsd) {
            final String fs_name = fsd.getFileSystemName();
            DistributionCollector<FileSystemData> filesystemUtilisationDistribution = fileSystemDistributionCollectors.get(fs_name);

            if (filesystemUtilisationDistribution == null) { // need to create a new entry for a new local file system.
                filesystemUtilisationDistribution = new DistributionCollector<FileSystemData>(new FileSystemClassifier(eventCollector, machine_id, sysInfoResults.getDefaultGatewayMAC()), machine_id, filesystemMeasurementsBeforeSummarization);
                fileSystemDistributionCollectors.put(fs_name, filesystemUtilisationDistribution);
            }

            filesystemUtilisationDistribution.addNewMeasurement(fsd);

            if (filesystemUtilisationDistribution.isFull()) {
                reporter.reportDistributionData(filesystemUtilisationDistribution);
                filesystemUtilisationDistribution.reset();
            }
        }
    }

    private void recordUtilizationDistribution(final MachineUtilisationData ud) throws Exception {

        machineUtilisationDistribution.addNewMeasurement(ud);
        if (machineUtilisationDistribution.isFull()) {
            reporter.reportDistributionData(machineUtilisationDistribution);

            machineUtilisationDistribution.reset();
        }
    }

    // Record Summaries of data

    private void recordFileSystemSummary(final List<FileSystemData> data) throws Exception {

        fileSystemSummary.addMeasurement(data);

        if (fileSystemSummary.isFull()) {
            reporter.reportFileSystemData(fileSystemSummary);
            fileSystemSummary.reset();
        }
    }

    private void recordH2OProcessSummary(final ProcessData data) throws Exception {

        if (data != null) {
            // Allow for H2O not running.

            h2oProcessSummary.addNewMeasurement(data);
            if (h2oProcessSummary.isFull()) {
                reporter.reportProcessData(h2oProcessSummary);
                h2oProcessSummary.reset();
            }
        }
    }

    private void recordProcessSummary(final List<ProcessData> data) throws Exception {

        allProcessSummary.addMeasurement(data);
        if (allProcessSummary.isFull()) {
            reporter.reportAllProcessData(allProcessSummary);
            allProcessSummary.reset();
        }
    }

    private void recordAllNetworkSummary(final List<NetworkData> data) throws Exception {

        allNetworkSummary.addMeasurement(data);

        if (allNetworkSummary.isFull()) {
            reporter.reportAllNetworkData(allNetworkSummary);
            allNetworkSummary.reset();
        }
    }

    private void recordPrimaryNetworkSummary(final NetworkData data) throws Exception {

        primaryNetworkSummary.addNewMeasurement(data);

        if (primaryNetworkSummary.isFull()) {
            reporter.reportNetworkData(primaryNetworkSummary);
            primaryNetworkSummary.reset();
        }
    }

    private void recordUtilizationSummary(final MachineUtilisationData data) throws Exception {

        machineSummary.addNewMeasurement(data);

        if (machineSummary.isFull()) {
            reporter.reportMachineUtilData(machineSummary);
            machineSummary.reset();
        }
    }

    private long getLastHeartBeatUpdate() {

        final File heart_beat_file = new File(HEARTBEAT_FILE_NAME);
        if (heart_beat_file.exists()) { return heart_beat_file.lastModified(); }
        return 0;
    }

    private void touchHeartBeatFile() {

        final File heart_beat_file = new File(HEARTBEAT_FILE_NAME);
        if (!heart_beat_file.exists()) {
            try {
                heart_beat_file.createNewFile();
            }
            catch (final IOException e) {
                ErrorHandling.error("Cannot create heart beat file: " + HEARTBEAT_FILE_NAME);
            }
        }
        else {
            if (!heart_beat_file.setLastModified(System.currentTimeMillis())) {
                ErrorHandling.error("Cannot touch heart beat file: " + HEARTBEAT_FILE_NAME);
            }
        }
    }

    private void reportMonitoringStatus() throws UnknownHostException {

        if (!(doing_fs_monitoring || all_process_summary_enabled || h2o_process_summary_enabled || all_network_summary_enabled || primary_network_summary_enabled || machine_utilisation_summary_enabled || all_network_distribution_enabled || primary_network_distribution_enabled
                        || machine_utilisation_distribution_enabled || file_system_distribution_enabled || event_collection_enabled || bandwidth_and_latency_enabled)) {
            ErrorHandling.hardError("Not monitoring anything - check configuration file (enable some kind of monitoring).");
        }

        Diagnostic.traceNoSource(DiagnosticLevel.FINAL, "\n~~ Numonic System Monitor ~~\nStarting monitoring on " + NetworkUtil.getLocalIPv4Address().getHostName() + " at " + new Date() + "\nTaking readings every " + sleepTime + " milliseconds.");

        if (file_system_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Summarizing local filesystems");
        }

        if (all_process_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Summarizing all system processed");
        }

        if (h2o_process_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Summarizing H2O process");
        }

        if (primary_network_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Summarizing primary network connection");
        }

        if (all_network_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Summarizing all network connections");
        }

        if (machine_utilisation_summary_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring machine utilisation");
        }

        if (primary_network_distribution_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring distributions of primary network connection");
        }

        if (all_network_distribution_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring distributions of all network connections");
        }

        if (machine_utilisation_distribution_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring distributions machine utilisation");
        }

        if (file_system_distribution_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring distributions local file system utilisation");
        }

        if (event_collection_enabled && !eventCollector.isEmpty()) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Recording events");
        }
        if (bandwidth_and_latency_enabled) {
            Diagnostic.traceNoEvent(DiagnosticLevel.RUN, "Monitoring bandwidth and latency utilisation");
        }
    }

    /**
     * Takes all distribution and summary measurements. Override this to take a subset of measurements.
     */

    private void takeMeasurements() {

        LOGGER.log(Level.INFO, "take measurements");

        try {
            checkPrimaryIPAddress();
            final ProcessData pmd = h2o_process_summary_enabled ? h2OProcessMonitor.getInfo() : null;
            final List<ProcessData> lpmd = all_process_summary_enabled ? processMonitor.getAllProcessData() : null;
            final List<FileSystemData> lfsd = doing_fs_monitoring ? fileSysMonitor.getLocalFileSystemData() : null;
            final NetworkData primary_net_data = doing_primary_network_monitoring ? networkMonitor.getPrimaryNetworkDeviceData() : null;
            final List<NetworkData> all_net_data = doing_all_network_monitoring ? networkMonitor.getAllNetworkDeviceData() : null;

            final MachineUtilisationData ud = doing_machine_utilisation_monitoring ? utilization.getInfo() : null;

            // First make sure we have not been sleeping

            final long last_running = getLastHeartBeatUpdate();
            final long milliseconds_elapsed = Duration.elapsed(last_running).getLength(TimeUnit.MILLISECONDS);

            if (milliseconds_elapsed > getSleepTime()) {
                if (eventCollector != null) {
                    final List<String> attributes = new ArrayList<String>();
                    attributes.add("0.0.0.0");
                    eventCollector.addEvent(new Event(Event.SUSPECTED_SYSTEM_SLEEP, machine_id, new Date(last_running), attributes, "00:00:00:00:00:00"));
                    eventCollector.addEvent(new Event(Event.SUSPECTED_SYSTEM_WAKE, machine_id, attributes, "00:00:00:00:00:00"));
                    updatePrimaryNetwork(); // update the primary network address
                }
            }

            touchHeartBeatFile(); // Record that we are still running

            if (file_system_summary_enabled) {
                recordFileSystemSummary(lfsd);
            }

            if (all_process_summary_enabled) {
                recordProcessSummary(lpmd);
            }

            if (h2o_process_summary_enabled && !all_process_summary_enabled) { // Don't do H2O if we are doing all processes
                recordH2OProcessSummary(pmd);
            }

            if (primary_network_summary_enabled) {
                recordPrimaryNetworkSummary(primary_net_data);
            }

            if (all_network_summary_enabled) {
                recordAllNetworkSummary(all_net_data);
            }

            if (machine_utilisation_summary_enabled) {
                recordUtilizationSummary(ud);
            }

            if (primary_network_distribution_enabled) {
                recordPrimaryNetworkDistribution(primary_net_data);
            }

            if (all_network_distribution_enabled) {
                recordAllNetworkDistribution(all_net_data);
            }

            if (machine_utilisation_distribution_enabled) {
                recordUtilizationDistribution(ud);
            }

            if (file_system_distribution_enabled) {
                recordFileSystemUtilizationDistribution(lfsd);
            }

            if (event_collection_enabled && !eventCollector.isEmpty()) {
                final List<Event> events = eventCollector.getEvents();
                reporter.reportEventData(events);
                eventCollector.reset();
            }

            if (bandwidth_and_latency_enabled) {
                if (counter >= 10) {
                    reporter.setHostActive(); // make sure the server appears as active in the database
                    counter = 0;
                }
                else {
                    counter++;
                }
            }

        }
        catch (final TimeoutException e) {
            LOGGER.log(Level.SEVERE, "Caught TimeoutException while taking measurements - Reset Connection", e);
            resetCollectors();
        }
        catch (final CommunicationsException e) {
            LOGGER.log(Level.SEVERE, "CommunicationsException thrown while taking measurements - Reset Connection", e);
            resetCollectors();
        }
        catch (final NullPointerException e) {
            LOGGER.log(Level.SEVERE, "NullPointerException while taking measurements - Reset Connection", e);
            resetCollectors();
        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "Exception thrown while taking measurements - Reset Connection", e);
            resetCollectors();
        }
    }

    private void resetCollectors() {

        if (file_system_summary_enabled) {
            fileSystemSummary.reset();
        }

        if (all_process_summary_enabled) {
            allProcessSummary.reset();
        }

        if (h2o_process_summary_enabled && !all_process_summary_enabled) { // Don't do H2O if we are doing all processes
            h2oProcessSummary.reset();
        }

        if (primary_network_summary_enabled) {
            primaryNetworkSummary.reset();
        }

        if (all_network_summary_enabled) {
            allNetworkSummary.reset();
        }

        if (machine_utilisation_summary_enabled) {
            machineSummary.reset();
        }

        if (primary_network_distribution_enabled) {
            primaryNetworkDistribution.reset();
        }

        if (all_network_distribution_enabled) {
            machineUtilisationDistribution.reset();
        }

        if (machine_utilisation_distribution_enabled) {
            machineUtilisationDistribution.reset();
        }

        if (file_system_distribution_enabled) {
            fileSystemDistributionCollectors.clear();

        }
    }

    private long getSleepTime() {

        return SLEEP_TIME_THRESHOLD_FACTOR + sleepTime;
    }

    /**
     * Take static system information (in a {@link SystemInfoData} class and send it to the numonic reporting class.
     */
    public boolean reportStaticSystemInfo() {

        try {
            sysInfoResults = sysInfo.getSystemData(machine_id);
            return reporter.reportSystemInfo(sysInfoResults);
        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "inserting system info into DB unsuccesful ... quitting");
            System.exit(1);
            //            Diagnostic.trace(DiagnosticLevel.FULL, "error recording system info, probably because the machine id is already present (this is OK): " + e.getMessage());
        }
        return false;
    }

    /**
     * checks the current primary ip address of the network the computer is connected 
     * to and fires an event if required
     * 
     */
    private void checkPrimaryIPAddress() {

        try {

            final String pIPAddress = sigar.getNetInterfaceConfig(null).getAddress();
            if (primaryIPAddress == null) {
                primaryIPAddress = pIPAddress;
                LOGGER.log(Level.INFO, "Assigned IP address in network: " + primaryIPAddress);
            }
            else if (pIPAddress.compareTo(primaryIPAddress) != 0) {
                LOGGER.log(Level.INFO, "Primary IP address changed: " + pIPAddress);

                // first record sleep and wake 
                final long last_running = getLastHeartBeatUpdate();
                final long milliseconds_elapsed = Duration.elapsed(last_running).getLength(TimeUnit.MILLISECONDS);

                if (milliseconds_elapsed > getSleepTime()) {
                    if (eventCollector != null) {
                        final List<String> attributes = new ArrayList<String>();
                        attributes.add("0.0.0.0");
                        eventCollector.addEvent(new Event(Event.SUSPECTED_SYSTEM_SLEEP, machine_id, new Date(last_running), attributes, sysInfoResults.getGatewayNoConnection()));
                        eventCollector.addEvent(new Event(Event.SUSPECTED_SYSTEM_WAKE, machine_id, attributes, sysInfoResults.getGatewayNoConnection()));
                    }
                }

                touchHeartBeatFile(); // Record that we are still running

                primaryIPAddress = pIPAddress;
                if (eventCollector != null) {
                    newSystemInfo();
                    final List<String> attributes = new ArrayList<String>();
                    attributes.add(sysInfoResults.getPrimaryIp());
                    eventCollector.addEvent(new Event(Event.IP_ADDRESS_CHANGE, machine_id, new Date(System.currentTimeMillis()), attributes, sysInfoResults.getDefaultGatewayMAC()));
                }

                if (bandwidth_and_latency_enabled) {
                    b_and_lclient.setLocationChange(true);
                }

            }
        }
        catch (final SigarException e) {
            LOGGER.log(Level.SEVERE, "Sigar Exception thrown while checking on primary IP address", e);
        }
        catch (final NullPointerException e) {
            LOGGER.log(Level.SEVERE, "NullPointer Exception thrown while checking on primary IP address", e);
        }
        catch (final Exception e) {
            LOGGER.log(Level.SEVERE, "exception thrown while checking on primary IP address", e);

        }

    }

    /**
     * sends a network address update regardless of the fact that the network has changed or not
     */
    private void updatePrimaryNetwork() {

        try {
            primaryIPAddress = sigar.getNetInterfaceConfig(null).getAddress();
            if (eventCollector != null) {
                newSystemInfo();
                final List<String> attributes = new ArrayList<String>();
                attributes.add(sysInfoResults.getPrimaryIp());
                eventCollector.addEvent(new Event(Event.IP_ADDRESS_CHANGE, machine_id, new Date(System.currentTimeMillis()), attributes, sysInfoResults.getDefaultGatewayMAC()));
            }
        }
        catch (final SigarException e) {
            LOGGER.log(Level.SEVERE, "Sigar Exception thrown while checking on primary IP address", e);
        }
    }

    /**
     * get new system information data
     */
    public void newSystemInfo() {

        sysInfo = new SystemMonitor();
        try {
            sysInfoResults = sysInfo.getSystemData(machine_id);
        }
        catch (final SigarException e) {
            e.printStackTrace();
        }
    }

}
