/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.monitor;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import org.hyperic.sigar.CpuInfo;
import org.hyperic.sigar.Mem;
import org.hyperic.sigar.NetInfo;
import org.hyperic.sigar.NetInterfaceConfig;
import org.hyperic.sigar.OperatingSystem;
import org.hyperic.sigar.SigarException;
import org.hyperic.sigar.Swap;

import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.main.NumonicLogger;

/**
 * Monitor for system data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class SystemMonitor extends MonitorBase {

    /**
     * Returns system data for the local machine.
     *
     * @param machine_id a label for the data
     * @return system data for the local machine
     * @throws SigarException if an error occurs accessing sigar data
     */

    private final String linux = "/sbin/arp -a";
    private final String mac = "arp -a";
    private final String windows = "arp -a";

    public final static Logger LOGGER = NumonicLogger.LOGGER;

    public SystemInfoData getSystemData(final String machine_id) throws SigarException {

        // OS Info.
        final OperatingSystem sys = OperatingSystem.getInstance();

        // Network Info.
        final NetInterfaceConfig config = sigar.getNetInterfaceConfig(null);
        final NetInfo netinfo = sigar.getNetInfo();
        String command = "";
        if (isWindows()) {
            command = windows;
        }
        else if (isMac()) {
            command = mac;
        }
        else {
            command = linux;
        }
        final String router_macAddress = getMACAddr(netinfo.getDefaultGateway().trim(), command);

        // CPU Info.
        final CpuInfo[] cpus = sigar.getCpuInfoList();
        final CpuInfo info = cpus[0];

        // Mem Info.
        final Mem mem = sigar.getMem();
        final Swap swap = sigar.getSwap();

        final SystemInfoData results = new SystemInfoData(machine_id, sys.getName(), sys.getVersion(), netinfo.getHostName(), config.getAddress(), netinfo.getDefaultGateway(), router_macAddress, info.getVendor(), info.getModel(), info.getTotalCores(), info.getTotalSockets(), info.getMhz(), info
                        .getCacheSize(), convertBytesToKBytes(mem.getTotal()), convertBytesToKBytes(swap.getTotal()), "true");

        return results;
    }

    /**
     * get the Mac address of the computer by running the execute command
     * and extract the resutlts to find the mac address of the router
     * @return the mac address or zero if there s none
     */
    private String getMACAddr(final String ipAddress, final String command) {

        String line = "";
        try {
            String buffer;
            final Process process = Runtime.getRuntime().exec(command);
            final BufferedReader input = new BufferedReader(new InputStreamReader(process.getInputStream()));
            while ((buffer = input.readLine()) != null) {
                line += buffer;
            }

        }
        catch (final IOException e) {
            //            LOGGER.log(Level.WARNING, "couldnt retrieve arp ", e);
            if (isUnix()) { return getMACAddr(ipAddress, mac); }
        }

        String mac = extract(line, ipAddress);
        if (mac == null) {
            mac = "00:00:00:00:00:00";
        }
        return mac;
    }

    private String extract(final String arp, final String ipAddress) {

        final StringTokenizer st = new StringTokenizer(arp);
        String token;
        while (st.hasMoreTokens()) {
            token = st.nextToken();
            if (token.indexOf(ipAddress) != -1) {
                if (isWindows()) {
                    return st.nextToken();
                }
                else {
                    st.nextToken();
                    return st.nextToken();
                }
            }
        }

        return null;
    }

    private static boolean isWindows() {

        final String os = System.getProperty("os.name").toLowerCase();
        return os.indexOf("win") >= 0;

    }

    public static boolean isMac() {

        final String os = System.getProperty("os.name").toLowerCase();
        //Mac
        return os.indexOf("mac") >= 0;

    }

    public static boolean isUnix() {

        final String os = System.getProperty("os.name").toLowerCase();
        //linux or unix
        return os.indexOf("nix") >= 0 || os.indexOf("nux") >= 0;

    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Prints out local system data.
     *
     * @param args ignored
     * @throws SigarException if an error occurs accessing sigar data
     */
    public static void main(final String[] args) throws SigarException {

        final SystemMonitor monitor = new SystemMonitor();
        System.out.println(monitor.getSystemData(""));
    }
}
