package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.StringTokenizer;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import uk.ac.standrews.cs.nds.p2p.util.RandomGenerator;
import uk.ac.standrews.cs.nds.p2p.util.SHA1Hash;
import uk.ac.standrews.cs.nds.util.Duration;

/**
 * Measures latency and bandwidth between this host and another, using TCP for both. Requires {@link BandwidthAndLatencyServer} to be running on the other host.
 * 
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class BandwidthAndLatencyClient extends BandwidthAndLatencyCommon {

    // -------------------------------------------------------------------------------------------------------

    private final InetAddress remote_address;

    private String remote_network_adapter_name = null;
    private String remote_machine_id = null;
    private String remote_gateway = null;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a client for a given remote server address.
     * 
     * @param remote_address the remote address
     * @throws IOException if there is an error obtaining the server configuration
     * @throws TimeoutException if the server does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public BandwidthAndLatencyClient(final InetAddress remote_address) throws IOException, TimeoutException, InterruptedException {

        super();
        this.remote_address = remote_address;
        getServerConfig();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Determines the latency in milliseconds between this host and the other specified in the constructor.
     * 
     * @return the latency in milliseconds between this host and the other
     * @throws IOException if an error occurs in communication with the remote host
     * @throws TimeoutException if the remote host does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public synchronized Duration getLatency() throws IOException, TimeoutException, InterruptedException {

        return getLatency(remote_address);
    }

    /**
     * Determines the bandwidth between this host and another specified in the constructor. This is based on the time to repeatedly send a
     * random data block between the two hosts.
     * 
     * @return the bandwidth in kB/s
     * @throws IOException if an error occurs in communication with the remote host
     * @throws TimeoutException if the remote host does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public synchronized long getBandwidth() throws IOException, TimeoutException, InterruptedException {

        return getBandwidth(remote_address);
    }

    /**
     * Returns the NIC name of the remote server.
     * @return the NIC name of the remote server
     */
    public String getRemoteNetworkAdapterName() {

        return remote_network_adapter_name;
    }

    /**
     * Returns the UUID of the remote server.
     * @return the UUID of the remote server
     */
    public String getRemoteMachineID() {

        return remote_machine_id;
    }

    /**
     * Returns the gateway address of the remote server.
     * @return the gateway address of the remote server
     */
    public String getRemoteGateway() {

        return remote_gateway;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Determines the latency between this host and the other.
     * 
     * @param remote_address the address of the remote host
     * @return the latency between the two hosts
     * 
     * @throws IOException if an error occurs in communication with the remote host
     * @throws TimeoutException if the remote host does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public static Duration getLatency(final InetAddress remote_address) throws IOException, TimeoutException, InterruptedException {

        final byte[] tiny_payload = new byte[]{0};

        final Duration elapsed_time = measureTimeToSendData(remote_address, tiny_payload, 1);

        return elapsed_time.dividedBy(2); // Since there and back messages - assumes symmetry in times.
    }

    /**
     * Determines the bandwidth in kB/s between this host and the other.
     * 
     * @param remote_address the address of the remote host
     * @return the bandwidth in kB/s between the two hosts
     * 
     * @throws IOException if an error occurs in communication with the remote host
     * @throws TimeoutException if the remote host does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public static synchronized long getBandwidth(final InetAddress remote_address) throws IOException, TimeoutException, InterruptedException {

        final byte[] payload = new RandomGenerator(new SHA1Hash()).randomBytes(BANDWIDTH_BLOCK_SIZE);

        final Duration duration = measureTimeToSendData(remote_address, payload, BANDWIDTH_BLOCK_COUNT);
        final long duration_in_ms = duration.getLength(TimeUnit.MILLISECONDS);
        final int bytes_transmitted = BANDWIDTH_BLOCK_COUNT * BANDWIDTH_BLOCK_SIZE;

        return kiloBytesPerSecond(duration_in_ms, bytes_transmitted);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Reports the latency and bandwidth between this host and another.
     * 
     * @param args a single parameter giving the target hostname e.g. mini.cs.st-andrews.ac.uk
     * 
     * @throws IOException if an error occurs in communication with the remote host
     * @throws TimeoutException if the remote host does not respond within the timeout interval (@link{#SOCKET_TIMEOUT})
     * @throws InterruptedException 
     */
    public static void main(final String[] args) throws IOException, TimeoutException, InterruptedException {

        if (args.length == 1) {

            final String host = args[0];
            final BandwidthAndLatencyClient client = new BandwidthAndLatencyClient(InetAddress.getByName(host));

            final Duration latency = client.getLatency();
            final long bandwidth = client.getBandwidth();
            final String remote_network_adapter_name = client.getRemoteNetworkAdapterName();
            final String remote_machine_id = client.getRemoteMachineID();
            final String remote_gateway = client.getRemoteGateway();

            System.out.println("Connection latency is: " + latency);
            System.out.println("Bandwidth: transmitted " + BANDWIDTH_BLOCK_COUNT + " packets, of " + BANDWIDTH_BLOCK_SIZE + " bytes, bandwidth = " + bandwidth + " KB/sec \n");
            System.out.println("Remote network adapter name is: " + remote_network_adapter_name);
            System.out.println("Remote machine id is: " + remote_machine_id);
            System.out.println("Remote gateway is: " + remote_gateway);
        }
        else {
            System.out.println("Requires a hostname as a parameter");
        }
    }

    // -------------------------------------------------------------------------------------------------------

    private static synchronized Duration measureTimeToSendData(final InetAddress remote_address, final byte[] payload, final int number_of_repetitions) throws IOException, TimeoutException, InterruptedException {

        Socket tcp_socket = null;
        InputStream input_stream = null;
        OutputStream output_stream = null;

        try {
            // Setup the connections.
            tcp_socket = makeSocket(remote_address, BandwidthAndLatencyCommon.DEFAULT_TCP_RECEIVE_PORT);
            input_stream = tcp_socket.getInputStream();
            output_stream = tcp_socket.getOutputStream();

            final DataInputStream in = new DataInputStream(input_stream);
            final DataOutputStream out = new DataOutputStream(output_stream);

            // Send timing request.
            out.write(DATA.getBytes());
            out.write('\n');

            // Send the number of blocks and the block length.
            out.write(intToByteArray(number_of_repetitions));
            out.write(intToByteArray(payload.length));
            out.flush();

            final Duration start = Duration.elapsed();

            // Send the data.
            for (int i = 0; i < number_of_repetitions; i++) {
                out.write(payload);
                out.flush();
            }

            // Wait for ack.
            readLine(in);

            return Duration.elapsed(start);
        }
        finally {

            // Tidy up.

            closeIfNotNull(input_stream);
            closeIfNotNull(output_stream);
            closeIfNotNull(tcp_socket);
        }
    }

    private synchronized void getServerConfig() throws IOException, TimeoutException, InterruptedException {

        Socket tcp_socket = null;
        InputStream input_stream = null;
        OutputStream output_stream = null;

        try {
            // Setup the connections.
            tcp_socket = makeSocket(remote_address, BandwidthAndLatencyCommon.DEFAULT_TCP_RECEIVE_PORT);
            input_stream = tcp_socket.getInputStream();
            output_stream = tcp_socket.getOutputStream();

            final DataInputStream in = new DataInputStream(input_stream);
            final DataOutputStream out = new DataOutputStream(output_stream);

            // Config request.
            out.write(CONFIG.getBytes());
            out.write('\n');

            // Read config.
            final String network_configuration = readLine(in);

            final StringTokenizer tokenizer = new StringTokenizer(network_configuration, SEPARATOR);

            remote_machine_id = tokenizer.nextToken();
            remote_network_adapter_name = tokenizer.nextToken();
            remote_gateway = tokenizer.nextToken();
        }
        finally {

            // Tidy up.

            closeIfNotNull(input_stream);
            closeIfNotNull(output_stream);
            closeIfNotNull(tcp_socket);
        }
    }

}
