/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import uk.ac.standrews.cs.numonic.main.NumonicLogger;

/**
 * Base class for bandwidth and client monitor.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public abstract class BandwidthAndLatencyClientControllerBase extends Thread {

    protected List<String> host_list;
    public final static Logger LOGGER = NumonicLogger.LOGGER;
    protected boolean locationChange = false;

    /**
     * Initialises the controller.
     */
    public BandwidthAndLatencyClientControllerBase() {

        host_list = new ArrayList<String>();
    }

    /**
     * Takes the appropriate network measurements. In the base class this does nothing.
     */
    protected void doMeasurement(@SuppressWarnings("unused") final String host_name) {

    }

    /**
     * Provides a hook to allow the process to sleep between measurements etc. The method is called each time around the thread loop after taking
     * measurements using {@link #doMeasurement(String)}. In the base class this sleeps for 1 second.
     * @throws InterruptedException 
     */
    protected void continueExecution() {

        try {
            Thread.sleep(5000);
        }
        catch (final InterruptedException e) {
            interrupt();
        }
    }

    /**
     * Called when all the hosts in the list have been processed. By default does nothing in base class. Intention is
     * that this could be used to pull down new hosts once all the known hosts have been processed or report results etc.
     */
    protected void refreshHosts() {

    }

    /**
     * Runs the client monitor, executing doMeasurement(String) against the list. Runs until desist() has been called. Calls continueExecution()
     * after each loop iteration.
     */
    @Override
    public synchronized void run() {

        while (true) {

            if (host_list == null) {
                host_list = new ArrayList<String>();
            }
            LOGGER.log(Level.INFO, "client measure on hosts: " + host_list);
            for (final String host : host_list) {

                doMeasurement(host);
            }

            continueExecution();
            refreshHosts();
        }

    }

    /**
     * Returns the list of monitored hosts.
     * @return the list of monitored hosts
     */
    public List<String> getMonitoredHosts() {

        return host_list;
    }
}
