/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.hyperic.sigar.NetInterfaceConfig;
import org.hyperic.sigar.Sigar;
import org.hyperic.sigar.SigarException;

import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.Duration;
import uk.ac.standrews.cs.nds.util.Timing;
import uk.ac.standrews.cs.numonic.util.IPropertiesWrapper;
import uk.ac.standrews.cs.numonic.util.PropertiesWrapper;

/**
 * Contains the code common to client and server.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class BandwidthAndLatencyCommon {

    /**
     * The default TCP port on which the server listens for bandwidth-measuring requests.
     */
    public static final int DEFAULT_TCP_RECEIVE_PORT = 32770;

    protected static final String ACK = "ACK";
    protected static final String DATA = "DATA";
    protected static final String CONFIG = "CONFIG";

    protected static final int SIZE_OF_INT = 4;
    protected static final String CONFIGURATION_NAME = "numonic-machineid.properties";

    protected static final String MACHINEID = "MACHINEID";
    protected static final String SEPARATOR = ";";

    // Test with 1 MB of data
    protected static final int BANDWIDTH_BLOCK_SIZE = 1024;
    protected static final int BANDWIDTH_BLOCK_COUNT = 1024;

    private static final int THOUSAND = 1000;
    private static final int KILO = 1024;

    private static final Duration SOCKET_RETRY_INTERVAL = new Duration(2, TimeUnit.SECONDS);
    private static final Duration SOCKET_TIMEOUT = new Duration(20, TimeUnit.SECONDS);

    private final String machine_id;
    private final Sigar sigar;

    // -------------------------------------------------------------------------------------------------------

    protected BandwidthAndLatencyCommon() {

        final IPropertiesWrapper pw = PropertiesWrapper.getWrapper(CONFIGURATION_NAME);
        try {
            pw.loadProperties();
        }
        catch (final IOException e) {
            // Ignore since the properties file will be created later.
        }
        machine_id = pw.getProperty(MACHINEID);
        sigar = new Sigar();
    }

    // -------------------------------------------------------------------------------------------------------

    protected String getLocalDefaultNetworkAdapterName() {

        try {
            final NetInterfaceConfig ifconfig = sigar.getNetInterfaceConfig(null); // null is the default interface.
            return ifconfig.getName();
        }
        catch (final SigarException e) {
            return "unknown";
        }
    }

    protected String getNetworkConfiguration() {

        final StringBuilder builder = new StringBuilder();
        builder.append(machine_id);
        builder.append(SEPARATOR);
        builder.append(getLocalDefaultNetworkAdapterName());
        builder.append(SEPARATOR);
        builder.append(getLocalGateway());
        return builder.toString();
    }

    protected String getLocalMachineID() {

        return machine_id;
    }

    protected String getLocalGateway() {

        try {
            return sigar.getNetInfo().getDefaultGateway();
        }
        catch (final SigarException e) {
            return "unknown";
        }
    }

    /**
     * Alternative to {@link Socket} constructors that calls {@link Socket#setReuseAddress(boolean)} to enable reuse of the fixed
     * server port even when there is a previous connection to that port in the timeout state.
     *
     * @param remote_address the server address
     * @param remote_port the server port
     * @return a socket connected to the given endpoint
     * @throws TimeoutException if the new socket can't be connected to
     * @throws SocketException if the attempt to enable port reuse fails
     * @throws InterruptedException 
     */
    protected static Socket makeSocket(final InetAddress remote_address, final int remote_port) throws TimeoutException, SocketException, InterruptedException {

        final Socket socket = new Socket();
        socket.setReuseAddress(true);
        final Callable<Socket> make_socket_action = new Callable<Socket>() {

            @Override
            public Socket call() throws Exception {

                socket.connect(new InetSocketAddress(remote_address, remote_port));
                return socket;
            }
        };

        try {
            return Timing.retry(make_socket_action, SOCKET_TIMEOUT, SOCKET_RETRY_INTERVAL, true, DiagnosticLevel.FULL);
        }
        catch (final Exception e) {
            launderException(e);
            return null;
        }

        //        final long start_time = System.currentTimeMillis();
        //
        //        while (System.currentTimeMillis() - start_time < SOCKET_TIMEOUT) {
        //            try {
        //                socket.connect(new InetSocketAddress(remote_address, remote_port));
        //                return socket;
        //            }
        //            catch (final IOException e) {
        //                try {
        //                    Thread.sleep(SOCKET_RETRY_INTERVAL);
        //                }
        //                catch (final InterruptedException e1) {
        //                    // Ignore and try again.
        //                }
        //            }
        //        }
        //
        //        throw new TimeoutException("Couldn't connect to remote socket at: " + remote_address + " : " + remote_port);
    }

    private static void launderException(final Exception e) throws TimeoutException, InterruptedException {

        if (e instanceof InterruptedException) { throw (InterruptedException) e; }
        if (e instanceof TimeoutException) { throw (TimeoutException) e; }
        if (e instanceof RuntimeException) { throw (RuntimeException) e; }

        throw new IllegalStateException("Unexpected checked exception", e);
    }

    protected static byte[] intToByteArray(final int n) {

        final ByteBuffer buffer = ByteBuffer.allocate(SIZE_OF_INT);
        buffer.putInt(n);
        return buffer.array();
    }

    protected static int byteArrayToInt(final byte[] buffer) {

        final ByteBuffer b = ByteBuffer.allocate(buffer.length);
        b.put(buffer);
        return b.getInt(0);
    }

    protected static long kiloBytesPerSecond(final long duration_in_ms, final int bytes_transmitted) {

        final long bytes_per_ms = bytes_transmitted / duration_in_ms;
        return bytes_per_ms * THOUSAND / KILO;
    }

    protected static String readLine(final DataInputStream in) throws IOException {

        final StringBuilder command = new StringBuilder();
        byte b = in.readByte();

        while (b != '\n') {

            command.append((char) b);
            b = in.readByte();
        }

        return command.toString();
    }

    protected static void closeIfNotNull(final Socket socket) throws IOException {

        if (socket != null) {
            socket.close();
        }
    }

    protected static void closeIfNotNull(final ServerSocket socket) throws IOException {

        if (socket != null) {
            socket.close();
        }
    }

    protected static void closeIfNotNull(final InputStream stream) throws IOException {

        if (stream != null) {
            stream.close();
        }
    }

    protected static void closeIfNotNull(final OutputStream stream) throws IOException {

        if (stream != null) {
            stream.close();
        }
    }
}
