package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.logging.Level;
import java.util.logging.Logger;

import uk.ac.standrews.cs.nds.util.NetworkUtil;
import uk.ac.standrews.cs.numonic.main.NumonicLogger;

/**
 * Server that accepts sequences of data blocks from other hosts, allowing them to estimate their latencies and
 * bandwidths relative to this host. Requires {@link BandwidthAndLatencyClient} to be running on other host.
 * 
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class BandwidthAndLatencyServer extends BandwidthAndLatencyCommon {

    private Server server;
    public final static Logger LOGGER = NumonicLogger.LOGGER;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises the server.
     */
    public BandwidthAndLatencyServer() {

        super();
        server = new Server();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Starts the service.
     */
    public void startServices() {

        server.start();
    }

    public void desist() {

        if (server != null) {
            server.desist();
        }
    }

    public void restartServer() {

        //        System.out.println("restart server");
        server = new Server();
        server.start();

    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Runs a server.
     * @param args ignored
     */
    public static void main(final String[] args) {

        final BandwidthAndLatencyServer server = new BandwidthAndLatencyServer();
        server.startServices();
    }

    // -------------------------------------------------------------------------------------------------------

    private class Server extends Thread {

        boolean running = true;
        ServerSocket server_socket = null;
        Socket tcp_socket = null;
        InputStream input_stream = null;
        OutputStream output_stream = null;

        @Override
        public synchronized void run() {

            while (running) {
                try {
                    //                    LOGGER.log(Level.INFO, "******RUN SERVER******");
                    receiveRequest();
                }
                catch (final SocketException e) {
                    LOGGER.log(Level.WARNING, "SQLException while receiving request in Server run");
                }
                catch (final IOException e) {
                    LOGGER.log(Level.WARNING, "Exception while receiving request in Server run");
                }

            }
        }

        public void desist() {

            try {
                running = false;
                closeIfNotNull(input_stream);
                closeIfNotNull(output_stream);
                closeIfNotNull(tcp_socket);
                closeIfNotNull(server_socket);
                //                receiveRequest();
            }
            catch (final SocketException e) {
                LOGGER.log(Level.INFO, "Catch SokcetException in desist ... ", e);
            }
            catch (final IOException e) {
                LOGGER.log(Level.WARNING, "Catch IOException in desist", e);
            }
        }

        /**
         * Receives and processes a request. The accepted requests are {@link #DATA} - which involves reading a specified number of data blocks,
         * each of a specified size - and {@link #CONFIG} - which involves returning this server's configuration.

         * @throws IOException if a network error occurs
         */
        void receiveRequest() throws IOException, SocketException {

            server_socket = null;
            tcp_socket = null;
            input_stream = null;
            output_stream = null;

            try {
                //                LOGGER.log(Level.INFO, "listen on Server for incoming requests");
                server_socket = NetworkUtil.makeReusableServerSocket(DEFAULT_TCP_RECEIVE_PORT);
                tcp_socket = server_socket.accept();
                //                LOGGER.log(Level.INFO, "Server received request");
                input_stream = tcp_socket.getInputStream();
                output_stream = tcp_socket.getOutputStream();

                final DataInputStream in = new DataInputStream(input_stream);
                final DataOutputStream out = new DataOutputStream(output_stream);

                final String command = readLine(in);

                if (command.equals(DATA)) {
                    doTimingRequest(in, out);
                }
                else if (command.equals(CONFIG)) {
                    doConfigRequest(out);
                }
            }
            finally {

                closeIfNotNull(input_stream);
                closeIfNotNull(output_stream);
                closeIfNotNull(tcp_socket);
                closeIfNotNull(server_socket);
            }
        }

        private void doTimingRequest(final DataInputStream in, final DataOutputStream out) throws IOException {

            final byte[] int_buffer = new byte[SIZE_OF_INT];

            // Read the number of blocks to be received.
            in.readFully(int_buffer);
            final int number_of_repetitions = byteArrayToInt(int_buffer);

            // Read the length of each block.
            in.readFully(int_buffer);
            final int payload_length = byteArrayToInt(int_buffer);

            final byte[] payload_buffer = new byte[payload_length];

            // Read the data.
            for (int i = 0; i < number_of_repetitions; i++) {
                in.readFully(payload_buffer);
            }

            // Acknowledge successful receipt.
            out.write(ACK.getBytes());
            out.write('\n');
            out.flush();
        }

        private void doConfigRequest(final DataOutputStream out) throws IOException {

            out.write(getNetworkConfiguration().getBytes());
            out.write('\n');
            out.flush();
        }
    }
}
