/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.net.InetAddress;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

import uk.ac.standrews.cs.nds.util.Duration;
import uk.ac.standrews.cs.numonic.data.LatencyAndBandwidthData;
import uk.ac.standrews.cs.numonic.main.NumonicLogger;
import uk.ac.standrews.cs.numonic.reporting.IRemoteHosts;
import uk.ac.standrews.cs.numonic.reporting.IReporting;
import uk.ac.standrews.cs.numonic.util.Time;

/**
 * Measures bandwidth and latency to specified hosts.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class SimpleBandwidthAndLatencyClient extends BandwidthAndLatencyClientControllerBase {

    protected Map<String, Data> recently_measured_hosts;
    private final IRemoteHosts remote_hosts;
    private final IReporting reporting;

    private int time_between_samples = Time.ONE_MINUTE_IN_MS;
    private final int time_between_samples2 = 5000;
    private long time_between_rechecks = Time.EIGHT_HOURS_IN_MS;
    private long forget_threshold = Time.FIVE_DAYS_IN_MS;
    public final static Logger LOGGER = NumonicLogger.LOGGER;

    /**
     * Initialises the client.
     *
     * @param remote_hosts the remote hosts
     * @param reporting the reporting interface
     */
    public SimpleBandwidthAndLatencyClient(final IRemoteHosts remote_hosts, final IReporting reporting) {

        super();
        this.remote_hosts = remote_hosts;
        this.reporting = reporting;
        recently_measured_hosts = new HashMap<String, Data>();
    }

    /**
     * if client is in a new location, make new measurements
     */
    public void reset() {

        LOGGER.log(Level.INFO, "$$$ RESET");
        host_list = new ArrayList<String>();
        recently_measured_hosts = new HashMap<String, Data>();
    }

    /**
     * Sets the time between network measurements.
     *
     * @param milliseconds time in milliseconds between each measurement
     */
    public void setTimeBetweenSamples(final int milliseconds) {

        time_between_samples = milliseconds;
    }

    /**
     * Sets the time between rechecks on the same NIC.
     *
     * @param milliseconds time in milliseconds between rechecks
     */
    public void setTimeBetweenRechecks(final long milliseconds) {

        time_between_rechecks = milliseconds;
    }

    /**
     * Sets the time before forgetting that we have probed a host.
     *
     * @param milliseconds time in milliseconds before forgetting that we have probed a host
     */
    public void setForgetTime(final long milliseconds) {

        forget_threshold = milliseconds;
    }

    /**
     * Take the appropriate network measurements.
     */
    @Override
    protected synchronized void doMeasurement(final String hostname) {

        try {
            if (hostname.trim().compareTo(InetAddress.getLocalHost().getHostAddress()) != 0) {
                LOGGER.log(Level.INFO, "measure on hostname: " + hostname);
                final Date time = new Date();
                final BandwidthAndLatencyClient b_and_l_client = new BandwidthAndLatencyClient(InetAddress.getByName(hostname));
                final String localMachineID = b_and_l_client.getLocalMachineID();
                final String localNetworkAdapterName = b_and_l_client.getLocalDefaultNetworkAdapterName();
                final String localGateway = b_and_l_client.getLocalGateway();

                if (locationChange) {
                    reset();
                    locationChange = false;
                }

                if (!recentlyProcessed(hostname, localNetworkAdapterName)) { // only contact if not recently contacted on current NIC or on hosts that we have recently contacted
                    final Duration latency = b_and_l_client.getLatency();
                    final long bandwidth = b_and_l_client.getBandwidth();

                    final String remoteMachineID = b_and_l_client.getRemoteMachineID();
                    final String remoteNetworkAdapterName = b_and_l_client.getRemoteNetworkAdapterName();
                    final String remoteGateway = b_and_l_client.getRemoteGateway();
                    final LatencyAndBandwidthData data = new LatencyAndBandwidthData(localMachineID, localNetworkAdapterName, localGateway, remoteMachineID, remoteNetworkAdapterName, remoteGateway, latency.getLength(TimeUnit.MILLISECONDS), bandwidth, time);
                    reporting.reportLatencyAndBandwidthData(data);
                    remember(hostname, localNetworkAdapterName, time);

                }
            }
        }
        catch (final TimeoutException e) {
            try {
                LOGGER.log(Level.INFO, "Timeout while trying to ping remote host " + hostname);
                reporting.setHostPassive(hostname);
            }
            catch (final SQLException e1) {
                LOGGER.log(Level.WARNING, "set host passive exception", e1);
            }
            catch (final Exception e2) {
                LOGGER.log(Level.WARNING, "set host passive exception", e2);
            }
        }
        catch (final Exception e) {
            try {
                //                LOGGER.log(Level.INFO, "Timeout while trying to ping remote host" + this);
                reporting.setHostPassive(hostname);
            }
            catch (final SQLException e1) {
                LOGGER.log(Level.WARNING, "set host passive exception", e1);
            }
            catch (final Exception e2) {
                LOGGER.log(Level.WARNING, "set host passive exception", e2);
            }

        }
    }

    private boolean recentlyProcessed(final String hostname, final String localNetworkAdapterName) {

        final Data d = recently_measured_hosts.get(hostname);// checked in the recent past
        //        System.out.println("D: " + d);

        if (d == null) { return false; }

        if (d.nics.contains(localNetworkAdapterName)) { // checked using current nic

            final long time_since_previous_check = System.currentTimeMillis() - d.checked.getTime();
            return time_since_previous_check < time_between_rechecks;
        }

        return false;
    }

    /**
     * Remember that we have checked the host using localNetworkAdapterName.
     */
    private void remember(final String hostname, final String localNetworkAdapterName, final Date date) {

        Data d = recently_measured_hosts.get(hostname);
        // Have we seen this host before
        if (d == null) { // we haven't looked at this host recently so add the information
            d = new Data(localNetworkAdapterName, date);
            recently_measured_hosts.put(hostname, d); // add a new entry to the system
        }
        else { // we have seen the host before so update the date and add the new nic.
            d.checked = new Date();
            d.nics.add(localNetworkAdapterName);
        }
    }

    /**
     * Forget about hosts that we haven't looked at for a long time.
     */
    private void forgetOldCheckedHosts() {

        final Set<Entry<String, Data>> items = recently_measured_hosts.entrySet();
        for (final Entry<String, Data> e : items) {
            final String key = e.getKey();
            final Data d = e.getValue();
            final long how_long = new Date().getTime() - d.checked.getTime();
            if (how_long > forget_threshold) {
                recently_measured_hosts.remove(key);
            }
        }
    }

    /**
     * Makes this process wait for {@link #DEFAULT_TIME_BETWEEN_SAMPLES} between samples by default. Time between samples is changed by
     * calling {@link #setTimeBetweenSamples(int)}
     */
    @Override
    protected void continueExecution() {

        try {
            sleep(time_between_samples2);
        }
        catch (final InterruptedException e) {
            LOGGER.log(Level.INFO, "interrupted exception", e);
        }
    }

    /**
     * set true if the location changed to take new bandwidth measurements
     * @param locationChange
     */
    public void setLocationChange(final boolean locationChange) {

        LOGGER.log(Level.INFO, "$$$ SET LOCATION CHANGE TRUE");
        this.locationChange = locationChange;
    }

    /**
     * This method is called when all the hosts in the hostList have been processed. Talk to the host_list_provider and refresh the list.
     */
    @Override
    protected void refreshHosts() {

        try {
            host_list = remote_hosts.getActiveKnownHosts();
            forgetOldCheckedHosts();

        }
        catch (final Exception e) {
            LOGGER.log(Level.INFO, "Failed to refresh hostlist", e);
        }
    }

    private static class Data {

        private final Set<String> nics;
        private Date checked;

        protected Data(final String nic, final Date date) {

            checked = date;
            nics = new HashSet<String>();
            nics.add(nic);
        }
    }
}
