/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.reporting;

import java.util.Collection;
import java.util.List;

import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.LatencyAndBandwidthData;
import uk.ac.standrews.cs.numonic.data.MachineUtilisationData;
import uk.ac.standrews.cs.numonic.data.NetworkData;
import uk.ac.standrews.cs.numonic.data.ProcessData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.distribution.DistributionCollector;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.sort.data.DistributionData;
import uk.ac.standrews.cs.numonic.summary.MultipleSummary;
import uk.ac.standrews.cs.numonic.summary.SingleSummary;

/**
 * Reporting interface.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public interface ILocalResourceMonitoring {

    /**
     * Reports distribution data.
     *
     * @param machineProbability a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportDistributionData(DistributionCollector<?> machineProbability) throws Exception;

    /**
     * Reports multiple file system data summaries.
     *
     * @param fileSystemSummary a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportFileSystemData(MultipleSummary<FileSystemData> fileSystemSummary) throws Exception;

    /**
     * Reports summary machine utilisation data.
     *
     * @param summary a summary of a set of measurements
     * @return 
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportMachineUtilData(SingleSummary<MachineUtilisationData> summary) throws Exception;

    /**
     * Reports summary network data.
     *
     * @param summary a summary of a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportNetworkData(SingleSummary<NetworkData> summary) throws Exception;

    /**
     * Reports multiple network data summaries.
     *
     * @param allNetworkSummary a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportAllNetworkData(MultipleSummary<NetworkData> allNetworkSummary) throws Exception;

    /**
     * Reports summary process data.
     *
     * @param summary a summary of a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportProcessData(SingleSummary<ProcessData> summary) throws Exception;

    /**
     * Reports multiple process data summaries.
     *
     * @param summaries a set of measurements
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportAllProcessData(MultipleSummary<ProcessData> summaries) throws Exception;

    /**
     * Reports system data.
     *
     * @param data the data
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportSystemInfo(SystemInfoData data) throws Exception;

    /**
     * Reports latency and bandwidth data.
     *
     * @param data the data
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportLatencyAndBandwidthData(LatencyAndBandwidthData data) throws Exception;

    /**
     * Reports events.
     *
     * @param events the events
     * @throws Exception if the data cannot be reported
     */
    public abstract boolean reportEventData(List<Event> events) throws Exception;

    /**
     * Gets file system data for a recent time window.
     *
     * @param minutes the number of minutes of data required
     * @return file system data for the window ending now and extending back the given number of minutes
     * @throws Exception if the data cannot be accessed
     */
    public abstract List<FileSystemData> getFileSystemData(int minutes) throws Exception;

    /**
     * Gets distribution data for a recent time window.
     *
     * @param minutes the number of minutes of data required
     * @return file system data for the window ending now and extending back the given number of minutes
     * @throws Exception if the data cannot be accessed
     */
    public abstract Collection<DistributionData> getDistributionData(int minutes) throws Exception;

    /**
     * Gets events for a recent time window.
     *
     * @param minutes the number of minutes of data required
     * @return file system data for the window ending now and extending back the given number of minutes
     * @throws Exception if the data cannot be accessed
     */
    public abstract Collection<Event> getEvents(int minutes) throws Exception;

    /**
     * set machineID in DB class
     * @param machineID
     */
    public abstract void setMachineID(final String machineID);
}
