/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.sort;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;

import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.reporting.IRemoteHosts;
import uk.ac.standrews.cs.numonic.reporting.IReporting;
import uk.ac.standrews.cs.numonic.sort.data.DistributionData;
import uk.ac.standrews.cs.numonic.sort.data.MachineInfo;
import uk.ac.standrews.cs.numonic.sort.data.SystemData;
import uk.ac.standrews.cs.numonic.sort.filter.MachineFilter;
import uk.ac.standrews.cs.numonic.sort.filter.Requirements;
import uk.ac.standrews.cs.numonic.sort.ranker.MachineRanker;
import uk.ac.standrews.cs.numonic.sort.ranker.MachineSortingMetric;

/**
 * Sorts machines.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public final class MachineSorter {

    private MachineSorter() {

    }

    /**
     * Given resource monitoring data from a database this method returns a sorted set of machines (for which monitoring data is available),
     * with the top machines being the most appropriate to execute the workload defined by the sort_metric parameter. Machines that don't
     * meet a set of basic requirements (the requirements parameter) are filtered out.
     *
     * @param requirements a set of minimum requirements for the machines being compared
     * @param sort_metric the metric by which machines will be sorted
     * @param reporting interface to the database containing monitoring information
     * @param remote_hosts the hosts being monitored
     * @return a sorted set of machines
     */
    public static SortedSet<MachineInfo> filterThenRankMachines(final Requirements requirements, final MachineSortingMetric sort_metric, final IReporting reporting, final IRemoteHosts remote_hosts) {

        final Map<String, MachineInfo> machines = addStaticMachineData(remote_hosts); // (key: machine ID; value: info on that machine).
        addFileSystemData(reporting, machines, sort_metric.getMinutes()); // for size of free space
        addDistributionData(reporting, machines, sort_metric.getMinutes());

        // Filter Machines.
        final Set<MachineInfo> filteredInstances = filterMachines(requirements, machines);

        // Rank Machines.
        final MachineRanker ranker = new MachineRanker(sort_metric);
        final SortedSet<MachineInfo> sortedMachines = ranker.sortMachines(filteredInstances);

        return sortedMachines;
    }

    /**
     * Given information on the availability of various machines (through the machines parameter), returns a sorted set of
     * machines (for which monitoring data is available), with the top machines being the most appropriate to execute the workload defined
     * by the sort_metric parameter. Machines that don't meet a set of basic requirements (the requirements parameter) are filtered out.
     *
     * @param requirements a set of minimum requirements for the machines being compared
     * @param sort_metric the metric by which machines will be sorted
     * @param machines the machines to be sorted
     * @return a sorted set of machines
     */
    public static SortedSet<MachineInfo> filterThenRankMachines(final Requirements requirements, final MachineSortingMetric sort_metric, final Set<MachineInfo> machines) {

        // Filter Machines.
        final Set<MachineInfo> filteredInstances = filterMachines(requirements, machines);

        // Rank Machines.
        final MachineRanker ranker = new MachineRanker(sort_metric);
        final SortedSet<MachineInfo> sortedMachines = ranker.sortMachines(filteredInstances);

        return sortedMachines;
    }

    /**
     * Filter out machines from the map of machines given that do not meet the specified requirements.
     *
     * @param requirements requirements that each machine must meet
     * @param machines machines to be checked against requirements
     * @return machines that meet or exceed requirements
     */
    private static Set<MachineInfo> filterMachines(final Requirements requirements, final Set<MachineInfo> machines) {

        final MachineFilter filter = new MachineFilter();

        final Set<MachineInfo> filteredInstances = filter.selectSuitableMachines(machines, requirements);

        for (final MachineInfo machineInfo : filteredInstances) {
            Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Meets requirements: " + machineInfo);
        }
        return filteredInstances;
    }

    /**
     * Filter out machines from the map of machines given that do not meet the specified requirements.
     *
     * @param requirements requirements that each machine must meet
     * @param machines machines to be checked against requirements
     * @return machines that meet or exceed requirements
     */
    private static Set<MachineInfo> filterMachines(final Requirements requirements, final Map<String, MachineInfo> machines) {

        return filterMachines(requirements, new HashSet<MachineInfo>(machines.values()));
    }

    /**
     * Get usage distribution data stored in the database and add it to the machine info for the relevant machine.
     *
     * @param reporting database where file system info is stored
     * @param machines existing machine info to which file system info will be added
     * @param minutes the time period of interest regarding the utilization of file system data. E.g. '5' = last 5 minutes.
     */
    private static void addDistributionData(final IReporting reporting, final Map<String, MachineInfo> machines, final int minutes) {

        try {
            final Collection<DistributionData> data = reporting.getDistributionData(minutes);

            for (final DistributionData distribution_data : data) {
                final MachineInfo machineInfo = machines.get(distribution_data.getMachineId());

                if (machineInfo == null) {
                    Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Machine info for this machine ID (" + distribution_data.getMachineId() + ") was null.");
                }
                else {
                    machineInfo.setDistributionData(distribution_data);
                }
            }
        }
        catch (final Exception e) {
            ErrorHandling.error("Failed to get distribution data");
        }
    }

    /**
     * Get usage distribution data stored in the database and add it to the machine info for the relevant machine.
     *
     * @param reporting database where file system info is stored
     * @param machines existing machine info to which file system info will be added
     * @param minutes the time period of interest regarding the utilization of file system data. E.g. '5' = last 5 minutes.
     */
    private static void addFileSystemData(final IReporting reporting, final Map<String, MachineInfo> machines, final int minutes) {

        try {
            final List<FileSystemData> fsData = reporting.getFileSystemData(minutes);

            // Add file system data to respective MachineInfo classes.
            for (final FileSystemData fs : fsData) {
                final MachineInfo machineInfo = machines.get(fs.getMachineId());

                if (machineInfo == null) {
                    ErrorHandling.errorNoEvent("Machine info for this machine ID (" + fs.getMachineId() + ") was null.");
                }
                machineInfo.getSystemData().addFileSystemData(fs);
                machineInfo.getSystemData();
            }
        }
        catch (final Exception e) {
            ErrorHandling.error("Failed to get filesystem data");
        }
    }

    private static Map<String, MachineInfo> addStaticMachineData(final IRemoteHosts rh) {

        final Map<String, MachineInfo> machines = new HashMap<String, MachineInfo>();
        /*
         * Get data from database.
         */
        // Get static system info.
        try {
            final List<SystemInfoData> lotsOfSystemInfo = rh.getSystemInfo();

            // Add static system info to new MachineInfo instance.
            for (final SystemInfoData sysInfo : lotsOfSystemInfo) {
                final SystemData sysData = new SystemData(sysInfo);

                machines.put(sysInfo.getMachineId(), new MachineInfo(sysInfo.getMachineId(), sysData));
            }

            Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Number of machines: " + lotsOfSystemInfo.size());
            return machines;
        }
        catch (final Exception e) {
            ErrorHandling.error("Failed to get System data");
            return machines;
        }
    }
}
