/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.sort.data;

import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.sort.filter.Requirements;

/**
 * A recent overview of the resources on a machine. Most of this information such as CPU and RAM capacity is static, but some information
 * such as free space in the file system is based on recent measurements.
 *
 * This information is used when filtering machines based on the requirements of a request.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class SystemData {

    /**
     * Contains the CPU and memory capacity of this machine.
     */
    private final SystemInfoData system_info;

    /**
     * Contains information on the total amount of free disk space.
     */
    private FileSystemData file_system_data;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a record.
     * @param system_info the system information
     */
    public SystemData(final SystemInfoData system_info) {

        this.system_info = system_info;
    }

    /**
     * Adds a file system data record.
     * @param file_system_data the record
     */
    public void addFileSystemData(final FileSystemData file_system_data) {

        // TODO handle multiple drives on a single machine.
        if (this.file_system_data != null) {
            System.err.println("Multiple drives.");
        }
        this.file_system_data = file_system_data;
    }

    /**
     * Tests whether the data in the record meets specified requirements.
     * @param requirements the requirements
     * @return true if the requirements are met
     */
    public boolean meetsRequirements(final Requirements requirements) {

        return (system_info == null || requirements.getCpuCapacity() > system_info.getCpuClockSpeed()) && (system_info == null || requirements.getMemoryCapacity() > system_info.getMemoryTotal()) && (file_system_data == null || requirements.getDiskCapacity() > file_system_data.getSpaceFree());
    }

    /**
     * Returns the system information.
     * @return the system information
     */
    public SystemInfoData getSystemInfo() {

        return system_info;
    }

    /**
     * Returns the file system information.
     * @return the file system information
     */
    public FileSystemData getFileSystemData() {

        return file_system_data;
    }

    @Override
    public String toString() {

        return "SystemData [machineInfo=" + system_info + ", fileSystemData=" + file_system_data + "]";
    }
}
