/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.sort.filter;

import java.util.Set;

import uk.ac.standrews.cs.numonic.sort.data.MachineInfo;

/**
 * Filters out suitable machines according to requirements, keeping a cache of previously results in case they can be reused.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class CachedMachineFilter implements IMachineFilter {

    private final MachineFilter machine_filter;

    private Requirements requirements = null;
    private Set<MachineInfo> previously_selected_machines = null;

    private int previous_set_hash_code = 0;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a filter with given requirements.
     *
     * @param requirements the requirements
     */
    public CachedMachineFilter(final Requirements requirements) {

        machine_filter = new MachineFilter();
        this.requirements = requirements;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Set<MachineInfo> selectSuitableMachines(final Set<MachineInfo> machines, final Requirements requirements) {

        if (machines.hashCode() == previous_set_hash_code && requirements.equals(this.requirements)) { return previously_selected_machines; }

        this.requirements = requirements;
        return selectSuitableMachines(machines);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Selects machines from a set according to previously specified requirements.
     *
     * @param machines a set of machines
     * @return the machines meeting the previously specified requirements
     */
    public Set<MachineInfo> selectSuitableMachines(final Set<MachineInfo> machines) {

        if (machines.hashCode() == previous_set_hash_code) { return previously_selected_machines; }

        final Set<MachineInfo> selected_machines = machine_filter.selectSuitableMachines(machines, requirements);

        previously_selected_machines = selected_machines;
        previous_set_hash_code = machines.hashCode();

        return selected_machines;
    }

    /**
     * Sets the requirements.
     * @param requirements the requirements
     */
    public void setRequirements(final Requirements requirements) {

        this.requirements = requirements;
    }
}
