/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.sort.filter;

/**
 * Used to filter out machines that don't meet the requirements of a particular request.
 *
 * <p>
 * Instances of this class specify the minimum amount of resources needed for a request.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class Requirements {

    private static final int BITS_SHIFTED = 32;

    /**
     * The amount of CPU capacity required, in MHz.
     */
    private final long cpu_capacity;

    /**
     * The quantity of memory required, in KB.
     */
    private final long memory_capacity;

    /**
     * The quantity of disk space required, in KB.
     */
    private final long disk_capacity;

    /**
     * The bandwidth of network connection required, in KB/s.
     */
    private final long network_capacity;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a set of requirements.
     *
     * @param cpu_capacity the amount of CPU capacity required, in MHz
     * @param memory_capacity the quantity of memory required, in KB
     * @param disk_capacity the quantity of disk space required, in KB
     * @param network_capacity the bandwidth of network connection required, in KB/s
     */
    public Requirements(final long cpu_capacity, final long memory_capacity, final long disk_capacity, final long network_capacity) {

        this.cpu_capacity = cpu_capacity;
        this.memory_capacity = memory_capacity;
        this.disk_capacity = disk_capacity;
        this.network_capacity = network_capacity;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the amount of CPU capacity required.
     * @return the amount of CPU capacity required, in MHz
     */
    public long getCpuCapacity() {

        return cpu_capacity;
    }

    /**
     * Returns the quantity of memory required.
     * @return the quantity of memory required, in KB
     */
    public long getMemoryCapacity() {

        return memory_capacity;
    }

    /**
     * Returns the quantity of disk space required.
     * @return the quantity of disk space required, in KB
     */
    public long getDiskCapacity() {

        return disk_capacity;
    }

    /**
     * Returns the bandwidth of network connection required.
     * @return the bandwidth of network connection required, in KB/s
     */
    public long getNetworkCapacity() {

        return network_capacity;
    }

    @Override
    public int hashCode() {

        final int prime = 31;
        int result = 1;
        result = prime * result + (int) (cpu_capacity ^ cpu_capacity >>> BITS_SHIFTED);
        result = prime * result + (int) (disk_capacity ^ disk_capacity >>> BITS_SHIFTED);
        result = prime * result + (int) (memory_capacity ^ memory_capacity >>> BITS_SHIFTED);
        result = prime * result + (int) (network_capacity ^ network_capacity >>> BITS_SHIFTED);
        return result;
    }

    @Override
    public boolean equals(final Object obj) {

        if (this == obj) { return true; }
        if (obj == null) { return false; }
        if (!(obj instanceof Requirements)) { return false; }

        final Requirements other = (Requirements) obj;

        return cpu_capacity == other.cpu_capacity && disk_capacity == other.disk_capacity && memory_capacity == other.memory_capacity && network_capacity == other.network_capacity;
    }
}
