/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.sort.ranker;

import java.util.Comparator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.sort.data.DistributionData;
import uk.ac.standrews.cs.numonic.sort.data.MachineInfo;
import uk.ac.standrews.cs.numonic.sort.data.SystemData;

/**
 * Responsible for ranking machines by a given sort metric.
 *
 * When this class is instantiated the sort metric must be specified. Then when the {@link #sortMachines(Set)} method is called the class
 * first iterates through each machine and stores the maximum CPU value and memory value for each - this is used to normalise the power of
 * each resource.
 *
 * <p>
 * The machines are then added to a sorted set, where the sort order is specified by the value assigned to each machine through
 * {@link MachineInfo#getValue(MachineSortingMetric, Bounds)}, which makes a call to
 * {@link #getMachineValue(MachineSortingMetric, SystemData, DistributionData, Bounds)} to obtain its value.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class MachineRanker implements Comparator<MachineInfo> {

    private static final double SWAP_WEIGHT = 0.2;
    private final MachineSortingMetric sorting_metric;
    private final Bounds maxima;

    /**
     * Initialises the ranker.
     *
     * @param sorting_metric the metric to be used
     */
    public MachineRanker(final MachineSortingMetric sorting_metric) {

        this.sorting_metric = sorting_metric;
        maxima = new Bounds(0, 0);
    }

    /**
     * Ranks machines based on a set of requirements (specifying which resources are most important to a request) and the availability
     * of these resources on individual machines.
     *
     * @param machines a set of machines
     * @return the ranked set
     */
    public SortedSet<MachineInfo> sortMachines(final Set<MachineInfo> machines) {

        for (final MachineInfo machine : machines) {
            maxima.setCpuMax(max(getCpuWeight(machine), maxima.getCpuMax()));
            maxima.setMemMax(max(getMemWeight(machine), maxima.getMemMax()));
        }

        final SortedSet<MachineInfo> sorted_machines = new TreeSet<MachineInfo>(this);
        sorted_machines.addAll(machines);

        return sorted_machines;
    }

    @Override
    public int compare(final MachineInfo m1, final MachineInfo m2) {

        final Double m1_value = m1.getValue(sorting_metric, maxima);
        final Double m2_value = m2.getValue(sorting_metric, maxima);

        return m1_value.compareTo(m2_value);
    }

    /**
     * Calculates a machine value.
     *
     * @param metric a sorting metric
     * @param system_data static system data used to weight more powerful machines favourably.
     * @param data the data distribution
     * @param maxima the values for the maximum cpu capacity and memory capacity among the machines being compared, used to normalize values.
     * @return the machine value
     */
    public static double getMachineValue(final MachineSortingMetric metric, final SystemData system_data, final DistributionData data, final Bounds maxima) {

        double totalValue = 1;

        final double cpu_weight = getCpuWeight(system_data.getSystemInfo().getCpuClockSpeed(), system_data.getSystemInfo().getNumberOfCores(), system_data.getSystemInfo().getNumberOfCpus());
        final double mem_weight = getMemWeight(system_data.getSystemInfo().getMemoryTotal(), system_data.getSystemInfo().getSwapTotal());

        final double normalized_cpu_weight = normalize(cpu_weight, maxima.getCpuMax(), 0);
        final double normalized_mem_weight = normalize(mem_weight, maxima.getMemMax(), 0);

        // Value of resource = normalised power of resource * probability resource is available * importance of this resource to request.
        totalValue *= normalized_cpu_weight * data.getCpuUtilization().getValue(metric.getCpuUtilizationInterval()) * metric.getCpuUtilization();
        totalValue *= normalized_mem_weight * data.getMemoryUtilization().getValue(metric.getMemoryUtilizationInterval()) * metric.getMemoryUtilization();

        totalValue *= data.getDiskUtilizationRead().getValue(metric.getDiskUtilizationReadInterval()) * metric.getDiskUtilizationRead();
        totalValue *= data.getDiskUtilizationWrite().getValue(metric.getDiskUtilizationWriteInterval()) * metric.getDiskUtilizationWrite();

        totalValue *= data.getNetworkUtilizationRX().getValue(metric.getNetworkUtilizationReadInterval()) * metric.getNetworkUtilizationRead();
        totalValue *= data.getNetworkUtilizationTX().getValue(metric.getNetworkUtilizationWriteInterval()) * metric.getNetworkUtilizationWrite();

        return totalValue;
    }

    /**
     * Returns a value representing the worth of this CPU based on its processing power, the number of cores, and the number of CPUs.
     * This value should be weighted so that small-MEDIUM differences don't matter, but large differences do.
     *
     * TODO the extent to which it is weighted may form part of the request, or be based on the absolute requirement of CPU power.
     *
     * @return value used to weight distribution information when comparing machines.
     */
    private static double getCpuWeight(final int cpu_mhz, final int num_cores, final int num_cpus) {

        return cpu_mhz * max(num_cpus, 1) * max(num_cores, 1);
    }

    private static double getMemWeight(final long memory_total, final long swap_total) {

        final double weight = memory_total + SWAP_WEIGHT * swap_total;

        return weight;
    }

    private static double getCpuWeight(final MachineInfo machine) {

        final SystemInfoData machineInfo = machine.getSystemData().getSystemInfo();
        return getCpuWeight(machineInfo.getCpuClockSpeed(), machineInfo.getNumberOfCores(), machineInfo.getNumberOfCpus());
    }

    private static double getMemWeight(final MachineInfo machine) {

        final SystemInfoData machineInfo = machine.getSystemData().getSystemInfo();
        return getMemWeight(machineInfo.getMemoryTotal(), machineInfo.getSwapTotal());
    }

    private static double normalize(final double value, final double max, final double min) {

        return (value - min) / (max - min);
    }

    private static int max(final int a, final int b) {

        return a > b ? a : b;
    }

    private static double max(final double a, final double b) {

        return a > b ? a : b;
    }
}
