/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.sort.ranker;

import uk.ac.standrews.cs.numonic.distribution.Distribution.Interval;
import uk.ac.standrews.cs.numonic.sort.data.MachineInfo;

/**
 * Metric which defines how {@link MachineInfo} instances should be sorted.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class MachineSortingMetric {

    /**
     * The number of minutes since the present that ranking results should be used on.
     */
    private final int minutes;

    /**
     * Values between 0-1, indicating the importance of the given measurement to this metric. The higher the value the more important it is.
     */
    private final double cpu_utilization, memory_utilization, swap_utilization, disk_utilization_read, disk_utilization_write, network_utilization_read, network_utilization_write;

    /**
     * The level of utilization that the metric is interested in for a given resource. For example, if the interval is MEDIUM_LOW, this
     * indicates that the metric cares about the probability that utilization is MEDIUM-LOW or less.
     */
    private final Interval cpu_utilization_interval, memory_utilization_interval, swap_utilization_interval, disk_utilization_write_interval, disk_utilization_read_interval, network_utilization_write_interval, network_utilization_read_interval;

    /**
     * Initializes the metric.
     *
     * @param minutes the number of minutes of distribution info that should be used in evaluating the metric
     * @param cpu_utilization the importance of CPU utilization to the requesting application, between 0-1
     * @param memory_utilization the importance of memory utilization to the requesting application, between 0-1
     * @param swap_utilization the importance of swap space utilization to the requesting application, between 0-1
     * @param disk_utilization_write the importance of disk utilization writes to the requesting application, between 0-1
     * @param disk_utilization_read the importance of disk utilization reads to the requesting application, between 0-1
     * @param network_utilization_write the importance of network utilization writes to the requesting application, between 0-1
     * @param network_utilization_read the importance of network utilization reads to the requesting application, between 0-1
     * @param cpu_utilization_interval the level of use that the metric is interested in
     * @param memory_utilization_interval the level of use that the metric is interested in
     * @param swap_utilization_interval the level of use that the metric is interested in
     * @param disk_utilization_write_interval the level of use that the metric is interested in
     * @param disk_utilization_read_interval the level of use that the metric is interested in
     * @param network_utilization_write_interval the level of use that the metric is interested in
     * @param network_utilization_read_interval the level of use that the metric is interested in
     */
    public MachineSortingMetric(final int minutes, final double cpu_utilization, final double memory_utilization, final double swap_utilization, final double disk_utilization_write, final double disk_utilization_read, final double network_utilization_write, final double network_utilization_read,
                    final Interval cpu_utilization_interval, final Interval memory_utilization_interval, final Interval swap_utilization_interval, final Interval disk_utilization_write_interval, final Interval disk_utilization_read_interval, final Interval network_utilization_write_interval,
                    final Interval network_utilization_read_interval) {

        this.minutes = minutes;

        this.cpu_utilization = cpu_utilization;
        this.memory_utilization = memory_utilization;
        this.swap_utilization = swap_utilization;
        this.disk_utilization_write = disk_utilization_write;
        this.disk_utilization_read = disk_utilization_read;
        this.network_utilization_write = network_utilization_write;
        this.network_utilization_read = network_utilization_read;
        this.cpu_utilization_interval = cpu_utilization_interval;
        this.memory_utilization_interval = memory_utilization_interval;
        this.swap_utilization_interval = swap_utilization_interval;
        this.disk_utilization_write_interval = disk_utilization_write_interval;
        this.disk_utilization_read_interval = disk_utilization_read_interval;
        this.network_utilization_read_interval = network_utilization_read_interval;
        this.network_utilization_write_interval = network_utilization_write_interval;
    }

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();
        builder.append("MachineSortingMetric [minutes=");
        builder.append(minutes);
        builder.append(", cpuUtilization=");
        builder.append(cpu_utilization);
        builder.append(", memoryUtilization=");
        builder.append(memory_utilization);
        builder.append(", swapUtilization=");
        builder.append(swap_utilization);
        builder.append(", diskUtilizationRead=");
        builder.append(disk_utilization_read);
        builder.append(", diskUtilizationWrite=");
        builder.append(disk_utilization_write);
        builder.append(", networkUtilizationRead=");
        builder.append(network_utilization_read);
        builder.append(", networkUtilizationWrite=");
        builder.append(network_utilization_write);
        builder.append(", cpuUtilizationInterval=");
        builder.append(cpu_utilization_interval);
        builder.append(", memoryUtilizationInterval=");
        builder.append(memory_utilization_interval);
        builder.append(", swapUtilizationInterval=");
        builder.append(swap_utilization_interval);
        builder.append(", diskUtilizationWriteInterval=");
        builder.append(disk_utilization_write_interval);
        builder.append(", diskUtilizationReadInterval=");
        builder.append(disk_utilization_read_interval);
        builder.append(", networkUtilizationTxInterval=");
        builder.append(network_utilization_write_interval);
        builder.append(", networkUtilizationRxInterval=");
        builder.append(network_utilization_read_interval);
        builder.append("]");
        return builder.toString();
    }

    /**
     * Returns the number of minutes of distribution info that should be used in evaluating the metric.
     * @return the number of minutes of distribution info that should be used in evaluating the metric
     */
    public int getMinutes() {

        return minutes;
    }

    /**
     * Returns the importance of CPU utilization to the requesting application, between 0-1.
     * @return the importance of CPU utilization to the requesting application, between 0-1
     */
    public double getCpuUtilization() {

        return cpu_utilization;
    }

    /**
     * Returns the importance of memory utilization to the requesting application, between 0-1.
     * @return the importance of memory utilization to the requesting application, between 0-1
     */
    public double getMemoryUtilization() {

        return memory_utilization;
    }

    /**
     * Returns the importance of swap space utilization to the requesting application, between 0-1.
     * @return the importance of swap space utilization to the requesting application, between 0-1
     */
    public double getSwapUtilization() {

        return swap_utilization;
    }

    /**
     * Returns the importance of disk utilization reads to the requesting application, between 0-1.
     * @return the importance of disk utilization reads to the requesting application, between 0-1
     */
    public double getDiskUtilizationRead() {

        return disk_utilization_read;
    }

    /**
     * Returns the importance of disk utilization writes to the requesting application, between 0-1.
     * @return the importance of disk utilization writes to the requesting application, between 0-1
     */
    public double getDiskUtilizationWrite() {

        return disk_utilization_write;
    }

    /**
     * Returns the importance of network utilization reads to the requesting application, between 0-1.
     * @return the importance of network utilization reads to the requesting application, between 0-1
     */
    public double getNetworkUtilizationRead() {

        return network_utilization_read;
    }

    /**
     * Returns the importance of network utilization writes to the requesting application, between 0-1.
     * @return the importance of network utilization writes to the requesting application, between 0-1
     */
    public double getNetworkUtilizationWrite() {

        return network_utilization_write;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getCpuUtilizationInterval() {

        return cpu_utilization_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getMemoryUtilizationInterval() {

        return memory_utilization_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getSwapUtilizationInterval() {

        return swap_utilization_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getDiskUtilizationWriteInterval() {

        return disk_utilization_write_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getDiskUtilizationReadInterval() {

        return disk_utilization_read_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getNetworkUtilizationWriteInterval() {

        return network_utilization_write_interval;
    }

    /**
     * Returns the level of use that the metric is interested in.
     * @return the level of use that the metric is interested in
     */
    public Interval getNetworkUtilizationReadInterval() {

        return network_utilization_read_interval;
    }
}
