/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.summary;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.IData;
import uk.ac.standrews.cs.numonic.main.Numonic;
import uk.ac.standrews.cs.numonic.monitor.FileSystemMonitor;

/**
 * Used to store a set of summaries for {@link Numonic} subclasses that return lists of data, rather than single data-sets. For example, the
 * {@link FileSystemMonitor} class returns multiple {@link FileSystemData} instances - one for each local file system. This class allows
 * this set of measurements to be added together.
 *
 * @param <T> the type of data being stored
 *
 * @author angus
 */
public class MultipleSummary<T extends IData<T>> extends Summary {

    private final Map<T, SingleSummary<T>> summaries = new HashMap<T, SingleSummary<T>>();
    private final Map<Integer, T> ident = new HashMap<Integer, T>();

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a summary.
     *
     * @param measurements_before_summary the number of measurements to be accumulated before summarization
     * @param machine_id the machine ID
     */
    public MultipleSummary(final int measurements_before_summary, final String machine_id) {

        super(measurements_before_summary, machine_id);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Adds a measurement.
     * @param items the measurement
     */
    public void addMeasurement(final List<T> items) {

        if (items != null && items.size() > 0) {

            incrementMeasurementCount();
            int id = 0;
            for (final T item : items) {
                SingleSummary<T> summary = summaries.get(ident.get(id));
                if (summary == null) {
                    summary = new SingleSummary<T>(item, getMachineID(), getNumberOfMeasurementsNeeded());
                }

                summary.addNewMeasurement(item);

                if (measurementsTaken() == 1) {
                    summaries.put(item, summary);
                    ident.put(id, item);

                }
                id++;
            }

        }
    }

    /**
     * Returns the current summaries for this type of measurement.
     * @return the current summaries for this type of measurement
     */
    public Collection<SingleSummary<T>> getSummaries() {

        return summaries.values();
    }

    /**
     * Resets all summary instances, removing all data. This should be called only after the previous summaries have been processed, or persisted to disk.
     */
    @Override
    public void reset() {

        summaries.clear();
        super.reset();
    }

    @Override
    public String toString() {

        return "MultipleSummary [summaries=" + summaries + "]";
    }

    public T getContentType() {

        final Set<T> values = summaries.keySet();
        for (final T key : values) {
            return key;
        }
        return null;
    }
}
