/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.summary;

import java.util.Date;
import java.util.logging.Level;

import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.numonic.data.IData;
import uk.ac.standrews.cs.numonic.main.NumonicLogger;

/**
 * Record a set of measurements for a specific type of monitor (a monitor that is a subclass of uk.ac.standrews.cs.numonic.monitor.MonitorBase).
 *
 * <p>
 * This class takes in new measurements (subclasses of the {@link IData} interface) and computes new minimums, maximums, and totals.
 *
 * <p>
 * The class can return the current totals, minimums, maximums, and averages. A number of measurements to be taken can be specified and used
 * by periodically checking the {@link SingleSummary#isFull()} method to see whether that number has been reached.
 *
 * Records a summary of monitoring data by comparing existing measurements to those that have already been taken and calculating a new
 * total, minimum, and maximum for each value.
 *
 * @param <T>
 * The monitoring data being measured. Any class that implements the interface can be used.
 * @author angus
 */
public class SingleSummary<T extends IData<T>> extends Summary {

    private T total, min, max, base;

    /**
     * When the first measurement was taken.
     */
    private Date start_time;

    boolean startDateSet = false;

    /**
     * When the last measurement was taken.
     */
    private Date end_time;

    /**
     * Create a new class to summarise a specific type of monitoring data.
     *
     * @param t an instance of the value object for this type of monitoring data.
     * @param machine_id the machine ID
     * @param measurements_before_summary the number of measurements to be accumulated before summarization
     */
    public SingleSummary(final T t, final String machine_id, final int measurements_before_summary) {

        super(measurements_before_summary, machine_id);

        total = t.getNewInstance();
        min = t.getNewInstance();
        max = t.getNewInstance();

    }

    /**
     * Adds a new measurement.
     * @param unadjusted_new_item the item
     */
    public void addNewMeasurement(final T unadjusted_new_item) {

        T item = unadjusted_new_item; // this will be adjusted for base, if needed.

        if (unadjusted_new_item == null) {
            ErrorHandling.errorNoEvent("A null value was passed to the summary instance.");
            return;
        }

        incrementMeasurementCount();

        if (measurementsTaken() == 1) { // record the first item as the base value.
            base = item;

            return;
        }
        else if (item.needsAdjustedForBase()) {

            item = item.adjustForBase(this.total.sum(base)); // get the sum of all previous entries (incl. base) and subtract it from newest
            // measurement (for incrementing measurements).
        }

        total = this.total.sum(item);

        if (measurementsTaken() == 2) {
            start_time = item.getDate();
            startDateSet = true;
            min = item; // first measurements always minimum.
        }
        else {
            min = min.min(item);
        }

        max = max.max(item);

        end_time = item.getDate();
    }

    /**
     * Returns the number of measurements that have been taken so far in this summary.
     * @return the number of measurements that have been taken so far in this summary
     */
    public int getNumberOfMeasurements() {

        return measurementsTaken() - 1; // Remove base measurement, which isn't included in the average.
    }

    /**
     * Returns the time when the first measurement was taken.
     * @return the time when the first measurement was taken
     */
    public Date getStartTime() {

        if (!startDateSet) {
            NumonicLogger.LOGGER.log(Level.SEVERE, "Start date not set.");
        }

        return (Date) start_time.clone();
    }

    /**
     * Returns the time when the last measurement was taken.
     * @return the time when the last measurement was taken
     */
    public Date getEndTime() {

        return (Date) end_time.clone();
    }

    /**
     * Clears the summary class of all measurements. This should be called only after the previous summaries have been processed, or
     * persisted to disk.
     */
    @Override
    public void reset() {

        total = total.getNewInstance();
        min = min.getNewInstance();
        max = max.getNewInstance();

        super.reset();
    }

    /**
     * Returns the total.
     * @return the total
     */
    public T getTotal() {

        return total;
    }

    /**
     * Returns the average.
     * @return the average
     */
    public T getAverage() {

        if (measurementsTaken() == 0) {
            // The process has been killed off, the network interface is no longer there, etc.
            return total;
        }
        return total.convertToAverage(measurementsTaken());
    }

    /**
     * Returns the minimum value.
     * @return the minimum value
     */
    public T getMin() {

        return min;
    }

    /**
     * Returns the maximum value.
     * @return the maximum value
     */
    public T getMax() {

        return max;
    }

    /**
     * Get a specific measurement (the miniumum, maximum, average, or total observed values)
     * @param measurement MIN, MAX, AVG, or TOTAL.
     * @return
     */
    public T get(final MeasurementType measurement) {

        switch (measurement) {
            case AVG:
                return getAverage();
            case MIN:
                return getMin();
            case MAX:
                return getMax();
            case TOTAL:
                return getTotal();
            default:
                return null;
        }

    }

    @Override
    public String toString() {

        return "SingleSummary [\n\ttotal=" + total + ", \n\tmin=" + min + ", \n\tmax=" + max + ", \n\tavg=" + getAverage() + ", \n\tstart_time=" + start_time + ", \n\tend_time=" + end_time + "\n]";
    }

    public T getContentType() {

        return max;
    }

}
