/***************************************************************************
 *                                                                         *
 * H2O                                                                     *
 * Copyright (C) 2010-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic/                                   *
 *                                                                         *
 * This file is part of H2O, a distributed database based on the open      *
 * source database H2 (www.h2database.com).                                *
 *                                                                         *
 * H2O is free software: you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * H2O is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with H2O.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.util;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;
import java.util.Set;

import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Wrapper for java properties class, which provides constructor allowing the creation of a properties file from a JAR resource.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class JarPropertiesWrapper implements IPropertiesWrapper {

    private final Properties properties;

    private String comment = "Properties File";
    private final InputStream inputStream;

    private static final String TRUE = "true";

    public JarPropertiesWrapper(final InputStream stream) {

        inputStream = stream;
        properties = new Properties();
    }

    /**
     * 
     * @param propertiesFileLocation
     */
    public JarPropertiesWrapper(final String propertiesFileLocation) {

        //Get the properties file from inside the H2O jar file (note: this will still work in eclipse, etc.
        final ClassLoader classLoader = this.getClass().getClassLoader();
        inputStream = classLoader.getResourceAsStream(propertiesFileLocation);
        properties = new Properties();
    }

    /**
     * Gets a properties wrapper for a given file.
     *
     * @param fileLocation the location of the file containing properties
     * @return the wrapper
     */
    public static synchronized IPropertiesWrapper getWrapper(final InputStream stream) {

        return new JarPropertiesWrapper(stream);

    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#loadProperties()
     */
    @Override
    public synchronized void loadProperties() throws IOException {

        try {

            properties.load(inputStream);
        }
        catch (final FileNotFoundException e) {
            ErrorHandling.exceptionError(e, "Unexpected file not found exception");
        }
        finally {
            if (inputStream != null) {
                inputStream.close();
            }
        }

    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#createNewFile()
     */
    @Override
    public synchronized void createNewFile() throws IOException {

        //Does nothing. Not possible inside JAR.
        throw new IOException("Not implemented in JarPropertiesWrapper because the file can't be created inside the JAR.");
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#getProperty(java.lang.String)
     */
    @Override
    public synchronized String getProperty(final String key) {

        return properties.getProperty(key);
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#isEnabled(java.lang.String)
     */
    @Override
    public synchronized boolean isEnabled(final String key) {

        return TRUE.equals(getProperty(key));
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#setProperty(java.lang.String, java.lang.String)
     */
    @Override
    public synchronized void setProperty(final String key, final String value) {

        properties.setProperty(key, value);
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#setPropertyComment(java.lang.String)
     */
    @Override
    public synchronized void setPropertyComment(final String propertyComment) {

        comment = propertyComment;
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#saveAndClose()
     */
    @Override
    public synchronized void saveAndClose() throws IOException {

        //Does nothing. Not possible inside JAR.
        throw new IOException("Not implemented in JarPropertiesWrapper because the file can't be created inside the JAR.");
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#getKeys()
     */
    @Override
    public synchronized Set<Object> getKeys() {

        return properties.keySet();
    }
}
