/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;

import uk.ac.standrews.cs.nds.madface.exceptions.UnequalArrayLengthsException;
import uk.ac.standrews.cs.nds.madface.exceptions.UnknownPlatformException;
import uk.ac.standrews.cs.nds.madface.interfaces.IStreamProcessor;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

import com.mindbright.ssh2.SSH2Exception;

/**
 * Describes a local or remote host and, optionally, SSH connection credentials and/or remote references to an application that is running or could run on it.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class HostDescriptor implements Cloneable, Comparable<HostDescriptor> {

    private static final String PROBE_COMMAND = "uname";

    private static final String PROBE_REPLY_LINUX = "Linux";
    private static final String PROBE_REPLY_MAC = "Darwin";
    private static final String LOCAL_HOST = "localhost";

    private static String local_host = null;
    private static InetAddress local_inet_address = null;
    private static final AtomicInteger next_id = new AtomicInteger();

    static {
        try {
            local_inet_address = InetAddress.getLocalHost();
            local_host = local_inet_address.getCanonicalHostName();
        }
        catch (final UnknownHostException e) {
            // No local IP address could be found.
        }
    }

    // -------------------------------------------------------------------------------------------------------

    // The host name.
    private volatile String host = null;

    // The host address.
    private volatile InetAddress inet_address = null;

    // The application port.
    private volatile int port = 0;

    // The host's platform.
    private volatile PlatformDescriptor platform_descriptor = null;

    // The classpath to be used by the application running on the host.
    private volatile ClassPath class_path = null;

    // The URLs of the library elements to be used by the application running on the host.
    private volatile URL[] application_urls = null;

    // The path of the Java executable on the host.
    private volatile File java_bin_path = null;

    // The Java library path on the host.
    private volatile File java_library_path = null;

    // Parameters for application deployment on the host.
    private volatile Object[] application_deployment_params;

    // The current view of the application state on the host.
    private volatile HostState host_state = HostState.UNKNOWN;

    // A reference to the application running on the host.
    private volatile Object application_reference = null;

    // A reference to the application process(es) running on the host.
    private final Set<Process> processes = Collections.synchronizedSet(new HashSet<Process>());

    // Map storing results of various scan operations.
    private volatile Map<String, String> scan_results = null;

    // Authentication credentials for the host.
    private volatile Credentials credentials = null;

    // Manager for executing processes on the host.
    private volatile ProcessManager process_manager = null;

    private volatile boolean deploy_in_local_process = false;

    private final int id;

    public boolean debug() {

        return host.equals("compute-0-13");
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a descriptor for the local host.
     */
    public HostDescriptor() {

        id = next_id.getAndIncrement();

        // Local host.
        initLocal();
    }

    /**
     * Initialises a descriptor with a given address.
     *
     * @param host the remote address
     */
    public HostDescriptor(final String host) {

        id = next_id.getAndIncrement();
        init(host);
    }

    /**
     * Initialises a descriptor using a host name and user credentials read from the user interactively.
     * The password or passphrase is input using a Swing dialog to prevent it being displayed on the screen.
     *
     * @param use_password true if password authentication should be used, false if public key authentication should be used
     * @throws IOException if an error occurs while reading user input
     */
    public HostDescriptor(final boolean use_password) throws IOException {

        id = next_id.getAndIncrement();
        final BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));

        System.out.print("enter machine name: ");
        final String host = reader.readLine();

        final Credentials credentials = new Credentials(use_password);

        init(host, credentials);
    }

    public HostDescriptor(final String host, final Credentials credentials) {

        id = next_id.getAndIncrement();
        init(host, credentials);
    }

    // -------------------------------------------------------------------------------------------------------

    /*
     * The setter methods are written in the style returning 'this', to enable a set of attributes to be set succinctly.
     *
     * For example: new HostDescriptor().credentials(cred).javaBinPath(bin_path).javaLibraryPath(lib_path)
     */

    /**
     * Sets a remote reference to the application running on the remote host.
     *
     * @param application_reference a remote reference to the application running on the remote host
     * @return this object
     */
    public HostDescriptor applicationReference(final Object application_reference) {

        this.application_reference = application_reference;
        return this;
    }

    /**
     * Sets the library elements used by the application running on the host.
     *
     * @param application_urls the library elements used by the application running on the host
     * @return this object
     */
    public HostDescriptor applicationURLs(final URL[] application_urls) {

        this.application_urls = application_urls.clone();
        return this;
    }

    /**
     * Sets the classpath used by the application running on the host.
     *
     * @param class_path the classpath used by the application running on the host
     * @return this object
     */
    public HostDescriptor classPath(final ClassPath class_path) {

        this.class_path = class_path;
        return this;
    }

    /**
     * Sets the authentication credentials for the host.
     *
     * @param credentials the authentication credentials for the host
     * @return this object
     */
    public HostDescriptor credentials(final Credentials credentials) {

        this.credentials = credentials;
        return this;
    }

    /**
     * Sets the state of the remote host and the application.
     *
     * @param host_state the state of the remote host and the application
     * @return this object
     */
    public HostDescriptor hostState(final HostState host_state) {

        //        if (this.host_state != host_state) {
        //            System.out.println("\nstate of " + getHost() + " changing from " + this.host_state + " to " + host_state);
        //        }
        this.host_state = host_state;
        return this;
    }

    /**
     * Sets the path of the Java executable on the host.
     *
     * @param java_bin_path the path of the Java executable on the host
     * @return this object
     */
    public HostDescriptor javaBinPath(final File java_bin_path) {

        this.java_bin_path = java_bin_path;
        return this;
    }

    /**
     * Sets the Java library path on the host.
     *
     * @param java_library_path the Java library path on the host
     * @return this object
     */
    public HostDescriptor javaLibraryPath(final File java_library_path) {

        this.java_library_path = java_library_path;
        return this;
    }

    /**
     * Sets the application deployment parameters on the host.
     *
     * @param application_deployment_params the application deployment parameters on the host
     * @return this object
     */
    public HostDescriptor applicationDeploymentParams(final Object[] application_deployment_params) {

        this.application_deployment_params = application_deployment_params.clone();
        return this;
    }

    /**
     * Sets the application port.
     *
     * @param port the application port
     * @return this object
     */
    public HostDescriptor port(final int port) {

        this.port = port;
        return this;
    }

    /**
     * Sets a handle on the application process running on the remote host.
     *
     * @param process a handle on the application process running on the remote host
     * @return this object
     */
    public synchronized HostDescriptor process(final Process process) {

        processes.add(process);
        return this;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns true if this descriptor represents the local host.
     * @return true if this descriptor represents the local host
     */
    public boolean local() {

        return local(host);
    }

    /**
     * Returns a remote reference to the application running on the remote host.
     * @return a remote reference to the application running on the remote host, or null if not set
     */
    public Object getApplicationReference() {

        return application_reference;
    }

    /**
     * Returns the URLs of the library elements used by the application running on the host.
     * @return the URLs of the library elements used by the application running on the host
     */
    public URL[] getApplicationURLs() {

        return application_urls != null ? application_urls.clone() : null;
    }

    /**
     * Returns the class path.
     * @return the class path
     */
    public synchronized ClassPath getClassPath() {

        if (class_path == null) {
            class_path = ClassPath.getCurrentClassPath();
        }

        return class_path;
    }

    /**
     * Returns the authentication credentials.
     * @return the authentication credentials
     */
    public Credentials getCredentials() {

        return credentials;
    }

    /**
     * Returns the network address.
     * @return the network address
     */
    public String getHost() {

        return host;
    }

    /**
     * Returns the state of the remote host and the application. Of course this is only an educated guess.
     * @return the state of the remote host and the application
     */
    public HostState getHostState() {

        return host_state;
    }

    /**
     * Returns the network address.
     * @return the network address
     */
    public InetAddress getInetAddress() {

        return inet_address;
    }

    /**
     * Returns the network address.
     * @return the network address
     * @throws UnknownHostException if the address cannot be resolved
     */
    public InetSocketAddress getInetSocketAddress() throws UnknownHostException {

        return NetworkUtil.getInetSocketAddress(host, port);
    }

    /**
     * Returns the path of the Java executable.
     * @return the path of the Java executable
     */
    public File getJavaBinPath() {

        return java_bin_path;
    }

    /**
     * Returns the Java library path.
     * @return the Java library path
     */
    public File getJavaLibraryPath() {

        return java_library_path;
    }

    /**
     * Returns the application deployment parameters on the host.
     * @return the application deployment parameters on the host
     */
    public Object[] getApplicationDeploymentParams() {

        return application_deployment_params.clone();
    }

    /**
     * Returns the network port.
     * @return the network port, or zero if not set
     */
    public int getPort() {

        return port;
    }

    /**
     * Returns the host's platform.
     * @return the host's platform
     */
    public synchronized PlatformDescriptor getPlatform() {

        if (platform_descriptor == null) {
            initPlatform();
        }
        return platform_descriptor;
    }

    /**
     * Returns a manager for executing processes on the host.
     * @return a manager for executing processes on the host
     */
    public ProcessManager getProcessManager() {

        return process_manager;
    }

    /**
     * Returns a map containing results recorded by application-specific scanners.
     * @return a map containing results recorded by application-specific scanners
     */
    public Map<String, String> getAttributes() {

        return scan_results;
    }

    public synchronized void killProcesses() throws Exception {

        for (final Process process : processes) {

            process.destroy();
            process.waitFor();
        }

        processes.clear();
    }

    public int getNumberOfProcesses() {

        return processes.size();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();

        builder.append("host: " + host + "\n");
        builder.append("port: " + port + "\n");
        builder.append("inet_address: " + inet_address + "\n");
        builder.append("platform_descriptor: " + platform_descriptor + "\n");
        builder.append("class_path: " + class_path + "\n");
        builder.append("host_state: " + getHostState() + "\n");
        builder.append("credentials: " + credentials + "\n");
        builder.append("id: " + id + "\n");
        builder.append("java_bin_path: " + java_bin_path + "\n");
        builder.append("local: " + local(host) + "\n");

        return builder.toString();
    }

    @Override
    public HostDescriptor clone() {

        return new HostDescriptor(host, credentials);
    }

    @Override
    public int compareTo(final HostDescriptor other) {

        final int compare_hosts = host.compareTo(other.host);
        return compare_hosts != 0 ? compare_hosts : id == other.id ? 0 : id < other.id ? -1 : 1;
    }

    public int getId() {

        return id;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a list of host descriptors from interactively entered username/password credentials for the specified hosts. Optionally the same credentials can be used for all hosts.
     *
     * @param hosts the hosts
     * @param same_credentials_for_all true if the same credentials should be used for all hosts
     * @return a list of host descriptors
     *
     * @throws IOException if an error occurs when trying to read in a username or password
     */
    public static SortedSet<HostDescriptor> createDescriptorsUsingPassword(final List<String> hosts, final boolean same_credentials_for_all) throws IOException {

        final SortedSet<HostDescriptor> connections = new ConcurrentSkipListSet<HostDescriptor>();

        final HostDescriptor first_connection = createUsernamePasswordConnection(hosts.get(0));
        connections.add(first_connection);

        for (int i = 1; i < hosts.size(); i++) {

            if (same_credentials_for_all) {
                connections.add(createConnection(hosts.get(i), first_connection.credentials));
            }
            else {
                connections.add(createUsernamePasswordConnection(hosts.get(i)));
            }
        }

        return connections;
    }

    /**
     * Creates a list of host descriptors from interactively entered username/public key credentials for the specified hosts. Optionally the same credentials can be used for all hosts.
     *
     * @param hosts the hosts
     * @param same_credentials_for_all true if the same credentials should be used for all hosts
     * @return a list of hosts descriptors
     *
     * @throws IOException if an error occurs when trying to read in a username or key passphrase
     */
    public static SortedSet<HostDescriptor> createDescriptorsUsingPublicKey(final List<String> hosts, final boolean same_credentials_for_all) throws IOException {

        final SortedSet<HostDescriptor> connections = new ConcurrentSkipListSet<HostDescriptor>();

        final HostDescriptor first_connection = createPublicKeyConnection(hosts.get(0));
        connections.add(first_connection);

        for (int i = 1; i < hosts.size(); i++) {

            if (same_credentials_for_all) {
                connections.add(createConnection(hosts.get(i), first_connection.credentials));
            }
            else {
                connections.add(createPublicKeyConnection(hosts.get(i)));
            }
        }

        return connections;
    }

    /**
     * Sets the class paths for a list of host descriptors.
     *
     * @param host_descriptors the host descriptors
     * @param class_paths the class paths
     * @throws UnequalArrayLengthsException if the lists are not of equal length
     */
    public static void setClassPaths(final SortedSet<HostDescriptor> host_descriptors, final List<ClassPath> class_paths) throws UnequalArrayLengthsException {

        if (host_descriptors.size() != class_paths.size()) { throw new UnequalArrayLengthsException(); }

        int i = 0;
        for (final HostDescriptor host_descriptor : host_descriptors) {
            host_descriptor.classPath(class_paths.get(i++));
        }
    }

    /**
     * Sets the application URLs for a list of host descriptors.
     *
     * @param host_descriptors the host descriptors
     * @param application_urls the URLs
     * @throws UnequalArrayLengthsException if the lists are not of equal length
     */
    public static void setApplicationURLs(final SortedSet<HostDescriptor> host_descriptors, final URL[] application_urls) throws UnequalArrayLengthsException {

        for (final HostDescriptor host_descriptor : host_descriptors) {
            host_descriptor.applicationURLs(application_urls);
        }
    }

    public void deployInLocalProcess(final boolean deploy_in_local_process) {

        this.deploy_in_local_process = deploy_in_local_process;
    }

    public boolean deployInLocalProcess() {

        return deploy_in_local_process;
    }

    public void shutdown() {

        process_manager.shutdown();
    }

    // -------------------------------------------------------------------------------------------------------

    private void init(final String host, final Credentials credentials) {

        init(host);
        this.credentials = credentials;
    }

    private void init(final String host) {

        if (local(host)) {
            initLocal();
        }
        else {
            initRemote(host);
        }
    }

    private void initLocal() {

        host = local_host;
        inet_address = local_inet_address;

        scan_results = Collections.synchronizedMap(new HashMap<String, String>());
        process_manager = new ProcessManager(this);

        initLocalPlatform();
    }

    private void initRemote(final String host) {

        this.host = host;
        try {
            inet_address = InetAddress.getByName(host);
        }
        catch (final UnknownHostException e) {
            // Ignore, and allow for host_address being potentially null.
        }

        scan_results = Collections.synchronizedMap(new HashMap<String, String>());
        process_manager = new ProcessManager(this);

        hostState(HostState.UNKNOWN);
    }

    private void initPlatform() {

        if (local(host)) {
            initLocalPlatform();
        }
        else {
            initRemotePlatform();
        }
    }

    private synchronized void initLocalPlatform() {

        String platform_name = System.getProperty("os.name");

        if (platform_name.startsWith(PlatformDescriptor.NAME_WINDOWS)) {
            platform_name = PlatformDescriptor.NAME_WINDOWS;
        }

        if (platform_name.equals(PlatformDescriptor.NAME_MAC) || platform_name.equals(PlatformDescriptor.NAME_LINUX) || platform_name.equals(PlatformDescriptor.NAME_WINDOWS)) {

            final String file_separator = System.getProperty("file.separator");
            final String class_path_separator = System.getProperty("path.separator");
            final String temp_path = System.getProperty("java.io.tmpdir");

            platform_descriptor = new PlatformDescriptor(platform_name, file_separator, class_path_separator, temp_path);
        }
        else {
            platform_descriptor = new PlatformDescriptor();
        }
    }

    private void initRemotePlatform() {

        // Probe remote machine to discover OS.

        // Examples of expected replies:
        //   Linux
        //   Darwin

        final CountDownLatch latch = new CountDownLatch(1);
        final StringBuffer remote_reply_buffer = new StringBuffer();

        // Processor to read the output from the remote process and extract the platform name.
        final IStreamProcessor extract_platform = new IStreamProcessor() {

            @Override
            public boolean processByte(final int byte_value) {

                // Accumulate the output.
                remote_reply_buffer.append((char) byte_value);

                if (byte_value == '\n') {
                    latch.countDown();
                }

                return false;
            }
        };

        ProcessDescriptor probe_platform_process_descriptor = null;
        try {
            // Default is unknown platform.
            platform_descriptor = new PlatformDescriptor();

            // Clone the SSH connection to avoid interference with other use of this one.
            final HostDescriptor cloned = this;
            probe_platform_process_descriptor = new ProcessDescriptor().command(PROBE_COMMAND).outputProcessor(extract_platform);
            cloned.getProcessManager().runProcess(probe_platform_process_descriptor);

            // Wait for the remote process to complete.
            latch.await();

            final String remote_reply = remote_reply_buffer.toString();

            if (remote_reply.contains(PROBE_REPLY_MAC)) {
                platform_descriptor = new PlatformDescriptor(PlatformDescriptor.NAME_MAC);
            }
            else if (remote_reply.contains(PROBE_REPLY_LINUX)) {
                platform_descriptor = new PlatformDescriptor(PlatformDescriptor.NAME_LINUX);
            }
            else {
                Diagnostic.trace("unexpected remote platform reply: " + remote_reply);
            }
        }
        catch (final TimeoutException e) {
            Diagnostic.trace("exception determining remote platform: " + e.getClass().getName() + " : " + e.getMessage());
        }
        catch (final InterruptedException e) {
            Diagnostic.trace("exception determining remote platform: " + e.getClass().getName() + " : " + e.getMessage());
        }
        catch (final IOException e) {
            Diagnostic.trace("exception determining remote platform: " + e.getClass().getName() + " : " + e.getMessage());
        }
        catch (final SSH2Exception e) {
            Diagnostic.trace("exception determining remote platform: " + e.getClass().getName() + " : " + e.getMessage());
        }
        catch (final UnknownPlatformException e) {
            Diagnostic.trace("exception determining remote platform: " + e.getClass().getName() + " : " + e.getMessage());
        }
        finally {
            if (probe_platform_process_descriptor != null) {
                probe_platform_process_descriptor.shutdown();
            }
        }
    }

    // -------------------------------------------------------------------------------------------------------

    private static HostDescriptor createUsernamePasswordConnection(final String host) throws IOException {

        final Credentials credentials = new Credentials(true);
        return new HostDescriptor(host, credentials);
    }

    private static HostDescriptor createPublicKeyConnection(final String host) throws IOException {

        final Credentials credentials = new Credentials(false);
        return new HostDescriptor(host, credentials);
    }

    private static HostDescriptor createConnection(final String host, final Credentials credentials) {

        return new HostDescriptor(host, credentials);
    }

    private static boolean local(final String host) {

        return host == null || host.equals("") || host.equals(LOCAL_HOST) || host.equals(local_host);
    }
}
