/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.TimeUnit;

import uk.ac.standrews.cs.nds.madface.interfaces.IApplicationManager;
import uk.ac.standrews.cs.nds.madface.interfaces.IAttributesCallback;
import uk.ac.standrews.cs.nds.madface.interfaces.IGlobalHostScanner;
import uk.ac.standrews.cs.nds.madface.interfaces.IHostScanner;
import uk.ac.standrews.cs.nds.madface.interfaces.IHostStatusCallback;
import uk.ac.standrews.cs.nds.madface.interfaces.IMadfaceManager;
import uk.ac.standrews.cs.nds.madface.interfaces.ISingleHostScanner;
import uk.ac.standrews.cs.nds.madface.scanners.DeployScanner;
import uk.ac.standrews.cs.nds.madface.scanners.DropScanner;
import uk.ac.standrews.cs.nds.madface.scanners.GlobalHostScannerThread;
import uk.ac.standrews.cs.nds.madface.scanners.HostScannerThread;
import uk.ac.standrews.cs.nds.madface.scanners.KillScanner;
import uk.ac.standrews.cs.nds.madface.scanners.SingleHostScannerThread;
import uk.ac.standrews.cs.nds.madface.scanners.StatusScanner;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.Duration;

/**
 * Madface server implementation.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class MadfaceManager implements IMadfaceManager {

    /**
     * Successful configuration message.
     */
    public static final String APPLICATION_CONFIGURED_SUCCESSFULLY = "application configured successfully";

    /**
     * Separator character for URL paths.
     */
    public static final String URL_PATH_SEPARATOR = ";";

    /**
     * Error message for invalid preference.
     */
    public static final String NO_PREFERENCE_WITH_NAME = "no preference with name: ";

    private static final String DISCARD_ERRORS = "Discard Errors";
    private static final String LIB_DIR_NAME = "lib";
    private static final String BIN_DIR_NAME = "bin";
    private static final String NEWLINE_REGEX = "[\\r\\n]+";

    // -------------------------------------------------------------------------------------------------------

    private static final long serialVersionUID = -8525716807644819016L;
    public static Duration STATE_WAIT_DELAY_DURATION = new Duration(1, TimeUnit.SECONDS);

    // -------------------------------------------------------------------------------------------------------

    private Class<? extends IApplicationManager> application_manager_class;
    private Set<URL> application_urls;

    private IApplicationManager application_manager;

    private StatusScanner status_scanner;
    private DeployScanner deploy_scanner;
    private DropScanner drop_scanner;
    private KillScanner kill_scanner;

    private Set<IHostStatusCallback> host_status_callbacks;
    private Set<IAttributesCallback> attributes_callbacks;

    private final SortedSet<HostDescriptor> host_state_list;
    private final List<HostScannerThread> scanner_list;
    private final Map<String, IHostScanner> scanner_map;
    private final Map<String, IHostScanner> headless_scanner_map;

    private boolean discard_errors;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the service implementation.
     */
    public MadfaceManager() {

        host_state_list = new ConcurrentSkipListSet<HostDescriptor>();
        scanner_list = Collections.synchronizedList(new ArrayList<HostScannerThread>());
        scanner_map = Collections.synchronizedMap(new HashMap<String, IHostScanner>());
        headless_scanner_map = Collections.synchronizedMap(new HashMap<String, IHostScanner>());

        application_urls = Collections.synchronizedSet(new HashSet<URL>());

        configureCallbacks();
        configureGenericScanners();
        startScanners();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the manager's scanner map.
     * @return the scanner map
     */
    public Map<String, IHostScanner> getScannerMap() {

        return scanner_map;
    }

    /**
     * Returns the application manager.
     * @return the application manager
     */
    public IApplicationManager getApplicationManager() {

        return application_manager;
    }

    /**
     * Returns the host status callbacks.
     * @return the host status callbacks.
     */
    public Set<IHostStatusCallback> getHostStatusCallbacks() {

        return host_status_callbacks;
    }

    /**
     * Returns the host attribute callbacks.
     * @return the host attribute callbacks.
     */
    public Set<IAttributesCallback> getAttributesCallbacks() {

        return attributes_callbacks;
    }

    /**
      * Sets the application manager.
      * @param application_manager the application manager
      */
    public void setApplicationManager(final IApplicationManager application_manager) {

        this.application_manager = application_manager;
        if (application_manager != null) {
            application_manager_class = application_manager.getClass();
        }
    }

    /**
     * Configures the application manager class.
     * 
     * @param application_manager_class_name the application manager class name
     * @throws ClassNotFoundException if the class cannot be resolved
     */
    public void configureApplicationManagerClass(final String application_manager_class_name) throws ClassNotFoundException {

        application_manager_class = getApplicationManagerClass(application_manager_class_name);
    }

    /**
     * Configures the URLS for the application.
     * 
     * @param application_url_classpath the URL classpath
     * @throws IOException if any of the URLs are invalid
     */
    public void configureApplicationUrls(final String application_url_classpath) throws IOException {

        setApplicationUrls(pathToUrls(application_url_classpath));
    }

    /**
     * Configures the application and any application-specific scanners.
     * 
     * @throws InstantiationException if the application manager cannot be instantiated
     * @throws IllegalAccessException if the application manager cannot be instantiated
     */
    public void configureApplication() throws InstantiationException, IllegalAccessException {

        configureApplicationManager();
        configureApplicationSpecificScanners();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void configureApplication(final Class<? extends IApplicationManager> application_entrypoint, final Set<URL> application_urls) throws IOException, InstantiationException, IllegalAccessException {

        application_manager_class = application_entrypoint;
        setApplicationUrls(application_urls);

        configureApplication();
    }

    @Override
    public void configureApplication(final Class<? extends IApplicationManager> application_entrypoint, final URL url_base, final Set<String> jar_names, final Set<String> lib_names) throws IOException, InstantiationException, IllegalAccessException {

        final Set<URL> application_url_classpath = makeURLs(url_base, jar_names, lib_names);

        configureApplication(application_entrypoint, application_url_classpath);
    }

    @Override
    public void configureApplication(final IApplicationManager application_manager) {

        setApplicationManager(application_manager);
        configureApplicationSpecificScanners();
    }

    @Override
    public Set<URL> getApplicationUrls() {

        return application_urls;
    }

    @Override
    public Class<? extends IApplicationManager> getApplicationEntrypoint() {

        return application_manager_class;
    }

    @Override
    public String getApplicationName() {

        return application_manager != null ? application_manager.getApplicationName() : "";
    }

    @Override
    public void add(final HostDescriptor host_descriptor) {

        host_state_list.add(host_descriptor);
        host_descriptor.applicationURLs(setToArray(application_urls));
    }

    @Override
    public void add(final String multi_line_host_patterns, final Credentials credentials) {

        final String[] host_patterns = splitAtNewLines(multi_line_host_patterns);

        add(host_patterns, credentials);
    }

    private void add(final String[] host_patterns, final Credentials credentials) {

        for (final String host_pattern : host_patterns) {

            // Each element of the array may contain a pattern that denotes multiple hosts.
            for (final String host : Patterns.resolveHostPattern(host_pattern)) {

                final HostDescriptor host_descriptor = new HostDescriptor(host).credentials(credentials);
                add(host_descriptor);
            }
        }
    }

    @Override
    public void drop(final HostDescriptor host_descriptor) {

        Diagnostic.traceNoSource(DiagnosticLevel.FINAL, "dropping: " + host_descriptor.getHost());

        synchronized (host_state_list) {
            host_state_list.remove(host_descriptor);
            host_descriptor.shutdown();
        }
    }

    @Override
    public void addHostStatusCallback(final IHostStatusCallback host_status_callback) {

        host_status_callbacks.add(host_status_callback);
    }

    @Override
    public void addAttributesCallback(final IAttributesCallback attributes_callback) {

        attributes_callbacks.add(attributes_callback);
    }

    @Override
    public void deploy(final HostDescriptor host_descriptor) throws Exception {

        Diagnostic.traceNoSource(DiagnosticLevel.FINAL, "deploying to: " + host_descriptor.getHost());
        application_manager.deployApplication(host_descriptor);
    }

    @Override
    public void kill(final HostDescriptor host_descriptor) throws Exception {

        Diagnostic.traceNoSource(DiagnosticLevel.FINAL, "killing application on: " + host_descriptor.getHost());
        application_manager.killApplication(host_descriptor);
    }

    @Override
    public void killAll() throws Exception {

        for (final HostDescriptor host_descriptor : host_state_list) {
            kill(host_descriptor);
        }
    }

    @Override
    public void deployAll() throws Exception {

        for (final HostDescriptor host_descriptor : host_state_list) {
            deploy(host_descriptor);
        }
    }

    @Override
    public void dropAll() {

        host_state_list.clear();
    }

    @Override
    public String setPrefEnabled(final String pref_name, final boolean enabled) {

        if (pref_name.equals(DISCARD_ERRORS)) {

            discard_errors = enabled;
            return Constants.RESULT_OK;
        }

        return setScannerEnabled(pref_name, enabled);
    }

    @Override
    public void setAutoDeploy(final boolean auto_deploy) {

        setPrefEnabled(DeployScanner.AUTO_DEPLOY_KEY, auto_deploy);
    }

    @Override
    public void setAutoDrop(final boolean auto_drop) {

        setPrefEnabled(DropScanner.AUTO_DROP_KEY, auto_drop);
    }

    @Override
    public void setDeployScannerInterval(final Duration min_cycle_time) {

        deploy_scanner.setMinCycleTime(min_cycle_time);
    }

    @Override
    public void setAutoKill(final boolean auto_kill) {

        setPrefEnabled(KillScanner.AUTO_KILL_KEY, auto_kill);
    }

    @Override
    public void setHostScanning(final boolean enabled) {

        setPrefEnabled(StatusScanner.STATUS_SCANNER_KEY, enabled);
    }

    @Override
    public void setStatusScannerInterval(final Duration min_cycle_time) {

        status_scanner.setMinCycleTime(min_cycle_time);
    }

    @Override
    public void waitForHostToReachState(final HostDescriptor host_descriptor, final HostState state_to_reach) {

        waitForHost(host_descriptor, state_to_reach, true);
    }

    @Override
    public void waitForHostToReachStateThatIsNot(final HostDescriptor host_descriptor, final HostState state_to_not_reach) {

        waitForHost(host_descriptor, state_to_not_reach, false);
    }

    @Override
    public void waitForAllToReachState(final HostState state_to_reach) {

        waitForAll(state_to_reach, true);
    }

    @Override
    public void waitForAllToReachStateThatIsNot(final HostState state_to_not_reach) {

        waitForAll(state_to_not_reach, false);
    }

    @Override
    public void shutdown() {

        shutdownScannerThreads();
        shutdownScanners();
        shutdownHostDescriptors();
        shutdownApplicationManager();
    }

    private void shutdownApplicationManager() {

        if (application_manager != null) {
            application_manager.shutdown();
        }
    }

    private void shutdownScannerThreads() {

        for (final HostScannerThread scanner_thread : scanner_list) {
            scanner_thread.shutdown();
        }
    }

    @Override
    public SortedSet<HostDescriptor> getHostDescriptors() {

        return host_state_list;
    }

    @Override
    public HostDescriptor getHostDescriptor(final String host) throws UnknownHostException {

        synchronized (host_state_list) {

            for (final HostDescriptor host_descriptor : host_state_list) {
                if (host_descriptor.getHost().equals(host)) { return host_descriptor; }
            }
        }

        throw new UnknownHostException();
    }

    private void shutdownHostDescriptors() {

        for (final HostDescriptor host_descriptor : host_state_list) {
            host_descriptor.shutdown();
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns true if error output should be discarded.
     * @return true if error output should be discarded
     */
    public boolean errorsAreDiscarded() {

        return discard_errors;
    }

    // -------------------------------------------------------------------------------------------------------

    private String setScannerEnabled(final String pref_name, final boolean enabled) {

        final IHostScanner scanner = getScannerMap().get(pref_name);
        if (scanner == null) { return NO_PREFERENCE_WITH_NAME + pref_name; }

        scanner.setEnabled(enabled);
        return Constants.RESULT_OK;
    }

    private void setApplicationUrls(final Set<URL> application_urls) throws IOException {

        validateUrls(application_urls);
        this.application_urls = application_urls;
    }

    private void validateUrls(final Set<URL> application_urls) throws IOException {

        for (final URL url : application_urls) {
            final URLConnection connection = url.openConnection();
            connection.connect();
            final InputStream stream = connection.getInputStream();
            stream.close();
        }
    }

    private void configureApplicationManager() throws InstantiationException, IllegalAccessException {

        application_manager = getApplicationManager(application_manager_class);
    }

    private void configureGenericScanners() {

        status_scanner = new StatusScanner(this);
        deploy_scanner = new DeployScanner(this);
        drop_scanner = new DropScanner(this);
        kill_scanner = new KillScanner(this);

        scanner_list.add(new SingleHostScannerThread(this, status_scanner));
        scanner_list.add(new SingleHostScannerThread(this, deploy_scanner));
        scanner_list.add(new SingleHostScannerThread(this, drop_scanner));
        scanner_list.add(new SingleHostScannerThread(this, kill_scanner));

        scanner_map.put(status_scanner.getToggleLabel(), status_scanner);
        scanner_map.put(deploy_scanner.getToggleLabel(), deploy_scanner);
        scanner_map.put(drop_scanner.getToggleLabel(), drop_scanner);
        scanner_map.put(kill_scanner.getToggleLabel(), kill_scanner);
    }

    private void configureApplicationSpecificScanners() {

        configureSingleScanners();
        configureGlobalScanners();
    }

    private void configureSingleScanners() {

        final List<ISingleHostScanner> single_host_scanners = application_manager.getSingleScanners();

        for (final ISingleHostScanner scanner : single_host_scanners) {

            final SingleHostScannerThread thread = new SingleHostScannerThread(this, scanner);
            scanner_list.add(thread);
            thread.start();

            final String toggle_label = scanner.getToggleLabel();
            if (toggle_label != null) {
                scanner_map.put(toggle_label, scanner);
            }
            else {
                headless_scanner_map.put(scanner.getName(), scanner);
            }
        }
    }

    private void configureGlobalScanners() {

        final List<IGlobalHostScanner> global_host_scanners = application_manager.getGlobalScanners();

        for (final IGlobalHostScanner scanner : global_host_scanners) {

            final GlobalHostScannerThread thread = new GlobalHostScannerThread(this, scanner);
            scanner_list.add(thread);
            thread.start();

            final String toggle_label = scanner.getToggleLabel();
            if (toggle_label != null) {
                scanner_map.put(toggle_label, scanner);
            }
            else {
                headless_scanner_map.put(scanner.getName(), scanner);
            }
        }
    }

    private void configureCallbacks() {

        host_status_callbacks = new HashSet<IHostStatusCallback>();
        attributes_callbacks = new HashSet<IAttributesCallback>();
    }

    private void startScanners() {

        for (final Thread scanner : scanner_list) {
            scanner.start();
        }
    }

    private void shutdownScanners() {

        shutdownScanners(scanner_map);
        shutdownScanners(headless_scanner_map);
    }

    private void shutdownScanners(final Map<String, IHostScanner> map) {

        for (final String scanner_name : map.keySet()) {
            map.get(scanner_name).shutdown();
        }
    }

    private String[] splitAtNewLines(final String hosts) {

        return hosts.split(NEWLINE_REGEX);
    }

    private Set<URL> pathToUrls(final String application_url_classpath) throws MalformedURLException {

        final String[] url_strings = application_url_classpath.split(URL_PATH_SEPARATOR);
        final Set<URL> result = new HashSet<URL>(url_strings.length);

        for (final String s : url_strings) {
            if (!s.equals("")) {
                result.add(new URL(s));
            }
        }

        return result;
    }

    private URL[] setToArray(final Set<URL> urls) {

        return urls.toArray(new URL[0]);
    }

    private Set<URL> makeURLs(final URL url_base, final Set<String> jar_names, final Set<String> lib_names) throws MalformedURLException {

        final Set<URL> urls = new HashSet<URL>();

        urls.addAll(makeUrls(url_base, jar_names, BIN_DIR_NAME));
        urls.addAll(makeUrls(url_base, lib_names, LIB_DIR_NAME));

        return urls;
    }

    private Set<URL> makeUrls(final URL url_base, final Set<String> names, final String directory) throws MalformedURLException {

        final Set<URL> urls = new HashSet<URL>();

        for (final String lib_name : names) {
            urls.add(new URL(new URL(url_base, directory), lib_name));
        }

        return urls;
    }

    @SuppressWarnings("unchecked")
    private Class<? extends IApplicationManager> getApplicationManagerClass(final String application_entrypoint_name) throws ClassNotFoundException {

        final ClassLoader parent_loader = Thread.currentThread().getContextClassLoader();
        final URL[] url_array = setToArray(application_urls);
        final ClassLoader url_class_loader = new URLClassLoader(url_array, parent_loader);

        return (Class<? extends IApplicationManager>) Class.forName(application_entrypoint_name, true, url_class_loader);
    }

    private IApplicationManager getApplicationManager(final Class<? extends IApplicationManager> application_entrypoint) throws InstantiationException, IllegalAccessException {

        return application_entrypoint.newInstance();
    }

    private void waitForHost(final HostDescriptor host_descriptor, final HostState state, final boolean match) {

        while (!Thread.currentThread().isInterrupted()) {

            if (!xor(match, host_descriptor.getHostState().equals(state))) { return; }
            STATE_WAIT_DELAY_DURATION.sleep();
        }
    }

    private void waitForAll(final HostState state, final boolean match) {

        while (!Thread.currentThread().isInterrupted()) {
            boolean all_at_termination_condition = true;

            for (final HostDescriptor host_descriptor : host_state_list) {
                if (xor(match, host_descriptor.getHostState().equals(state))) {
                    all_at_termination_condition = false;
                    break;
                }
            }

            if (all_at_termination_condition) { return; }
            STATE_WAIT_DELAY_DURATION.sleep();
        }
    }

    private boolean xor(final boolean first, final boolean second) {

        return first && !second || !first && second;
    }
}
