/***************************************************************************
 *                                                                         *
 * MADFACE Library                                                         *
 * Copyright (C) 2010-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/madface/                               *
 *                                                                         *
 * This file is part of MADFACE, a package providing                       *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * MADFACE is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * MADFACE is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with MADFACE.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface.interfaces;

import java.io.IOException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Set;
import java.util.SortedSet;

import uk.ac.standrews.cs.nds.madface.Credentials;
import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.HostState;
import uk.ac.standrews.cs.nds.madface.Patterns;
import uk.ac.standrews.cs.nds.util.Duration;

/**
 * Madface service.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public interface IMadfaceManager {

    /**
     * Configures the application to be monitored.
     *
     * @param application_manager_class the application manager class
     * @param application_urls the URLs for for the application's jar and library files
     * @throws IOException if a connection to an application URL cannot be established
     * @throws IllegalAccessException if the application manager cannot be instantiated
     * @throws InstantiationException if the application manager cannot be instantiated
     */
    void configureApplication(Class<? extends IApplicationManager> application_manager_class, Set<URL> application_urls) throws IOException, InstantiationException, IllegalAccessException;

    /**
     * Configures the application to be monitored.
     *
     * @param application_manager_class the entry point class of the application to be managed
     * @param url_base the common base for the application's jar and library URLs
     * @param jar_names the names of the jar files used by the application
     * @param lib_names the names of the library files used by the application
     * @throws IOException if an application URL cannot be formed, or a connection to an application URL cannot be established
     * @throws IllegalAccessException if the application manager cannot be instantiated
     * @throws InstantiationException if the application manager cannot be instantiated
     */
    void configureApplication(Class<? extends IApplicationManager> application_manager_class, URL url_base, Set<String> jar_names, Set<String> lib_names) throws IOException, InstantiationException, IllegalAccessException;

    /**
     * Configures the application to be monitored.
     *
     * @param application_manager the manager for the application
     */
    void configureApplication(IApplicationManager application_manager);

    /**
     * Sets a specified preference flag.
     *
     * @param pref_name the name of the preference to be set
     * @param enabled true if the preference should be enabled
     * @return a status message
     */
    String setPrefEnabled(String pref_name, boolean enabled);

    /**
     * Returns the library URLs used by the configured application.
     * @return the library URLs used by the configured application
     */
    Set<URL> getApplicationUrls();

    /**
     * Returns the entrypoint class for the configured application.
     * @return the entrypoint class for the configured application
     */
    Class<? extends IApplicationManager> getApplicationEntrypoint();

    /**
     * Returns the name for the configured application.
     * @return the name for the configured application
     */
    String getApplicationName();

    /**
     * Adds the set of hosts specified by a given address pattern.
     *
     * @param multi_line_host_patterns a newline-separated list, each entry containing a single host address or a pattern as defined by {@link Patterns#resolveHostPattern}
     * @param credentials the credentials to be used with the hosts
     */
    void add(String multi_line_host_patterns, Credentials credentials);

    /**
     * Adds a host.
     *
     * @param host_descriptor the host descriptor
     */
    void add(HostDescriptor host_descriptor);

    /**
     * Drops a specified host from the manager.
     * @param host_descriptor the host descriptor
     */
    void drop(HostDescriptor host_descriptor);

    /**
     * Attempts to deploy the application to a specified host.
     * @param host_descriptor the host descriptor
     * @throws Exception if the attempt fails
     */
    void deploy(HostDescriptor host_descriptor) throws Exception;

    /**
     * Attempts to kill the application on a specified host.
     * @param host_descriptor the host descriptor
     * @throws Exception if the attempt fails
     */
    void kill(HostDescriptor host_descriptor) throws Exception;

    /**
     * Drops all hosts from the manager.
     */
    void dropAll();

    /**
     * Attempts to start the application an all hosts and waits until this appears to have succeeded.
     * @throws Exception if the attempt fails
     */
    void deployAll() throws Exception;

    /**
     * Attempts to kill the application an all hosts and waits until this appears to have succeeded.
     * @throws Exception if the attempt fails
     */
    void killAll() throws Exception;

    /**
     * Blocks until a given host is in a specified state.
     * 
     * @param host_descriptor the host descriptor
     * @param state_to_reach the state
     */
    void waitForHostToReachState(HostDescriptor host_descriptor, HostState state_to_reach);

    /**
     * Blocks until a given host is not in a specified state.
     * 
     * @param host_descriptor the host descriptor
     * @param state_to_not_reach the state
     */
    void waitForHostToReachStateThatIsNot(HostDescriptor host_descriptor, HostState state_to_not_reach);

    /**
     * Blocks until all hosts are in a specified state.
     * @param state_to_reach the state
     */
    void waitForAllToReachState(HostState state_to_reach);

    /**
     * Blocks until all hosts are not in a specified state.
     * @param state_to_not_reach the state
     */
    void waitForAllToReachStateThatIsNot(HostState state_to_not_reach);

    /**
      * Sets whether auto-deployment should be performed. If enabled, an attempt is made to start the application on any host on
      * which the application is not detected.
      *
      * @param auto_deploy true if auto-deployment should be performed
      */
    void setAutoDeploy(boolean auto_deploy);

    /**
     * Sets the minimum cycle time between successive scans of the deployment scanner.
     * @param min_cycle_time the minimum cycle time between successive scans of the deployment scanner
     */
    void setDeployScannerInterval(Duration min_cycle_time);

    /**
     * Sets whether auto-kill should be performed. If enabled, an attempt is made to kill the application on any host on
     * which the application is detected.
     *
     * @param auto_kill true if auto-kill should be performed
     */
    void setAutoKill(boolean auto_kill);

    /**
     * Sets whether auto-deploy should be performed. If enabled, an attempt is made to deploy the application on any host on
     * which it is not currently detected, and SSH authentication is possible.
     *
     * @param auto_drop true if auto-auto_drop should be performed
     */
    void setAutoDrop(boolean auto_drop);

    /**
     * Sets whether host scanning should be performed.
     *
     * @param enabled true if host scanning should be performed
     */
    void setHostScanning(boolean enabled);

    /**
     * Sets the minimum cycle time between successive scans of the status scanner.
     * @param min_cycle_time the minimum cycle time between successive scans of the status scanner
     */
    void setStatusScannerInterval(Duration min_cycle_time);

    /**
     * Gets the host descriptor for a given host.
     *
     * @param host a host
     * @return the host descriptor for the host
     * @throws UnknownHostException if the host is not known by the manager
     */
    HostDescriptor getHostDescriptor(String host) throws UnknownHostException;

    /**
     * Gets the host descriptors for all currently managed hosts.
     *
     * @return the host descriptors
     */
    SortedSet<HostDescriptor> getHostDescriptors();

    /**
     * Adds a callback for host status changes.
     * @param host_status_callback the callback
     */
    void addHostStatusCallback(IHostStatusCallback host_status_callback);

    /**
     * Adds a callback for attribute changes.
     * @param attributes_callback the callback
     */
    void addAttributesCallback(IAttributesCallback attributes_callback);

    /**
     * Shuts down the manager.
     */
    void shutdown();
}
