/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface.scanners;

import java.util.SortedSet;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;

import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.interfaces.IHostScanner;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.Duration;

/**
 * Thread that continually monitors the status of the hosts in a given list,
 * the contents of which may vary dynamically.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class HostScannerThread extends Thread {

    protected SortedSet<HostDescriptor> host_state_list;
    protected String name;

    public HostScannerThread(final IHostScanner scanner) {

        name = scanner.getName();
    }

    /**
     * Shuts down the scanning thread.
     */
    public void shutdown() {

        interrupt();
    }

    protected void executeRepeatedly(final Callable<Void> action, final Duration minimum_cycle_time) {

        long previous_loop_end = 0;
        final long min_cycle_time_in_millis = minimum_cycle_time.getLength(TimeUnit.MILLISECONDS);

        while (!isInterrupted()) {

            try {
                action.call();
            }
            catch (final InterruptedException e) {
                interrupt();
                break;
            }
            catch (final Exception e) {
                Diagnostic.trace("unexpected exception: " + e.getMessage());
            }

            final long current = System.currentTimeMillis();
            final long earliest_finish = previous_loop_end + min_cycle_time_in_millis;
            final long delay_required = earliest_finish - current;

            if (delay_required > 0) {
                try {
                    TimeUnit.MILLISECONDS.sleep(delay_required);
                }
                catch (final InterruptedException e) {
                    interrupt();
                    break;
                }
            }

            previous_loop_end = System.currentTimeMillis();
        }
    }
}
