/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface.scanners;

import java.util.Set;
import java.util.concurrent.Callable;

import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.MadfaceManager;
import uk.ac.standrews.cs.nds.madface.interfaces.IAttributesCallback;
import uk.ac.standrews.cs.nds.madface.interfaces.ISingleHostScanner;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.TimeoutExecutor;

/**
 * Thread that continually monitors the status of the hosts in a given list, the contents of which may vary dynamically.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class SingleHostScannerThread extends HostScannerThread {

    // -------------------------------------------------------------------------------------------------------

    private final ISingleHostScanner scanner;
    private final Set<IAttributesCallback> attributes_callbacks;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a monitor for the host list.
     *
     * @param manager the manager
     * @param scanner an application-specific scanner
     */
    public SingleHostScannerThread(final MadfaceManager manager, final ISingleHostScanner scanner) {

        super(scanner);
        host_state_list = manager.getHostDescriptors();
        attributes_callbacks = manager.getAttributesCallbacks();
        this.scanner = scanner;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void run() {

        final TimeoutExecutor executor = scanner.getTimeoutExecutor();

        final Callable<Void> check_all_hosts = new Callable<Void>() {

            @Override
            public Void call() throws Exception {

                for (final HostDescriptor host_descriptor : host_state_list) {

                    final Callable<Void> check = new Callable<Void>() {

                        @Override
                        public Void call() throws Exception {

                            try {
                                scanner.check(host_descriptor, attributes_callbacks);
                            }
                            catch (final InterruptedException e) {
                                throw e;
                            }
                            catch (final Exception e) {
                                Diagnostic.trace("inner error in scanner check: ", e.getClass().getName(), " : ", e.getMessage());
                                final Throwable cause = e.getCause();
                                if (cause != null) {
                                    Diagnostic.trace("caused by: ", cause.getClass().getName(), " : ", cause.getMessage());
                                }
                            }
                            return null;
                        }
                    };

                    try {
                        executor.executeWithTimeout(check);
                    }
                    catch (final InterruptedException e) {
                        throw e;
                    }
                    catch (final Exception e) {
                        Diagnostic.trace(DiagnosticLevel.FULL, "error in scanner check: " + scanner.getAttributeName() + " : " + e.getClass().getName() + " : " + e.getMessage());
                    }
                }
                return null;
            }
        };

        executeRepeatedly(check_all_hosts, scanner.getMinCycleTime());
    }
}
