/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.madface.scanners;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.HostState;
import uk.ac.standrews.cs.nds.madface.MadfaceManager;
import uk.ac.standrews.cs.nds.madface.ProcessDescriptor;
import uk.ac.standrews.cs.nds.madface.exceptions.UnknownPlatformException;
import uk.ac.standrews.cs.nds.madface.interfaces.IApplicationManager;
import uk.ac.standrews.cs.nds.madface.interfaces.IAttributesCallback;
import uk.ac.standrews.cs.nds.madface.interfaces.IHostStatusCallback;
import uk.ac.standrews.cs.nds.madface.interfaces.ISingleHostScanner;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.Duration;
import uk.ac.standrews.cs.nds.util.TimeoutExecutor;

import com.mindbright.ssh2.SSH2Exception;

/**
 * Thread that continually monitors the status of the machines in a given list,
 * the contents of which may vary dynamically. Machines are probed for the
 * presence of a particular application, and for their willingness to accept an
 * SSH connection with specified credentials. The results of these tests are
 * recorded within the list.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class StatusScanner extends Scanner implements ISingleHostScanner {

    private static final String MINIMAL_COMMAND = "cd /";

    /**
     * Key for the host scanning operation.
     */
    public static final String STATUS_SCANNER_KEY = "Status Scanner";

    // The timeout for attempted SSH connections.
    public static Duration SSH_CHECK_TIMEOUT = new Duration(30, TimeUnit.SECONDS);

    private static final Duration STATUS_CHECK_TIMEOUT = new Duration(30, TimeUnit.SECONDS);
    public static int STATUS_CHECK_THREADS = 10;
    public static int SSH_CHECK_THREADS = 7; // Seems to give best throughput in tests.

    private final MadfaceManager manager;
    private final Set<IHostStatusCallback> host_status_callbacks;

    private final ProcessDescriptor process_descriptor;
    private final TimeoutExecutor ssh_timeout_executor;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a status scanner for the given manager.
     *
     * @param manager the manager
     */
    public StatusScanner(final MadfaceManager manager) {

        this.manager = manager;
        host_status_callbacks = manager.getHostStatusCallbacks();

        ssh_timeout_executor = new TimeoutExecutor(SSH_CHECK_THREADS, SSH_CHECK_TIMEOUT, false, "SSH executor for status scanner");
        process_descriptor = new ProcessDescriptor().command(MINIMAL_COMMAND).executor(ssh_timeout_executor);
    }

    @Override
    public TimeoutExecutor makeExecutor() {

        return new TimeoutExecutor(STATUS_CHECK_THREADS, STATUS_CHECK_TIMEOUT, false, "status scanner");
    }

    @Override
    public void shutdown() {

        super.shutdown();
        ssh_timeout_executor.shutdown();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String getName() {

        return "Status";
    }

    @Override
    public String getAttributeName() {

        return "Status";
    }

    @Override
    public String getToggleLabel() {

        return STATUS_SCANNER_KEY;
    }

    @Override
    public void check(final HostDescriptor host_descriptor, final Set<IAttributesCallback> attribute_callbacks) {

        final IApplicationManager application_manager = manager.getApplicationManager();

        if (application_manager != null && enabled) {

            final HostState original_state = host_descriptor.getHostState();

            // First try to contact the application directly. If that doesn't work, try to make an SSH connection to
            // see whether it would be possible to launch the application on the machine.

            // TODO try separate scanner for machine liveness, using address verification? Could run at higher concurrency?

            try {
                application_manager.attemptApplicationCall(host_descriptor);
                setHostState(host_descriptor, HostState.RUNNING);
            }
            catch (final UnknownHostException e) {
                // Machine address couldn't be resolved.
                setHostState(host_descriptor, HostState.INVALID);
            }
            catch (final Exception e) {
                try {

                    // Application call failed, so try SSH connection.
                    attemptSSHConnection(host_descriptor);

                    setHostState(host_descriptor, HostState.AUTH);
                }
                catch (final SSH2Exception e1) {

                    // Couldn't make SSH connection with specified credentials.
                    setHostState(host_descriptor, HostState.NO_AUTH);
                }
                catch (final UnknownHostException e1) {

                    // Machine address couldn't be resolved.
                    setHostState(host_descriptor, HostState.INVALID);
                }
                catch (final IOException e1) {

                    // Network error trying to make SSH connection.
                    setHostState(host_descriptor, HostState.UNREACHABLE);
                }
                catch (final TimeoutException e1) {

                    // SSH connection timed out.
                    setHostState(host_descriptor, HostState.UNREACHABLE);
                }
                catch (final InterruptedException e1) {

                    // SSH connection timed out.
                    setHostState(host_descriptor, HostState.UNREACHABLE);
                }
                catch (final UnknownPlatformException e1) {

                    // This shouldn't happen...
                    Diagnostic.trace("Unexpected unknown platform exception");
                    setHostState(host_descriptor, HostState.UNKNOWN);
                }
            }

            final HostState new_state = host_descriptor.getHostState();
            Diagnostic.trace(DiagnosticLevel.FULL, "state: " + new_state);

            if (new_state != original_state && host_status_callbacks.size() > 0) {
                for (final IHostStatusCallback callback : host_status_callbacks) {
                    callback.hostStatusChange(host_descriptor, original_state);
                }
            }
        }
    }

    @Override
    public boolean enabledByDefault() {

        return true;
    }

    // -------------------------------------------------------------------------------------------------------

    private void attemptSSHConnection(final HostDescriptor host_descriptor) throws SSH2Exception, IOException, TimeoutException, UnknownPlatformException, InterruptedException {

        // Try to execute a 'cd /' shell command on the machine.
        // This is selected as a command that produces no output and doesn't require a functioning home directory.

        host_descriptor.getProcessManager().runProcess(process_descriptor);
    }

    private void setHostState(final HostDescriptor host_descriptor, final HostState new_state) {

        final HostState old_state = host_descriptor.getHostState();
        host_descriptor.hostState(new_state);

        if (new_state != old_state) {
            Diagnostic.traceNoSource(DiagnosticLevel.RUN, host_descriptor.getHost() + ": " + new_state);
        }
    }
}
