/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.interfaces;

import java.math.BigInteger;

/**
 * Interface defining keys.
 *
 * @author Stephanie Anderson
 * @author Stuart Norcross (stuart@cs.st-andrews.ac.uk)
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public interface IKey extends Comparable<IKey> {

    /**
     * Returns the key's value.
     * @return the key's value
     */
    BigInteger keyValue();

    /**
     * Returns a string representation of the key's value.
     *
     * @param radix the radix
     * @return a string representation of the key value using the given radix
     */
    String toString(int radix);

    /**
     * Returns the number of prefix digits shared with a given key, using a given radix.
     *
     * @param key the other key to compare
     * @param radix the radix to be used
     * @return the number of shared prefix digits
     */
    int countSharedPrefixDigits(IKey key, int radix);

    /**
     * Returns a given digit of the key, using a given radix.
     *
     * @param index the index of the required digit
     * @param radix the radix to be used
     * @return the character at that index
     */
    char digit(int index, int radix);

    /**
     * Returns a string representing the fraction of the key space represented by this key.
     * @return a string representing the fraction of the key space represented by this key
     */
    String toStringAsKeyspaceFraction();

    /**
     * Returns a string representing the angle represented by this key relative to a 360 degree key space.
     * @return a string representing the angle represented by this key relative to a 360 degree key space
     */
    String toStringAsAngle();

    /**
     * Returns the number of bits in the minimal two's-complement representation of this key, excluding sign bit.
     * @return the number of bits in the minimal two's-complement representation of this key, excluding sign bit
     */
    int bitLength();

    /**
     * Returns the number of bits in the minimal two's-complement representation of the biggest key possible.
     * @return the number of bits in the minimal two's-complement representation of the biggest key possible
     */
    int maxBitLength();
}
