/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.keys;

import java.io.Serializable;
import java.math.BigInteger;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;

/**
 * Implementation of key.
 *
 * @author Stephanie Anderson
 * @author Stuart Norcross (stuart@cs.st-andrews.ac.uk)
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class Key implements IKey, Serializable {

    private static final long serialVersionUID = 7221912189212816937L;

    private static final BigInteger ONE_HUNDRED = new BigInteger("100");
    private static final BigInteger THREE_SIXTY = new BigInteger("360");

    /**
     * Degree symbol defined using unicode.
     */
    public static final String DEGREE_SYMBOL = String.valueOf('\u02DA');

    // -------------------------------------------------------------------------------------------------------

    /**
     * The length of a key in bits.
     */
    public static final int KEY_LENGTH = 160;

    /**
     * Big integer representation of two.
     */
    public static final BigInteger TWO = BigInteger.ONE.add(BigInteger.ONE);

    /**
     * The size of the keyspace.
     */
    public static final BigInteger KEYSPACE_SIZE = TWO.pow(KEY_LENGTH);

    /**
     * The value of the largest key.
     */
    public static final BigInteger MAX_KEY_VALUE = KEYSPACE_SIZE.subtract(BigInteger.ONE);

    /**
     * The default radix used in converting a key's value to a string.
     */
    public static final int DEFAULT_RADIX = 16;

    // ***********************************************************************

    private BigInteger key_value;

    // ***********************************************************************

    /**
     * Creates a new key using the given value modulo the key space size.
     *
     * @param key_value the value of the key
     */
    public Key(final BigInteger key_value) {

        this.key_value = key_value.remainder(KEYSPACE_SIZE);

        // Allow for negative key value.
        if (this.key_value.compareTo(BigInteger.ZERO) < 0) {
            this.key_value = this.key_value.add(KEYSPACE_SIZE);
        }
    }

    /**
     * Creates a new key using a string representation of a BigInteger with radix {@link #DEFAULT_RADIX}.
     *
     * @param string_representation a representation of the key value
     */
    public Key(final String string_representation) {

        this(new BigInteger(string_representation, DEFAULT_RADIX));
    }

    /**
     * Creates a new key using a byte array representation of a BigInteger with radix {@link #DEFAULT_RADIX}.
     *
     * @param bytes a representation of the key value
     */
    public Key(final byte[] bytes) {

        this(Radix.toHexString(bytes));
    }

    // ***********************************************************************

    @Override
    public BigInteger keyValue() {

        return key_value;
    }

    @Override
    public String toString(final int radix) {

        final int bits_per_digit = Radix.bitsPerDigit(radix);
        final int pad_to_length = KEY_LENGTH / bits_per_digit;

        return toString(radix, pad_to_length);
    }

    @Override
    public int countSharedPrefixDigits(final IKey key, final int radix) {

        final String representation1 = toString(radix);
        final String representation2 = key.toString(radix);

        int shared_digit_count = 0;
        final int key_length = representation1.length();

        while (shared_digit_count < key_length && representation1.charAt(shared_digit_count) == representation2.charAt(shared_digit_count)) {
            shared_digit_count++;
        }

        return shared_digit_count;
    }

    @Override
    public char digit(final int index, final int radix) {

        return toString(radix).charAt(index);
    }

    @Override
    public String toStringAsKeyspaceFraction() {

        return key_value.multiply(ONE_HUNDRED).divide(KEYSPACE_SIZE) + "%";
    }

    @Override
    public String toStringAsAngle() {

        return key_value.multiply(THREE_SIXTY).divide(KEYSPACE_SIZE) + DEGREE_SYMBOL;
    }

    @Override
    public int bitLength() {

        return key_value.bitLength();
    }

    @Override
    public int maxBitLength() {

        return MAX_KEY_VALUE.bitLength();
    }

    // ***********************************************************************

    @Override
    public int compareTo(final IKey k) {

        return key_value.compareTo(k.keyValue());
    }

    // ***********************************************************************

    /**
     * Returns a string representation of the key value with radix {@link #DEFAULT_RADIX}.
     *
     * @return a string representation of the key value
     */
    @Override
    public String toString() {

        return toStringAsAngle();
    }

    @Override
    public boolean equals(final Object o) {

        try {
            final IKey k = (IKey) o;

            return o != null && key_value.equals(k.keyValue());

        }
        catch (final ClassCastException e) {
            return false;
        }
    }

    @Override
    public int hashCode() {

        return toString().hashCode();
    }

    // ***********************************************************************

    private String toString(final int radix, final int pad_to_length) {

        final StringBuilder result = new StringBuilder(key_value.toString(radix));
        while (result.length() < pad_to_length) {
            result.insert(0, '0');
        }
        return result.toString();
    }
}
