/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.nds.p2p.keys;

import java.math.BigInteger;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;

/**
 * Ring arithmetic calculations.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class RingArithmetic {

    /**
     * Prevent instantiation of utility class.
     */
    private RingArithmetic() {

    }

    /**
     * Calculates the ring distance from k1 to k2, where ringDistance(k, k) = 0, and in general ringDistance(k1, k2) != ringDistance(k2, k1).
     *
     * @param k1 the first key
     * @param k2 the second key
     * @return the ring distance from k1 to k2
     */
    public static BigInteger ringDistance(final IKey k1, final IKey k2) {

        BigInteger distance = k2.keyValue().subtract(k1.keyValue());

        // Check for ring wrap-around, indicated by negative distance.
        if (distance.compareTo(BigInteger.ZERO) < 0) {

            // The segment wraps around, so add the keyspace size.
            distance = distance.add(Key.KEYSPACE_SIZE);
        }

        return distance;
    }

    /**
     * Tests whether three keys lie in ring order.
     *
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     * @return true if, when traversing the key space in increasing key direction from k1, and allowing for wrap-around at zero, k3 is not encountered before k2
     */
    public static boolean inRingOrder(final IKey k1, final IKey k2, final IKey k3) {

        // If the first and last keys are the same then the keys are in order regardless of the second.
        return k1.equals(k3) || !ringDistanceFurther(k1, k2, k3);
    }

    /**
     * Tests whether the second key is further from the first in ring distance than the third is from the first.
     *
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     * @return true if the ring distance from k1 to k2 is greater than the ring distance from k1 to k3
     */
    public static boolean ringDistanceFurther(final IKey k1, final IKey k2, final IKey k3) {

        final BigInteger first_to_second = ringDistance(k1, k2);
        final BigInteger first_to_third = ringDistance(k1, k3);

        // If the first and last keys are the same then the keys are in order regardless of the second.
        return first_to_second.compareTo(first_to_third) > 0;
    }

    /**
     * Tests whether a key lies within a specified half-open ring segment.
     *
     * @param k1 the key after which the segment starts
     * @param k2 the key being tested
     * @param k3 the key at which the segment ends
     * @return true if k2 lies within the key segment starting after k1 and ending at k3
     */
    public static boolean inSegment(final IKey k1, final IKey k2, final IKey k3) {

        // If k1 = k3 then the segment is the whole ring so it doesn't matter what k2 is.
        // Otherwise, if k1 = k2 then the target lies just before the segment.
        return k1.equals(k3) || !k1.equals(k2) && inRingOrder(k1, k2, k3);
    }
}
