/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.keys.test;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;

/**
 * Tests for keys as bit strings. Tests for ring arithmetic are in {@link RingArithmeticTest}.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class KeyTest {

    private static final int HEX_RADIX = 16;

    private static final IKey ZERO_KEY = RingArithmeticTest.ZERO_KEY;
    private static final IKey ONE_KEY = RingArithmeticTest.ONE_KEY;
    private static final IKey ARBITRARY_KEY = RingArithmeticTest.ARBITRARY_KEY;
    private static final IKey MAX_KEY = RingArithmeticTest.MAX_KEY;

    private static final IKey QUARTER_KEY = RingArithmeticTest.QUARTER_KEY;
    private static final IKey HALF_KEY = RingArithmeticTest.HALF_KEY;
    private static final IKey THREE_QUARTER_KEY = RingArithmeticTest.THREE_QUARTER_KEY;

    /**
     * Tests whether the bit lengths of the keys are as expected.
     */
    @Test
    public void bitLength() {

        final int key_length = ZERO_KEY.maxBitLength();
        final int arbitrary_length = 42;

        assertEquals(ZERO_KEY.bitLength(), 0);
        assertEquals(ONE_KEY.bitLength(), 1);
        assertEquals(ARBITRARY_KEY.bitLength(), arbitrary_length);
        assertEquals(MAX_KEY.bitLength(), key_length);
    }

    /**
     * Tests whether key comparison works as expected.
     */
    @Test
    public void comparison() {

        assertEquals(MAX_KEY.compareTo(ZERO_KEY), 1);
        assertEquals(MAX_KEY.compareTo(ONE_KEY), 1);
        assertEquals(MAX_KEY.compareTo(ARBITRARY_KEY), 1);

        assertEquals(ZERO_KEY.compareTo(MAX_KEY), -1);
        assertEquals(ZERO_KEY.compareTo(ONE_KEY), -1);
        assertEquals(ZERO_KEY.compareTo(ARBITRARY_KEY), -1);

        assertEquals(ZERO_KEY.compareTo(ZERO_KEY), 0);
        assertEquals(MAX_KEY.compareTo(MAX_KEY), 0);

        // Miscellaneous pairs.
        assertEquals(ARBITRARY_KEY.compareTo(MAX_KEY), -1);
        assertEquals(ARBITRARY_KEY.compareTo(ZERO_KEY), 1);
        assertEquals(ARBITRARY_KEY.compareTo(ONE_KEY), 1);
        assertEquals(ARBITRARY_KEY.compareTo(ARBITRARY_KEY), 0);
    }

    /**
     * Tests whether the hex representations of keys are as expected.
     */
    @Test
    public void toStringTest() {

        // 40 hex digits.
        assertEquals(ZERO_KEY.toString(HEX_RADIX), "0000000000000000000000000000000000000000");
        assertEquals(ONE_KEY.toString(HEX_RADIX), "0000000000000000000000000000000000000001");
        assertEquals(ARBITRARY_KEY.toString(HEX_RADIX), "000000000000000000000000000002f4315d8102");
        assertEquals(QUARTER_KEY.toString(HEX_RADIX), "4000000000000000000000000000000000000000");
        assertEquals(HALF_KEY.toString(HEX_RADIX), "8000000000000000000000000000000000000000");
        assertEquals(THREE_QUARTER_KEY.toString(HEX_RADIX), "c000000000000000000000000000000000000000");
        assertEquals(MAX_KEY.toString(HEX_RADIX), "ffffffffffffffffffffffffffffffffffffffff");
    }

    /**
     * Tests whether the percentage representations of keys are as expected.
     */
    @Test
    public void toStringAsKeyspaceFraction() {

        assertEquals(ZERO_KEY.toStringAsKeyspaceFraction(), "0%");
        assertEquals(ONE_KEY.toStringAsKeyspaceFraction(), "0%");
        assertEquals(ARBITRARY_KEY.toStringAsKeyspaceFraction(), "0%");
        assertEquals(QUARTER_KEY.toStringAsKeyspaceFraction(), "25%");
        assertEquals(HALF_KEY.toStringAsKeyspaceFraction(), "50%");
        assertEquals(THREE_QUARTER_KEY.toStringAsKeyspaceFraction(), "75%");
        assertEquals(MAX_KEY.toStringAsKeyspaceFraction(), "99%");
    }

    /**
     * Tests whether the angular representations of keys are as expected.
     */
    @Test
    public void toStringAsAngle() {

        assertEquals(ZERO_KEY.toStringAsAngle(), "0" + Key.DEGREE_SYMBOL);
        assertEquals(ONE_KEY.toStringAsAngle(), "0" + Key.DEGREE_SYMBOL);
        assertEquals(ARBITRARY_KEY.toStringAsAngle(), "0" + Key.DEGREE_SYMBOL);
        assertEquals(QUARTER_KEY.toStringAsAngle(), "90" + Key.DEGREE_SYMBOL);
        assertEquals(HALF_KEY.toStringAsAngle(), "180" + Key.DEGREE_SYMBOL);
        assertEquals(THREE_QUARTER_KEY.toStringAsAngle(), "270" + Key.DEGREE_SYMBOL);
        assertEquals(MAX_KEY.toStringAsAngle(), "359" + Key.DEGREE_SYMBOL);
    }

    /**
     * Tests digit prefix calculation.
     */
    @Test
    public void countSharedPrefixDigits() {

        // Check that zero and one have 39 hex digits in common.
        final int all_but_one_digit = 39;
        final int arbitrary_digits = 29;

        assertEquals(ZERO_KEY.countSharedPrefixDigits(ONE_KEY, HEX_RADIX), all_but_one_digit);

        // Check that FFF..F and one have 0 digits in common.
        assertEquals(MAX_KEY.countSharedPrefixDigits(ZERO_KEY, HEX_RADIX), 0);

        // Check that 2f4315d8102 and one have 1 digits in common.
        assertEquals(ARBITRARY_KEY.countSharedPrefixDigits(ONE_KEY, HEX_RADIX), arbitrary_digits);

        // Same checks in reverse.

        assertEquals(ONE_KEY.countSharedPrefixDigits(ZERO_KEY, HEX_RADIX), all_but_one_digit);
        assertEquals(ZERO_KEY.countSharedPrefixDigits(MAX_KEY, HEX_RADIX), 0);
        assertEquals(ONE_KEY.countSharedPrefixDigits(ARBITRARY_KEY, HEX_RADIX), arbitrary_digits);
    }

    /**
     * Tests digit extraction.
     */
    @Test
    public void digit() {

        final int last_but_2 = 37;
        final int last_but_1 = 38;
        final int last = 39;

        assertEquals(ARBITRARY_KEY.digit(0, HEX_RADIX), '0');
        assertEquals(ARBITRARY_KEY.digit(1, HEX_RADIX), '0');
        assertEquals(ARBITRARY_KEY.digit(last_but_2, HEX_RADIX), '1');
        assertEquals(ARBITRARY_KEY.digit(last_but_1, HEX_RADIX), '0');
        assertEquals(ARBITRARY_KEY.digit(last, HEX_RADIX), '2');
    }
}
