/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.nds.p2p.keys.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertTrue;

import java.math.BigInteger;

import org.junit.Test;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;
import uk.ac.standrews.cs.nds.p2p.keys.RingArithmetic;

/**
 * Tests ring arithmetic.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RingArithmeticTest {

    private static final BigInteger MINUS_ONE = new BigInteger("-1");
    private static final BigInteger ZERO = BigInteger.ZERO;
    private static final BigInteger ONE = BigInteger.ONE;
    private static final BigInteger TWO = new BigInteger("2");
    private static final BigInteger THREE = new BigInteger("3");
    private static final BigInteger FOUR = new BigInteger("4");

    private static final BigInteger ARBITRARY_VALUE = new BigInteger("3247823487234");
    private static final BigInteger MAX_VALUE = Key.KEYSPACE_SIZE.subtract(BigInteger.ONE);

    private static final BigInteger QUARTER_VALUE = Key.KEYSPACE_SIZE.divide(FOUR);
    private static final BigInteger HALF_VALUE = QUARTER_VALUE.multiply(TWO);
    private static final BigInteger THREE_QUARTER_VALUE = QUARTER_VALUE.multiply(THREE);

    static final IKey ZERO_KEY = new Key(BigInteger.ZERO);
    static final IKey ONE_KEY = new Key(BigInteger.ONE);
    static final IKey ARBITRARY_KEY = new Key(ARBITRARY_VALUE);
    static final IKey MAX_KEY = new Key(MAX_VALUE);

    static final IKey QUARTER_KEY = new Key(QUARTER_VALUE);
    static final IKey HALF_KEY = new Key(HALF_VALUE);
    static final IKey THREE_QUARTER_KEY = new Key(THREE_QUARTER_VALUE);

    /**
     * Tests whether keys created with values within the key range retain their original values.
     */
    @Test
    public void creation() {

        assertEquals(ZERO_KEY.keyValue(), BigInteger.ZERO);
        assertEquals(ONE_KEY.keyValue(), BigInteger.ONE);
        assertEquals(ARBITRARY_KEY.keyValue(), ARBITRARY_VALUE);
        assertEquals(MAX_KEY.keyValue(), MAX_VALUE);
    }

    /**
     * Tests whether keys created with values outwith the key range are wrapped as expected.
     */
    @Test
    public void creationWithValuesOutwithKeyRange() {

        // -1 should wrap to keyspace_size - 1.
        assertEquals(new Key(new BigInteger("-1")).keyValue(), MAX_VALUE);

        // keyspace_size should wrap to 0.
        assertEquals(new Key(Key.KEYSPACE_SIZE).keyValue(), ZERO);

        // keyspace_size + 1 should wrap to 1.
        assertEquals(new Key(Key.KEYSPACE_SIZE.add(ONE)).keyValue(), ONE);

        // 2 * keyspace_size + 1 should wrap to 1.
        assertEquals(new Key(Key.KEYSPACE_SIZE.multiply(TWO).add(ONE)).keyValue(), ONE);

        // -2 * keyspace_size + 1 should wrap to 1.
        assertEquals(new Key(Key.KEYSPACE_SIZE.multiply(MINUS_ONE).multiply(TWO).add(ONE)).keyValue(), ONE);
    }

    /**
     * Tests whether ring distance calculation works as expected.
     */
    @Test
    public void ringDistanceGeneral() {

        final IKey[] test_keys = new IKey[]{ZERO_KEY, ONE_KEY, ARBITRARY_KEY, MAX_KEY, QUARTER_KEY, HALF_KEY, THREE_QUARTER_KEY};

        assertRingDistanceInvariantsForSingleKeys(test_keys);
        assertRingDistanceInvariantsForPairsOfDifferentKeys(test_keys);
    }

    /**
     * Tests whether ring distance calculation works as expected.
     */
    @Test
    public void ringDistanceSpecific() {

        // Distances from keyspace_size - 1.
        assertEquals(RingArithmetic.ringDistance(MAX_KEY, ZERO_KEY), BigInteger.ONE);
        assertEquals(RingArithmetic.ringDistance(MAX_KEY, ONE_KEY), Key.TWO);
        assertEquals(RingArithmetic.ringDistance(MAX_KEY, ARBITRARY_KEY), ARBITRARY_VALUE.add(ONE));

        // Distances from 0.
        assertEquals(RingArithmetic.ringDistance(ZERO_KEY, ONE_KEY), BigInteger.ONE);
        assertEquals(RingArithmetic.ringDistance(ZERO_KEY, ARBITRARY_KEY), ARBITRARY_VALUE);
        assertEquals(RingArithmetic.ringDistance(ZERO_KEY, MAX_KEY), MAX_VALUE);

        // Miscellaneous pairs.
        assertEquals(RingArithmetic.ringDistance(QUARTER_KEY, HALF_KEY), QUARTER_VALUE);
        assertEquals(RingArithmetic.ringDistance(HALF_KEY, THREE_QUARTER_KEY), QUARTER_VALUE);
        assertEquals(RingArithmetic.ringDistance(THREE_QUARTER_KEY, ZERO_KEY), QUARTER_VALUE);
        assertEquals(RingArithmetic.ringDistance(QUARTER_KEY, THREE_QUARTER_KEY), HALF_VALUE);
        assertEquals(RingArithmetic.ringDistance(THREE_QUARTER_KEY, QUARTER_KEY), HALF_VALUE);
        assertEquals(RingArithmetic.ringDistance(HALF_KEY, QUARTER_KEY), THREE_QUARTER_VALUE);
    }

    /**
     * Tests whether ring ordering calculation works as expected.
     */
    @Test
    public void ringOrderingGeneral() {

        final IKey[] test_keys = new IKey[]{ZERO_KEY, ONE_KEY, ARBITRARY_KEY, MAX_KEY, QUARTER_KEY, HALF_KEY, THREE_QUARTER_KEY};

        assertRingOrderingInvariantsForSingleKeys(test_keys);
        assertRingOrderingInvariantsForTriplesOfDifferentKeys(test_keys);
    }

    /**
     * Tests whether ring ordering calculation works as expected.
     */
    @Test
    public void ringOrderingSpecific() {

        assertRingOrderingInvariantsForOrderedKeys(MAX_KEY, ZERO_KEY, ONE_KEY);
        assertRingOrderingInvariantsForOrderedKeys(MAX_KEY, ONE_KEY, ARBITRARY_KEY);
        assertRingOrderingInvariantsForOrderedKeys(MAX_KEY, ZERO_KEY, ARBITRARY_KEY);

        assertRingOrderingInvariantsForOrderedKeys(ZERO_KEY, ONE_KEY, ARBITRARY_KEY);
        assertRingOrderingInvariantsForOrderedKeys(ZERO_KEY, ARBITRARY_KEY, MAX_KEY);
        assertRingOrderingInvariantsForOrderedKeys(ZERO_KEY, ONE_KEY, MAX_KEY);

        assertRingOrderingInvariantsForOrderedKeys(ARBITRARY_KEY, MAX_KEY, ZERO_KEY);
        assertRingOrderingInvariantsForOrderedKeys(ARBITRARY_KEY, ZERO_KEY, ONE_KEY);
        assertRingOrderingInvariantsForOrderedKeys(ARBITRARY_KEY, MAX_KEY, ONE_KEY);

        assertRingOrderingInvariantsForOrderedKeys(QUARTER_KEY, HALF_KEY, THREE_QUARTER_KEY);
    }

    /**
     * Tests whether ring ordering calculation works as expected.
     */
    @Test
    public void ringSegmentsGeneral() {

        final IKey[] test_keys = new IKey[]{ZERO_KEY, ONE_KEY, ARBITRARY_KEY, MAX_KEY, QUARTER_KEY, HALF_KEY, THREE_QUARTER_KEY};

        assertRingSegmentInvariantsForSingleKeys(test_keys);
        assertRingSegmentInvariantsForTriplesOfDifferentKeys(test_keys);
    }

    /**
     * Tests whether segment calculation works as expected.
     */
    @Test
    public void segmentsSpecific() {

        assertRingSegmentInvariantsForOrderedDifferentKeys(MAX_KEY, ZERO_KEY, ONE_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(MAX_KEY, ONE_KEY, ARBITRARY_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(MAX_KEY, ZERO_KEY, ARBITRARY_KEY);

        assertRingSegmentInvariantsForOrderedDifferentKeys(ZERO_KEY, ONE_KEY, ARBITRARY_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(ZERO_KEY, ARBITRARY_KEY, MAX_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(ZERO_KEY, ONE_KEY, MAX_KEY);

        assertRingSegmentInvariantsForOrderedDifferentKeys(ARBITRARY_KEY, MAX_KEY, ZERO_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(ARBITRARY_KEY, ZERO_KEY, ONE_KEY);
        assertRingSegmentInvariantsForOrderedDifferentKeys(ARBITRARY_KEY, MAX_KEY, ONE_KEY);

        assertRingSegmentInvariantsForOrderedDifferentKeys(QUARTER_KEY, HALF_KEY, THREE_QUARTER_KEY);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests conditions that should hold for all key values.
     * @param k a key
     */
    private void assertRingDistanceInvariants(final IKey k) {

        assertEquals(RingArithmetic.ringDistance(k, k), ZERO);
        assertEquals(RingArithmetic.ringDistance(k, offsetKey(k, ONE)), ONE);
        assertEquals(RingArithmetic.ringDistance(k, offsetKey(k, MINUS_ONE)), MAX_VALUE);
    }

    /**
     * Tests conditions that should hold for all pairs of different key values.
     * @param k1 the first key
     * @param k2 the second key
     */
    private void assertRingDistanceInvariantsForDifferentKeys(final IKey k1, final IKey k2) {

        // Check assumption for this test.
        assertNotSame(k1, k2);

        final BigInteger distance_forwards = RingArithmetic.ringDistance(k1, k2);
        final BigInteger distance_backwards = RingArithmetic.ringDistance(k2, k1);

        // Distances must themselves lie within the key range.
        assertWithinKeyRange(distance_forwards);
        assertWithinKeyRange(distance_backwards);

        // For different keys, the sum of the two distances must equal the ring size.
        assertTrue(distance_forwards.add(distance_backwards).equals(Key.KEYSPACE_SIZE));
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests conditions that should hold for all key values.
     * @param k a key
     */
    private void assertRingOrderingInvariants(final IKey k) {

        final IKey k_plus_one = offsetKey(k, ONE);
        final IKey k_plus_two = offsetKey(k, TWO);

        // Elaborate the combinations of k/k+1 in each of the 3 positions.

        assertTrue(RingArithmetic.inRingOrder(k, k, k));
        assertTrue(RingArithmetic.inRingOrder(k, k, k_plus_one));
        assertTrue(RingArithmetic.inRingOrder(k, k_plus_one, k));
        assertTrue(RingArithmetic.inRingOrder(k, k_plus_one, k_plus_one));
        assertTrue(RingArithmetic.inRingOrder(k_plus_one, k, k));
        assertTrue(RingArithmetic.inRingOrder(k_plus_one, k, k_plus_one));
        assertTrue(RingArithmetic.inRingOrder(k_plus_one, k_plus_one, k));
        assertTrue(RingArithmetic.inRingOrder(k_plus_one, k_plus_one, k_plus_one));

        assertTrue(RingArithmetic.inRingOrder(k, k_plus_one, k_plus_two));
        assertFalse(RingArithmetic.inRingOrder(k, k_plus_two, k_plus_one));
    }

    /**
     * Tests conditions that should hold for all triples of different key values.
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     */
    private void assertRingOrderingInvariantsForArbitraryKeys(final IKey k1, final IKey k2, final IKey k3) {

        assertTrue(xor(RingArithmetic.inRingOrder(k1, k2, k3), RingArithmetic.inRingOrder(k1, k3, k2)));

        if (RingArithmetic.inRingOrder(k1, k2, k3)) {
            assertRingOrderingInvariantsForOrderedKeys(k1, k2, k3);
        }
        else {
            assertRingOrderingInvariantsForOrderedKeys(k1, k3, k2);
        }
    }

    /**
     * Tests conditions that should hold for triples of different key values that are ordered.
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     */
    private void assertRingOrderingInvariantsForOrderedKeys(final IKey k1, final IKey k2, final IKey k3) {

        assertTrue(RingArithmetic.inRingOrder(k1, k2, k3));
        assertTrue(RingArithmetic.inRingOrder(k2, k3, k1));
        assertTrue(RingArithmetic.inRingOrder(k3, k1, k2));

        assertTrue(RingArithmetic.ringDistanceFurther(k1, k3, k2));
        assertTrue(RingArithmetic.ringDistanceFurther(k3, k2, k1));
        assertTrue(RingArithmetic.ringDistanceFurther(k2, k1, k3));

        assertFalse(RingArithmetic.ringDistanceFurther(k1, k2, k3));
        assertFalse(RingArithmetic.ringDistanceFurther(k2, k3, k1));
        assertFalse(RingArithmetic.ringDistanceFurther(k3, k1, k2));
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests conditions that should hold for all key values.
     * @param k a key
     */
    private void assertRingSegmentInvariants(final IKey k) {

        final IKey k_plus_one = offsetKey(k, ONE);
        final IKey k_plus_two = offsetKey(k, TWO);

        // Elaborate the combinations of k/k+1 in each of the 3 positions.

        assertTrue(RingArithmetic.inSegment(k, k, k));
        assertFalse(RingArithmetic.inSegment(k, k, k_plus_one));
        assertTrue(RingArithmetic.inSegment(k, k_plus_one, k));
        assertTrue(RingArithmetic.inSegment(k, k_plus_one, k_plus_one));
        assertTrue(RingArithmetic.inSegment(k_plus_one, k, k));
        assertTrue(RingArithmetic.inSegment(k_plus_one, k, k_plus_one));
        assertFalse(RingArithmetic.inSegment(k_plus_one, k_plus_one, k));
        assertTrue(RingArithmetic.inSegment(k_plus_one, k_plus_one, k_plus_one));

        assertTrue(RingArithmetic.inSegment(k, k_plus_one, k_plus_two));
        assertFalse(RingArithmetic.inSegment(k, k_plus_two, k_plus_one));
    }

    /**
     * Tests conditions that should hold for all triples of different key values.
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     */
    private void assertRingSegmentInvariantsForArbitraryKeys(final IKey k1, final IKey k2, final IKey k3) {

        assertTrue(xor(RingArithmetic.inSegment(k1, k2, k3), RingArithmetic.inSegment(k1, k3, k2)));

        if (RingArithmetic.inRingOrder(k1, k2, k3)) {
            assertRingSegmentInvariantsForOrderedDifferentKeys(k1, k2, k3);
        }
        else {
            assertRingSegmentInvariantsForOrderedDifferentKeys(k1, k3, k2);
        }
    }

    /**
     * Tests conditions that should hold for triples of different key values that are ordered.
     * @param k1 the first key
     * @param k2 the second key
     * @param k3 the third key
     */
    private void assertRingSegmentInvariantsForOrderedDifferentKeys(final IKey k1, final IKey k2, final IKey k3) {

        // Check assumption for this test.
        assertTrue(allDifferent(k1, k2, k3));

        assertTrue(RingArithmetic.inSegment(k1, k2, k3));
        assertTrue(RingArithmetic.inSegment(k2, k3, k1));
        assertTrue(RingArithmetic.inSegment(k3, k1, k2));
    }

    // -------------------------------------------------------------------------------------------------------

    private boolean xor(final boolean a, final boolean b) {

        return a && !b || !a && b;
    }

    private void assertWithinKeyRange(final BigInteger value) {

        assertTrue(value.compareTo(ZERO) >= 0);
        assertTrue(value.compareTo(Key.KEYSPACE_SIZE) < 0);
    }

    private void assertRingDistanceInvariantsForSingleKeys(final IKey[] keys) {

        for (final IKey k : keys) {
            assertRingDistanceInvariants(k);
        }
    }

    private void assertRingDistanceInvariantsForPairsOfDifferentKeys(final IKey[] keys) {

        for (final IKey k1 : keys) {
            for (final IKey k2 : keys) {

                if (!k1.equals(k2)) {
                    assertRingDistanceInvariantsForDifferentKeys(k1, k2);
                }
            }
        }
    }

    private void assertRingOrderingInvariantsForSingleKeys(final IKey[] keys) {

        for (final IKey k : keys) {
            assertRingOrderingInvariants(k);
        }
    }

    private void assertRingOrderingInvariantsForTriplesOfDifferentKeys(final IKey[] keys) {

        for (final IKey k1 : keys) {
            for (final IKey k2 : keys) {
                for (final IKey k3 : keys) {
                    if (allDifferent(k1, k2, k3)) {
                        assertRingOrderingInvariantsForArbitraryKeys(k1, k2, k3);
                    }
                }
            }
        }
    }

    private void assertRingSegmentInvariantsForSingleKeys(final IKey[] keys) {

        for (final IKey k : keys) {
            assertRingSegmentInvariants(k);
        }
    }

    private void assertRingSegmentInvariantsForTriplesOfDifferentKeys(final IKey[] keys) {

        for (final IKey k1 : keys) {
            for (final IKey k2 : keys) {
                for (final IKey k3 : keys) {
                    if (allDifferent(k1, k2, k3)) {
                        assertRingSegmentInvariantsForArbitraryKeys(k1, k2, k3);
                    }
                }
            }
        }
    }

    private boolean allDifferent(final IKey k1, final IKey k2, final IKey k3) {

        return !(k1.equals(k2) || k1.equals(k3) || k2.equals(k3));
    }

    private IKey offsetKey(final IKey k, final BigInteger i) {

        return new Key(k.keyValue().add(i));
    }
}
