/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.nds.p2p.network;

import java.math.BigInteger;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;

/**
 * Options for distribution of a set of keys in key space.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public enum KeyDistribution {

    /**
     * Keys randomly distributed in the key space.
     */
    RANDOM,

    /**
     * Keys evenly distributed around the key space.
     */
    EVEN,

    /**
     * Keys clustered tightly in one region of the key space.
     */
    CLUSTERED;

    // -------------------------------------------------------------------------------------------------------

    public IKey[] generateKeys(final int number_of_keys) {

        final IKey[] node_keys = new IKey[number_of_keys];

        // Leave null entries for RANDOM since the keys will be generated by hashing addresses.
        if (!equals(RANDOM)) {

            BigInteger key_value = null;
            BigInteger node_increment = null;

            if (equals(EVEN)) {

                key_value = new BigInteger("0");
                node_increment = Key.KEYSPACE_SIZE.divide(new BigInteger(String.valueOf(number_of_keys)));
            }

            if (equals(CLUSTERED)) {

                // Span the wrap-over point just to be awkward.
                key_value = Key.KEYSPACE_SIZE.subtract(new BigInteger(String.valueOf(number_of_keys / 2)));
                node_increment = new BigInteger("1");
            }

            for (int i = 0; i < number_of_keys; i++) {
                node_keys[i] = new Key(key_value);
                key_value = key_value.add(node_increment);
            }
        }

        return node_keys;
    }
}
