/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.network;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import uk.ac.standrews.cs.nds.madface.ApplicationManager;
import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.HostState;
import uk.ac.standrews.cs.nds.madface.exceptions.DeploymentException;
import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class P2PNodeManager extends ApplicationManager {

    protected abstract P2PNodeFactory getP2PNodeFactory();

    private final boolean local_deployment_only;
    private Map<String, String> java_bin_paths;

    /**
     * Initializes a node manager.
     * 
     * @param local_deployment_only true if nodes are only to be deployed to the local node
     */
    public P2PNodeManager(final boolean local_deployment_only) {

        this.local_deployment_only = local_deployment_only;
        initJavaBinPaths();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void deployApplication(final HostDescriptor host_descriptor, final Object... args) throws Exception {

        Diagnostic.traceNoSource(DiagnosticLevel.FULL, "deploying application to: " + host_descriptor.getHost());
        setJavaBinPath(host_descriptor);

        getP2PNodeFactory().createNode(host_descriptor, getKey(args));

        // If locally deployed, assume deployment was successful and set the host descriptor state directly to speed things up.
        if (local_deployment_only) {
            host_descriptor.hostState(HostState.RUNNING);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    private void setJavaBinPath(final HostDescriptor host_descriptor) {

        if (host_descriptor.getJavaBinPath() == null) {
            final String path = java_bin_paths.get(host_descriptor.getHost());
            if (path != null) {
                host_descriptor.javaBinPath(new File(path));
            }
        }
    }

    private IKey getKey(final Object... args) throws DeploymentException {

        if (args == null || args.length == 0 || args[0] == null) { return null; }

        final Object arg = args[0];
        if (arg instanceof IKey) { return (IKey) arg; }
        throw new DeploymentException("argument not of type IKey");
    }

    private void initJavaBinPaths() {

        java_bin_paths = new HashMap<String, String>();

        java_bin_paths.put("teaching-1.cs.st-andrews.ac.uk", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-2.cs.st-andrews.ac.uk", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-3.cs.st-andrews.ac.uk", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-4.cs.st-andrews.ac.uk", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-1", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-2", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-3", "/usr/local/jdk/bin");
        java_bin_paths.put("teaching-4", "/usr/local/jdk/bin");

        for (int i = 0; i < 60; i++) {
            java_bin_paths.put("compute-0-" + i, "/usr/java/latest/bin");
        }
    }
}
