/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.util;

import java.net.InetSocketAddress;

import uk.ac.standrews.cs.nds.p2p.interfaces.IHash;
import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;

/**
 * Provides various ways to generate keys.
 *
 * @author Stuart Norcross (stuart@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class HashBasedKeyFactory {

    private final IHash hash;
    private final RandomGenerator random_generator;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a key factory using a given hashing function.
     *
     * @param hash the hash function
     */
    public HashBasedKeyFactory(final IHash hash) {

        this.hash = hash;
        random_generator = new RandomGenerator(hash);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a key with a value generated by hashing the given byte array.
     *
     * @param bytes the array from which to generate the key's value
     * @return a key with a value generated from bytes
     */
    public IKey generateKey(final byte[] bytes) {

        final byte[] hashed = hash.hash(bytes);
        return new Key(hashed);
    }

    /**
     * Creates a key with a value generated by hashing the given string.
     *
     * @param s the string from which to generate the key's value
     * @return a key with a value generated from s
     */
    public IKey generateKey(final String s) {

        return generateKey(s.getBytes());
    }

    /**
     * Creates a key with a value generated by hashing the given IP address.
     *
     * @param ip_address the IP address from which to generate the key's value
     * @return a key with a value generated from ip_address
     */
    public IKey generateKey(final InetSocketAddress ip_address) {

        final String s = ip_address.getAddress().getHostAddress() + ip_address.getPort();
        return generateKey(s);
    }

    /**
     * Creates a key with a value represented by the given string. The representation is in the same
     * format as obtained by calling toString() on a key.
     *
     * @param s the representation of a key value
     * @return a key with the given value
     */
    public IKey recreateKey(final String s) {

        return new Key(s);
    }

    /**
     * Creates a key with a pseudo-random value. The method {@link #setDeterministic(boolean)} controls whether the sequence
     * generated by successive calls is deterministic.
     *
     * @return a key with a pseudo-random value
     */
    public IKey generateRandomKey() {

        final byte[] bytes = random_generator.randomBytes(hash.getHashSize());

        return generateKey(bytes);
    }

    /**
     * Controls whether the sequence of values generated by successive calls to {@link #generateRandomKey()} is deterministic.
     *
     * @param deterministic true if the sequence should be deterministic
     */
    public void setDeterministic(final boolean deterministic) {

        random_generator.setDeterministic(deterministic);
    }
}
