/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.registry;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Provides access to the registry for a local or remote machine.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Angus Macdonald (angus@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class AbstractRegistryFactory {

    private static final long SLEEP_TIME = 100;
    private static final long TIMEOUT = 10000;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the currently running registry for the local machine.
     *
     * @return the local registry
     * @throws UnknownHostException if no local network device is available
     * @throws RegistryUnavailableException if the local registry is not accessible
     */
    public IRegistry getRegistry() throws UnknownHostException, RegistryUnavailableException {

        return getRegistry(false);
    }

    /**
     * Returns the currently running registry for the a remote machine.
     *
     * @param host the remote address
     * @return the remote registry
     * @throws RegistryUnavailableException if the remote registry is not accessible
     */
    public abstract IRegistry getRegistry(final InetAddress host) throws RegistryUnavailableException;

    /**
     * Returns the currently running registry for the local machine, optionally attempting to start the registry if it is not accessible.
     *
     * @param start_if_unavailable true if the registry should be started if not accessible
     * @return the local registry
     * @throws UnknownHostException if no local network device is available
     * @throws RegistryUnavailableException if the local registry is not accessible and cannot be started within the timeout period @{link {@link #TIMEOUT}.
     */
    public IRegistry getRegistry(final boolean start_if_unavailable) throws UnknownHostException, RegistryUnavailableException {

        final long start_time = System.currentTimeMillis();

        while (!Thread.currentThread().isInterrupted()) {

            // Try to bind to a registry that is already running.
            try {
                return getLocalRegistry();

                // There is a local registry running and we've made a successful RPC call to it.
            }
            catch (final UnknownHostException e) {
                ErrorHandling.hardExceptionError(e, "Can't bind to local device");
            }
            catch (final RegistryUnavailableException e) {

                if (!start_if_unavailable) { throw e; }

                try {
                    // No registry currently running so try to start one.
                    return startNewRegistry();
                }
                catch (final Exception e1) {
                    Diagnostic.trace(DiagnosticLevel.FULL, "Exception while trying to start registry: ", e1.getMessage());
                }
                finally {

                    // Cases:
                    // 1. Another registry may have started since we first tried to bind to registry.
                    // 2. Something else is bound to the registry port.
                    // In either case, sleep and retry up to time limit.

                    sleepAndCheckTimeout(start_time);
                }
            }
        }
        throw new RegistryUnavailableException("interrupted");
    }

    // -------------------------------------------------------------------------------------------------------

    protected abstract IRegistry startNewRegistry() throws IOException, RPCException, AlreadyBoundException, RegistryUnavailableException;

    private static void sleepAndCheckTimeout(final long start_time) throws RegistryUnavailableException {

        try {
            Thread.sleep(SLEEP_TIME);
        }
        catch (final InterruptedException e2) {
            // Ignore.
        }

        final long time_elapsed = System.currentTimeMillis() - start_time;
        if (time_elapsed > TIMEOUT) { throw new RegistryUnavailableException("Can't create local registry or bind to existing registry"); }
    }

    private IRegistry getLocalRegistry() throws UnknownHostException, RegistryUnavailableException {

        return getRegistry(NetworkUtil.getLocalIPv4Address());
    }
}
