/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.registry.nostream;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.nostream.Marshaller;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;

/**
 * Provides methods serializing and deserializing registry types to/from JSON strings.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RegistryMarshaller extends Marshaller {

    private static final String KEYS = "keys";
    private static final String VALUES = "values";

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a set of strings to an array.
     *
     * @param source the set
     * @return a JSON object
     */
    public JSONValue serializeSetString(final Set<String> source) {

        final JSONArray array = new JSONArray();
        for (final String s : source) {
            array.put(s);
        }
        return array;
    }

    /**
     * Deserializes a set of strings.
     *
     * @param array a JSON array containing the values
     * @return a set of strings
     * @throws DeserializationException if the representation is invalid
     */
    public Set<String> deserializeSetString(final JSONArray array) throws DeserializationException {

        if (array == null) { return null; }

        try {
            final Set<String> result = new HashSet<String>();
            for (int i = 0; i < array.length(); i++) {
                result.add(array.getString(i));
            }
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a collection of ints to an array.
     *
     * @param source the collection
     * @return a JSON object
     */
    public JSONValue serializeCollectionInteger(final Collection<Integer> source) {

        final JSONArray array = new JSONArray();
        for (final Integer instance : source) {
            array.put(instance);
        }
        return array;
    }

    /**
     * Deserializes a collection of ints.
     *
     * @param array a JSON array containing the values
     * @return a collection of ints
     * @throws DeserializationException if the representation is invalid
     */
    public Collection<Integer> deserializeCollectionInteger(final JSONArray array) throws DeserializationException {

        if (array == null) { return null; }

        try {
            final List<Integer> result = new ArrayList<Integer>();
            for (int i = 0; i < array.length(); i++) {
                result.add(array.getInt(i));
            }
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a map from strings to ints, to an object containing an array of keys and an array of values.
     *
     * @param source the map
     * @return a JSON object
     */
    public JSONValue serializeMapStringInt(final Map<String, Integer> source) {

        final JSONObject object = new JSONObject();

        final JSONValue key_array = serializeSetString(source.keySet());
        final JSONValue value_array = serializeCollectionInteger(source.values());

        object.put(KEYS, key_array);
        object.put(VALUES, value_array);

        return object;
    }

    /**
     * Deserializes a map from strings to ints.
     *
     * @param object a JSON object containing keys and values
     * @return a map from strings to ints
     * @throws DeserializationException if the representation is invalid
     */
    public Map<String, Integer> deserializeMapStringInt(final JSONObject object) throws DeserializationException {

        if (object == null) { return null; }

        try {
            final Set<String> keys = deserializeSetString(object.getJSONArray(KEYS));
            final Collection<Integer> values = deserializeCollectionInteger(object.getJSONArray(VALUES));

            final Map<String, Integer> result = new HashMap<String, Integer>();

            final Iterator<String> it1 = keys.iterator();
            final Iterator<Integer> it2 = values.iterator();
            while (it1.hasNext()) {
                if (!it2.hasNext()) { throw new DeserializationException("unbalanced elements in Map keys and values"); }
                final String arg0 = it1.next();
                final Integer arg1 = it2.next();
                result.put(arg0, arg1);
            }

            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }
}
