/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.registry.nostream;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Map;

import uk.ac.standrews.cs.nds.registry.AlreadyBoundException;
import uk.ac.standrews.cs.nds.registry.IRegistry;
import uk.ac.standrews.cs.nds.registry.NotBoundException;
import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.nostream.Marshaller;
import uk.ac.standrews.cs.nds.rpc.nostream.NoStreamProxy;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;

/**
 * Proxy for remotely accessible registry.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Angus Macdonald (angus@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class RegistryProxy extends NoStreamProxy implements IRegistry {

    private static final Map<InetSocketAddress, RegistryProxy> PROXY_MAP;

    private final RegistryMarshaller marshaller = new RegistryMarshaller();

    static {
        PROXY_MAP = new HashMap<InetSocketAddress, RegistryProxy>();
    }

    // -------------------------------------------------------------------------------------------------------

    private RegistryProxy(final InetSocketAddress node_address) {

        super(node_address);
    }

    // -------------------------------------------------------------------------------------------------------

    public static synchronized RegistryProxy getProxy(final InetSocketAddress proxy_address) {

        RegistryProxy proxy = PROXY_MAP.get(proxy_address);
        if (proxy == null) {
            proxy = new RegistryProxy(proxy_address);
            PROXY_MAP.put(proxy_address, proxy);
        }
        return proxy;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Marshaller getMarshaller() {

        return marshaller;
    }

    @Override
    public void bind(final String key, final int port) throws RPCException, AlreadyBoundException {

        try {
            final JSONArray args = new JSONArray();
            args.put(key);
            args.put(port);
            makeCall("bind", args);
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final AlreadyBoundException e) {
            throw e;
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }

    @Override
    public void rebind(final String key, final int port) throws RPCException {

        try {
            final JSONArray args = new JSONArray();
            args.put(key);
            args.put(port);
            makeCall("rebind", args);
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }

    @Override
    public int lookup(final String key) throws RPCException {

        try {
            final JSONArray args = new JSONArray();
            args.put(key);
            return (Integer) makeCall("lookup", args).getValue();
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            dealWithException(e);
            return -1; //not reached.
        }
    }

    @Override
    public void unbind(final String key) throws RPCException, NotBoundException {

        try {
            final JSONArray args = new JSONArray();
            args.put(key);
            makeCall("unbind", args);
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final NotBoundException e) {
            throw e;
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }

    @Override
    public Map<String, Integer> getEntries() throws RPCException {

        try {
            return marshaller.deserializeMapStringInt((JSONObject) makeCall("getEntries"));
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            dealWithException(e);
            return null; // not reached
        }
    }

    @Override
    public void shutdown() throws RPCException {

        try {
            makeCall("shutdown");
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }
}
