/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.registry.stream;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.stream.Marshaller;

/**
 * Provides methods serializing and deserializing registry types to/from JSON strings.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class RegistryMarshaller extends Marshaller {

    private static final String KEYS = "keys";
    private static final String VALUES = "values";

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a set of strings to an array.
     *
     * @param source the set
     * @param writer writes the serialzed string set
     * @throws JSONException If the value is out of sequence
     */
    public void serializeSetString(final Set<String> source, final JSONWriter writer) throws JSONException {

        writer.array();
        for (final String s : source) {
            writer.value(s);
        }
        writer.endArray();
    }

    /**
     * Deserializes a set of strings.
     *
     * @param reader a stream starting with a JSON array containing the values
     * @return a set of strings
     * @throws DeserializationException if the representation is invalid
     */
    public Set<String> deserializeSetString(final JSONReader reader) throws DeserializationException {

        try {
            reader.array();
            final Set<String> result = new HashSet<String>();
            while (!reader.have(JSONReader.ENDARRAY)) {
                result.add(reader.stringValue());
            }
            reader.endArray();
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a collection of ints to an array.
     *
     * @param source the collection
     * @param writer the writer
     * @throws JSONException If the value is out of sequence
     */
    public void serializeCollectionInteger(final Collection<Integer> source, final JSONWriter writer) throws JSONException {

        writer.array();
        for (final Integer instance : source) {
            writer.value(instance);
        }
        writer.endArray();
    }

    /**
     * Deserializes a collection of ints.
     *
     * @param reader a stream starting with a JSON array containing the values
     * @return a collection of ints
     * @throws DeserializationException if the representation is invalid
     */
    public Collection<Integer> deserializeCollectionInteger(final JSONReader reader) throws DeserializationException {

        try {
            reader.array();
            final List<Integer> result = new ArrayList<Integer>();
            while (!reader.have(JSONReader.ENDARRAY)) {
                result.add(reader.intValue());
            }
            reader.endArray();
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a map from strings to ints, to an object containing an array of keys and an array of values.
     *
     * @param source the map
     * @param writer the writer
     * @throws JSONException If the value is out of sequence
     */
    public void serializeMapStringInt(final Map<String, Integer> source, final JSONWriter writer) throws JSONException {

        if (source == null) {
            writer.value(null);

        }
        else {
            writer.object();

            writer.key(KEYS);
            serializeSetString(source.keySet(), writer);
            writer.key(VALUES);
            serializeCollectionInteger(source.values(), writer);
            writer.endObject();
        }
    }

    /**
     * Deserializes a map from strings to ints.
     *
     * @param reader a stream starting with a JSON object containing keys and values
     * @return a map from strings to ints
     * @throws DeserializationException if the representation is invalid
     */
    public Map<String, Integer> deserializeMapStringInt(final JSONReader reader) throws DeserializationException {

        try {

            if (reader.checkNull()) { return null; }

            reader.object();
            reader.key(KEYS);
            final Set<String> keys = deserializeSetString(reader);
            reader.key(VALUES);
            final Collection<Integer> values = deserializeCollectionInteger(reader);

            final Map<String, Integer> result = new HashMap<String, Integer>();

            final Iterator<String> it1 = keys.iterator();
            final Iterator<Integer> it2 = values.iterator();
            while (it1.hasNext()) {
                if (!it2.hasNext()) { throw new DeserializationException("unbalanced elements in Map keys and values"); }
                final String arg0 = it1.next();
                final Integer arg1 = it2.next();
                result.put(arg0, arg1);
            }
            reader.endObject();
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }
}
