/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.registry.stream;

import java.net.UnknownHostException;

import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.registry.IRegistry;
import uk.ac.standrews.cs.nds.rpc.stream.ApplicationServer;
import uk.ac.standrews.cs.nds.rpc.stream.IHandler;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.stream.Marshaller;
import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Registry-specific server-side RPC mechanism.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RegistryServer extends ApplicationServer {

    private final IRegistry registry;
    private final RegistryMarshaller marshaller;

    static final int REGISTRY_PORT = 6828;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a server for a given registry.
     *
     * @param registry the registry
     */
    public RegistryServer(final IRegistry registry) {

        super();
        this.registry = registry;

        try {
            setLocalAddress(NetworkUtil.getLocalIPv4Address());
            setPort(REGISTRY_PORT);
        }
        catch (final UnknownHostException e) {
            ErrorHandling.hardExceptionError(e, "Couldn't find local IP address.");
        }

        marshaller = new RegistryMarshaller();
        initHandlers();
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Marshaller getMarshaller() {

        return marshaller;
    }

    @Override
    public String getApplicationRegistryKey() {

        // The null value is checked in ApplicationServer.start() to prevent recursive registry creation.
        return null;
    }

    // -------------------------------------------------------------------------------------------------------

    private void initHandlers() {

        handler_map.put("bind", new BindHandler());
        handler_map.put("lookup", new LookupHandler());
        handler_map.put("unbind", new UnbindHandler());
        handler_map.put("rebind", new RebindHandler());
        handler_map.put("getEntries", new GetEntriesHandler());
        handler_map.put("shutdown", new ShutdownHandler());
    }

    // -------------------------------------------------------------------------------------------------------

    private final class BindHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            final String key = args.stringValue();
            final Integer port = args.intValue();
            registry.bind(key, port);

            writer.value(""); // no response
        }
    }

    private final class RebindHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            final String key = args.stringValue();
            final Integer port = args.intValue();
            registry.rebind(key, port);

            writer.value(""); // no response
        }
    }

    private final class LookupHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            final String key = args.stringValue();

            writer.value(registry.lookup(key));
        }
    }

    private final class UnbindHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            final String key = args.stringValue();

            registry.unbind(key);

            writer.value(""); // no response
        }
    }

    private final class GetEntriesHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            marshaller.serializeMapStringInt(registry.getEntries(), writer);
        }
    }

    private final class ShutdownHandler implements IHandler {

        @Override
        public void execute(final JSONReader args, final JSONWriter writer) throws Exception {

            registry.shutdown();

            writer.value(""); // no response

            stop();
        }
    }
}
