/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;

import uk.ac.standrews.cs.nds.util.Duration;

/**
 * Manages persistent socket connections. The connection pool is a map containing an entry for various server addresses (instances of InetSocketAddress). Each entry maps to a set of
 * free connections to the corresponding address. When a connection to a given address is required, it is taken from the connection pool if one exists. When a connection is released,
 * it is added to the connection pool for future reuse.
 * 
 * This class is thread-safe.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class AbstractConnectionPool {

    private static final Duration SOCKET_CREATION_TIMEOUT = new Duration(2, TimeUnit.SECONDS);
    private static final Duration SOCKET_READ_TIMEOUT = new Duration(20, TimeUnit.SECONDS);
    private static final Duration INTERVAL_BETWEEN_COLLECTIONS = new Duration(5, TimeUnit.SECONDS);

    /**
     * Number of connections to be active in the connection pool.
     */
    public static int MAX_FREE_CONNECTIONS_PER_ADDRESS = 10;

    /*
    * The connection pool map is a ConcurrentMap, which is thread safe for manipulation without explicit synchronization, and should be faster than
    * locking for frequent contention. The set of free connections for a given address is also implemented using a ConcurrentMap, since standard
    * libraries provide no equivalent concurrent set.
    */
    private final ConcurrentMap<InetSocketAddress, ConcurrentMap<AbstractConnection, Object>> connection_pool = new ConcurrentHashMap<InetSocketAddress, ConcurrentMap<AbstractConnection, Object>>();

    /*
     * The garbage map contains connections that have been identified as garbage due to having remained unused in the connection pool for a duration
     * defined by Connection#GARBAGE_TIMEOUT. Again, it is implemented using a concurrent map to avoid locking.
     */
    private final ConcurrentMap<AbstractConnection, Object> garbage = new ConcurrentHashMap<AbstractConnection, Object>();

    private final Object dummy = new Object();

    // -------------------------------------------------------------------------------------------------------

    protected AbstractConnectionPool() {

        startGarbageCollector();
    }

    // -------------------------------------------------------------------------------------------------------

    public void shutdown() {

        stopGarbageCollector();
        closeAllConnections();
        clearDataStructures();

    }

    private void clearDataStructures() {

        connection_pool.clear();
        garbage.clear();
    }

    private void closeAllConnections() {

        for (final Map<AbstractConnection, Object> connection_set : connection_pool.values()) {
            for (final AbstractConnection connection : connection_set.keySet()) {
                closeConnection(connection);
            }
        }
    }

    private void stopGarbageCollector() {

        garbage_collector.shutdown();

    }

    /**
     * Returns the number of free connections for a given address. Used for testing.
     * 
     * @param address an address
     * @return the number of free connections for the address
     */
    public int getNumberOfFreeConnections(final InetSocketAddress address) {

        final Map<AbstractConnection, Object> free_connections = connection_pool.get(address);

        return free_connections == null ? 0 : free_connections.size();
    }

    /**
     * Gets a connection for the given address.
     * 
     * @param address an address
     * @return an existing free connection for the address, if one is available, otherwise a new connection
     * @throws IOException if a new connection cannot be created
     */
    public AbstractConnection getConnection(final InetSocketAddress address) throws IOException {

        // Get the free connections for the address.
        final ConcurrentMap<AbstractConnection, Object> free_connections = connection_pool.get(address);

        // If there are none, make a new connection.
        if (free_connections == null) { return makeConnection(address); }

        // Atomically remove an element from the map if there is one.
        final AbstractConnection free_connection = pickElement(free_connections);

        // If the map was empty, make a new connection.
        if (free_connection == null) { return makeConnection(address); }

        return free_connection;
    }

    /**
     * Releases a connection for reuse.
     * 
     * @param connection a connection
     */
    public void releaseConnection(final AbstractConnection connection) {

        final InetSocketAddress address = connection.getAddress();
        ConcurrentMap<AbstractConnection, Object> free_connections = connection_pool.get(address);

        if (free_connections == null) {

            free_connections = new ConcurrentHashMap<AbstractConnection, Object>();
            final ConcurrentMap<AbstractConnection, Object> other_set = connection_pool.putIfAbsent(address, free_connections);

            // Possible that another set was created and inserted since the initial get.
            if (other_set != null) {
                free_connections = other_set;
            }
        }

        // Possible race condition could lead to additional connections being added beyond limit, but doesn't seem to
        // matter enough to be worth the cost of locking.
        if (free_connections.size() < MAX_FREE_CONNECTIONS_PER_ADDRESS) {

            tidyConnection(connection);
            free_connections.put(connection, dummy);
            connection.recordReleaseTime();
        }
        else {
            closeConnection(connection);
        }
    }

    protected void tidyConnection(@SuppressWarnings("unused") final AbstractConnection connection) {

        // Override if any specific tidying required.
    }

    public void closeConnection(final AbstractConnection connection) {

        if (connection != null) {
            connection.tearDown();
        }
    }

    // -------------------------------------------------------------------------------------------------------

    private void collectGarbage() {

        identifyGarbage();
        discardGarbage();
        clearGarbage();
    }

    private void identifyGarbage() {

        // Examine the set of free connections for each address.
        for (final ConcurrentMap<AbstractConnection, Object> connections_for_address : connection_pool.values()) {

            // Examine each free connection.
            for (final AbstractConnection free_connection : connections_for_address.keySet()) {

                if (free_connection.isGarbage()) {
                    garbage.put(free_connection, dummy);
                }
            }
        }
    }

    private void discardGarbage() {

        for (final AbstractConnection garbage_connection : garbage.keySet()) {

            final ConcurrentMap<AbstractConnection, Object> set_containing_garbage = getSetContainingConnection(garbage_connection);

            // The connection may no longer be in the set, if it's been reused in the interim.
            // The 'remove' call atomically removes the key iff it is currently mapped to the value.
            if (set_containing_garbage.remove(garbage_connection, dummy)) {
                closeConnection(garbage_connection);
            }
        }
    }

    private ConcurrentMap<AbstractConnection, Object> getSetContainingConnection(final AbstractConnection connection) {

        return connection_pool.get(connection.getAddress());
    }

    private void clearGarbage() {

        garbage.clear();
    }

    private AbstractConnection pickElement(final ConcurrentMap<AbstractConnection, Object> connections) {

        while (true) {

            if (connections.isEmpty()) { return null; }

            AbstractConnection free_connection = null;
            for (final AbstractConnection connection : connections.keySet()) {
                free_connection = connection;
                break;
            }

            if (free_connection != null) {
                if (connections.remove(free_connection, dummy)) { return free_connection; }
            }
        }
    }

    protected abstract AbstractConnection makeConnection(final InetSocketAddress address) throws IOException;

    public Socket makeSocket(final InetSocketAddress address) throws IOException {

        final Socket socket = new Socket();
        socket.connect(address, (int) SOCKET_CREATION_TIMEOUT.getLength(TimeUnit.MILLISECONDS));
        socket.setSoTimeout((int) SOCKET_READ_TIMEOUT.getLength(TimeUnit.MILLISECONDS));
        return socket;
    }

    private GarbageCollector garbage_collector = null;

    private void startGarbageCollector() {

        garbage_collector = new GarbageCollector();
        garbage_collector.start();
    }

    // -------------------------------------------------------------------------------------------------------

    private class GarbageCollector extends Thread {

        public GarbageCollector() {

            setDaemon(true);
        }

        @Override
        public void run() {

            while (!isInterrupted()) {

                collectGarbage();
                INTERVAL_BETWEEN_COLLECTIONS.sleep();
            }
        }

        // TODO shouldn't need this since it's a daemon thread.
        public void shutdown() {

            interrupt();
        }
    }
}
