/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.security.SignatureException;
import java.util.concurrent.atomic.AtomicInteger;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.madface.IPingable;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundRemoteException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * Generic client-side RPC mechanism.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public abstract class AbstractProxy implements IPingable {

    /** The JSON key for the method name. */
    public static final String METHOD_KEY = "method";

    /** The JSON key for the method parameters. */
    public static final String PARAMS_KEY = "params";

    /** The JSON key for the version. */
    public static final String VERSION_KEY = "jsonrpc";

    /** The JSON key for the ID. */
    public static final String ID_KEY = "id";

    protected final InetSocketAddress node_address;
    protected final AtomicInteger method_call_number = new AtomicInteger(0);

    protected volatile boolean connection_changed; // For testing only.

    // -------------------------------------------------------------------------------------------------------------------------------

    protected AbstractProxy(final InetSocketAddress node_address) {

        this.node_address = node_address;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Returns the address of the service being proxied.
     * 
     * @return the address of the service being proxied
     */
    public InetSocketAddress getProxiedAddress() {

        return node_address;
    }

    /**
     * Checks whether the stream pair used for the last call was different than for the call before. Used primarily for testing.
     * 
     * @return true if the stream pair changed
     */
    public boolean connectionChangedOnLastCall() {

        return connection_changed;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    public void dealWithException(final Exception e) throws RPCException {

        if (e instanceof ClassCastException || e instanceof RPCException || e instanceof SignatureException || e instanceof PublicKeyNotFoundLocalException || e instanceof PublicKeyNotFoundRemoteException || e instanceof IOException || e instanceof JSONException) {

            // Preserve the stack trace of the original exception.
            final RPCException newException = new RPCException(stringWithServerLocation(e), e);
            newException.setStackTrace(e.getStackTrace());
            throw newException;
        }

        if (e instanceof RuntimeException) { throw (RuntimeException) e; }

        Diagnostic.trace(DiagnosticLevel.NONE, e, "unexpected checked exception: " + e.getClass().getName() + "; error message: " + e.getMessage());
    }

    /**
     * Returns a string describing which server this proxy is connected to. This helps to determine where an error such as a ConnectException is thrown from.
     * 
     * @param e The exception that was thrown
     * @return a string describing which server this proxy is connected to
     */
    private String stringWithServerLocation(final Exception e) {

        return "Originating Exception: '" + e.getClass().getName() + "', Message: '" + e.getMessage() + "'. Occurred on proxy connecting to " + node_address + ".";
    }
}
