/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import uk.ac.standrews.cs.nds.registry.AlreadyBoundException;
import uk.ac.standrews.cs.nds.registry.IRegistry;
import uk.ac.standrews.cs.nds.registry.RegistryUnavailableException;
import uk.ac.standrews.cs.nds.registry.nostream.RegistryFactory;
import uk.ac.standrews.cs.nds.rpc.AbstractApplicationServer;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Generic server-side RPC mechanism.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class ApplicationServer extends AbstractApplicationServer {

    private ServerThread server;

    protected final Map<String, IHandler> handler_map;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new application server.
     */
    public ApplicationServer() {

        handler_map = new HashMap<String, IHandler>();
        initHandlers();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Returns the marshaller used by the server.
     * @return the marshaller used by the server
     */
    public abstract Marshaller getMarshaller();

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Starts the server.
     *
     * @throws IOException if a server socket cannot be created using the specified address and port
     * @throws RPCException if an error occurs binding the application to the registry
     * @throws AlreadyBoundException if another instance of the application is already bound in the registry
     * @throws RegistryUnavailableException if the registry is unavailable
     */
    public void start() throws IOException, RPCException, AlreadyBoundException, RegistryUnavailableException {

        start(false);
    }

    /**
     * Starts the server.
     * @param allow_registry_overwrite true if any existing registry entry for the given key should be overwritten, false if such a clash should throw an exception
     *
     * @throws IOException if a server socket cannot be created using the specified address and port
     * @throws RPCException if an error occurs binding the application to the registry
     * @throws AlreadyBoundException if another instance of the application is already bound in the registry and allow_registry_overwrite is false
     * @throws RegistryUnavailableException if the registry is unavailable
     */
    public void start(final boolean allow_registry_overwrite) throws IOException, RPCException, AlreadyBoundException, RegistryUnavailableException {

        start(getApplicationRegistryKey(), allow_registry_overwrite);
    }

    /**
     * Starts the server and registers it with the local registry.
     *
     * @param application_registry_key the registry key to be used
     * @param allow_registry_overwrite true if any existing registry entry for the given key should be overwritten, false if such a clash should throw an exception
     *
     * @throws IOException if a server socket cannot be created using the specified address and port
     * @throws RPCException if an error occurs binding the application to the registry
     * @throws AlreadyBoundException if another instance of the application is already bound in the registry and allow_registry_overwrite is false
     * @throws RegistryUnavailableException if the registry is unavailable
     */
    public void start(final String application_registry_key, final boolean allow_registry_overwrite) throws IOException, RPCException, AlreadyBoundException, RegistryUnavailableException {

        setupServerSocket();
        startServer();

        // The key will be null if the application starting up is the registry itself.
        if (application_registry_key != null) {

            final IRegistry registry = RegistryFactory.FACTORY.getRegistry(true);

            if (allow_registry_overwrite) {
                registry.rebind(application_registry_key, port);
            }
            else {
                registry.bind(application_registry_key, port);
            }
        }
    }

    /**
     * Stops the server.
     *
     * @throws IOException if an error occurs when closing the server socket
     */
    @Override
    public void stop() throws IOException {

        stopServer();
    }

    /**
     * Gets the handler for a given call.
     *
     * @param method_name the method name
     * @return the handler
     */
    public IHandler getHandler(final String method_name) {

        return handler_map.get(method_name);
    }

    // -------------------------------------------------------------------------------------------------------

    private void setupServerSocket() throws IOException {

        server_socket = NetworkUtil.makeReusableServerSocket(local_address, port);
    }

    private void startServer() {

        server = new ServerThread(this, server_socket);
        server.startServer();
    }

    private void stopServer() {

        if (server != null) {
            server.stopServer();
            server = null;
        }
    }

    protected JSONValue dispatch(final String method_name, final JSONArray args) throws Exception {

        final IHandler method_handler = getHandler(method_name);

        if (method_handler == null) { throw new RPCException("unknown method name: " + method_name); }

        return method_handler.execute(args);
    }

    private void initHandlers() {

        handler_map.put(AbstractApplicationServer.PING_NAME, new PingHandler());
    }

    private static class PingHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) {

            return JSONObject.NULL;
        }
    }

    protected RequestHandler makeRequest(final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        return new RequestHandler(this, socket, extant_handlers);
    }

    @Override
    public void runServer(final InetAddress address, final int port) throws Exception {

        Diagnostic.trace("Not implemented - only present for ease of testing");
    }

    @Override
    public void runServer(final String[] args) throws Exception {

        Diagnostic.trace("Not implemented - only present for ease of testing");
    }
}
