/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;

import uk.ac.standrews.cs.nds.rpc.AbstractConnection;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * Utility class binding together input and output streams for a given socket.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class Connection extends AbstractConnection {

    private final PrintStream output_stream;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a stream pair for a given socket.
     *
     * @param socket the socket
     * @throws IOException if the streams cannot be created
     */
    public Connection(final Socket socket) throws IOException {

        super(socket);
        input_stream = new BufferedReader(new InputStreamReader(socket.getInputStream()));
        output_stream = new PrintStream(socket.getOutputStream(), true);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
      * Closes the input and output streams.
      */
    @Override
    public void tearDown() {

        try {
            socket.close();
        }
        catch (final IOException e) {
            Diagnostic.trace(DiagnosticLevel.RUN, "error closing socket");
        }
    }

    /**
     * Returns the input stream.
     * @return the input stream
     */
    public BufferedReader getInputStream() {

        return input_stream;
    }

    /**
     * Returns the output stream.
     * @return the output stream
     */
    public PrintStream getOutputStream() {

        return output_stream;
    }

    /**
     * Reads a line from the input stream.
     *
     * @return the line
     * @throws IOException if an error occurs while reading from the stream
     */
    public String readLine() throws IOException {

        return input_stream.readLine();
    }

    /**
     * Prints a newline to the output stream.
     */
    public void println() {

        output_stream.println();
    }

    /**
     * Prints a string followed by a newline to the output stream.
     *
     * @param s the string
     */
    public void println(final String s) {

        output_stream.println(s);
        output_stream.flush();
    }

}
