/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream;

import java.lang.reflect.Constructor;
import java.net.InetAddress;
import java.net.InetSocketAddress;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;
import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONString;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Provides methods serializing and deserializing common types to/from JSON strings.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class Marshaller {

    /**
     * The JSON key for error codes.
     */
    public static final String CODE_KEY = "code";

    /**
     * The JSON key for error messages.
     */
    public static final String MESSAGE_KEY = "message";

    /**
     * Serializes an exception to a JSON object containing a code (required by JSON-RPC spec but unused) and a
     * message. The message itself contains a serialized JSON array containing the exception class name and
     * the exception message.
     *
     * @param e the exception to be serialized
     * @return the serialized exception
     */
    public JSONObject serializeException(final Exception e) {

        final JSONObject error = new JSONObject();
        error.put(CODE_KEY, 0);

        final JSONArray array = new JSONArray();
        array.put(e.getClass().getName());
        array.put(e.getMessage());

        error.put(MESSAGE_KEY, array.toString());

        return error;
    }

    /**
     * Deserializes an exception from a JSON representation.
     *
     * @param error the serialized representation
     * @return the exception
     */
    public Exception deserializeException(final JSONObject error) {

        try {
            final String serialized_exception = error.getString(MESSAGE_KEY);
            return instantiateException(serialized_exception);
        }
        catch (final JSONException e) {
            return e;
        }
    }

    /**
     * Instantiates an exception from a JSON string containing the exception class name and the exception message.
     *
     * @param serialized_exception the serialized representation
     * @return the instantiated exception, or an instance of RuntimeException if the serialized exception cannot be instantiated
     */
    public Exception instantiateException(final String serialized_exception) {

        try {
            final JSONArray array = new JSONArray(serialized_exception);

            final String exception_class_name = array.getString(0);
            final String message = array.getString(1);

            @SuppressWarnings("unchecked")
            final Class<? extends Exception> exception_class = (Class<? extends Exception>) Class.forName(exception_class_name);

            final Class<String> arg0 = String.class;
            final Constructor<? extends Exception> constructor = exception_class.getConstructor(arg0);
            final Exception e = constructor.newInstance(message);
            return e;
        }
        catch (final Exception e) {
            return new RuntimeException("could not instantiate serialized exception: " + serialized_exception);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a network address to a string.
     *
     * @param inet_socket_address the network address
     * @return a JSON object
     */
    public JSONValue serializeInetSocketAddress(final InetSocketAddress inet_socket_address) {

        return new JSONString(inet_socket_address.toString());
    }

    /**
     * Deserializes a network address.
     *
     * @param serialized_inet_socket_address a string containing the address and port separated by a colon
     * @return a network address
     * @throws DeserializationException if the representation is invalid
     */
    public InetSocketAddress deserializeInetSocketAddress(final String serialized_inet_socket_address) throws DeserializationException {

        if (serialized_inet_socket_address == null || serialized_inet_socket_address.equals("null")) { return null; }

        final String[] components = serialized_inet_socket_address.split(":", -1);
        try {
            final String host = components[0];
            final int port = Integer.parseInt(components[1]);

            final String name = getName(host);
            final byte[] address_bytes = getBytes(host);

            final InetAddress addr = name.equals("") ? InetAddress.getByAddress(address_bytes) : InetAddress.getByAddress(name, address_bytes);

            return new InetSocketAddress(addr, port);
        }
        catch (final Exception e) {
            ErrorHandling.exceptionErrorNoEvent(e, "Failed to deserialize InetSocketAddress: '" + serialized_inet_socket_address + "'");
            throw new DeserializationException(e);
        }
    }

    /**
     * Serializes a key to a string.
     *
     * @param key the key
     * @return a JSON object
     */
    public JSONValue serializeKey(final IKey key) {

        return new JSONString(key.toString(Key.DEFAULT_RADIX));
    }

    /**
     * Deserializes a key.
     *
     * @param serialized_key a string containing representing the integer value of the key
     * @return a key
     * @throws DeserializationException if the representation is invalid
     */
    public IKey deserializeKey(final String serialized_key) throws DeserializationException {

        try {
            return new Key(serialized_key);
        }
        catch (final NumberFormatException e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    private static String getName(final String host) {

        final String[] name_address = host.split("/", -1);
        return name_address[0];
    }

    private static byte[] getBytes(final String host) {

        final String[] name_address = host.split("/", -1);
        final String[] byte_strings = name_address[1].split("\\.", -1);
        final byte[] bytes = new byte[byte_strings.length];

        for (int i = 0; i < byte_strings.length; i++) {
            final Integer j = Integer.valueOf(byte_strings[i]);
            bytes[i] = j.byteValue();
        }

        return bytes;
    }
}
