/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream;

import java.io.IOException;
import java.net.InetSocketAddress;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.rpc.AbstractApplicationServer;
import uk.ac.standrews.cs.nds.rpc.AbstractProxy;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * Generic client-side RPC mechanism, using explicit JSON data structures.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public abstract class NoStreamProxy extends AbstractProxy {

    public static final NoStreamConnectionPool CONNECTION_POOL = new NoStreamConnectionPool();

    private volatile Connection previous_connection = null; // For testing only.

    protected NoStreamProxy(final InetSocketAddress node_address) {

        super(node_address);
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Returns the marshaller used by this proxy.
     * @return the marshaller used by this proxy
     */
    public abstract Marshaller getMarshaller();

    /**
     * Checks whether the proxied server is reachable.
     * @throws RPCException if the proxied server is not reachable
     */
    @Override
    public final void ping() throws RPCException {

        try {
            makeCall(AbstractApplicationServer.PING_NAME);
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }

    protected JSONValue makeCall(final String method_name) throws Exception {

        return makeCall(method_name, new JSONArray());
    }

    protected JSONValue makeCall(final String method_name, final JSONArray args) throws Exception {

        Connection connection = null;
        boolean successful = false;
        try {
            connection = setupConnection();

            final JSONValue result = sendAndReceive(method_name, args, connection);
            successful = true;
            return result;
        }
        catch (final IOException e) {
            try {
                // Try once more in case the error was due to a server thread terminating through lack of activity.
                CONNECTION_POOL.closeConnection(connection);

                connection = setupConnection();

                final JSONValue result = sendAndReceive(method_name, args, connection);
                successful = true;
                return result;
            }
            catch (final IOException e1) {

                throw new RPCException(e1);
            }
        }
        catch (final JSONException e) {
            throw new RPCException(e);
        }
        finally {
            if (successful) {
                releaseConnection(connection);
            }
            else {
                closeConnection(connection);
            }
        }
    }

    protected void releaseConnection(final Connection connection) {

        if (connection != null) {
            CONNECTION_POOL.releaseConnection(connection);
        }
    }

    protected void closeConnection(final Connection connection) {

        if (connection != null) {
            CONNECTION_POOL.closeConnection(connection);
        }
    }

    protected void sendRequest(final Connection connection, final String method_name, final JSONArray args) throws Exception {

        // Send as JSON-RPC request.
        // E.g. { "method": "echo", "params": ["Hello JSON-RPC"], "id": 1}

        final JSONObject json_request = new JSONObject();
        json_request.put(METHOD_KEY, method_name);
        json_request.put(PARAMS_KEY, args);
        json_request.put(ID_KEY, method_call_number.incrementAndGet());
        json_request.put(VERSION_KEY, "2.0");

        final String request = json_request.toString();

        Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Request being sent from proxy: ", request);

        connection.println(request);
    }

    protected JSONObject readReply(final Connection connection) throws IOException, JSONException {

        final String reply = connection.readLine();

        final JSONObject json_reply = new JSONObject(reply);
        Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Reply returned from server: ", reply);

        return json_reply;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    private JSONValue sendAndReceive(final String method_name, final JSONArray args, final Connection streams) throws Exception {

        sendRequest(streams, method_name, args);

        return getResult(streams);
    }

    private Connection setupConnection() throws IOException {

        final Connection connection = (Connection) CONNECTION_POOL.getConnection(node_address);
        handleConnectionChange(connection);
        return connection;
    }

    protected void handleConnectionChange(final Connection connection) {

        connection_changed = connection != previous_connection;
        previous_connection = connection;
    }

    private JSONValue getResult(final Connection connection) throws Exception {

        JSONObject reply;
        try {
            reply = readReply(connection);
        }
        catch (final NullPointerException e) {
            // readReply() can throw NPE if connection has broken.
            throw new RPCException("server connection error");
        }

        if (reply.has(RequestHandler.RESULT_KEY)) { return reply.get(RequestHandler.RESULT_KEY); }

        if (reply.has(ApplicationServer.ERROR_KEY)) { throw getMarshaller().deserializeException(reply.getJSONObject(ApplicationServer.ERROR_KEY)); }

        Diagnostic.trace(this, "Unable to parse the received reply: ", reply);

        throw new RPCException("Unable to parse reply: " + reply);
    }

}
