/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream;

import java.io.IOException;
import java.net.Socket;
import java.util.Set;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.rpc.AbstractRequestHandler;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Handler for a specific request.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RequestHandler extends AbstractRequestHandler {

    private final ApplicationServer application_server;
    protected Connection connection;
    protected final Set<RequestHandler> extant_handlers;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a request handler.
     *
     * @param application_server the application server
     * @param socket the socket associated with request
     * @param extant_handlers 
     * @throws IOException if input and output streams for the socket cannot be created
     */
    public RequestHandler(final ApplicationServer application_server, final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        this.application_server = application_server;
        this.socket = socket;
        this.extant_handlers = extant_handlers;

        setupConnection();
        extant_handlers.add(this);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void run() {

        while (!Thread.currentThread().isInterrupted()) {

            JSONObject request;
            final String method_name;
            final JSONArray args;
            final int id;
            final JSONValue method_result;

            // Read request.
            try {
                request = readRequest();
            }
            catch (final Exception e) {
                sendApplicationError(e);
                break;
            }

            // Parse request.
            try {
                method_name = request.getString(NoStreamProxy.METHOD_KEY);
                args = request.getJSONArray(NoStreamProxy.PARAMS_KEY);
                id = request.getInt(NoStreamProxy.ID_KEY);
            }
            catch (final JSONException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "error parsing request: " + e.getMessage());
                break;
            }

            // Execute application method.

            try {
                method_result = application_server.dispatch(method_name, args);
            }
            catch (final Exception e) {

                // Return error to client if possible.

                Diagnostic.trace(DiagnosticLevel.FULL, "error in application method: " + e.getMessage());
                sendApplicationError(e);
                break;
            }

            // Return result to client.

            try {
                sendReply(method_name, method_result, id, connection);
            }
            catch (final Exception e) {
                ErrorHandling.error("error sending the reply piggybacked information: " + e.getMessage());
                break;
            }
        }

        shutDown();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    protected void sendReply(@SuppressWarnings("unused") final String method_name, final JSONValue result, final int id, final Connection streams) throws Exception {

        // Send as JSON-RPC reply.
        // E.g. {"result": 1, "error": null, "id": 99}

        final JSONObject json_reply = new JSONObject();

        json_reply.put(RESULT_KEY, result);
        json_reply.put(NoStreamProxy.ID_KEY, id);
        json_reply.put(NoStreamProxy.VERSION_KEY, JSON_VERSION);

        final String reply = json_reply.toString();

        Diagnostic.trace(DiagnosticLevel.FULL, "Response from Application Server: ", reply);
        streams.println(reply);
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    protected void shutDown() {

        try {
            tearDownConnection();
        }
        finally {
            try {
                tearDownSocket();
            }
            catch (final IOException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "error tearing down socket: " + e.getMessage());
            }
        }

        extant_handlers.remove(this);
    }

    private void setupConnection() throws IOException {

        // Set timeout so that request handler thread will terminate after specified time blocked waiting for next request.
        socket.setSoTimeout(SERVER_READ_TIMEOUT_IN_MILLIS);
        connection = new Connection(socket);
    }

    protected void tearDownConnection() {

        connection.tearDown();
    }

    protected void tearDownSocket() throws IOException {

        if (socket != null) {
            socket.close();
            socket = null;
        }
    }

    protected JSONObject readRequest() throws IOException, JSONException {

        final String request = connection.readLine();

        if (request == null) { throw new IOException("Incoming request was null. Possibly an external process such as a port scanner."); }

        final JSONObject json_request = new JSONObject(request);

        return json_request;
    }

    private void sendApplicationError(final Exception e) {

        final JSONObject json_reply = new JSONObject();
        final JSONObject error = application_server.getMarshaller().serializeException(e);

        json_reply.put(ApplicationServer.ERROR_KEY, error);
        json_reply.put(NoStreamProxy.ID_KEY, 0);
        json_reply.put(NoStreamProxy.VERSION_KEY, JSON_VERSION);

        final String reply = json_reply.toString();
        Diagnostic.trace(DiagnosticLevel.FULL, "Response from Application Server: [error]", json_reply);
        connection.println(reply);

    }
}
