/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream.json;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Representation of JSON array, backed by org.json version.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class JSONArray extends JSONValue {

    private final org.json.JSONArray json_array;

    /**
     * Representation of null array.
     */
    public static final JSONArray NULL = new JSONArray();

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes an empty array.
     */
    public JSONArray() {

        this(new org.json.JSONArray());
    }

    /**
     * Initializes an array from the given serialized representation.
     * @param serialized_array the serialized form
     * @throws JSONException if the serialized form cannot be deserialized
     */
    public JSONArray(final String serialized_array) throws JSONException {

        json_array = new org.json.JSONArray(serialized_array);
    }

    // -------------------------------------------------------------------------------------------------------

    protected JSONArray(final org.json.JSONArray json_array) {

        this.json_array = json_array;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Object getValue() {

        return this;
    }

    @Override
    public void putSelfIntoObject(final String key, final JSONObject parent) {

        parent.putArray(key, this);
    }

    @Override
    public void putSelfIntoArray(final int index, final JSONArray parent) {

        parent.putArray(index, this);
    }

    @Override
    public String toString() {

        return json_array.toString();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final JSONValue value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final JSONValue value) {

        if (value == null) {
            try {
                json_array.put(index, org.json.JSONObject.NULL);
            }
            catch (final JSONException e) {
                ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
            }
        }
        else {
            value.putSelfIntoArray(index, this);
        }
    }

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final int value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final int value) {

        try {
            json_array.put(index, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final long value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final long value) {

        try {
            json_array.put(index, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final double value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final double value) {

        try {
            json_array.put(index, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final boolean value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final boolean value) {

        try {
            json_array.put(index, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value at the end of the array.
     * @param value the value
     */
    public void put(final String value) {

        put(length(), value);
    }

    /**
     * Inserts the given value at the given position in the array.
     * @param index the position
     * @param value the value
     */
    public void put(final int index, final String value) {

        try {
            json_array.put(index, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Gets the JSONObject at the given position.
     *
     * @param index the position
     * @return the JSONObject at the given position, or {@link JSONObject#NULL} if the value is null
     * @throws JSONException if the value at the given position is not a JSONObject or the index is out of range
     */
    public JSONObject getJSONObject(final int index) throws JSONException {

        if (index < 0 || index >= json_array.length()) { throw new JSONException("index out of range"); }

        if (json_array.isNull(index)) { return JSONObject.NULL; }
        if (json_array.get(index).equals(org.json.JSONObject.NULL)) { return JSONObject.NULL; }

        return new JSONObject(json_array.getJSONObject(index));
    }

    /**
     * Gets the JSONArray at the given position.
     *
     * @param index the position
     * @return the JSONArray at the given position, or {@link #NULL} if the value is null
     * @throws JSONException if the value at the given position is not a JSONArray or the index is out of range
     */
    public JSONArray getJSONArray(final int index) throws JSONException {

        if (index < 0 || index >= json_array.length()) { throw new JSONException("index out of range"); }

        if (json_array.isNull(index)) { return NULL; }
        if (json_array.get(index).equals(org.json.JSONObject.NULL)) { return NULL; }

        return new JSONArray(json_array.getJSONArray(index));
    }

    /**
     * Gets the int at the given position.
     *
     * @param index the position
     * @return the int at the given position
     * @throws JSONException if the value at the given position is not an int or the index is out of range
     */
    public int getInt(final int index) throws JSONException {

        return json_array.getInt(index);
    }

    /**
     * Gets the boolean at the given position.
     *
     * @param index the position
     * @return the boolean at the given position
     * @throws JSONException if the value at the given position is not a boolean or the index is out of range
     */
    public boolean getBoolean(final int index) throws JSONException {

        return json_array.getBoolean(index);
    }

    /**
     * Gets the double at the given position.
     *
     * @param index the position
     * @return the double at the given position
     * @throws JSONException if the value at the given position is not a double or the index is out of range
     */
    public double getDouble(final int index) throws JSONException {

        return json_array.getDouble(index);
    }

    /**
     * Gets the long at the given position.
     *
     * @param index the position
     * @return the long at the given position
     * @throws JSONException if the value at the given position is not a long or the index is out of range
     */
    public long getLong(final int index) throws JSONException {

        return json_array.getLong(index);
    }

    /**
     * Gets the string at the given position.
     *
     * @param index the position
     * @return the string at the given position
     * @throws JSONException if the value at the given position is not a string or the index is out of range
     */
    public String getString(final int index) throws JSONException {

        if (json_array.isNull(index)) { return null; }
        return json_array.getString(index);
    }

    /**
     * Returns the length of the array.
     *
     * @return the length of the array
     */
    public int length() {

        return json_array.length();
    }

    /**
     * Removes the element at the given position.
     * @param index the position
     */
    public void remove(final int index) {

        json_array.remove(index);
    }

    // -------------------------------------------------------------------------------------------------------

    protected org.json.JSONArray getRealJSONArray() {

        return json_array;
    }

    protected void putObject(final int index, final JSONObject value) {

        try {
            json_array.put(index, value == null || value == JSONObject.NULL ? org.json.JSONObject.NULL : value.getRealJSONObject());
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    private void putArray(final int index, final JSONArray value) {

        try {
            json_array.put(index, value == null || value == JSONArray.NULL ? org.json.JSONObject.NULL : value.getRealJSONArray());
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }
}
