/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream.json;

import java.io.IOException;

import javax.mail.MessagingException;

import uk.ac.standrews.cs.nds.util.Encoding;

/**
 * Presents JSON Byte array value. The byte array is converted to {@link JSONObject} with three values: encoded bytes as string with the key of {@link #ENCODED_BYTES_KEY}, length of byte array as integer with the key of {@link #LENGTH_KEY} and the encoding as String with the key of {@link #ENCODING_KEY}.
 *
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class JSONBytes extends JSONValue {

    /** The default encoding used to encode the byte array to string if no encoding was specified. */
    public static final String DEFAULT_ENCODING = "base64";
    /** The key to encoded bytes. */
    public static final String ENCODED_BYTES_KEY = "encodedbytes";
    /** The length of byte array before encoding. */
    public static final String LENGTH_KEY = "length";
    /** The encoding of encoded byte array. */
    public static final String ENCODING_KEY = "encoding";

    private final JSONObject json_bytes_value;
    private final String encoding;
    private final int length;
    private final String encoded_bytes;

    /**
     * Instantiates a new JSON byte array. The byte array is converted to {@link JSONObject} with three values: encoded bytes as string with the key of {@link #ENCODED_BYTES_KEY}, length of byte array as integer with the key of {@link #LENGTH_KEY} and the encoding as String with the key of {@link #ENCODING_KEY}.
     *
     * @param value the value
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     */
    public JSONBytes(final byte[] value) throws MessagingException, IOException {

        this(value, DEFAULT_ENCODING);
    }

    /**
     * Instantiates a new JSON byte array. The byte array is converted to {@link JSONObject} with three values: encoded bytes as string with the key of {@link #ENCODED_BYTES_KEY}, length of byte array as integer with the key of {@link #LENGTH_KEY} and the encoding as String with the key of {@link #ENCODING_KEY}.
     *
     * @param value the value
     * @param encoding the encoding used to encode the value
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     */
    public JSONBytes(final byte[] value, final String encoding) throws MessagingException, IOException {

        //super(value);

        this.encoding = encoding;
        length = value.length;
        encoded_bytes = Encoding.encodeByteArray(value, encoding);

        json_bytes_value = new JSONObject();
        json_bytes_value.put(ENCODED_BYTES_KEY, encoded_bytes);
        json_bytes_value.put(LENGTH_KEY, length);
        json_bytes_value.put(ENCODING_KEY, encoding);
    }

    @Override
    public Object getValue() {

        return json_bytes_value;
    }

    @Override
    protected void putSelfIntoObject(final String key, final JSONObject parent) {

        parent.put(key, json_bytes_value);

    }

    @Override
    protected void putSelfIntoArray(final int index, final JSONArray parent) {

        parent.put(index, json_bytes_value);
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Gets the encodin used to encode the bytes.
     *
     * @return the encoding
     */
    public String getEncoding() {

        return encoding;
    }

    /**
     * Gets the length of the byte array.
     *
     * @return the length
     */
    public int getLength() {

        return length;
    }

    /**
     * Gets the encoded bytes in the given encoding.
     *
     * @return the encoded bytes
     */
    public String getEncodedBytes() {

        return encoded_bytes;
    }
}
