/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream.json;

import java.io.IOException;

import javax.mail.MessagingException;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.util.Encoding;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Representation of JSON object, backed by org.json version.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class JSONObject extends JSONValue {

    private final org.json.JSONObject json_object;

    /**
     * Representation of null object.
     */
    public static final JSONObject NULL = new JSONObject();

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes an empty object.
     */
    public JSONObject() {

        this(new org.json.JSONObject());
    }

    /**
     * Initializes an object from the given serialized representation.
     * @param serialized_object the serialized form
     * @throws JSONException if the serialized form cannot be deserialized
     */
    public JSONObject(final String serialized_object) throws JSONException {

        json_object = new org.json.JSONObject(serialized_object);
    }

    // -------------------------------------------------------------------------------------------------------

    protected JSONObject(final org.json.JSONObject json_object) {

        this.json_object = json_object;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        return json_object.toString();
    }

    @Override
    public Object getValue() {

        return this;
    }

    @Override
    public void putSelfIntoObject(final String key, final JSONObject parent) {

        parent.putObject(key, this);
    }

    @Override
    public void putSelfIntoArray(final int index, final JSONArray parent) {

        parent.putObject(index, this);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final JSONValue value) {

        if (value == null) {
            try {
                json_object.put(key, org.json.JSONObject.NULL);
            }
            catch (final JSONException e) {
                ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
            }
        }
        else {
            value.putSelfIntoObject(key, this);
        }
    }

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final int value) {

        try {
            json_object.put(key, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final long value) {

        try {
            json_object.put(key, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final double value) {

        try {
            json_object.put(key, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final boolean value) {

        try {
            json_object.put(key, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Inserts the given value with the given key.
     *
     * @param key the key
     * @param value the value
     */
    public void put(final String key, final String value) {

        try {
            json_object.put(key, value);
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    /**
     * Checks whether a given key is present.
     *
     * @param key the key
     * @return true if the key is present
     */
    public boolean has(final String key) {

        return json_object.has(key);
    }

    /**
     * Gets the value associated with the given key.
     *
     * @param key the key
     * @return the value at the given position
     * @throws JSONException if the key is not present
     */
    public JSONValue get(final String key) throws JSONException {

        return JSONValue.makeValue(json_object.get(key));
    }

    /**
     * Gets the JSONObject associated with the given key.
     *
     * @param key the key
     * @return the JSONObject at the given position, or {@link #NULL} if the key is not present
     * @throws JSONException if the value at the given position is not a JSONObject
     */
    public JSONObject getJSONObject(final String key) throws JSONException {

        if (json_object.isNull(key)) { return NULL; }
        if (json_object.get(key).equals(org.json.JSONObject.NULL)) { return NULL; }

        return new JSONObject(json_object.getJSONObject(key));
    }

    /**
     * Gets the JSONArray associated with the given key.
     *
     * @param key the key
     * @return the JSONArray at the given position, or {@link JSONArray#NULL} if the key is not present
     * @throws JSONException if the value at the given position is not a JSONArray
     */
    public JSONArray getJSONArray(final String key) throws JSONException {

        if (json_object.isNull(key)) { return JSONArray.NULL; }
        if (json_object.get(key).equals(org.json.JSONObject.NULL)) { return JSONArray.NULL; }

        return new JSONArray(json_object.getJSONArray(key));
    }

    /**
     * Gets the int associated with the given key.
     *
     * @param key the key
     * @return the int at the given position
     * @throws JSONException if the value at the given position is not an int or the key is not present
     */
    public int getInt(final String key) throws JSONException {

        return json_object.getInt(key);
    }

    /**
     * Gets the boolean associated with the given key.
     *
     * @param key the key
     * @return the boolean at the given position
     * @throws JSONException if the value at the given position is not a boolean or the key is not present
     */
    public boolean getBoolean(final String key) throws JSONException {

        return json_object.getBoolean(key);
    }

    /**
     * Gets the string associated with the given key.
     *
     * @param key the key
     * @return the string at the given position, or null if the key is not present
     * @throws JSONException if the value at the given position is not a string
     */
    public String getString(final String key) throws JSONException {

        if (json_object.isNull(key)) { return null; }
        return json_object.getString(key);
    }

    /**
     * Gets the double associated with the given key.
     *
     * @param key the key
     * @return the double at the given position
     * @throws JSONException if the value at the given position is not a double or the key is not present
     */
    public double getDouble(final String key) throws JSONException {

        return json_object.getDouble(key);
    }

    /**
     * Gets the long associated with the given key.
     *
     * @param key the key
     * @return the long at the given position
     * @throws JSONException if the value at the given position is not a long or the key is not present
     */
    public long getLong(final String key) throws JSONException {

        return json_object.getLong(key);
    }

    /**
     * Gets the byte array associated with the given key.
     *
     * @param key the key
     * @return the byte array at the given position
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     * @throws JSONException if the value at the given position is not a long or the key is not present
     */
    public byte[] getBytes(final String key) throws MessagingException, IOException, JSONException {

        final JSONObject json_bytes = getJSONObject(key);
        return Encoding.decodeByteArray(json_bytes.getString(JSONBytes.ENCODED_BYTES_KEY), json_bytes.getInt(JSONBytes.LENGTH_KEY), json_bytes.getString(JSONBytes.ENCODING_KEY));
    }

    // -------------------------------------------------------------------------------------------------------

    protected org.json.JSONObject getRealJSONObject() {

        return json_object;
    }

    protected void putArray(final String key, final JSONArray value) {

        try {
            json_object.put(key, value == null || value == JSONArray.NULL ? org.json.JSONObject.NULL : value.getRealJSONArray());
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }

    private void putObject(final String key, final JSONObject value) {

        try {
            json_object.put(key, value == null || value == JSONObject.NULL ? org.json.JSONObject.NULL : value.getRealJSONObject());
        }
        catch (final JSONException e) {
            ErrorHandling.hardExceptionError(e, "unexpected JSON exception");
        }
    }
}
