/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.nostream.json;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertThat;

import org.json.JSONException;
import org.junit.Before;
import org.junit.Test;

/**
 * Tests for JSON manipulation.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class JSONTest {

    private static final int THREE = 3;

    private JSONObject object1;
    private JSONObject object2;
    private JSONArray array1;
    private JSONArray array2;

    /**
     * Sets up test.
     */
    @Before
    public void setup() {

        object1 = new JSONObject();
        object2 = new JSONObject();
        array1 = new JSONArray();
        array2 = new JSONArray();
    }

    /**
     * Tests insertion and retrieval of an object field.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void insertAndRetrieveObjectField1() throws JSONException {

        object1.put("a", THREE);
        assertThat(object1.getInt("a"), is(THREE));
    }

    /**
     * Tests insertion and retrieval of an object field as a {@link JSONValue}.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void insertAndRetrieveObjectField2() throws JSONException {

        object1.put("a", new JSONInteger(THREE));
        assertThat(object1.getInt("a"), is(THREE));
    }

    /**
     * Tests that insertion of an object field and attempted retrieval of a different field throws the expected exception.
     * @throws JSONException expected
     */
    @Test(expected = JSONException.class)
    public void insertObjectFieldAndRetrieveDifferent() throws JSONException {

        object1.put("a", THREE);
        object1.getInt("b");
    }

    /**
     * Tests that insertion of an object field and attempted retrieval as a different type throws the expected exception.
     * @throws JSONException expected
     */
    @Test(expected = JSONException.class)
    public void insertObjectFieldAndRetrieveAsDifferentType() throws JSONException {

        object1.put("a", THREE);
        object1.getBoolean("a");
    }

    /**
     * Tests manipulation of nested objects.
     * @throws JSONException if a field cannot be retrieved
     */
    @Test
    public void nestedObjects() throws JSONException {

        object1.put("a", THREE);
        object2.put("b", object1);

        assertThat(object2.getJSONObject("b").getInt("a"), is(THREE));
    }

    /**
     * Tests serialization of nested objects.
     */
    @Test
    public void serializeNestedObjects() {

        object1.put("a", THREE);
        object2.put("b", object1);

        assertThat(object2.toString(), is("{\"b\":{\"a\":3}}"));
    }

    /**
     * Tests serialization of an array nested in a object.
     */
    @Test
    public void serializeArrayNestedInObject1() {

        object1.put("a", array1);
        array1.put("b");
        array1.put("c");

        assertThat(object1.toString(), is("{\"a\":[\"b\",\"c\"]}"));
    }

    /**
     * Tests serialization of an array nested in a object.
     */
    @Test
    public void serializeArrayNestedInObject2() {

        array1.put("b");
        array1.put("c");
        object1.put("a", array1);

        assertThat(object1.toString(), is("{\"a\":[\"b\",\"c\"]}"));
    }

    /**
     * Tests that null object stored in an object is retrieved as {@link JSONObject#NULL}.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void storeNullObjectInObject() throws JSONException {

        object1.put("a", (JSONObject) null);
        assertThat(object1.getJSONObject("a"), is(JSONObject.NULL));
    }

    /**
     * Tests that null array stored in an object is retrieved as {@link JSONArray#NULL}.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void storeNullArrayInObject() throws JSONException {

        object1.put("a", (JSONArray) null);
        assertThat(object1.getJSONArray("a"), is(JSONArray.NULL));
    }

    /**
     * Tests that attempted retrieval of a non-existent object field yields {@link JSONObject#NULL}.
     * @throws JSONException if an error occurs
     */
    @Test
    public void getNonExistentObjectField() throws JSONException {

        assertThat(object1.has("a"), is(false));
        assertThat(object1.getJSONObject("a"), is(JSONObject.NULL));
    }

    /**
     * Tests insertion and retrieval of an array element.
     * @throws JSONException if the element cannot be retrieved
     */
    @Test
    public void insertAndRetrieveArrayElement1() throws JSONException {

        array1.put(0, THREE);
        assertThat(array1.getInt(0), is(THREE));
    }

    /**
     * Tests insertion and retrieval of an array element as a {@link JSONValue}.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void insertAndRetrieveArrayElement2() throws JSONException {

        array1.put(0, new JSONInteger(THREE));
        assertThat(array1.getInt(0), is(THREE));
    }

    /**
     * Tests that insertion of an array element and attempted retrieval of a different element throws the expected exception.
     * @throws JSONException expected
     */
    @Test(expected = JSONException.class)
    public void insertArrayElementAndRetrieveDifferent() throws JSONException {

        array1.put(0, THREE);
        array1.getInt(1);
    }

    /**
     * Tests that insertion of an array element and attempted retrieval as a different type throws the expected exception.
     * @throws JSONException expected
     */
    @Test(expected = JSONException.class)
    public void insertArrayElementAndRetrieveAsDifferentType() throws JSONException {

        array1.put(0, THREE);
        array1.getBoolean(0);
    }

    /**
     * Tests manipulation of nested arrays.
     * @throws JSONException if an element cannot be retrieved
     */
    @Test
    public void nestedArrays() throws JSONException {

        array1.put(0, THREE);
        array2.put(0, array1);

        assertThat(array2.getJSONArray(0).getInt(0), is(THREE));
    }

    /**
     * Tests serialization of nested arrays.
     */
    @Test
    public void serializeNestedArrays() {

        array1.put(0, THREE);
        array2.put(0, array1);

        assertThat(array2.toString(), is("[[3]]"));
    }

    /**
     * Tests that null object stored in an array is retrieved as {@link JSONObject#NULL}.
     * @throws JSONException if the field cannot be retrieved
     */
    @Test
    public void storeNullObjectInArray() throws JSONException {

        array1.put(0, (JSONObject) null);
        assertThat(array1.getJSONObject(0), is(JSONObject.NULL));
    }

    /**
     * Tests that attempted retrieval of a non-existent array element yields {@link JSONObject#NULL}.
     * @throws JSONException if an error occurs
     */
    @Test
    public void getNonExistentArrayElement() throws JSONException {

        assertThat(array1.getJSONObject(0), is(JSONObject.NULL));
    }
}
