package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;

/**
 * This class buffers up data and permits the possibility of aborting all data written to it - for example in the case of an exception.
 * 
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class AbortableBufferedWriter extends Writer {

    protected static final int BUFFER_SIZE = 8192; // Maximum size of the buffer.

    protected final byte[] buffer;
    protected int index = 0;

    protected final OutputStream output_stream;

    /**
     * Instantiates a new abortable buffered writer.
     * 
     * @param outputStream the output stream
     */
    public AbortableBufferedWriter(final OutputStream outputStream) {

        output_stream = outputStream;
        buffer = new byte[BUFFER_SIZE];
    }

    @Override
    public void close() throws IOException {

        output_stream.close();

    }

    @Override
    public void flush() throws IOException {

        // FIXME for testing only.
        // final byte[] b = new byte[index];
        // System.arraycopy(buffer, 0, b, 0, index);
        // System.out.println(new String(b));

        output_stream.write(buffer, 0, index);
        output_stream.flush();

    }

    @Override
    public void write(final char[] src, final int offset, final int length) throws IOException {

        if (index + length > BUFFER_SIZE) { throw new IOException("buffer full"); }

        arraycopy(src, offset, buffer, index, length);
        index += length;
    }

    /**
     * Resets the buffer index to <code>zero</code>.
     * 
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public void reset() throws IOException {

        index = 0;
    }

    /**
     * Coppies an array of <code>char</code> into an array of <code>byte</code>.
     * 
     * @param src the source
     * @param srcPos the source offset
     * @param dest the destination
     * @param destPos the destination offset
     * @param length the length of data to be coppied
     */
    public static void arraycopy(final char[] src, int srcPos, final byte[] dest, int destPos, int length) {

        while (length > 0) {
            dest[destPos] = (byte) src[srcPos];
            srcPos++;
            destPos++;
            length--;
        }
    }
}
