package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;

import uk.ac.standrews.cs.nds.util.Diagnostic;

/**
 * A class which provides a Reader over InputStreams.
 * Intended for use over sockets to provide a blocking read.
 * It is the smallest possible shim  over InputStreams.
 * It provides no marks and no buffering.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class BlockingReader extends Reader {

    private static final int BUFFER_SIZE = 8192; // 8k should be enough

    private final InputStream in;
    private final byte[] buff;
    private boolean closed = false;

    /**
     * Constructs a BlockingReader over the supplied InputStream and using an internal 8K buffer.
     *
     * @param in the input stream
     */
    public BlockingReader(final InputStream in) {

        this.in = in;
        buff = new byte[BUFFER_SIZE]; // 8K buffer to use for input by this reader only
    }

    /**
     * Sets the closed flag and closes the underlying InputStream.
     *
     * @throws IOException Signals that an I/O exception has occurred.
     * @see java.io.Reader#close()
     */
    @Override
    public void close() throws IOException {

        synchronized (lock) {
            closed = true;
            in.close();
        }

    }

    @Override
    public int read() throws IOException { // TODO AL delete me - for debugging only

        int x = 0;
        System.out.println("in read");
        try {
            x = super.read();
        }
        catch (final IOException e) {
            System.out.println("exception in super.read");
            e.printStackTrace();
            throw e;
        }
        System.out.println("read returning value=" + (char) x);
        System.out.flush();
        return x;

    }

    /**
     * Reads from the InputStream and returns what is read in the char array supplied as a parameter.
     *
     * @param buffer the buffer
     * @param off the offset
     * @param number_bytes_to_read the number of bytes to read
     * @return the the length of data, which has been read
     * @throws IOException Signals that an I/O exception has occurred.
     *
     * @see java.io.Reader#read()
     */
    @Override
    public int read(final char[] buffer, final int off, final int number_bytes_to_read) throws IOException {

        synchronized (lock) {

            if (isClosed()) { throw new IOException("Read from closed Reader"); }

            final int to_read = number_bytes_to_read >= BUFFER_SIZE ? BUFFER_SIZE : number_bytes_to_read;

            final int bytes_read = in.read(buff, 0, to_read); // should be blocking if we we get InputStream from a socket
            if (bytes_read > 0) { // copy into result buffer if we have successfully read any bytes

                arraycopy(buff, 0, buffer, 0, bytes_read);
                Diagnostic.trace("***: returning>" + new String(buffer) + "< count=" + bytes_read);
                return bytes_read;
            }
            else { // no more data to be read
                return -1;
            }
        }
    }

    /**
     * This class does not support marks; it is the smallest shim over InputStreams possible.
     *
     * @return false
     * @see java.io.Reader#markSupported()
     */
    @Override
    public boolean markSupported() {

        return false;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Coppies an array of <code>char</code> into an array of <code>byte</code>.
     *
     * @param src the source
     * @param srcPos the source offset
     * @param dest the destination
     * @param destPos the destination offset
     * @param length the length of data to be coppied
     */
    public static void arraycopy(final byte[] src, int srcPos, final char[] dest, int destPos, int length) {

        while (length > 0) {
            dest[destPos] = (char) src[srcPos];
            srcPos++;
            destPos++;
            length--;
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    private boolean isClosed() {

        return closed;
    }
}
