/***************************************************************************
 * * nds Library * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group * University of St Andrews, Scotland *
 * http://www-systems.cs.st-andrews.ac.uk/ * * This file is part of nds, a package of utility classes. * * nds is free software: you can
 * redistribute it and/or modify * it under the terms of the GNU General Public License as published by * the Free Software Foundation,
 * either version 3 of the License, or * (at your option) any later version. * * nds is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the * GNU
 * General Public License for more details. * * You should have received a copy of the GNU General Public License * along with nds. If not,
 * see <http://www.gnu.org/licenses/>. * *
 ***************************************************************************/

package uk.ac.standrews.cs.nds.rpc.stream;

import java.lang.reflect.Constructor;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.rpc.RPCException;

//TODO consider moving and eliminating this class.

/**
 * Utility class for handling JSON messages.
 *
 * @author Alan Dearle  (al@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public final class JSONMessageUtil {

    /**
     * Utility class; no need for public consturctor.
     */
    private JSONMessageUtil() {

    }

    /**
     * Reads message id key and value.
     *
     * @param reader the reader
     * @return the message id
     * @throws Exception the exception
     */
    public static int readMessageId(final JSONReader reader) throws Exception {

        reader.object();
        reader.key(StreamProxy.ID_KEY);
        final int id = reader.intValue();
        return id;
    }

    /**
     * Reads up to result payload if a non error. <br/>
     * Deals with reply strings such as: <code>{"id":4,"result":false,"jsonrpc":"2.0"}</code>
     * and exception strings such as: <code>{"id":0,"error":{"message":"[\"uk.ac.standrews.cs.nds.rpc.RPCException\",\"JSONException\"]","code":0},"jsonrpc":"2.0"}</code>
     *
     * @param reader the reader
     * @return the message id
     * @throws Exception the exception
     */
    public static int readReplyHeader(final JSONReader reader) throws Exception {

        final int id;
        final boolean result;

        try {
            id = readMessageId(reader);
            result = isResultNext(reader);
        }

        catch (final NullPointerException e) {
            //  can throw NPE if connection has broken. Need to differentiate this from a NPE from the application.
            throw new RPCException("server connection error");
        }

        if (!result) {
            readError(reader); //must be error
        }

        return id;
    }

    /**
     * Checks if {@link RequestHandler#RESULT_KEY}} is next.
     *
     * @param reader the reader
     * @return true, if is result next
     * @throws Exception the exception
     */
    public static boolean isResultNext(final JSONReader reader) throws Exception {

        return reader.key().equals(RequestHandler.RESULT_KEY);
    }

    /**
     * Read result or error.
     *
     * @param reader the reader
     * @param key the key
     * @throws Exception the exception
     */
    public static void readResultOrError(final JSONReader reader, final String key) throws Exception {

        if (key.equals(ApplicationServer.ERROR_KEY)) { // Check if the given key is equal to error key
            readError(reader); // Read error and throw the relative exception
        }
        else if (!key.equals(RequestHandler.RESULT_KEY)) { throw new JSONException("Expected result or error key, found " + key); }

        // If the given key is equal to the 'result key', there is nothing that needs to be done.
    }

    /**
     * Reads json rpc trailer: version key and value.
     *
     * @param reader the reader
     * @throws JSONException the jSON exception
     */
    public static void readJsonRpcTrailer(final JSONReader reader) throws JSONException {

        reader.key(StreamProxy.VERSION_KEY);
        reader.stringValue(); // 2.0
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    private static void readError(final JSONReader reader) throws Exception {

        // Needs to deal with {"message":"[\"uk.ac.standrews.cs.nds.rpc.RPCException\",\"JSONException\"]","code":0}

        reader.object();
        reader.key(Marshaller.MESSAGE_KEY);
        reader.array();
        final String exception_class = reader.stringValue();
        String exception_value = null;
        if (!reader.checkNull()) {
            exception_value = reader.stringValue();
        }

        throw instantiateException(exception_class, exception_value);
    }

    private static Exception instantiateException(final String exception_class_name, final String message) {

        try {
            @SuppressWarnings("unchecked")
            final Class<? extends Exception> exception_class = (Class<? extends Exception>) Class.forName(exception_class_name);

            final Class<String> arg0 = String.class;
            final Constructor<? extends Exception> constructor = exception_class.getConstructor(arg0);
            final Exception e = constructor.newInstance(message);
            return e;
        }
        catch (final Exception e) {
            return new RuntimeException("could not instantiate serialized exception: " + exception_class_name);
        }
    }
}
