/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.util.UnknownFormatConversionException;

import javax.mail.MessagingException;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.keys.Key;
import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.Encoding;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Provides methods serializing and deserializing common types to/from JSON strings.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Alan Dearle (al@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class Marshaller {

    /** The JSON key for error codes. */
    public static final String CODE_KEY = "code";
    /** The JSON key for error messages. */
    public static final String MESSAGE_KEY = "message";
    private static final String ENCODED_BYTES_KEY = "encodedbytes";
    private static final String FORMAT_KEY = "format";
    private static final String ALGORITHM_KEY = "algorithm";
    private static final String X509_FORMAT = "X.509";
    private static final String BYTE_ENCODING_KEY = "encoding";
    private static final String BYTE_ENCODING = "base64";
    private static final String LENGTH_KEY = "length";

    /**
     * Serializes an exception to a JSON stream containing a code (required by JSON-RPC spec but unused) and a
     * message. The message itself contains a serialized JSON array containing the exception class name and
     * the exception message.
     *
     * @param e the exception to be serialized
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeException(final Exception e, final JSONWriter writer) throws JSONException {

        Diagnostic.trace(DiagnosticLevel.FULL, "Serialised exception ", e);
        writer.object();
        writer.key(MESSAGE_KEY);
        writer.array();
        writer.value(e.getClass().getName());
        writer.value(e.getMessage());
        writer.endArray();
        writer.endObject();
    }

    /**
     * Deserializes an exception from a JSON representation.
     *
     * @param reader the stream containing the serialized representation
     * @return the exception
     */
    public Exception deserializeException(final JSONReader reader) {

        String exception_class_name = "";

        try {
            reader.object();
            reader.key(MESSAGE_KEY);
            reader.array();

            exception_class_name = reader.stringValue();
            System.out.println("DESERIALISING EX: " + exception_class_name);
            final String message = reader.stringValue();
            reader.endArray();
            reader.endObject();

            @SuppressWarnings("unchecked")
            final Class<? extends Exception> exception_class = (Class<? extends Exception>) Class.forName(exception_class_name);

            final Class<String> arg0 = String.class;
            final Constructor<? extends Exception> constructor = exception_class.getConstructor(arg0);
            final Exception e = constructor.newInstance(message);
            return e;
        }
        catch (final JSONException e) {
            return e;
        }
        catch (final Exception e) {
            return new RuntimeException("could not instantiate serialized exception: " + exception_class_name);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes a network address to a string and writes it using the given writer.
     *
     * @param inet_socket_address the network address
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeInetSocketAddress(final InetSocketAddress inet_socket_address, final JSONWriter writer) throws JSONException {

        writer.value(inet_socket_address.toString());
    }

    /**
     * Deserializes a network address.
     *
     * @param reader the reader
     * @return a network address
     * @throws DeserializationException if the representation is invalid
     */
    public InetSocketAddress deserializeInetSocketAddress(final JSONReader reader) throws DeserializationException {

        try {
            if (reader.checkNull()) { return null; }

            final String serialized_inet_socket_address = reader.stringValue();

            return getAddress(serialized_inet_socket_address);
        }
        catch (final Exception e) {
            ErrorHandling.exceptionErrorNoEvent(e, "Failed to deserialize InetSocketAddress.");
            throw new DeserializationException(e);
        }
    }

    /**
     * Serializes a key to a string and writes using the given writer.
     *
     * @param key the key
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeKey(final IKey key, final JSONWriter writer) throws JSONException {

        if (key == null) {
            writer.value(null);
        }
        else {
            writer.value(key.toString(Key.DEFAULT_RADIX));
        }
    }

    /**
     * Deserializes a key.
     *
     * @param reader the reader
     * @return a key
     * @throws DeserializationException if the representation is invalid
     */
    public IKey deserializeKey(final JSONReader reader) throws DeserializationException {

        try {
            if (reader.checkNull()) { return null; }

            final String serialized_key = reader.stringValue();

            return new Key(serialized_key);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    /**
     * Serialise public key.
     *
     * @param public_key the public_key
     * @param writer the writer
     * @throws MessagingException the messaging exception
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws JSONException the jSON exception
     */
    public void serializePublicKey(final PublicKey public_key, final JSONWriter writer) throws MessagingException, IOException, JSONException {

        if (public_key != null) {
            writer.object();

            writer.key(ENCODED_BYTES_KEY);
            serializeBytes(public_key.getEncoded(), writer);

            writer.key(ALGORITHM_KEY);
            writer.value(public_key.getAlgorithm());

            writer.key(FORMAT_KEY);
            writer.value(public_key.getFormat());

            writer.endObject();
        }
        else {
            writer.value(null);
        }
    }

    /**
     * Deserialise public key.
     *
     * @param reader the reader
     * @return the public key
     * @throws DeserializationException the deserialization exception
     */
    public PublicKey deserializePublicKey(final JSONReader reader) throws DeserializationException {

        try {

            if (reader.checkNull()) { return null; }

            reader.object();

            reader.key(ENCODED_BYTES_KEY);
            final byte[] encoded_public_key = deserializeBytes(reader);

            reader.key(ALGORITHM_KEY);
            final String algorithm_name = reader.stringValue();

            reader.key(FORMAT_KEY);
            final String format = reader.stringValue();

            reader.endObject();

            if (format.equals(X509_FORMAT)) { return convertToX509PublicKey(encoded_public_key, algorithm_name); }

            throw new UnknownFormatConversionException("Unable to convert public key with the format of" + format);
        }
        catch (final Exception e) {
            throw new DeserializationException("Unable to deserialise the given serialised Public Key, because:" + e.getMessage());
        }
    }

    /**
     * Serialise bytes as string and writes the serialised bytes using the given writer.
     *
     * @param bytes the bytes top be serialised
     * @param writer the writer by which the serialised bytes are written
     * @throws MessagingException the messaging exception
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws JSONException the jSON exception
     */
    public void serializeBytes(final byte[] bytes, final JSONWriter writer) throws MessagingException, IOException, JSONException {

        if (bytes != null) {
            writer.object();

            writer.key(ENCODED_BYTES_KEY);
            writer.value(Encoding.encodeByteArray(bytes, BYTE_ENCODING));

            writer.key(LENGTH_KEY);
            writer.value(bytes.length);

            writer.key(BYTE_ENCODING_KEY);
            writer.value(BYTE_ENCODING);

            writer.endObject();
        }
        else {
            writer.value(null);
        }
    }

    /**
     * Deserialise signature.
     *
     * @param reader the reader
     * @return the byte[]
     * @throws DeserializationException the deserialisation exception
     */
    public byte[] deserializeBytes(final JSONReader reader) throws DeserializationException {

        try {

            if (reader.checkNull()) { return null; }

            reader.object();

            reader.key(ENCODED_BYTES_KEY);
            final String encoded_bytes = reader.stringValue();

            reader.key(LENGTH_KEY);
            final int length = reader.intValue();

            reader.key(BYTE_ENCODING_KEY);
            final String encoding = reader.stringValue();

            reader.endObject();

            return Encoding.decodeByteArray(encoded_bytes, length, encoding);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Gets the address from an String representation.
     *
     * @param address_in_string the address in string
     * @return the address
     * @throws UnknownHostException the unknown host exception
     */
    public static InetSocketAddress getAddress(final String address_in_string) throws UnknownHostException {

        final String serialized_inet_socket_address = address_in_string;

        final String[] components = serialized_inet_socket_address.split(":", -1);

        if (serialized_inet_socket_address == null || serialized_inet_socket_address.equals("null")) { return null; }
        final String host = components[0];
        final int port = Integer.parseInt(components[1]);

        final String name = getName(host);
        final byte[] address_bytes = getBytes(host);

        final InetAddress addr = name.equals("") ? InetAddress.getByAddress(address_bytes) : InetAddress.getByAddress(name, address_bytes);

        return new InetSocketAddress(addr, port);
    }

    /**
     * Converts an array of {@link Byte} to {@link PublicKey}.
     *
     * @param public_key_bytes the public_key_bytes
     * @param key_pair_algorithm the key_pair_algorithm
     * @return the public key
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     * @throws InvalidKeySpecException the invalid key spec exception
     */
    public static PublicKey convertToX509PublicKey(final byte[] public_key_bytes, final String key_pair_algorithm) throws NoSuchAlgorithmException, NoSuchProviderException, InvalidKeySpecException {

        final X509EncodedKeySpec public_key_specification = new X509EncodedKeySpec(public_key_bytes);
        final KeyFactory keyFactory = KeyFactory.getInstance(key_pair_algorithm);

        return keyFactory.generatePublic(public_key_specification);
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    private static String getName(final String host) {

        final String[] name_address = host.split("/", -1);
        return name_address[0];
    }

    private static byte[] getBytes(final String host) {

        final String[] name_address = host.split("/", -1);
        final String[] byte_strings = name_address[1].split("\\.", -1);
        final byte[] bytes = new byte[byte_strings.length];

        for (int i = 0; i < byte_strings.length; i++) {
            final Integer j = Integer.valueOf(byte_strings[i]);
            bytes[i] = j.byteValue();
        }

        return bytes;
    }
}
