/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.net.Socket;
import java.util.Set;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.AbstractRequestHandler;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Handler for a specific request.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Alan Dearle (al@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class RequestHandler extends AbstractRequestHandler {

    protected final ApplicationServer application_server;
    protected AbstractStreamConnection connection;
    protected final Set<RequestHandler> extant_handlers;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a request handler.
     * 
     * @param application_server the application server
     * @param socket the socket associated with request
     * @param extant_handlers the extant handlers
     * @throws IOException if input and output streams for the socket cannot be created
     */
    public RequestHandler(final ApplicationServer application_server, final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        this.application_server = application_server;
        this.socket = socket;
        this.extant_handlers = extant_handlers;

        setupConnection();
        extant_handlers.add(this);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public void run() {

        int id = -1;
        final JSONReader reader = connection.getJSONreader();

        while (!Thread.currentThread().isInterrupted()) {

            // Read request.
            try {
                id = readRequestHeader(reader);
            }
            catch (final Exception e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "Exception reading from input stream server side", e.getMessage());
                break;
            }

            // Parse request.
            final String method_name;
            try {

                method_name = parseRequest(id, reader);
            }
            catch (final JSONException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "error parsing request: " + e.getMessage());
                break;
            }

            // Reply to the request.
            reply(method_name, id);

            // Read the rest of the request, e.g. RPC trailer, piggyback, etc.
            try {

                readRequestTrailer(reader);
            }
            catch (final JSONException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "JSONException tidying up trailer from server: " + e.getMessage());
                // Nothing we can do here; we have replied already.
                break;
            }
        }

        tearDownConnection();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    protected int readRequestHeader(final JSONReader reader) throws Exception {

        reader.nextSymbol();
        return JSONMessageUtil.readMessageId(reader); // throw away cruft at the start and recover the message id
    }

    protected String parseRequest(final int id, final JSONReader reader) throws JSONException {

        // The 'id' can be used for sending error back to the client.

        reader.key(StreamProxy.METHOD_KEY);
        return reader.stringValue();
    }

    protected void reply(final String method_name, final int id) {

        JSONWriter writer = null;

        try {
            writer = connection.getJSONwriter(); // was newJSONwriter

            writer.object();
            writeReplyHeader(id, writer);

            application_server.dispatch(method_name, id, connection.getJSONreader(), writer); // fills in result

            writeReplyTrailer(writer); // Write RPC Trailer and piggyback
            writer.endObject();

            connection.completeWriteRequest();
        }
        catch (final Exception e) {

            // Return error to client if possible.
            Diagnostic.trace(DiagnosticLevel.FULL, "In Exception handling in application method: " + e.getMessage());
            try {

                sendApplicationError(e, id, writer);
            }
            catch (final JSONException e1) {
                // Nested exception - we are not going to be able to send this!
                Diagnostic.trace(DiagnosticLevel.FULL, e.getMessage(), "Error in serializing an error message", (socket == null ? "" : " back to " + socket.getRemoteSocketAddress()));
            }
            catch (final IOException e1) {
                // Error within error situation give up
                ErrorHandling.errorNoEvent(e.getMessage(), "Error attempting to send error message", (socket == null ? "" : " back to requesting instance at " + socket.getRemoteSocketAddress()));
            }
        }
    }

    protected void readRequestTrailer(final JSONReader reader) throws JSONException {

        JSONMessageUtil.readJsonRpcTrailer(reader); // Reads "jsonrpc":"2.0"

        if (application_server.hasPiggyback(reader)) {
            application_server.handlePiggyback(reader);
        }

        reader.endObjectNoLookahead(); // Expect json end-of-object symbol here.
    }

    // -------------------------------------------------------------------------------------------------------

    protected void shutDown() {

        try {
            tearDownConnection();
        }
        finally {
            try {
                tearDownSocket();
            }
            catch (final IOException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "error tearing down socket: " + e.getMessage());
            }
        }

        extant_handlers.remove(this);
    }

    protected void setupConnection() throws IOException {

        // Set timeout so that request handler thread will terminate after specified time blocked waiting for next request.
        socket.setSoTimeout(SERVER_READ_TIMEOUT_IN_MILLIS);
        connection = new Connection(socket);
    }

    protected void tearDownConnection() {

        connection.tearDown();
    }

    protected void tearDownSocket() throws IOException {

        if (socket != null) {
            socket.close();
            socket = null;
        }
    }

    protected void sendApplicationError(final Exception e, final int id, final JSONWriter writer) throws JSONException, IOException {

        connection.resetWriteBuffer();
        writer.object();
        writeErrorHeader(id, writer);

        application_server.getMarshaller().serializeException(e, writer);

        writeErrorTrailer(writer);
        writer.endObject();
        connection.completeWriteRequest();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    protected void writeReplyHeader(final int id, final JSONWriter writer) throws JSONException {

        writeId(id, writer);
        writer.key(RESULT_KEY);
    }

    protected void writeReplyTrailer(final JSONWriter writer) throws JSONException {

        writeVersion(writer);
        application_server.addPiggyback(writer);
    }

    protected void writeErrorHeader(final int id, final JSONWriter writer) throws JSONException {

        writeId(id, writer);
        writer.key(ApplicationServer.ERROR_KEY);
    }

    protected void writeErrorTrailer(final JSONWriter writer) throws JSONException {

        writeVersion(writer);
    }

    protected void writeId(final int id, final JSONWriter writer) throws JSONException {

        writer.key(StreamProxy.ID_KEY);
        writer.value(id);
    }

    protected void writeVersion(final JSONWriter writer) throws JSONException {

        writer.key(StreamProxy.VERSION_KEY);
        writer.value(JSON_VERSION);
    }
}
