package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.io.OutputStream;
import java.security.InvalidKeyException;
import java.security.SignatureException;

import uk.ac.standrews.cs.nds.rpc.security.ISignable;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.IByteDecorator;

/**
 * This class buffers up data and permits the possibility of aborting all data written to it - for example in the case of an exception. It also permits the signing of outgoing communications.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class SignableBufferedWriter extends AbortableBufferedWriter implements ISignable {

    private ISigner signer = null;
    private boolean signing_initialised = false;
    private int signed_payload_offset;
    private IByteDecorator signature_decorator;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new signable buffered writer. By default the signing is disabled.
     * 
     * @param output_stream the output stream
     */
    public SignableBufferedWriter(final OutputStream output_stream) {

        this(output_stream, null);

    }

    /**
     * Instantiates a new signable buffered writer.
     * 
     * @param output_stream the output_stream
     * @param signer the signer
     */
    public SignableBufferedWriter(final OutputStream output_stream, final ISigner signer) {

        super(output_stream);
        this.signer = signer;
        resetSignedPayloadOffest();
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // AbortableBufferedWriter overriden methods

    @Override
    public void write(final char[] src, final int offset, final int len) throws IOException {

        super.write(src, offset, len);

        if (isSigningEnabled()) {
            try {
                signer.update(buffer, index - len, len); // Update the signature from the byte buffer.
            }
            catch (final Exception e) {
                Diagnostic.trace(DiagnosticLevel.NONE, SignableBufferedWriter.class, "Unable to update signature with the buffered data, because: ", e.getMessage());
                throw new IOException(e);
            }
        }
    }

    @Override
    public void reset() throws IOException {

        if (isSigningEnabled()) {
            try {
                initSign();
            }
            catch (final Exception e) {
                throw new IOException(e);
            }
        }
        resetSignedPayloadOffest();
        super.reset();
    }

    @Override
    public void flush() throws IOException {

        if (isSigningEnabled()) {

            final byte[] signature;
            try {
                signature = signer.sign();
            }
            catch (final SignatureException e) {
                throw new IOException(e);
            }

            // FIXME for debug only. remove.
            // byte[] temp_buffer = new byte[signed_payload_offset];
            // System.arraycopy(buffer, 0, temp_buffer, 0, signed_payload_offset);
            // System.out.print("\n  About to be flushed out: " + new String(temp_buffer));
            // System.out.print(new String(isSignatureDecorationEnabled() ? signature_decorator.decorate(signature) : signature));
            // temp_buffer = new byte[index - signed_payload_offset];
            // System.arraycopy(buffer, signed_payload_offset, temp_buffer, 0, index - signed_payload_offset);
            // System.out.print(new String(temp_buffer) + "\n");

            output_stream.write(buffer, 0, signed_payload_offset); // write the unsigned payload
            output_stream.write(isSignatureDecorationEnabled() ? signature_decorator.decorate(signature) : signature); // write the signature, decorate before writing if decoration is enabled
            output_stream.write(buffer, signed_payload_offset, index - signed_payload_offset); // write the rest of the payload
            output_stream.flush();

        }
        else {
            super.flush();
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // ISignable overriden methods.

    @Override
    public boolean isSigningEnabled() {

        return signer != null && signing_initialised;
    }

    @Override
    public void enableSigning(final ISigner signer) throws InvalidKeyException {

        this.signer = signer;
        initSign();
        signed_payload_offset = index;
    }

    @Override
    public void disableSigning() {

        signer = null;
        signing_initialised = false;
        resetSignedPayloadOffest();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Enable signature decoration.
     * 
     * @param signature_decorator the new signature decorator
     */
    public void enableSignatureDecoration(final IByteDecorator signature_decorator) {

        this.signature_decorator = signature_decorator;
    }

    /**
     * Checks if is signature decoration enabled.
     * 
     * @return true, if is signature decoration enabled
     */
    public boolean isSignatureDecorationEnabled() {

        return signature_decorator != null;
    }

    /**
     * Gets the signed payload size.
     * 
     * @return the signed payload size
     */
    public int getSignedPayloadSize() {

        final int signed_payload_size;
        if (isSigningEnabled()) {
            signed_payload_size = index - signed_payload_offset;
        }
        else {
            signed_payload_size = 0;
        }

        return signed_payload_size;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    private void initSign() throws InvalidKeyException {

        signer.initSign();
        signing_initialised = true;
    }

    private void resetSignedPayloadOffest() {

        signed_payload_offset = 0;
    }

}
