/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.Socket;
import java.security.InvalidKeyException;
import java.security.SignatureException;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISignable;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifiable;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * The Class SignableStreamPair.
 *
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class SignableConnection extends AbstractStreamConnection implements ISignable, IVerifiable {

    private final VerifiableJSONReader verifiable_json_reader;
    private final VerifiableBufferedReader verifiable_reader;

    private SignableJSONWriter signable_json_writer;
    private final SignableBufferedWriter signable_writer;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new signable connection.
     *
     * @param socket the socket
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public SignableConnection(final Socket socket) throws IOException {

        super(socket);

        verifiable_reader = new VerifiableBufferedReader(new BufferedInputStream(socket.getInputStream()));
        verifiable_json_reader = new VerifiableJSONReader(verifiable_reader, this);

        signable_writer = new SignableBufferedWriter(socket.getOutputStream());
        newJSONwriter();
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IStreamPair overridden methods.

    @Override
    public void newJSONwriter() throws IOException {

        signable_json_writer = new SignableJSONWriter(signable_writer);
    }

    @Override
    public JSONWriter getJSONwriter() {

        return signable_json_writer;
    }

    @Override
    public JSONReader getJSONreader() {

        return verifiable_json_reader;
    }

    @Override
    public void resetWriteBuffer() throws IOException {

        signable_writer.reset();
    }

    // -------------------------------------------------------------------------------------------------------
    // AbstractStreamPair overridden methods.

    @Override
    public void completeWriteRequest() throws IOException {

        signable_writer.flush();
        super.completeWriteRequest();
    }

    @Override
    public void tearDown() {

        try {
            signable_writer.close();
        }
        catch (final IOException e) {
            Diagnostic.trace(DiagnosticLevel.RUN, "error closing signable writer");
        }
        finally {
            super.tearDown();
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IVerifiable overridden methods.

    @Override
    public boolean isVerificationEnabled() {

        return verifiable_json_reader.isVerificationEnabled();
    }

    @Override
    public void enableVerification(final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        verifiable_json_reader.enableVerification(verifier, public_key_manager);
    }

    @Override
    public void disableVerification() {

        verifiable_json_reader.disableVerification();
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // ISignable overriden methods.

    @Override
    public boolean isSigningEnabled() {

        return signable_json_writer.isSigningEnabled();
    }

    @Override
    public void enableSigning(final ISigner signer) throws InvalidKeyException, JSONException {

        signable_json_writer.enableSigning(signer);

    }

    @Override
    public void disableSigning() {

        signable_json_writer.disableSigning();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Verify.
     *
     * @return true, if successful
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws JSONException the jSON exception
     * @throws InvalidKeyException the invalid key exception
     * @throws SignatureException the signature exception
     * @throws PublicKeyNotFoundLocalException the public key not found exception
     */
    public boolean verify() throws IOException, JSONException, InvalidKeyException, SignatureException, PublicKeyNotFoundLocalException {

        return verifiable_json_reader.verify();
    }

    /**
     * Skip signature.
     *
     * @throws JSONException the jSON exception
     */
    public void skipSignature() throws JSONException {

        verifiable_json_reader.skipSignature();
    }
}
