/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.security.InvalidKeyException;

import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.security.ISignable;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.util.IByteDecorator;

/**
 * The Class SignableJSONWriter.
 *
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class SignableJSONWriter extends JSONWriter implements ISignable, IByteDecorator {

    /** The json key of signature. */
    public static final String SIGNATURE_KEY = "signature";
    /** The json key of payload size. */
    public static final String PAYLOAD_SIZE_KEY = "payloadsize";
    /** The json key of signature. */
    public static final String SIGNER_ID_KEY = "signerid";
    /** The default digital signature encoding. */
    public static final String DEFAULT_SIGNATURE_ENCODING = "base64";

    private static final String COLON = ":";
    private static final char COMMA = ',';

    private final SignableBufferedWriter signable_writer;

    /**
     * Instantiates a new signable json writer.
     *
     * @param writer the writer
     */
    public SignableJSONWriter(final SignableBufferedWriter writer) {

        super(writer);
        signable_writer = writer;
        signable_writer.disableSigning();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    @Override
    public boolean isSigningEnabled() {

        return signable_writer.isSigningEnabled();
    }

    @Override
    public void enableSigning(final ISigner signer) throws InvalidKeyException, JSONException {

        key(SIGNATURE_KEY);
        signable_writer.enableSigning(signer);
        signable_writer.enableSignatureDecoration(this);

        // Set the mode to 'key'.
        // Note:
        //  - The JSON value of SIGNATURE_KEY will be written by SignableBuggeredWriter#flush().
        //  - The signature is decorated by #decorator(byte[]) before getting written on the wire.
        //  - A comma character (",") must be added to the end of the SIGNATURE_KEY's value, which will be added by #decorator(byte[]).
        mode = 'k';

        key(SIGNER_ID_KEY);
        value(signer.getId());
    }

    @Override
    public void disableSigning() {

        signable_writer.disableSigning();
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IByteDecorator overriden method.

    /**
     * Decorates the digital signature bytes into a JSON friendly version. This done by encoing the bytes to BASE64 and add a quotation to the start and end, plus an aditional comma to the end.
     *
     * @param bytes_to_decorate the signature to decorate
     * @return a JSON friendly of the digital signature
     */
    @Override
    public byte[] decorate(final byte[] bytes_to_decorate) {

        // FIXME consider making decorator static?
        // FIXME use nds Encoding class
        final String encoded_signature = new sun.misc.BASE64Encoder().encode(bytes_to_decorate); //Encoding.encodeByteArray(signature, encoding);
        final StringBuilder sb = new StringBuilder();
        sb.append(JSONObject.quote(encoded_signature)); // Value for signature key
        sb.append(COMMA);
        sb.append(JSONObject.quote(PAYLOAD_SIZE_KEY));
        sb.append(COLON);
        sb.append(signable_writer.getSignedPayloadSize());
        sb.append(COMMA); // need an extra comma, since the mode is set to 'k' at #enableSigning(ISigner).
        return sb.toString().getBytes();
    }
}
