/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.security.InvalidKeyException;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISignable;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifiable;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundRemoteException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * The Class SignableProxy.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public abstract class SignableProxy extends StreamProxy implements ISignable, IVerifiable {

    /** The JSON key of the node key. */
    public static final String NODE_KEY_KEY = "nodekey";
    /** The connection pool. */
    public static final SignableConnectionPool CONNECTION_POOL = new SignableConnectionPool();

    protected ISigner signer;
    protected IVerifier verifier;
    protected IPublicKeyManager public_key_manager;

    // -------------------------------------------------------------------------------------------------------------------------------

    protected SignableProxy(final InetSocketAddress node_address) {

        this(node_address, null, null, null);
    }

    protected SignableProxy(final InetSocketAddress node_address, final ISigner signer, final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        super(node_address);

        this.signer = signer;
        this.verifier = verifier;
        this.public_key_manager = public_key_manager;
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // StreamProxy overridden methods.

    /**
     * Fetches a signable stream pair with <i>signing</i> and <i>verification</i> disabled from {@link SignableConnectionPool}.
     * 
     * @return a signable stream pair with signing and verification disabled
     */
    @Override
    protected AbstractStreamConnection setupConnection() throws IOException {

        final SignableConnection signable_connection = (SignableConnection) CONNECTION_POOL.getConnection(node_address);
        handleConnectionChange(signable_connection);
        return signable_connection;
    }

    @Override
    public synchronized AbstractStreamConnection startCall(final String method_name) throws IOException, JSONException {

        final SignableConnection signable_connection = (SignableConnection) setupConnection();

        final JSONWriter writer = signable_connection.getJSONwriter();
        writer.object();
        writeMessageId(writer);

        // Enable signing and verification if signer and verifier is provided by the user.
        if (isSigningEnabled()) {
            try {
                signable_connection.enableSigning(signer);
            }
            catch (final InvalidKeyException e) {
                throw new IOException(e);
            }
        }

        initMethodCall(writer, method_name);
        return signable_connection;
    }

    @Override
    protected JSONReader getReplyReader(final AbstractStreamConnection connection) throws Exception {

        final SignableConnection signable_connection = (SignableConnection) connection;
        if (isVerificationEnabled()) {
            signable_connection.enableVerification(verifier, public_key_manager);
        }

        final JSONReader reader = signable_connection.getJSONreader();
        reader.nextSymbol();

        JSONMessageUtil.readMessageId(reader);

        final String next_key;

        final String result_or_error_or_signature_key = reader.key();

        if (result_or_error_or_signature_key.equals(SignableJSONWriter.SIGNATURE_KEY)) {

            if (isVerificationEnabled()) {
                final boolean verified = signable_connection.verify();

                if (!verified) { throw new JSONException("Verification has failed. Provided signature does not match the Public key"); }
            }
            else {
                Diagnostic.trace(DiagnosticLevel.RUN, "Reply is signed but verification is not required, skipping the signature.");
                signable_connection.skipSignature();
            }

            next_key = reader.key();
        }
        else if (isVerificationEnabled()) {
            Diagnostic.trace(DiagnosticLevel.RUN, "Reply is not but verification is required, LETTING GO TEST ONLY"); // FIXME TEST ONLY
            next_key = result_or_error_or_signature_key;
        }
        else {
            next_key = result_or_error_or_signature_key;
        }

        JSONMessageUtil.readResultOrError(reader, next_key);
        return reader;
    }

    @Override
    protected void releaseConnection(final AbstractStreamConnection connection) {

        if (connection != null) {
            CONNECTION_POOL.releaseConnection(connection);
        }
    }

    @Override
    public void dealWithException(final Exception e) throws RPCException {

        if (e instanceof PublicKeyNotFoundRemoteException) { throw new RPCException(e); }
        if (e instanceof PublicKeyNotFoundLocalException) { throw new RPCException(e); }
        super.dealWithException(e);
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IVerifiable overridden methods.

    @Override
    public boolean isVerificationEnabled() {

        return verifier != null && public_key_manager != null;
    }

    @Override
    public void enableVerification(final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        this.verifier = verifier;
        this.public_key_manager = public_key_manager;
    }

    @Override
    public void disableVerification() {

        verifier = null;
        public_key_manager = null;
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // ISignable overridden methods.

    @Override
    public boolean isSigningEnabled() {

        return signer != null;
    }

    @Override
    public void enableSigning(final ISigner signer) {

        this.signer = signer;
    }

    @Override
    public void disableSigning() {

        signer = null;
    }
}
