/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.net.Socket;
import java.security.InvalidKeyException;
import java.util.Set;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISignable;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifiable;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundRemoteException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * Handler for a specific request.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class SignableRequestHandler extends RequestHandler implements Runnable, ISignable, IVerifiable {

    protected final SignableConnection signable_connection;

    private ISigner signer;
    private IVerifier verifier;
    private IPublicKeyManager public_key_manager;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new signable request handler.
     * 
     * @param application_server the application_server
     * @param socket the socket
     * @param extant_handlers the extant_handlers
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public SignableRequestHandler(final ApplicationServer application_server, final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        this(application_server, socket, null, null, null, extant_handlers);
    }

    /**
     * Instantiates a new signable request handler.
     * 
     * @param application_server the application_server
     * @param socket the socket
     * @param signer the signer
     * @param verifier the verifier
     * @param public_key_manager the public_key_manager
     * @param extant_handlers the extant_handlers
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public SignableRequestHandler(final ApplicationServer application_server, final Socket socket, final ISigner signer, final IVerifier verifier, final IPublicKeyManager public_key_manager, final Set<RequestHandler> extant_handlers) throws IOException {

        super(application_server, socket, extant_handlers);

        this.signer = signer;
        this.verifier = verifier;
        this.public_key_manager = public_key_manager;

        signable_connection = (SignableConnection) connection;
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // RequestHandler verridden methods.

    @Override
    protected String parseRequest(final int id, final JSONReader reader) throws JSONException {

        final String method_name;
        final String method_or_signature_key = reader.key();

        // There are five cases
        if (method_or_signature_key.equals(SignableJSONWriter.SIGNATURE_KEY)) {

            // Case 1: Verification is enabled and request is signed.
            if (isVerificationEnabled()) {
                final boolean verified;
                try {
                    verified = signable_connection.verify();
                }
                catch (final PublicKeyNotFoundLocalException e) {

                    Diagnostic.trace(DiagnosticLevel.RUN, e);

                    try {
                        sendApplicationError(new PublicKeyNotFoundRemoteException(e), id, connection.getJSONwriter()); // Notify client about public key not found.
                        Diagnostic.trace(DiagnosticLevel.RUN, "Notified the client about : " + e);
                    }
                    catch (final IOException e1) {
                        Diagnostic.trace(DiagnosticLevel.RUN, "Error while notifying the client about " + e.getMessage(), "caused by ", e1);
                    }

                    throw new JSONException(e);

                }
                catch (final Exception e) {
                    throw new JSONException(e);
                }

                if (!verified) { throw new JSONException("Verification has failed. Provided signature does not match with the Public key"); }
            }
            // Case 2: Verification is not enabled but request is signed.
            else {
                Diagnostic.trace(DiagnosticLevel.RUN, "Request is signed but verification is not required, skipping the signature.");
                signable_connection.skipSignature();
            }

            reader.key(StreamProxy.METHOD_KEY);
            method_name = reader.stringValue();
        }
        else if (method_or_signature_key.equals(StreamProxy.METHOD_KEY)) {
            // Case 3: Verification is enabled but request is not signed.
            if (isVerificationEnabled()) {

                // throw new JSONException("Server requires signed messages, the recieved request is not signed. dropping the connection...");
                method_name = reader.stringValue();
                Diagnostic.trace(DiagnosticLevel.RUN, "Server requires signed messages, the recieved request is not signed. LETTING GO, TEST ONLY -  method name: ", method_name);
            }
            // Case 4: Verification is not enabled and request is not signed.
            else {
                method_name = reader.stringValue();
            }
        }
        else {
            // Case 5: Unexpected key was found.
            throw new JSONException("expected signature key or method name key, found " + method_or_signature_key + ", dropping the connection...");
        }

        return method_name;
    }

    @Override
    protected void writeReplyHeader(final int id, final JSONWriter writer) throws JSONException {

        writeId(id, writer);

        if (isSigningEnabled()) {
            try {
                signable_connection.enableSigning(signer);
            }
            catch (final InvalidKeyException e) {
                throw new JSONException(e);
            }
        }

        writer.key(RESULT_KEY);
    }

    @Override
    protected void setupConnection() throws IOException {

        // Set timeout so that request handler thread will terminate after specified time blocked waiting for next request.
        socket.setSoTimeout(SERVER_READ_TIMEOUT_IN_MILLIS);
        connection = new SignableConnection(socket);
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IVerifiable overriden methods.

    @Override
    public boolean isVerificationEnabled() {

        return verifier != null && public_key_manager != null;
    }

    @Override
    public void enableVerification(final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        this.verifier = verifier;
        this.public_key_manager = public_key_manager;
    }

    @Override
    public void disableVerification() {

        verifier = null;
        public_key_manager = null;
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // ISignable overriden methods.

    @Override
    public boolean isSigningEnabled() {

        return signer != null;
    }

    @Override
    public void enableSigning(final ISigner signer) throws InvalidKeyException {

        this.signer = signer;
    }

    @Override
    public void disableSigning() {

        signer = null;
    }

}
