/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.net.InetSocketAddress;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.AbstractApplicationServer;
import uk.ac.standrews.cs.nds.rpc.AbstractProxy;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * Generic client-side RPC mechanism, manipulating JSON strings directly on streams. Supports piggybacking and digital signing of messages. Example use of foo with params <code>123</code>,<code>"abc"</code> and <code>true</code> as in <code>foo( 123,"abc",true );</code>:<br/>
 * <code>StreamPair streams = myproxy.startCall( "foo" ); <br/>
 * myproxy.addIntegerParam( 123); <br/>
 * myproxy.addStringParam( "abc" ); <br/>
 * myproxy.addBooleanParam( true ); <br/>
 * myproxy.makeCall(); <br/>
 * myproxy.finishCall();</code>.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 * @author Alan Dearle (al@st-andrews.ac.uk)
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public abstract class StreamProxy extends AbstractProxy {

    /** The JSON key for piggybacked data. */
    public static final String PIGGYBACK_KEY = "piggyback";

    /** The shared connection pool for stream-based proxies. */
    public static final StreamConnectionPool CONNECTION_POOL = new StreamConnectionPool();

    // -------------------------------------------------------------------------------------------------------

    private AbstractStreamConnection previous_connection = null; // For testing only.

    // -------------------------------------------------------------------------------------------------------

    protected StreamProxy(final InetSocketAddress node_address) {

        super(node_address);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the marshaller used by this proxy.
     * @return the marshaller used by this proxy
     */
    public abstract Marshaller getMarshaller();

    // -------------------------------------------------------------------------------------------------------
    /**
     * Checks whether the proxied server is reachable.
     * @throws RPCException if the proxied server is not reachable
     */
    @Override
    public void ping() throws RPCException {

        try {
            final AbstractStreamConnection connection = startCall(AbstractApplicationServer.PING_NAME);
            handleVoidCall(makeCall(connection));
            finishCall(connection);
        }
        catch (final Exception e) {
            dealWithException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Starts a call for a given method.
     * 
     * @param method_name the method name
     * @return the connection used for the call
     * @throws IOException if the connection cannot be established
     * @throws JSONException if the start of the JSON stream cannot be output to the connection
     */
    public AbstractStreamConnection startCall(final String method_name) throws IOException, JSONException {

        final AbstractStreamConnection connection = setupConnection();

        final JSONWriter writer = connection.getJSONwriter();
        writer.object();
        writeMessageId(writer);
        initMethodCall(writer, method_name);
        return connection;
    }

    /**
     * Makes the call.
     * @param connection the streams
     * @return the reader to read the result from
     * @throws Exception if JSON syntax error
     */
    public JSONReader makeCall(final AbstractStreamConnection connection) throws Exception {

        try {
            completeRequest(connection);
            final JSONReader reader = getReplyReader(connection);
            return reader;
        }
        catch (final Exception e) {
            Diagnostic.trace(DiagnosticLevel.FULL, "Exception in makeCall: ", e.getMessage());
            throw e;
        }
    }

    /**
     * Reads the JSON trailer from the end of the reply and releases the streams.
     * 
     * @param connection the connection to be released
     */
    public void finishCall(final AbstractStreamConnection connection) {

        try {
            final JSONReader reader = connection.getJSONreader();

            JSONMessageUtil.readJsonRpcTrailer(reader); // Reads the version.

            if (hasPiggyback(reader)) {
                handlePiggyback(reader);
            }

            reader.endObjectNoLookahead(); // Expect JSON end-of-object symbol here.
            releaseConnection(connection); // Release the connection for reuse.
        }
        catch (final Exception e) {
            Diagnostic.trace(DiagnosticLevel.FULL, "Exception finishing call: ", e.getMessage());
            closeConnection(connection); // Discard the connection since the streams are in an unpredictable state.
        }
    }

    protected void completeRequest(final AbstractStreamConnection connection) {

        // Send as JSON-RPC request.
        // E.g. { "method": "echo", "params": ["Hello JSON-RPC"], "id": 1}

        final JSONWriter writer = connection.getJSONwriter();

        try {
            writer.endArray(); // finish the params

            writer.key(VERSION_KEY);
            writer.value("2.0");

            addPiggyback(writer);

            writer.endObject();

            connection.completeWriteRequest();
        }
        catch (final JSONException e) {
            Diagnostic.trace(DiagnosticLevel.FULL, e.getMessage(), " Error whilst sending JSON-RPC request");
        }
        catch (final IOException e) {
            Diagnostic.trace(DiagnosticLevel.FULL, e.getMessage(), " Error whilst flushing stream for JSON-RPC request");
        }
    }

    /**
     * Consumes an empty string value from a JSONReader. This is used for void methods, because the server returns an empty string to the proxy (along with the standard JSON header) to indicate that the methods invocation has been completed, so the proxy must consume the string.
     * <p>
     * Example use: <code>handleVoidCall(makeCall(streams));</code>, in comparison to a method call which returns a value, where the proxy class should make use of the JSONReader: '<code>final JSONReader reader = makeCall(streams);</code>
     * 
     * @param reader The reader which is currently tokenized on an empty string.
     * @throws JSONException If it tries to consume a string but another type was found.
     */
    public void handleVoidCall(final JSONReader reader) throws JSONException {

        reader.stringValue();
    }

    protected void releaseConnection(final AbstractStreamConnection connection) {

        if (connection != null) {
            CONNECTION_POOL.releaseConnection(connection);
        }
    }

    protected void closeConnection(final AbstractStreamConnection connection) {

        if (connection != null) {
            CONNECTION_POOL.closeConnection(connection);
        }
    }

    /**
     * Adds piggyback. By default this method does not do anything. If piggybacking is required, this method must be overriden. The override of this method is expected to write {@link #PIGGYBACK_KEY} and value.
     * 
     * @param writer the writer
     * @throws JSONException if unexpected json syntax has been read.
     */
    protected void addPiggyback(@SuppressWarnings("unused") final JSONWriter writer) throws JSONException {

        // Must be overridden if piggyback is required.
    }

    /**
     * Handles piggyback. This method is called only if piggyback was found. This is checked using {@link #hasPiggyback(JSONReader)}. By default, this method does not do anything.If piggybacking is required, this method must be overriden. The override of this method shall expect the value only, since the {@link #PIGGYBACK_KEY} has already been read.
     * 
     * @param reader the reader
     * @throws JSONException if unexpected JSON syntax has been read.
     */
    protected void handlePiggyback(@SuppressWarnings("unused") final JSONReader reader) throws JSONException {

        // Must be overridden if piggyback is required.
    }

    /**
     * Checks whether the piggyback exists. By default always returns <code>false</code>. This method must be overridden if piggybacking is required.
     * 
     * @param reader the reader
     * @return false
     * @throws JSONException if unexpected json syntax has been read.
     */
    protected boolean hasPiggyback(@SuppressWarnings("unused") final JSONReader reader) throws JSONException {

        // Must be overridden if piggyback is required.
        return false;
    }

    protected AbstractStreamConnection setupConnection() throws IOException {

        final Connection connection = (Connection) CONNECTION_POOL.getConnection(node_address);
        handleConnectionChange(connection);
        return connection;
    }

    protected void writeMessageId(final JSONWriter writer) throws JSONException {

        writer.key(ID_KEY);
        writer.value(method_call_number.incrementAndGet());
    }

    protected void initMethodCall(final JSONWriter writer, final String method_name) throws JSONException {

        writer.key(METHOD_KEY);
        writer.value(method_name);
        writer.key(PARAMS_KEY);
        writer.array();
    }

    protected void handleConnectionChange(final AbstractStreamConnection connection) {

        connection_changed = connection != previous_connection;
        previous_connection = connection;
    }

    protected JSONReader getReplyReader(final AbstractStreamConnection connection) throws Exception {

        final JSONReader reader = connection.getJSONreader();
        reader.nextSymbol();
        JSONMessageUtil.readReplyHeader(reader);

        return reader;
    }
}
