/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.security.InvalidKeyException;
import java.security.PublicKey;
import java.security.SignatureException;

import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.IVerifiable;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;

/**
 * Verifiable buffered reader, which has the ability to verify the inputstream.
 *
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class VerifiableBufferedReader extends Reader implements IVerifiable {

    // FIXME implement your own buffer rather than using buffered inputstream
    // private static final int BUFFER_SIZE = 8192; // Maximum size of the buffer.
    //
    // protected final byte[] buffer;
    // protected int index = 0;
    private final BufferedInputStream input_stream;

    private IVerifier verifier = null;
    private IPublicKeyManager public_key_manager;

    /**
     * Instantiates a new verifiable buffered reader. By default the verification is disabled.
     *
     * @param input_stream the input_stream
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public VerifiableBufferedReader(final BufferedInputStream input_stream) throws IOException {

        this(input_stream, null);
    }

    /**
     * Instantiates a new verifiable buffered reader.
     *
     * @param input_stream the input_stream
     * @param verifier the verifier
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public VerifiableBufferedReader(final BufferedInputStream input_stream, final IVerifier verifier) throws IOException {

        super(new InputStreamReader(input_stream));
        this.input_stream = input_stream;
        this.verifier = verifier;

        // buffer = new byte[BUFFER_SIZE];
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // Reader overriden methods.

    @Override
    public void close() throws IOException {

        input_stream.close();

    }

    @Override
    public int read(final char[] cbuf, final int off, final int len) throws IOException {

        final byte[] temp_buff = new byte[len];
        final int read = input_stream.read(temp_buff, off, len);
        arraycopy(temp_buff, off, cbuf, off, len);
        return read;
    }

    @Override
    public void mark(final int read_ahead_limit) throws IOException {

        input_stream.mark(read_ahead_limit);
    }

    @Override
    public boolean markSupported() {

        return input_stream.markSupported();
    }

    @Override
    public void reset() throws IOException {

        input_stream.reset();
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IVerifiable overridden methods.

    @Override
    public boolean isVerificationEnabled() {

        return verifier != null && public_key_manager != null;
    }

    @Override
    public void enableVerification(final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        this.verifier = verifier;
        this.public_key_manager = public_key_manager;
    }

    @Override
    public void disableVerification() {

        verifier = null;
        public_key_manager = null;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Coppies an array of <code>byte</code> into an array of <code>char</code>.
     *
     * @param src the source
     * @param srcPos the source offset
     * @param dest the destination
     * @param destPos the destination offset
     * @param length the length of data to be coppied
     */
    public static void arraycopy(final byte[] src, int srcPos, final char[] dest, int destPos, int length) {

        while (length > 0) {
            dest[destPos] = (char) src[srcPos];
            srcPos++;
            destPos++;
            length--;
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Verifies from current position up to the end of the message. The end of the message must be identified by a <code>\n</code> character.
     *
     * @param signature the digital signature
     * @param signer_id the signer_id
     * @param payload_size the payload_size
     * @param payload_is_marked the payload_is_marked
     * @return true, if successful, false if verification has failed or the verification is not enabled
     * @throws PublicKeyNotFoundLocalException the public key not found local exception
     * @throws InvalidKeyException the invalid key exception
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws SignatureException the signature exception
     */
    public boolean verify(final byte[] signature, final String signer_id, final int payload_size, final boolean payload_is_marked) throws PublicKeyNotFoundLocalException, InvalidKeyException, IOException, SignatureException {

        boolean verification_result = false;

        if (isVerificationEnabled()) {

            // System.out.println("about to verify payload len " + payload_size);
            final PublicKey public_key = public_key_manager.getPublicKey(signer_id);

            verifier.initVerify(public_key);

            if (!payload_is_marked) { // Check if payload not marked
                input_stream.mark(payload_size); // Mark the payload
            }

            final byte[] payload = new byte[payload_size];
            final int read_payload_size = input_stream.read(payload);

            if (read_payload_size != payload_size) { throw new IOException("expected payload size of " + payload_size + " bytes, but read " + read_payload_size + " bytes."); }

            input_stream.reset();

            verifier.update(payload, 0, payload.length);

            verification_result = verifier.verify(signature);
        }

        return verification_result;
    }
}
