/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.stream;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.SignatureException;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.rpc.AbstractConnection;
import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.IVerifiable;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.security.exception.PublicKeyNotFoundLocalException;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

/**
 * The Class VerifiableJSONReader.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class VerifiableJSONReader extends JSONReader implements IVerifiable {

    private final VerifiableBufferedReader verifiable_reader;

    /**
     * Instantiates a new verifiable JSON reader.
     * 
     * @param verifiable_reader the reader
     */
    public VerifiableJSONReader(final VerifiableBufferedReader verifiable_reader, final AbstractConnection connection) {

        super(verifiable_reader, connection);
        this.verifiable_reader = verifiable_reader;
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // IVerifiable overridden methods.

    @Override
    public boolean isVerificationEnabled() {

        return verifiable_reader.isVerificationEnabled();
    }

    @Override
    public void enableVerification(final IVerifier verifier, final IPublicKeyManager public_key_manager) {

        verifiable_reader.enableVerification(verifier, public_key_manager);
    }

    @Override
    public void disableVerification() {

        verifiable_reader.disableVerification();
    }

    @Override
    protected void closeConnection() {

        SignableProxy.CONNECTION_POOL.closeConnection(connection);
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    /**
     * Verify.
     * 
     * @return true, if successful
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws JSONException the jSON exception
     * @throws InvalidKeyException the invalid key exception
     * @throws SignatureException the signature exception
     * @throws PublicKeyNotFoundLocalException the public key not found local exception
     */
    public boolean verify() throws IOException, JSONException, InvalidKeyException, SignatureException, PublicKeyNotFoundLocalException {

        return verify(SignableJSONWriter.DEFAULT_SIGNATURE_ENCODING);
    }

    /**
     * Verify.
     * 
     * @param encoding the encoding
     * @return true, if successful
     * @throws IOException Signals that an I/O exception has occurred.
     * @throws JSONException the jSON exception
     * @throws InvalidKeyException the invalid key exception
     * @throws SignatureException the signature exception
     * @throws PublicKeyNotFoundLocalException the public key not found local exception
     */
    public boolean verify(final String encoding) throws IOException, JSONException, InvalidKeyException, SignatureException, PublicKeyNotFoundLocalException {

        final String encoded_signatrue = stringValue();
        final byte[] signature = new sun.misc.BASE64Decoder().decodeBuffer(encoded_signatrue);// Encoding.decodeByteArray(encoded_signatrue, encoding);

        key(SignableJSONWriter.PAYLOAD_SIZE_KEY);
        final int payloadsize = intValueNoLookahead(); // find out how much to read
        verifiable_reader.mark(payloadsize); // mark where the singed payload statrs.

        nextSymbol(); // consume the payload size value since it was read with no lookahead
        nextSymbol(); // consume the comma
        key(SignableJSONWriter.SIGNER_ID_KEY);
        final String signer_id = stringValue();
        verifiable_reader.reset(); // reset to the begining of the signed payload

        final boolean verification_result = verifiable_reader.verify(signature, signer_id, payloadsize, true);
        nextSymbol(); // consume comma
        skipSignerId(); // skip the signature id. this is neaded since the inputstream is reset to the beginging of the signed payload and signerID is part of the signed payload.
        return verification_result;
    }

    /**
     * Skip signature.
     * 
     * @throws JSONException the jSON exception
     */
    public void skipSignature() throws JSONException {

        Diagnostic.trace(DiagnosticLevel.RUN, "Skipping signature...");
        stringValue(); // consume signature value
        skipPayloadSize();
        skipSignerId();
    }

    /**
     * Consumes an integer value from the stream.
     * 
     * @return the value from the stream
     * @throws JSONException if some other item is found
     */
    public int intValueNoLookahead() throws JSONException {

        if (the_symbol == INTEGER) {
            final int result = theInteger;
            return result;
        }
        throw new JSONException("expected int, found: " + decode(the_symbol));
    }

    private void skipPayloadSize() throws JSONException {

        key(SignableJSONWriter.PAYLOAD_SIZE_KEY); // Skip payload size key
        intValue(); // Skip payload size
    }

    private void skipSignerId() throws JSONException {

        key(SignableJSONWriter.SIGNER_ID_KEY); // skip signer_id key
        stringValue(); // Skip signer_id value
    }
}
