package uk.ac.standrews.cs.nds.rpc.test;

import static org.hamcrest.core.Is.is;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.Assert.assertThat;

import java.lang.reflect.InvocationTargetException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;

import org.junit.After;
import org.junit.Before;

import uk.ac.standrews.cs.nds.madface.HostDescriptor;
import uk.ac.standrews.cs.nds.madface.JavaProcessDescriptor;
import uk.ac.standrews.cs.nds.madface.ProcessDescriptor;
import uk.ac.standrews.cs.nds.madface.ProcessManager;
import uk.ac.standrews.cs.nds.registry.NotBoundException;
import uk.ac.standrews.cs.nds.registry.RegistryUnavailableException;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter3;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

public abstract class RPCTestBase {

    protected static final int THREE = 3;
    protected static final int FIVE = 5;
    protected static final int SEVEN = 7;
    protected static final int NOT_THREE = -345987235;

    protected static final int INITIAL_SERVER_PORT = 35050;
    protected static final int TIMEOUT_INTERVAL = 30000;
    protected static final int RETRY_INTERVAL = 2000;

    protected static final AtomicInteger server_port = new AtomicInteger(INITIAL_SERVER_PORT);

    protected ITest server;
    protected Process server_process;
    protected InetSocketAddress server_address;

    protected List<String> server_args = new ArrayList<String>();

    protected abstract Class<?> getServerClass();

    protected abstract ITest bindToServer(final InetSocketAddress server_address) throws RPCException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException;

    protected abstract void shutdownRegistry() throws UnknownHostException, RegistryUnavailableException, RPCException, NotBoundException;

    // -------------------------------------------------------------------------------------------------------

    protected void setServerAddress() throws UnknownHostException {

        server_address = new InetSocketAddress(NetworkUtil.getLocalIPv4Address(), server_port.incrementAndGet());
    }

    protected void bindToServer() throws Exception {

        server = bindToServerWithRetry(server_address);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Starts the server in a new process and connects to it.
     * 
     * @throws Exception if the test fails
     */
    @Before
    public void setup() throws Exception {

        setServerAddress();

        final ProcessManager process_manager = new HostDescriptor().getProcessManager();
        server_args.add("-p" + server_port.get());
        final ProcessDescriptor java_process_descriptor = new JavaProcessDescriptor().classToBeInvoked(getServerClass()).args(server_args);
        server_process = process_manager.runProcess(java_process_descriptor);

        bindToServer();
    }

    /**
     * Shuts down the server.
     * 
     * @throws InterruptedException if the test fails
     * @throws RegistryUnavailableException
     * @throws UnknownHostException
     * @throws NotBoundException
     * @throws RPCException
     */
    @After
    public void teardown() throws InterruptedException, UnknownHostException, RegistryUnavailableException, RPCException, NotBoundException {

        server_process.destroy();
        server_process.waitFor();

        try {
            shutdownRegistry();
        }
        catch (final Exception e) {
            // Ignore error due to server shutting down.
        }
        Thread.sleep(500); // FIXME investigating test failure due to the address already in use exception, appears that process#waitFor doesnt work properly.
    }

    // -------------------------------------------------------------------------------------------------------

    protected ITest bindToServerWithRetry(final InetSocketAddress server_address) throws TimeoutException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, InterruptedException {

        final long start_time = System.currentTimeMillis();

        while (!Thread.currentThread().isInterrupted()) {

            try {
                return bindToServer(server_address);
            }
            catch (final Exception e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "remote binding failed: " + e.getMessage());
            }

            try {
                Thread.sleep(RETRY_INTERVAL);
            }
            catch (final InterruptedException e) {
                // Ignore.
            }

            final long duration = System.currentTimeMillis() - start_time;
            if (duration > TIMEOUT_INTERVAL) { throw new TimeoutException(); }
        }
        throw new InterruptedException();
    }

    public void passTwoObjects(final ITest server) throws RPCException {

        assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(THREE))), is(true));
        assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(THREE + 1))), is(false));
        assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], null)), is(false));

        assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[0], null)), is(true));
        assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[]{THREE, SEVEN}, null), new TestParameter2("hello", false, new int[]{THREE, SEVEN}, null)), is(true));
        assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, null, null)), is(false));
        assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[]{THREE}, null)), is(false));
    }

    public void returnObject(final ITest server) throws RPCException {

        assertThat(server.returnsWithX3().getX(), is(equalTo(3)));
    }

    public void timeWork(final ITest server) throws RPCException {

        final TestParameter1 t12 = new TestParameter1(THREE);
        final TestParameter3 t3 = new TestParameter3(THREE);
        final TestParameter2 t2 = new TestParameter2("hello", false, new int[0], t3);
        final TestParameter3 t32 = new TestParameter3(THREE + 1);
        final TestParameter2 t22 = new TestParameter2("hello", false, new int[0], t32);
        final TestParameter2 t23 = new TestParameter2("hello", false, new int[0], null);
        final TestParameter2 t13 = new TestParameter2("hello", false, new int[]{THREE, SEVEN}, null);
        final TestParameter2 t24 = new TestParameter2("hello", false, null, null);
        final TestParameter2 t25 = new TestParameter2("hello", false, new int[]{THREE}, null);
        final TestParameter1[] t = new TestParameter1[]{};
        final TestParameter1[] t4 = new TestParameter1[]{t12};
        final TestParameter1[] t5 = new TestParameter1[]{t12, t12};
        final TestParameter1 testParameter1 = new TestParameter1(NOT_THREE);
        final TestParameter1[] t6 = new TestParameter1[]{testParameter1, t12};
        final TestParameter1[] t7 = new TestParameter1[]{t12, null};

        final long start = System.currentTimeMillis();

        for (int i = 0; i < 1000; i++) {

            // returnObject
            assertThat(server.returnsWithX3().getX(), is(THREE));

            // xSameAsY pass 2 objects
            assertThat(server.xSameAsY(t12, t2), is(true));
            assertThat(server.xSameAsY(t12, t22), is(false));
            assertThat(server.xSameAsY(t12, t23), is(false));

            // arrays equal
            assertThat(server.arraysEqual(t23, t23), is(true));
            assertThat(server.arraysEqual(t13, t13), is(true));
            assertThat(server.arraysEqual(t23, t24), is(false));
            assertThat(server.arraysEqual(t23, t25), is(false));

            // passArray
            assertThat(server.allElementsHaveSameX(t), is(true));
            assertThat(server.allElementsHaveSameX(t4), is(true));
            assertThat(server.allElementsHaveSameX(t5), is(true));
            assertThat(server.allElementsHaveSameX(t6), is(false));
            assertThat(server.allElementsHaveSameX(null), is(false));
            assertThat(server.allElementsHaveSameX(t7), is(false));

            // returnArray
            final TestParameter1[] array = server.returnsArrayLength2With5And7();
            assertThat(array.length, is(2));
            assertThat(array[0].getX(), is(FIVE));
            assertThat(array[1].getX(), is(SEVEN));
        }

        final long end = System.currentTimeMillis();

        final long elapsed = end - start;

        System.out.println("Elapsed time: " + elapsed + " ms");
    }

    protected void outputTestName() {

        System.out.println(Diagnostic.getMethodInCallChain(2));
    }
}
