package uk.ac.standrews.cs.nds.rpc.test;

import java.util.concurrent.TimeoutException;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.app.TestException;

public abstract class RPCTestErrorBase extends RPCTestBase {

    protected static final int THREE = 3;
    private static final long SERVER_FAULT_TEST_TIMEOUT = 30000;

    public static final String ACCEPT_REQUEST_BUT_DONT_REPLY = "-a";
    public static final String RETURN_INVALID_JSON = "-i";
    public static final String RETURN_BLANK_LINE = "-b";
    public static final String RETURN_WRONG_TYPE = "-w";
    public static final String TERMINATE_CONNECTION_BEFORE_REPLY = "-t";

    public static final String SEND_INVALID_JSON = "-i";
    public static final String SEND_BLANK_LINE = "-b";
    public static final String SEND_REQUEST_OF_WRONG_TYPE = "-w";
    public static final String TERMINATE_CONNECTION_BEFORE_REQUEST = "-t";
    public static final String TERMINATE_CONNECTION_AFTER_REQUEST = "-a";

    protected boolean faulty_client = false;
    protected boolean faulty_server = false;

    // -------------------------------------------------------------------------------------------------------

    // Suppress setup/teardown from superclass.

    @Override
    @Before
    public void setup() throws Exception {

        faulty_client = false;
        faulty_server = false;
    }

    @Override
    @After
    public void teardown() throws InterruptedException {

    }

    protected void doSetup() throws Exception {

        super.setup();
    }

    protected void doTeardown() throws Exception {

        super.teardown();
    }

    /**
     * Tests correct handling of application-level exceptions.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = TestException.class)
    public void applicationException() throws Exception {

        outputTestName();
        try {
            super.setup();
            server.throwsApplicationException();
        }
        finally {
            super.teardown();
        }
    }

    /**
     * Tests correct handling of runtime exceptions.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = NullPointerException.class)
    public void nullPointerException() throws Exception {

        outputTestName();
        try {
            super.setup();
            server.throwsNullPointerException();
        }
        finally {
            super.teardown();
        }
    }

    /**
     * Tests client handling of case where a connection to a server cannot be established in the first place.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = TimeoutException.class)
    public void serverNeverRunning() throws Exception {

        outputTestName();
        setServerAddress();
        bindToServerWithRetry(server_address);
    }

    /**
     * Tests client handling of case where a connection to a server is established but it subsequently crashes.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    public void serverCrashes() throws Exception {

        outputTestName();
        super.setup();
        super.teardown();

        server.check();
    }

    /**
     * Tests client handling of case where a server accepts a request but never replies.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    // , timeout = SERVER_FAULT_TEST_TIMEOUT)
    public void serverAcceptsRequestButDoesntReply() throws Exception {

        outputTestName();
        faultyServerTest(ACCEPT_REQUEST_BUT_DONT_REPLY);
    }

    /**
     * Tests client handling of case where a server returns invalid JSON.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    public void serverReturnsInvalidJSON() throws Exception {

        outputTestName();
        faultyServerTest(RETURN_INVALID_JSON);
    }

    /**
     * Tests client handling of case where a server returns a blank line.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    public void serverReturnsBlankLine() throws Exception {

        outputTestName();
        faultyServerTest(RETURN_BLANK_LINE);
    }

    /**
     * Tests client handling of case where a server closes the connection before replying.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    public void serverTerminatesConnectionBeforeReply() throws Exception {

        outputTestName();
        faultyServerTest(TERMINATE_CONNECTION_BEFORE_REPLY);
    }

    /**
     * Tests client handling of case where a server returns a result of the wrong type.
     * 
     * @throws Exception if the test fails
     */
    @Test(expected = RPCException.class)
    public void serverReturnsResultOfWrongType() throws Exception {

        outputTestName();
        try {
            faulty_server = true;
            server_args.add(RETURN_WRONG_TYPE);

            super.setup();
            server.equalToThree(THREE);
        }
        finally {
            super.teardown();
        }
    }

    /**
     * Tests server handling of case where a client sends invalid JSON.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void clientSendsInvalidJSON() throws Exception {

        outputTestName();
        faultyClientTest(SEND_INVALID_JSON);
    }

    /**
     * Tests server handling of case where a client sends a blank line.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void clientSendsBlankLine() throws Exception {

        outputTestName();
        faultyClientTest(SEND_BLANK_LINE);
    }

    /**
     * Tests server handling of case where a client closes the connection before sending the request.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void clientTerminatesConnectionBeforeRequest() throws Exception {

        outputTestName();
        faultyClientTest(TERMINATE_CONNECTION_BEFORE_REQUEST);
    }

    /**
     * Tests server handling of case where a client closes the connection after sending the request.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void clientTerminatesConnectionAfterRequest() throws Exception {

        outputTestName();
        faultyClientTest(TERMINATE_CONNECTION_AFTER_REQUEST);
    }

    protected abstract void faultyServerTest(final String fault_flag) throws Exception;

    protected abstract void faultyClientTest(final String fault_flag) throws Exception;

}
