/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test;

import static org.hamcrest.core.Is.is;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.hamcrest.core.IsNull.nullValue;
import static org.junit.Assert.assertThat;

import java.net.BindException;
import java.net.InetSocketAddress;
import java.util.concurrent.TimeUnit;

import org.junit.Test;

import uk.ac.standrews.cs.nds.rpc.AbstractApplicationServer;
import uk.ac.standrews.cs.nds.rpc.AbstractConnectionPool;
import uk.ac.standrews.cs.nds.rpc.AbstractProxy;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.util.Duration;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Base class for RPC mechanism tests.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class RPCTestNormalBase extends RPCTestBase {

    protected static final Duration SLEEP_DELAY = new Duration(2, TimeUnit.SECONDS);
    protected static final long GARBAGE_COLLECTION_TEST_TIMEOUT = 60000;
    private static final int NUMBER_OF_SERVERS = 20;

    protected abstract AbstractConnectionPool getConnectionPool();

    protected abstract AbstractApplicationServer instantiateServer() throws Exception;

    /**
     * Tests basic connectivity.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void check() throws RPCException {

        outputTestName();
        server.check();
    }

    /**
     * Tests correct passing of integer parameter.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passInteger() throws RPCException {

        outputTestName();
        assertThat(server.equalToThree(THREE), is(true));
        assertThat(server.equalToThree(NOT_THREE), is(false));
    }

    /**
     * Tests correct passing of boolean parameter.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passBoolean() throws RPCException {

        outputTestName();
        assertThat(server.not(false), is(true));
        assertThat(server.not(true), is(false));
    }

    /**
     * Tests correct passing of object parameters.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passObject() throws RPCException {

        outputTestName();
        assertThat(server.xIsThree(new TestParameter1(THREE)), is(true));
        assertThat(server.xIsThree(new TestParameter1(NOT_THREE)), is(false));
    }

    /**
     * Tests correct passing of object parameters.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passAndReturnString() throws RPCException {

        outputTestName();
        final String test_string = "abc\n!@$%^&*(){}[]/>";
        assertThat(server.returnsParam(test_string), is(equalTo(test_string)));
    }

    /**
     * Tests correct passing of null parameters.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passNull() throws RPCException {

        outputTestName();
        assertThat(server.xIsThree(null), is(false));
    }

    /**
     * Tests correct returning of null results.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passEmptyString() throws RPCException {

        outputTestName();
        assertThat(server.isEmptyString(""), is(true));
    }

    /**
     * Tests correct passing of multiple object parameters.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passTwoObjects() throws RPCException {

        outputTestName();
        passTwoObjects(server);
    }

    /**
     * Tests correct passing of array parameters.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void passArray() throws RPCException {

        outputTestName();
        assertThat(server.allElementsHaveSameX(new TestParameter1[]{}), is(true));
        assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE)}), is(true));
        assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE), new TestParameter1(THREE)}), is(true));
        assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(NOT_THREE), new TestParameter1(THREE)}), is(false));
        assertThat(server.allElementsHaveSameX(null), is(false));
        assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE), null}), is(false));
    }

    /**
     * Tests correct returning of object results.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void returnObject() throws RPCException {

        outputTestName();
        returnObject(server);
    }

    /**
     * Tests correct returning of array results.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void returnArray() throws RPCException {

        outputTestName();
        final TestParameter1[] array = server.returnsArrayLength2With5And7();

        assertThat(array.length, is(2));
        assertThat(array[0].getX(), is(FIVE));
        assertThat(array[1].getX(), is(SEVEN));
    }

    /**
     * Tests correct returning of null results.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void returnNull() throws RPCException {

        outputTestName();
        assertThat(server.returnsNull(), is(nullValue()));
    }

    /**
     * Tests correct returning of empty string.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void returnEmptyString() throws RPCException {

        outputTestName();
        assertThat(server.returnsEmptyString(), is(equalTo("")));
    }

    /**
     * Tests correct running of multiple successive calls.
     * 
     * @throws RPCException if the test fails
     */
    @Test
    public void multipleSuccessiveCalls() throws RPCException {

        outputTestName();
        passTwoObjects();
        returnObject();
        passTwoObjects();
        returnObject();
    }

    /**
     * Tests socket reuse between calls.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void socketReuse() throws Exception {

        final AbstractProxy proxy = (AbstractProxy) server;
        server.check();
        server.check();
        server.check();
        server.check();
        assertThat(proxy.connectionChangedOnLastCall(), is(false));
    }

    /**
     * Tests that sockets are eventually garbage collected.
     * 
     * @throws RPCException if the test fails
     */
    @Test(timeout = GARBAGE_COLLECTION_TEST_TIMEOUT)
    public void socketGarbageCollection() throws RPCException {

        outputTestName();
        server.check();

        while (getConnectionPool().getNumberOfFreeConnections(server_address) > 0) {
            SLEEP_DELAY.sleep();
        }

        server.check();

        while (getConnectionPool().getNumberOfFreeConnections(server_address) > 0) {
            SLEEP_DELAY.sleep();
        }
    }

    /**
     * Tests whether an attempt to start two servers on the same port fails as expected.
     * 
     * @throws Exception if an error occurs
     */
    @Test(expected = BindException.class)
    public void serverSocketPortClashes() throws Exception {

        outputTestName();

        AbstractApplicationServer server1 = null;
        AbstractApplicationServer server2 = null;
        try {
            server1 = instantiateServer();
            server1.runServer(NetworkUtil.getLocalIPv4Address(), server_port.incrementAndGet());

            server2 = instantiateServer();
            server2.runServer(NetworkUtil.getLocalIPv4Address(), server_port.get());
        }
        finally {
            if (server1 != null) {
                server1.stop();
            }
            if (server2 != null) {
                server2.stop();
            }
        }
    }

    @Test
    public void multipleServers() throws Exception {

        outputTestName();

        final AbstractApplicationServer[] servers = new AbstractApplicationServer[NUMBER_OF_SERVERS];
        final ITest[] proxies = new ITest[NUMBER_OF_SERVERS];

        for (int i = 0; i < NUMBER_OF_SERVERS; i++) {

            final AbstractApplicationServer server = instantiateServer();
            server.runServer(new String[]{"-p" + server_port.incrementAndGet()});
            servers[i] = server;
            proxies[i] = bindToServerWithRetry(new InetSocketAddress(NetworkUtil.getLocalIPv4Address(), server_port.get()));
        }

        for (int i = 0; i < NUMBER_OF_SERVERS; i++) {
            passTwoObjects(proxies[i]);
        }

        for (int i = 0; i < NUMBER_OF_SERVERS; i++) {
            returnObject(proxies[i]);
        }

        for (int i = 0; i < NUMBER_OF_SERVERS; i++) {
            for (int j = 0; j < NUMBER_OF_SERVERS; j++) {
                assertThat(proxies[i].passByReference(proxies[j], 3), is(true));
            }
        }

        for (int i = 0; i < NUMBER_OF_SERVERS; i++) {
            servers[i].stop();
        }
    }
}
