/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test;

import static org.hamcrest.core.Is.is;
import static org.hamcrest.core.IsNull.nullValue;
import static org.junit.Assert.assertThat;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicReference;

import org.junit.Test;

import uk.ac.standrews.cs.nds.rpc.AbstractConnectionPool;
import uk.ac.standrews.cs.nds.rpc.RPCException;

/**
 * Base class for slow-running RPC mechanism tests.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public abstract class RPCTestSlowBase extends RPCTestBase {

    private static final int NUMBER_OF_ITERATIONS = 1000;
    private static final int NUMBER_OF_CONCURRENT_THREADS = 7;

    protected abstract AbstractConnectionPool getConnectionPool();

    /**
     * Tests that stream pooling works with multiple concurrent threads
     * using the same server.
     * 
     * <p>This test fails to find a bug that occurs when there are multiple concurrent connections
     * write to a socket stream and produce invalid JSON strings. This test shows that for roughly n threads,
     * n+1 stream pairs are created, which seems to show that the stream pooling is working correctly -- existing
     * streams are re-used where possible, and new streams are created for the remaining cases.
     * 
     * @author Angus Macdonald (angus.macdonald@st-andrews.ac.uk)
     * @throws RPCException if the test fails
     * @throws InterruptedException if one of the threads is interrupted
     */
    @Test
    public void multiThreadTest() throws RPCException, InterruptedException {

        final CountDownLatch start_latch = new CountDownLatch(1);
        final CountDownLatch end_latch = new CountDownLatch(NUMBER_OF_CONCURRENT_THREADS);
        final AtomicReference<Throwable> exception_holder = new AtomicReference<Throwable>(null);

        for (int i = 0; i < NUMBER_OF_CONCURRENT_THREADS; i++) {

            new RPCTestThread(NUMBER_OF_ITERATIONS, server, start_latch, end_latch, exception_holder).start();
        }

        start_latch.countDown();
        end_latch.await();
        assertThat(exception_holder.get(), is(nullValue()));
    }
}
