package uk.ac.standrews.cs.nds.rpc.test;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertThat;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicReference;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter3;
import uk.ac.standrews.cs.nds.rpc.test.stream.RPCTestNormalOperation;

/**
 * Threaded class, used to run JUnit tests concurrently with other running JUnit tests. Create multiple instances
 * of this class and run them at the same time to check for any concurrency problems with the JSON code.
 * @author Angus Macdonald (angus.macdonald@st-andrews.ac.uk)
 */
public class RPCTestThread extends Thread {

    private final int maxIterations;
    private final ITest server;
    private final CountDownLatch start_latch;
    private final CountDownLatch end_latch;
    private final AtomicReference<Throwable> exception_holder;

    /**
     * 
     * @param maxIterations The number of times the tests in this thread should be run. The tests will loop this number of times, 
     *  or until one of them fails.
     * @param server The JSON server class that messages are to be sent to.
     * @param end_latch 
     * @param start_latch 
     * @param exception_holder 
     */
    public RPCTestThread(final int maxIterations, final ITest server, final CountDownLatch start_latch, final CountDownLatch end_latch, final AtomicReference<Throwable> exception_holder) {

        this.maxIterations = maxIterations;
        this.server = server;
        this.start_latch = start_latch;
        this.end_latch = end_latch;
        this.exception_holder = exception_holder;
    }

    @Override
    public void run() {

        try {
            start_latch.await();
        }
        catch (final InterruptedException e) {
            exception_holder.set(e);
        }

        for (int i = 0; i < maxIterations; i++) {
            try {
                assertThat(server.xSameAsY(new TestParameter1(RPCTestNormalOperation.THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(RPCTestNormalOperation.THREE))), is(true));
                assertThat(server.xSameAsY(new TestParameter1(RPCTestNormalOperation.THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(RPCTestNormalOperation.THREE + 1))), is(false));
                assertThat(server.xSameAsY(new TestParameter1(RPCTestNormalOperation.THREE), new TestParameter2("hello", false, new int[0], null)), is(false));

                assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[0], null)), is(true));
                assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[]{RPCTestNormalOperation.THREE, RPCTestNormalOperation.SEVEN}, null), new TestParameter2("hello", false, new int[]{RPCTestNormalOperation.THREE, RPCTestNormalOperation.SEVEN}, null)), is(true));
                assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, null, null)), is(false));
                assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[]{RPCTestNormalOperation.THREE}, null)), is(false));
            }
            catch (final AssertionError e) {
                exception_holder.set(e);
            }
            catch (final RPCException e) {
                exception_holder.set(e);
            }
        }

        end_latch.countDown();
    }
}
