/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app;

import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Signature;

import uk.ac.standrews.cs.nds.p2p.util.SHA1Hash;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.util.Encoding;

/**
 * Implementation of test service.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public abstract class AbstractSignableTestImpl extends TestImpl implements ISignableTest {

    /** Default Algorithm used to generate {@link KeyPair}. */
    public static final String KEY_PAIR_ALGORITHM = "DSA";
    /** Default Algorithm provider used to generate {@link KeyPair}. */
    public static final String KEY_PAIR_ALGORITHM_PROVIDER = "SUN";
    /** The Algorithm used to generate a {@link Signature} object. */
    public static final String SIGNATURE_ALGORITHM = "SHA1withDSA";
    /** The Algorithm provider of {@link Signature} object. */
    public static final String SIGNATURE_ALGORITHM_PROVIDER = "SUN";

    private static final int KEY_SIZE = 1024;
    private static final SHA1Hash SHA1_HASHER = new SHA1Hash();
    private static KeyPairGenerator key_pair_generator;

    protected final IPublicKeyManager public_key_manager;
    protected final KeyPair key_pair;

    /**
     * Instantiates a new signable test impl.
     * 
     * @param public_key_manager the public_key_manager
     */
    public AbstractSignableTestImpl(final IPublicKeyManager public_key_manager) {

        this.public_key_manager = public_key_manager;
        key_pair = makeKeyPair();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Make signer.
     * 
     * @return the signer
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     */
    public abstract ISigner makeSigner() throws NoSuchAlgorithmException, NoSuchProviderException;

    /**
     * Make verifier.
     * 
     * @return the verifier
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     */
    public abstract IVerifier makeVerifier() throws NoSuchAlgorithmException, NoSuchProviderException;

    // -------------------------------------------------------------------------------------------------------------------------------
    // TestImpl overridden methods.

    @Override
    public boolean passByReference(final ITest other, final int i) throws RPCException {

        return ((ISignableTest) other).equalToThree(i);
    }

    @Override
    public PublicKey getPublicKey() throws RPCException {

        return key_pair.getPublic();
    }

    @Override
    public void notifyPublicKey(final PublicKey public_key) throws RPCException {

        final String signer_id = generateSHA1SignerId(public_key);
        public_key_manager.storePublicKey(signer_id, public_key);
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Generates signer id from sha1 hashing of the given public key.
     * 
     * @param public_key the public_key
     * @return the signer id
     */
    public static String generateSHA1SignerId(final PublicKey public_key) {

        return Encoding.byteArrayToHex(SHA1_HASHER.hash(public_key.getEncoded()));
    }

    /**
     * Makes key pair.
     * 
     * @return the key pair
     */
    public static KeyPair makeKeyPair() {

        KeyPair key_pair;
        try {
            if (key_pair_generator == null) {
                key_pair_generator = KeyPairGenerator.getInstance(KEY_PAIR_ALGORITHM, KEY_PAIR_ALGORITHM_PROVIDER);
                key_pair_generator.initialize(KEY_SIZE, new SecureRandom("".getBytes()));
            }
            key_pair = key_pair_generator.generateKeyPair();
        }
        catch (final Exception e) {
            key_pair = null;
        }

        return key_pair;
    }
}
