/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.nostream;

import java.net.InetSocketAddress;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.nostream.Marshaller;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONObject;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter3;

/**
 * Marshaller for the test application.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class TestMarshaller extends Marshaller {

    /**
     * Serializes the parameter.
     * @param t the parameter
     * @return the serialized representation
     */
    public JSONValue serializeTestParameter1(final TestParameter1 t) {

        if (t == null) { return JSONObject.NULL; }

        final JSONObject object = new JSONObject();
        object.put("x", t.getX());
        return object;
    }

    /**
     * Deserializes the representation.
     * @param object a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter1 deserializeTestParameter1(final JSONObject object) throws DeserializationException {

        if (object == JSONObject.NULL) { return null; }

        try {
            final int x = object.getInt("x");
            return new TestParameter1(x);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * @param t the parameter
     * @return the serialized representation
     */
    public JSONValue serializeTestParameter1Array(final TestParameter1[] t) {

        if (t == null) { return JSONObject.NULL; }

        final JSONArray array = new JSONArray();
        for (final TestParameter1 element : t) {
            array.put(serializeTestParameter1(element));
        }
        return array;
    }

    /**
     * Deserializes the representation.
     * @param array a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter1[] deserializeTestParameter1Array(final JSONArray array) throws DeserializationException {

        if (array == JSONArray.NULL) { return null; }

        try {
            final TestParameter1[] t = new TestParameter1[array.length()];
            for (int i = 0; i < t.length; i++) {
                t[i] = deserializeTestParameter1(array.getJSONObject(i));
            }
            return t;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * @param t2 the parameter
     * @return the serialized representation
     */
    public JSONValue serializeTestParameter2(final TestParameter2 t2) {

        if (t2 == null) { return JSONObject.NULL; }

        final JSONObject object = new JSONObject();

        object.put("s", t2.getS());
        object.put("b", t2.getB());
        object.put("i", serializeIntArray(t2.getI()));
        object.put("t3", serializeTestParameter3(t2.getT3()));

        return object;
    }

    /**
     * Deserializes the representation.
     * @param object a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter2 deserializeTestParameter2(final JSONObject object) throws DeserializationException {

        if (object == JSONObject.NULL) { return null; }

        try {
            final String s = object.getString("s");
            final boolean b = object.getBoolean("b");
            final int[] i = deserializeIntArray(object.getJSONArray("i"));
            final TestParameter3 t3 = deserializeTestParameter3(object.getJSONObject("t3"));

            return new TestParameter2(s, b, i, t3);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * @param t3 the parameter
     * @return the serialized representation
     */
    public JSONValue serializeTestParameter3(final TestParameter3 t3) {

        if (t3 == null) { return JSONObject.NULL; }

        final JSONObject object = new JSONObject();
        object.put("y", t3.getY());
        return object;
    }

    /**
     * Deserializes the representation.
     * @param object a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter3 deserializeTestParameter3(final JSONObject object) throws DeserializationException {

        if (object == JSONObject.NULL) { return null; }

        try {
            final int y = object.getInt("y");

            return new TestParameter3(y);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * @param i the parameter
     * @return the serialized representation
     */
    public JSONValue serializeIntArray(final int[] i) {

        if (i == null) { return JSONObject.NULL; }

        final JSONArray array = new JSONArray();

        for (final int j : i) {
            array.put(j);
        }

        return array;
    }

    /**
     * Deserializes the representation.
     * @param array a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public int[] deserializeIntArray(final JSONArray array) throws DeserializationException {

        if (array == JSONArray.NULL) { return null; }

        try {
            final int[] i = new int[array.length()];
            for (int j = 0; j < i.length; j++) {
                i[j] = array.getInt(j);
            }

            return i;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    public JSONValue serializeITest(final ITest other) {

        final JSONValue serialized_proxy = serializeInetSocketAddress(((TestProxy) other).getProxiedAddress());
        final JSONObject object = new JSONObject();

        object.put("proxy", serialized_proxy);

        return object;
    }

    public ITest deserializeITest(final JSONObject object) throws DeserializationException {

        try {
            final String serialized_address = object.getString("proxy");
            final InetSocketAddress address = deserializeInetSocketAddress(serialized_address);
            return new TestProxy(address);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }
}
