/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.nostream;

import java.net.InetAddress;

import org.json.JSONException;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.nostream.ApplicationServer;
import uk.ac.standrews.cs.nds.rpc.nostream.IHandler;
import uk.ac.standrews.cs.nds.rpc.nostream.Marshaller;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONArray;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONBoolean;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONString;
import uk.ac.standrews.cs.nds.rpc.nostream.json.JSONValue;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestException;
import uk.ac.standrews.cs.nds.rpc.test.app.TestImpl;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.util.CommandLineArgs;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Server for test service.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class TestServer extends ApplicationServer {

    private static final String APPLICATION_REGISTRY_KEY = "TEST";

    private final TestImpl impl;
    private final TestMarshaller marshaller;

    /**
     * Creates a server for the given implementation.
     * @param impl the implementation
     */
    public TestServer(final TestImpl impl) {

        this.impl = impl;
        marshaller = new TestMarshaller();

        initHandlers();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Runs the test service.
     *
     * @param args command line arguments including the port specified by '-p'
     * @throws Exception if the server cannot be started
     */
    public static void main(final String[] args) throws Exception {

        final TestImpl impl = new TestImpl();
        final TestServer server = new TestServer(impl);

        server.runServer(args);
    }

    /**
     * Runs the server at a given network address.
     * @param args args containing the port specified by -p
     * @throws Exception if the server cannot be started
     */
    public void runServer(final String[] args) throws Exception {

        runServer(NetworkUtil.getLocalIPv4Address(), Integer.parseInt(CommandLineArgs.getArg(args, "-p")));
    }

    /**
     * Runs the server at a given network address.
     * @param address the address
     * @param port the port
     * @throws Exception if the server cannot be started
     */
    public void runServer(final InetAddress address, final int port) throws Exception {

        setLocalAddress(address);
        setPort(port);
        start(true);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public Marshaller getMarshaller() {

        return marshaller;
    }

    @Override
    public String getApplicationRegistryKey() {

        return APPLICATION_REGISTRY_KEY;
    }

    // -------------------------------------------------------------------------------------------------------

    private void initHandlers() {

        handler_map.put("allElementsHaveSameX", new AllElementsHaveSameXHandler());
        handler_map.put("arraysEqual", new ArraysEqualHandler());
        handler_map.put("check", new CheckHandler());
        handler_map.put("equalToThree", new EqualToThreeHandler());
        handler_map.put("isEmptyString", new IsEmptyStringHandler());
        handler_map.put("not", new NotHandler());
        handler_map.put("passByReference", new PassByReferenceHandler());
        handler_map.put("returnsArrayLength2With5And7", new ReturnsArrayLength2With5And7Handler());
        handler_map.put("returnsEmptyString", new ReturnsEmptyStringHandler());
        handler_map.put("returnsNull", new ReturnsNullHandler());
        handler_map.put("returnsParam", new ReturnsParamHandler());
        handler_map.put("returnsWithX3", new ReturnsWithX3Handler());
        handler_map.put("throwsApplicationException", new ThrowsApplicationExceptionHandler());
        handler_map.put("throwsNullPointerException", new ThrowsNullPointerExceptionHandler());
        handler_map.put("xIsThree", new XIsThreeHandler());
        handler_map.put("xSameAsY", new XSameAsYHandler());
    }

    // -------------------------------------------------------------------------------------------------------

    private final class AllElementsHaveSameXHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final TestParameter1[] t = marshaller.deserializeTestParameter1Array(args.getJSONArray(0));
                return new JSONBoolean(impl.allElementsHaveSameX(t));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException: " + e.getMessage());
            }
            catch (final DeserializationException e) {
                throw new RPCException(e);
            }
        }
    }

    private final class ArraysEqualHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final TestParameter2 t1 = marshaller.deserializeTestParameter2(args.getJSONObject(0));
                final TestParameter2 t2 = marshaller.deserializeTestParameter2(args.getJSONObject(1));

                return new JSONBoolean(impl.arraysEqual(t1, t2));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException");
            }
            catch (final DeserializationException e) {
                throw new RPCException(e);
            }
        }
    }

    private final class CheckHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            impl.check();
            return JSONString.EMPTY_STRING;
        }
    }

    private final class EqualToThreeHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final int i = args.getInt(0);
                return new JSONBoolean(impl.equalToThree(i));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException");
            }
        }
    }

    private final class IsEmptyStringHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                return new JSONBoolean(impl.isEmptyString(args.getString(0)));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException: " + e.getMessage());
            }
        }
    }

    private final class NotHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final boolean b = args.getBoolean(0);
                return new JSONBoolean(impl.not(b));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException");
            }
        }
    }

    private final class PassByReferenceHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final ITest other = marshaller.deserializeITest(args.getJSONObject(0));
                final int i = args.getInt(1);
                return new JSONBoolean(impl.passByReference(other, i));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException: " + e.getMessage());
            }
            catch (final DeserializationException e) {
                throw new RPCException(e);
            }
        }
    }

    private final class ReturnsArrayLength2With5And7Handler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            return marshaller.serializeTestParameter1Array(impl.returnsArrayLength2With5And7());
        }
    }

    private final class ReturnsEmptyStringHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            return new JSONString(impl.returnsEmptyString());
        }
    }

    private final class ReturnsNullHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            return marshaller.serializeTestParameter1(impl.returnsNull());
        }
    }

    private final class ReturnsParamHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                return new JSONString(impl.returnsParam(args.getString(0)));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException: " + e.getMessage());
            }
        }
    }

    private final class ReturnsWithX3Handler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            return marshaller.serializeTestParameter1(impl.returnsWithX3());
        }
    }

    private final class ThrowsApplicationExceptionHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException, TestException {

            impl.throwsApplicationException();
            return JSONString.EMPTY_STRING;
        }
    }

    private final class ThrowsNullPointerExceptionHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            impl.throwsNullPointerException();
            return JSONString.EMPTY_STRING;
        }
    }

    private final class XIsThreeHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                return new JSONBoolean(impl.xIsThree(marshaller.deserializeTestParameter1(args.getJSONObject(0))));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException: " + e.getMessage());
            }
            catch (final DeserializationException e) {
                throw new RPCException(e);
            }
        }
    }

    private final class XSameAsYHandler implements IHandler {

        @Override
        public JSONValue execute(final JSONArray args) throws RPCException {

            try {
                final TestParameter1 t1 = marshaller.deserializeTestParameter1(args.getJSONObject(0));
                final TestParameter2 t2 = marshaller.deserializeTestParameter2(args.getJSONObject(1));

                return new JSONBoolean(impl.xSameAsY(t1, t2));
            }
            catch (final JSONException e) {
                throw new RPCException("JSONException");
            }
            catch (final DeserializationException e) {
                throw new RPCException(e);
            }
        }
    }
}
