/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.stream;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.stream.Marshaller;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter3;

/**
 * Marshaller for the test application.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class TestMarshaller extends Marshaller {

    /**
     * Serializes the parameter.
     * 
     * @param t the parameter
     * @param jw the jw
     */
    public void serializeTestParameter1(final TestParameter1 t, final JSONWriter jw) {

        try {
            if (t == null) {
                jw.value(t);
                return;
            }

            jw.object();
            jw.key("x");
            jw.value(t.getX());
            jw.endObject();
        }
        catch (final JSONException e) {
            e.printStackTrace();
        }
    }

    /**
     * Deserializes the representation.
     * 
     * @param reader - the stream of values containing an object
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter1 deserializeTestParameter1(final JSONReader reader) throws DeserializationException {

        try {
            if (reader.checkNull()) { return null; }

            reader.object();
            reader.key("x");
            final int x = reader.intValue();
            reader.endObject();
            return new TestParameter1(x);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * 
     * @param t the parameter
     * @param jw the jw
     */
    public void serializeTestParameter1Array(final TestParameter1[] t, final JSONWriter jw) {

        try {
            if (t == null) {
                jw.value(t);
                return;
            }

            jw.array(); // final JSONArray array = new JSONArray();
            for (final TestParameter1 element : t) {
                serializeTestParameter1(element, jw);
            }
            jw.endArray(); // return array;
        }
        catch (final JSONException e) {
            e.printStackTrace();
        }
    }

    /**
     * Deserializes the representation.
     * 
     * @param reader a stream starting with an array a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     */
    public TestParameter1[] deserializeTestParameter1Array(final JSONReader reader) throws DeserializationException {

        final TestParameter1[] dummy = new TestParameter1[0];

        try {
            if (reader.checkNull()) { return null; }

            reader.array();
            final List<TestParameter1> l = new ArrayList<TestParameter1>();
            while (!reader.have(JSONReader.ENDARRAY)) {
                l.add(deserializeTestParameter1(reader));
            }
            reader.endArray();
            return l.toArray(dummy);

        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * 
     * @param t2 the parameter
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeTestParameter2(final TestParameter2 t2, final JSONWriter writer) throws JSONException {

        if (t2 == null) {
            writer.value(t2);
            return;
        }

        writer.object();
        writer.key("s");
        writer.value(t2.getS());
        writer.key("b");
        writer.value(t2.getB());
        writer.key("i");
        serializeIntArray(t2.getI(), writer);
        writer.key("t3");
        serializeTestParameter3(t2.getT3(), writer);
        writer.endObject();
    }

    /**
     * Deserializes the representation.
     * 
     * @param reader = a stream containing a serialized representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     * @throws JSONException the jSON exception
     */
    public TestParameter2 deserializeTestParameter2(final JSONReader reader) throws DeserializationException, JSONException {

        if (reader.checkNull()) { return null; }

        try {
            reader.object();
            reader.key("s");
            final String s = reader.stringValue();
            reader.key("b");
            final boolean b = reader.booleanValue();
            reader.key("i");
            final int[] i = deserializeIntArray(reader);
            reader.key("t3");
            final TestParameter3 t3 = deserializeTestParameter3(reader);
            reader.endObject();
            return new TestParameter2(s, b, i, t3);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * 
     * @param t3 the parameter
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeTestParameter3(final TestParameter3 t3, final JSONWriter writer) throws JSONException {

        if (t3 == null) {
            writer.value(t3);
            return;
        }

        writer.object();
        writer.key("y");
        writer.value(t3.getY());
        writer.endObject();
    }

    /**
     * Deserializes the representation.
     * 
     * @param reader a stream containing a serialized object representation
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     * @throws JSONException the jSON exception
     */
    public TestParameter3 deserializeTestParameter3(final JSONReader reader) throws DeserializationException, JSONException {

        if (reader.checkNull()) { return null; }

        try {
            reader.object();
            reader.key("y");
            final int y = reader.intValue();
            reader.endObject();

            return new TestParameter3(y);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serializes the parameter.
     * 
     * @param i the parameter
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeIntArray(final int[] i, final JSONWriter writer) throws JSONException {

        if (i == null) {
            writer.value(i);
            return;
        }

        writer.array();
        for (final int j : i) {
            writer.value(j);
        }
        writer.endArray();
    }

    /**
     * Deserializes the representation.
     * 
     * @param reader the reader
     * @return the deserialized value
     * @throws DeserializationException if the deserialization fails
     * @throws JSONException the jSON exception
     */
    public int[] deserializeIntArray(final JSONReader reader) throws DeserializationException, JSONException {

        if (reader.have(JSONReader.NULL)) {
            reader.nullValue();
            return null;
        }

        final List<Integer> list = new ArrayList<Integer>();

        try {
            reader.array();

            while (!reader.have(JSONReader.ENDARRAY)) {
                list.add(reader.intValue());
            }
            reader.endArray();
            final int[] result = new int[list.size()];
            int index = 0;
            for (final int i : result) {
                result[index++] = i;
            }
            return result;
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Serialize i test.
     * 
     * @param other the other
     * @param writer the writer
     * @throws JSONException the jSON exception
     */
    public void serializeITest(final ITest other, final JSONWriter writer) throws JSONException {

        writer.object();
        writer.key("proxy");
        serializeInetSocketAddress(((TestProxy) other).getProxiedAddress(), writer);
        writer.endObject();

    }

    /**
     * Deserialize i test.
     * 
     * @param reader the reader
     * @return the test
     * @throws DeserializationException the deserialization exception
     * @throws JSONException the jSON exception
     */
    public ITest deserializeITest(final JSONReader reader) throws DeserializationException, JSONException {

        if (reader.checkNull()) { return null; }

        try {
            reader.object();
            reader.key("proxy");
            final InetSocketAddress address = deserializeInetSocketAddress(reader);
            reader.endObject();

            return new TestProxy(address);
        }
        catch (final Exception e) {
            throw new DeserializationException(e);
        }

    }
}
