/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.stream;

import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.stream.AbstractStreamConnection;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.stream.StreamProxy;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestException;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;

/**
 * Implementes the ITest methods.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class TestProxyImpl implements ITest {

    private final StreamProxy proxy;
    private final TestMarshaller marshaller;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new test stream impl.
     * 
     * @param proxy the proxy
     */
    public TestProxyImpl(final StreamProxy proxy) {

        this.proxy = proxy;
        marshaller = (TestMarshaller) proxy.getMarshaller();
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    @Override
    public void check() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("check");
            proxy.handleVoidCall(proxy.makeCall(connection));
            proxy.finishCall(connection);
        }
        catch (final Exception e) {
            e.printStackTrace();
            proxy.dealWithException(e);
        }
    }

    @Override
    public boolean equalToThree(final int i) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("equalToThree");
            final JSONWriter jw = connection.getJSONwriter();
            jw.value(i);
            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean not(final boolean b) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("not");
            final JSONWriter jw = connection.getJSONwriter();
            jw.value(b);
            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean xIsThree(final TestParameter1 t) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("xIsThree");
            final JSONWriter jw = connection.getJSONwriter();
            marshaller.serializeTestParameter1(t, jw);
            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean isEmptyString(final String s) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("isEmptyString");
            final JSONWriter jw = connection.getJSONwriter();
            jw.value(s);
            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean xSameAsY(final TestParameter1 t1, final TestParameter2 t2) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("xSameAsY");
            final JSONWriter jw = connection.getJSONwriter();

            marshaller.serializeTestParameter1(t1, jw);
            marshaller.serializeTestParameter2(t2, jw);

            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean arraysEqual(final TestParameter2 t1, final TestParameter2 t2) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("arraysEqual");
            final JSONWriter jw = connection.getJSONwriter();

            marshaller.serializeTestParameter2(t1, jw);
            marshaller.serializeTestParameter2(t2, jw);

            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public boolean allElementsHaveSameX(final TestParameter1[] t) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("allElementsHaveSameX");
            final JSONWriter jw = connection.getJSONwriter();

            marshaller.serializeTestParameter1Array(t, jw);

            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }

    @Override
    public TestParameter1 returnsWithX3() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("returnsWithX3");

            final JSONReader reader = proxy.makeCall(connection);

            final TestParameter1 result = marshaller.deserializeTestParameter1(reader);
            proxy.finishCall(connection);
            return result;
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public TestParameter1[] returnsArrayLength2With5And7() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("returnsArrayLength2With5And7");

            final JSONReader reader = proxy.makeCall(connection);
            final TestParameter1[] result = marshaller.deserializeTestParameter1Array(reader);
            proxy.finishCall(connection);
            return result;
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public void throwsApplicationException() throws RPCException, TestException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("throwsApplicationException");
            proxy.handleVoidCall(proxy.makeCall(connection));
            proxy.finishCall(connection);
        }
        catch (final TestException e) {
            throw e;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
        }
    }

    @Override
    public void throwsNullPointerException() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("throwsNullPointerException");
            proxy.handleVoidCall(proxy.makeCall(connection));
            proxy.finishCall(connection);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
        }
    }

    @Override
    public TestParameter1 returnsNull() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("returnsNull");
            final JSONReader reader = proxy.makeCall(connection);

            final TestParameter1 result = marshaller.deserializeTestParameter1(reader);
            proxy.finishCall(connection);
            return result;
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public String returnsEmptyString() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("returnsEmptyString");
            final JSONReader reader = proxy.makeCall(connection);

            final String result = reader.stringValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public String returnsParam(final String s) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("returnsParam");
            final JSONWriter jw = connection.getJSONwriter();
            jw.value(s);
            final JSONReader reader = proxy.makeCall(connection);
            final String result = reader.stringValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public boolean passByReference(final ITest other, final int i) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("passByReference");
            final JSONWriter jw = connection.getJSONwriter();

            marshaller.serializeITest(other, jw);
            jw.value(i);

            final JSONReader reader = proxy.makeCall(connection);
            final boolean result = reader.booleanValue();
            proxy.finishCall(connection);
            return result;
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return false;
        }
    }
}
