/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.stream;

import java.io.IOException;
import java.net.Socket;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.util.Set;

import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.stream.ApplicationServer;
import uk.ac.standrews.cs.nds.rpc.stream.RequestHandler;
import uk.ac.standrews.cs.nds.rpc.stream.SignableRequestHandler;
import uk.ac.standrews.cs.nds.rpc.test.app.AbstractSignableTestImpl;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestSigner;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestVerifier;

/**
 * The Class SignableStreamTestImpl.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class TestSignableImpl extends AbstractSignableTestImpl {

    private final String self_signer_id;
    private final IVerifier verifier;
    private final ISigner signer;

    /**
     * Instantiates a new signable stream test impl.
     * 
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     */
    public TestSignableImpl() throws NoSuchAlgorithmException, NoSuchProviderException {

        super(new TestPublicKeyManager());
        self_signer_id = generateSHA1SignerId(key_pair.getPublic());
        public_key_manager.storePublicKey(self_signer_id, key_pair.getPublic()); // store self public key

        signer = new TestSigner(self_signer_id, key_pair.getPrivate(), SIGNATURE_ALGORITHM, SIGNATURE_ALGORITHM_PROVIDER);
        verifier = new TestVerifier(SIGNATURE_ALGORITHM, SIGNATURE_ALGORITHM_PROVIDER);
    }

    // -------------------------------------------------------------------------------------------------------------------------------
    // AbstractSignableTestImpl overridden methods.

    @Override
    public ISigner makeSigner() {

        return signer;
    }

    @Override
    public IVerifier makeVerifier() {

        return verifier;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Make signable request handler.
     * 
     * @param server the server
     * @param socket the socket
     * @param extant_handlers the extant_handlers
     * @return the signable request handler
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public SignableRequestHandler makeSignableRequestHandler(final ApplicationServer server, final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        final SignableRequestHandler signable_request_handler = new SignableRequestHandler(server, socket, extant_handlers);

        try {
            signable_request_handler.enableSigning(signer);
        }
        catch (final InvalidKeyException e) {
            e.printStackTrace();
            throw new IOException(e);
        }

        signable_request_handler.enableVerification(verifier, public_key_manager);

        return signable_request_handler;
    }
}
