/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.app.stream;

import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;

import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.DeserializationException;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;
import uk.ac.standrews.cs.nds.rpc.stream.AbstractStreamConnection;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.test.app.AbstractSignableTestImpl;
import uk.ac.standrews.cs.nds.rpc.test.app.ISignableTest;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestSigner;
import uk.ac.standrews.cs.nds.rpc.test.stream.TestVerifier;

/**
 * Implementes the ITest methods.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
public class TestSignableProxyImpl extends TestProxyImpl implements ISignableTest {

    private final TestSignableProxy proxy;
    private final TestSignableMarshaller marshaller;
    private final String signer_id;
    private final KeyPair key_pair;
    private final IPublicKeyManager public_key_manager;
    private final ISigner signer;
    private final IVerifier verifier;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instantiates a new test stream impl.
     * 
     * @param proxy the proxy
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     */
    public TestSignableProxyImpl(final TestSignableProxy proxy) throws NoSuchAlgorithmException, NoSuchProviderException {

        super(proxy);
        this.proxy = proxy;
        marshaller = proxy.getMarshaller();

        key_pair = AbstractSignableTestImpl.makeKeyPair();

        signer_id = AbstractSignableTestImpl.generateSHA1SignerId(key_pair.getPublic());

        public_key_manager = new TestPublicKeyManager();
        public_key_manager.storePublicKey(signer_id, key_pair.getPublic()); // store self public key

        signer = new TestSigner(signer_id, key_pair.getPrivate(), AbstractSignableTestImpl.SIGNATURE_ALGORITHM, AbstractSignableTestImpl.SIGNATURE_ALGORITHM_PROVIDER);
        verifier = new TestVerifier(AbstractSignableTestImpl.SIGNATURE_ALGORITHM, AbstractSignableTestImpl.SIGNATURE_ALGORITHM_PROVIDER);
    }

    @Override
    public PublicKey getPublicKey() throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("getPublicKey");
            final JSONReader reader = proxy.makeCall(connection);

            final PublicKey result = proxy.getMarshaller().deserializePublicKey(reader);
            proxy.finishCall(connection);
            return result;
        }
        catch (final DeserializationException e) {
            throw new RPCException(e);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
            return null;
        }
    }

    @Override
    public void notifyPublicKey(final PublicKey public_key) throws RPCException {

        try {
            final AbstractStreamConnection connection = proxy.startCall("notifyPublicKey");
            final JSONWriter writer = connection.getJSONwriter();
            marshaller.serializePublicKey(key_pair.getPublic(), writer);

            proxy.handleVoidCall(proxy.makeCall(connection));

            proxy.finishCall(connection);
        }
        catch (final Exception e) {
            proxy.dealWithException(e);
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * Handle public key not found remote exception.
     * 
     * @throws RPCException the rPC exception
     */
    public void handlePublicKeyNotFoundRemoteException() throws RPCException {

        proxy.disableSigning();
        proxy.disableVerification();
        notifyPublicKey(key_pair.getPublic());
        proxy.enableSigning(signer);
        proxy.enableVerification(verifier, public_key_manager);

    }

    /**
     * Handle public key not found local exception.
     * 
     * @throws RPCException the rPC exception
     */
    public void handlePublicKeyNotFoundLocalException() throws RPCException {

        proxy.disableSigning();
        proxy.disableVerification();
        final PublicKey public_key = getPublicKey();
        public_key_manager.storePublicKey(AbstractSignableTestImpl.generateSHA1SignerId(public_key), public_key);
        proxy.enableSigning(signer);
        proxy.enableVerification(verifier, public_key_manager);
    }
}
