/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.nostream;

import java.io.IOException;
import java.net.Socket;
import java.util.Set;

import uk.ac.standrews.cs.nds.rpc.nostream.RequestHandler;
import uk.ac.standrews.cs.nds.rpc.test.app.TestImpl;
import uk.ac.standrews.cs.nds.rpc.test.app.nostream.TestServer;
import uk.ac.standrews.cs.nds.util.CommandLineArgs;

/**
 * Server sub-class allowing various faults for testing.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class TestServerFaulty extends TestServer {

    private boolean dont_reply_to_ping = false;
    private boolean returns_invalid_json = false;
    private boolean returns_blank_line = false;
    private boolean returns_wrong_type = false;
    private boolean terminates_connection_before_reply = false;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a test server.
     *
     * @param impl the implementation
     * @param dont_reply_to_ping if the server shouldn't reply to ping requests
     * @param returns_invalid_json if the server should return invalid JSON to ping requests
     * @param returns_blank_line if the server should return a blank line to ping requests
     * @param returns_wrong_type if the server should return a value of the wrong type to equalToThree requests
     * @param terminates_connection_before_reply if the server should close the connection before replying to ping requests
     */
    public TestServerFaulty(final TestImpl impl, final boolean dont_reply_to_ping, final boolean returns_invalid_json, final boolean returns_blank_line, final boolean returns_wrong_type, final boolean terminates_connection_before_reply) {

        super(impl);

        this.dont_reply_to_ping = dont_reply_to_ping;
        this.returns_invalid_json = returns_invalid_json;
        this.returns_blank_line = returns_blank_line;
        this.returns_wrong_type = returns_wrong_type;
        this.terminates_connection_before_reply = terminates_connection_before_reply;
    }

    // -------------------------------------------------------------------------------------------------------

    public boolean dontReplyToCheck() {

        return dont_reply_to_ping;
    }

    protected boolean returnsInvalidJson() {

        return returns_invalid_json;
    }

    public boolean returnsBlankLine() {

        return returns_blank_line;
    }

    protected boolean returnsWrongType() {

        return returns_wrong_type;
    }

    protected boolean terminatesConnectionBeforeReply() {

        return terminates_connection_before_reply;
    }

    @Override
    protected RequestHandler makeRequest(final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        return new RequestHandlerFaulty(this, socket, extant_handlers);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Runs the test service.
     *
     * @param args command line arguments including the port specified by '-p'
     * @throws Exception if the server cannot be started
     */
    public static void main(final String[] args) throws Exception {

        final boolean dont_reply_to_ping = CommandLineArgs.containsArg(args, RPCTestErrorHandling.ACCEPT_REQUEST_BUT_DONT_REPLY);
        final boolean returns_invalid_json = CommandLineArgs.containsArg(args, RPCTestErrorHandling.RETURN_INVALID_JSON);
        final boolean returns_blank_line = CommandLineArgs.containsArg(args, RPCTestErrorHandling.RETURN_BLANK_LINE);
        final boolean returns_wrong_type = CommandLineArgs.containsArg(args, RPCTestErrorHandling.RETURN_WRONG_TYPE);
        final boolean terminates_connection_before_reply = CommandLineArgs.containsArg(args, RPCTestErrorHandling.TERMINATE_CONNECTION_BEFORE_REPLY);

        final TestImpl impl = new TestImpl();
        final TestServer server = new TestServerFaulty(impl, dont_reply_to_ping, returns_invalid_json, returns_blank_line, returns_wrong_type, terminates_connection_before_reply);

        server.runServer(args);
    }
}
