package uk.ac.standrews.cs.nds.rpc.test.stream;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertThat;

import java.lang.reflect.InvocationTargetException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.concurrent.TimeoutException;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.app.ITest;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter1;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter2;
import uk.ac.standrews.cs.nds.rpc.test.app.TestParameter3;
import uk.ac.standrews.cs.nds.rpc.test.app.stream.TestProxy;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

public class ClientRunner {

    static final int SERVER_PORT = 35050;

    static final int TIMEOUT_INTERVAL = 20000;
    static final int RETRY_INTERVAL = 2000;

    static final int THREE = 3;
    static final int FIVE = 5;
    static final int SEVEN = 7;
    static final int NOT_THREE = -345987235;

    private InetSocketAddress server_address;

    private ITest server;

    public static void main(final String[] args) throws Exception {

        final ClientRunner cr = new ClientRunner();
        cr.setServerAddress();
        cr.bindToServer();
        cr.timeWork();
    }

    protected void setServerAddress() throws UnknownHostException {

        server_address = new InetSocketAddress(NetworkUtil.getLocalIPv4Address(), SERVER_PORT);
    }

    protected void bindToServer() throws Exception {

        server = bindToServerWithRetry(server_address);
    }

    ITest bindToServer(final InetSocketAddress server_address) throws RPCException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {

        final TestProxy server = new TestProxy(server_address);

        server.check();
        return server;
    }

    ITest bindToServerWithRetry(final InetSocketAddress server_address) throws TimeoutException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {

        final long start_time = System.currentTimeMillis();

        while (true) {

            try {
                return bindToServer(server_address);
            }
            catch (final RPCException e) {
                Diagnostic.trace(DiagnosticLevel.FULL, "remote binding failed: " + e.getMessage());
            }

            try {
                Thread.sleep(RETRY_INTERVAL);
            }
            catch (final InterruptedException e) {
                // Ignore.
            }

            final long duration = System.currentTimeMillis() - start_time;
            if (duration > TIMEOUT_INTERVAL) { throw new TimeoutException(); }
        }
    }

    public void timeWork() throws RPCException {

        final long t1 = System.currentTimeMillis();

        for (int i = 0; i < 1000; i++) {
            // returnObject
            assertThat(server.returnsWithX3().getX(), is(THREE));

            // xSameAsY pass 2 objects
            assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(THREE))), is(true));
            assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], new TestParameter3(THREE + 1))), is(false));
            assertThat(server.xSameAsY(new TestParameter1(THREE), new TestParameter2("hello", false, new int[0], null)), is(false));

            // arrays equal
            assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[0], null)), is(true));
            assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[]{THREE, SEVEN}, null), new TestParameter2("hello", false, new int[]{THREE, SEVEN}, null)), is(true));
            assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, null, null)), is(false));
            assertThat(server.arraysEqual(new TestParameter2("hello", false, new int[0], null), new TestParameter2("hello", false, new int[]{THREE}, null)), is(false));

            // passArray
            assertThat(server.allElementsHaveSameX(new TestParameter1[]{}), is(true));
            assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE)}), is(true));
            assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE), new TestParameter1(THREE)}), is(true));
            assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(NOT_THREE), new TestParameter1(THREE)}), is(false));
            assertThat(server.allElementsHaveSameX(null), is(false));
            assertThat(server.allElementsHaveSameX(new TestParameter1[]{new TestParameter1(THREE), null}), is(false));
            // returnArray
            final TestParameter1[] array = server.returnsArrayLength2With5And7();
            assertThat(array.length, is(2));
            assertThat(array[0].getX(), is(FIVE));
            assertThat(array[1].getX(), is(SEVEN));
        }

        final long t2 = System.currentTimeMillis();

        final long elapsed = t2 - t1;

        System.out.println("Stream version - elapsed time (ms) = " + elapsed);

    }

}
