/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.stream;

import static org.junit.Assert.fail;

import java.lang.reflect.InvocationTargetException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;

import org.junit.Test;

import uk.ac.standrews.cs.nds.registry.NotBoundException;
import uk.ac.standrews.cs.nds.registry.RegistryUnavailableException;
import uk.ac.standrews.cs.nds.registry.stream.RegistryFactory;
import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.test.RPCTestErrorBase;
import uk.ac.standrews.cs.nds.rpc.test.app.ISignableTest;
import uk.ac.standrews.cs.nds.rpc.test.app.stream.TestSignableProxy;
import uk.ac.standrews.cs.nds.rpc.test.app.stream.TestSignableServer;

/**
 * Tests for RPC mechanism in error cases.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RPCTestSignableErrorHandling extends RPCTestErrorBase {

    @Override
    protected Class<? extends TestSignableServer> getServerClass() {

        return faulty_server ? TestSignableServerFaulty.class : TestSignableServer.class;
    }

    @Override
    protected ISignableTest bindToServer(final InetSocketAddress server_address) throws RPCException, InstantiationException, IllegalAccessException, InvocationTargetException {

        TestSignableProxy server;
        try {
            server = faulty_client ? new TestSignableProxyFaulty(server_address) : new TestSignableProxy(server_address);
            server.ping();
            return server;
        }
        catch (final NoSuchAlgorithmException e) {
            fail(e.getMessage());
        }
        catch (final NoSuchProviderException e) {
            fail(e.getMessage());
        }
        return null;
    }

    @Override
    protected void shutdownRegistry() throws UnknownHostException, RegistryUnavailableException, RPCException, NotBoundException {

        RegistryFactory.FACTORY.getRegistry().shutdown();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests server handling of case where a client sends a request of the wrong type.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void clientSendsRequestOfWrongType() throws Exception {

        try {
            faulty_client = true;
            doSetup();

            ((TestSignableProxyFaulty) server).setFaultFlag(SEND_REQUEST_OF_WRONG_TYPE);

            try {
                server.equalToThree(THREE);
            }
            catch (final RPCException e) {

                // Expected exception.
                // Check server still working properly.
                faulty_client = false;
                bindToServer();
                server.check();
                return;
            }

            fail("expected RPC exception when sending request of wrong type");
        }
        catch (final Exception e) {
            e.printStackTrace();
            throw e;
        }
        finally {
            doTeardown();
        }
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    protected void faultyServerTest(final String fault_flag) throws Exception {

        try {
            faulty_server = true;
            server_args.add(fault_flag);

            doSetup();
            server.check();
        }
        finally {
            doTeardown();
        }
    }

    @Override
    protected void faultyClientTest(final String fault_flag) throws Exception {

        try {
            faulty_client = true;
            doSetup();

            ((TestSignableProxyFaulty) server).setFaultFlag(fault_flag);

            try {
                server.check();
            }
            catch (final Exception e) { // RPCException

                // Expected exception.
                // Check server still working properly.
                faulty_client = true;
                bindToServer();
                server.check();
                return;
            }

            fail("expected RPC exception");
        }
        finally {
            doTeardown();
        }
    }
}
