/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.stream;

import java.io.IOException;
import java.net.Socket;
import java.util.Set;

import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.stream.RequestHandler;
import uk.ac.standrews.cs.nds.rpc.stream.StreamProxy;

/**
 * Request handler sub-class allowing various faults for testing.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class RequestHandlerFaulty extends RequestHandler {

    /** The number three. */
    public static final int THREE = 3;

    private final TestServerFaulty faulty_server;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises a request handler.
     * 
     * @param faulty_server the application server
     * @param socket the socket associated with request
     * @param extant_handlers the extant_handlers
     * @throws IOException if input and output streams for the socket cannot be created
     */
    public RequestHandlerFaulty(final TestServerFaulty faulty_server, final Socket socket, final Set<RequestHandler> extant_handlers) throws IOException {

        super(faulty_server, socket, extant_handlers);
        this.faulty_server = faulty_server;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    protected void reply(final String method_name, final int id) {

        if (method_name.equals("check")) {

            if (faulty_server.dontReplyToCheck()) { return; }

            if (faulty_server.returnsInvalidJson()) {
                sendFaultyJson();
                return;
            }

            if (faulty_server.returnsBlankLine()) {
                sendBlankLine();
                return;
            }

            if (faulty_server.terminatesConnectionBeforeReply()) {
                try {
                    super.tearDownConnection();
                    super.tearDownSocket();
                }
                catch (final IOException e) {
                    // Ignore.
                }
                return;
            }
        }

        if (method_name.equals("equalToThree")) {

            if (faulty_server.returnsWrongType()) {
                badReply(method_name, id);
                return;
            }
        }

        super.reply(method_name, id);
    }

    protected void badReply(final String method_name, final int id) {

        try {
            final JSONWriter writer = connection.getJSONwriter(); // was newJSONwriter

            writer.object();
            writer.key(RESULT_KEY);
            writer.value(THREE); // Illegal result
            writer.key(StreamProxy.ID_KEY);
            writer.value(id);
            writer.key(StreamProxy.VERSION_KEY);
            writer.value(JSON_VERSION);
            writer.endObject();
            connection.completeWriteRequest();
        }
        catch (final Exception e) {
            // Ignore.
        }
    }

    protected void sendFaultyJson() {

        try {
            final JSONWriter writer = connection.getJSONwriter(); // was newJSONwriter

            writer.value("not a valid JSON string");
            connection.completeWriteRequest();
        }
        catch (final Exception e) {
            // This will result in an error too!
        }

    }

    protected void sendBlankLine() {

        try {
            final JSONWriter writer = connection.getJSONwriter(); // was newJSONwriter

            writer.value("\n");

            connection.completeWriteRequest();
        }
        catch (final Exception e) {
            // This will result in an error too!
        }
    }

    @Override
    protected void tearDownConnection() {

    }

    @Override
    protected void tearDownSocket() throws IOException {

    }
}
