/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.stream;

import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.Signature;

import uk.ac.standrews.cs.nds.rpc.security.IPublicKeyManager;
import uk.ac.standrews.cs.nds.rpc.security.ISigner;
import uk.ac.standrews.cs.nds.rpc.security.IVerifier;

/**
 * A factory for creating Test Signed objects.
 * 
 * @author Masih Hajiarabderkani (masih@cs.st-andrews.ac.uk)
 */
final class TestSignableFactory {

    /** Default Algorithm used to generate {@link KeyPair}. */
    public static final String KEY_PAIR_ALGORITHM = "DSA";
    /** Default Algorithm provider used to generate {@link KeyPair}. */
    public static final String KEY_PAIR_ALGORITHM_PROVIDER = "SUN";
    /** The Algorithm used to generate a {@link Signature} object. */
    public static final String SIGNATURE_ALGORITHM = "SHA1withDSA";
    /** The Algorithm provider of {@link Signature} object. */
    public static final String SIGNATURE_ALGORITHM_PROVIDER = "SUN";

    private static final int KEY_SIZE = 1024;

    private static KeyPairGenerator key_pair_generator;

    // -------------------------------------------------------------------------------------------------------------------------------

    private TestSignableFactory() {

    }

    // -------------------------------------------------------------------------------------------------------------------------------

    static ISigner makeSigner(final String signer_id, final PrivateKey private_key) throws NoSuchAlgorithmException, NoSuchProviderException {

        return new TestSigner(signer_id, private_key, SIGNATURE_ALGORITHM, SIGNATURE_ALGORITHM_PROVIDER);
    }

    static IVerifier makeVerifier() throws NoSuchAlgorithmException, NoSuchProviderException {

        return new TestVerifier(SIGNATURE_ALGORITHM, SIGNATURE_ALGORITHM_PROVIDER);
    }

    static IPublicKeyManager makePublicKeyManager() {

        return new TestPublicKeyManager();
    }

    static KeyPair makeKeyPair() {

        KeyPair key_pair;
        try {
            if (key_pair_generator == null) {
                key_pair_generator = KeyPairGenerator.getInstance(KEY_PAIR_ALGORITHM, KEY_PAIR_ALGORITHM_PROVIDER);
                key_pair_generator.initialize(KEY_SIZE, new SecureRandom("".getBytes()));
            }
            key_pair = key_pair_generator.generateKeyPair();
        }
        catch (final Exception e) {
            key_pair = null;
        }

        return key_pair;
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    @Override
    public Object clone() throws CloneNotSupportedException {

        throw new CloneNotSupportedException("not clonable, move along...");
    }

}
