/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.rpc.test.stream;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;

import org.json.JSONException;
import org.json.JSONWriter;

import uk.ac.standrews.cs.nds.rpc.RPCException;
import uk.ac.standrews.cs.nds.rpc.stream.AbstractStreamConnection;
import uk.ac.standrews.cs.nds.rpc.stream.JSONReader;
import uk.ac.standrews.cs.nds.rpc.test.app.stream.TestSignableProxy;
import uk.ac.standrews.cs.nds.util.Diagnostic;

/**
 * Proxy sub-class allowing various faults for testing.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class TestSignableProxyFaulty extends TestSignableProxy {

    private boolean send_invalid_json = false;
    private boolean send_blank_line = false;
    private boolean send_request_of_wrong_type = false;
    private boolean terminate_connection_before_request = false;
    private boolean terminate_connection_after_request = false;
    private String method_name = "";

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a proxy for a given server address.
     *
     * @param node_address the address
     * @throws NoSuchAlgorithmException the no such algorithm exception
     * @throws NoSuchProviderException the no such provider exception
     */
    public TestSignableProxyFaulty(final InetSocketAddress node_address) throws NoSuchAlgorithmException, NoSuchProviderException {

        super(node_address);
    }

    // -------------------------------------------------------------------------------------------------------

    protected void setFaultFlag(final String fault_flag) {

        if (fault_flag.equals(RPCTestErrorHandling.SEND_INVALID_JSON)) {
            send_invalid_json = true;
        }
        if (fault_flag.equals(RPCTestErrorHandling.SEND_BLANK_LINE)) {
            send_blank_line = true;
        }
        if (fault_flag.equals(RPCTestErrorHandling.SEND_REQUEST_OF_WRONG_TYPE)) {
            send_request_of_wrong_type = true;
        }
        if (fault_flag.equals(RPCTestErrorHandling.TERMINATE_CONNECTION_BEFORE_REQUEST)) {
            terminate_connection_before_request = true;
        }
        if (fault_flag.equals(RPCTestErrorHandling.TERMINATE_CONNECTION_AFTER_REQUEST)) {
            terminate_connection_after_request = true;
        }
    }

    @Override
    public AbstractStreamConnection startCall(final String method_name) throws IOException, JSONException {

        this.method_name = method_name;

        final AbstractStreamConnection connection = setupConnection();
        final JSONWriter jw = connection.getJSONwriter();

        if (method_name.equals("check")) {

            if (send_invalid_json) {
                jw.value("not a valid JSON string");
                return connection;
            }

            if (send_blank_line) {
                jw.value("\n");
                return connection;
            }
        }

        if (method_name.equals("equalToThree")) {

            if (send_request_of_wrong_type) {

                jw.array();
                super.startCall(method_name);
                return connection;
            }
        }

        return super.startCall(method_name); // normal case
    }

    @Override
    public JSONReader makeCall(final AbstractStreamConnection connection) throws Exception {

        Diagnostic.trace("Make faulty call " + method_name);

        if (method_name.equals("check")) {
            if (terminate_connection_before_request) {
                closeConnection(connection);
                throw new RPCException("test has made this gone wrong on purpose");
            }
            if (terminate_connection_after_request) {

                completeRequest(connection);
                closeConnection(connection);
                return getReplyReader(connection);
            }

            final JSONReader result = super.makeCall(connection); // normal case
            finishCall(connection);
            return result;
        }
        return super.makeCall(connection); // normal case
    }
}
