/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.mail.MessagingException;
import javax.mail.internet.MimeUtility;

/**
 * Provides utility methods for encoding and decoding byte arrays.
 *
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public final class Encoding {

    private static final int HEX_A_VALUE = 10;
    private static final int HIGHEST_DIGIT = 9;
    private static final int MASK = 0xf;
    private static final int ROTATE = 4;

    /**
     * Prevent instantiation of utility class.
     */
    private Encoding() {

    }

    /**
     * Translates the given byte array to a string using the specified encoding.
     *
     * @param bytes a byte array
     * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
     * @return a string representing the encoding of the given bytes
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     */
    public static String encodeByteArray(final byte[] bytes, final String encoding) throws MessagingException, IOException {

        final ByteArrayOutputStream encoded_bytes_stream = new ByteArrayOutputStream();
        final OutputStream encoding_stream = MimeUtility.encode(encoded_bytes_stream, encoding);

        encoding_stream.write(bytes);
        encoding_stream.close();

        return new String(encoded_bytes_stream.toByteArray());
    }

    /**
     * Translates the given byte array to a string using the specified encoding, with the option of suppressing certain characters.
     * Note that there is no facility to reverse an encoding if character suppression is used.
     *
     * @param bytes a byte array
     * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
     * @param suppressed_characters a string containing characters that should be suppressed from the output
     * @return a string representing the encoding of the given bytes
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     */
    public static String encodeByteArray(final byte[] bytes, final String encoding, final String suppressed_characters) throws MessagingException, IOException {

        final String encoded = encodeByteArray(bytes, encoding);
        final StringBuilder buffer = new StringBuilder();

        final int len = encoded.length();
        for (int i = 0; i < len; i++) {
            final char c = encoded.charAt(i);
            if (suppressed_characters.indexOf(c) == -1) {
                buffer.append(c);
            }
        }

        return buffer.toString();
    }

    /**
     * Translates the given encoded string to a decoded byte array using the specified encoding.
     *
     * @param text an encoded string
     * @param result_size the size of the decoded array to be returned
     * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
     * @return an array representing the decoding of the given string
     * @throws MessagingException if the encoding is not recognised
     * @throws IOException if an error occurs during translation
     */
    public static byte[] decodeByteArray(final String text, final int result_size, final String encoding) throws MessagingException, IOException {

        final byte[] encoded_bytes = text.getBytes();
        final InputStream decoding_stream = MimeUtility.decode(new ByteArrayInputStream(encoded_bytes), encoding);

        final byte[] decoded_bytes = new byte[result_size];
        if (decoding_stream.read(decoded_bytes, 0, result_size) < result_size) { throw new IOException("couldn't read enough bytes"); }

        return decoded_bytes;
    }

    /**
     * Converts a byte array to a hex representation.
     *
     * This method derived from code in:
     *
     * SHA1.java - An implementation of the SHA-1 Algorithm obtained from
     * http://jabberapplet.jabberstudio.org/doxygen/SHA1_8java-source.html
     *
     * This version by Chuck McManis (cmcmanis@netcom.com) and
     * still public domain.
     *
     * Adapted for JabberApplet by M.Kiesel
     *
     * Based on the C code that Steve Reid wrote his header
     * was :
     *      SHA-1 in C
     *      By Steve Reid <steve@edmweb.com>
     *      100% Public Domain
     *
     * @param bytes a byte array
     * @return a string representation of the bytes in hex
     */
    public static String byteArrayToHex(final byte[] bytes) {

        final StringBuilder sb = new StringBuilder();

        for (final byte digestBit : bytes) {

            char c1 = (char) (digestBit >>> ROTATE & MASK);
            char c2 = (char) (digestBit & MASK);

            c1 = (char) (c1 > HIGHEST_DIGIT ? 'a' + (c1 - HEX_A_VALUE) : '0' + c1);
            c2 = (char) (c2 > HIGHEST_DIGIT ? 'a' + (c2 - HEX_A_VALUE) : '0' + c2);

            sb.append(c1);
            sb.append(c2);
        }

        return sb.toString();
    }
}
