package uk.ac.standrews.cs.nds.util;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

public class TimeoutExecutor {

    private final ExecutorService first_stage_executor, second_stage_executor;
    private final Duration timeout;
    private final boolean include_queued_time_in_timeout;
    private final String name;
    private static final AtomicInteger NEXT_ID = new AtomicInteger(1);

    public TimeoutExecutor(final int thread_pool_size, final Duration timeout, final boolean include_queued_time_in_timeout, final String name) {

        first_stage_executor = Executors.newFixedThreadPool(thread_pool_size);
        second_stage_executor = include_queued_time_in_timeout ? null : Executors.newFixedThreadPool(thread_pool_size);

        this.timeout = timeout;
        this.include_queued_time_in_timeout = include_queued_time_in_timeout;
        this.name = name + " " + NEXT_ID.getAndIncrement();
    }

    public void executeWithTimeout(final Runnable action) throws Exception {

        executeWithTimeout(Executors.callable(action));
    }

    public <T> T executeWithTimeout(final Callable<T> action) throws Exception {

        return executeWithTimeout(false, action);
    }

    public <T> T executeWithTimeout(final boolean debug, final Callable<T> action) throws Exception {

        if (include_queued_time_in_timeout) { return executeWithTimeout(action, first_stage_executor, true); }

        final Callable<T> action_with_timeout = new Callable<T>() {

            @Override
            public T call() throws Exception {

                return executeWithTimeout(action, second_stage_executor, true);
            }
        };

        return executeWithTimeout(action_with_timeout, first_stage_executor, false);
    }

    public void shutdown() {

        first_stage_executor.shutdownNow();
        if (second_stage_executor != null) {
            second_stage_executor.shutdownNow();
        }
    }

    private <T> T executeWithTimeout(final Callable<T> action, final ExecutorService executor, final boolean do_timeout) throws Exception {

        final Future<T> task = executor.submit(action);
        try {
            return do_timeout ? task.get(timeout.getLength(), timeout.getTimeUnit()) : task.get();
        }
        catch (final ExecutionException e) {
            throw launderThrowable(e.getCause());
        }
        finally {
            task.cancel(true);
        }
    }

    private static Exception launderThrowable(final Throwable t) {

        if (t instanceof Exception) {
            return (Exception) t;
        }
        else if (t instanceof Error) {
            throw (Error) t;
        }
        else {
            throw new IllegalStateException("Throwable not exception or error", t);
        }
    }
}
