/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.data;

import java.lang.reflect.Field;
import java.util.Date;

import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Abstract data class.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public abstract class Data {

    private final Date date;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a data item with the current time.
     */
    public Data() {

        date = new Date();
    }

    /**
     * Initializes a data item with the given time.
     * @param date the time
     */
    public Data(final java.sql.Date date) {

        // DODO check - is this right? Previously, date parameter was ignored.
        this.date = (Date) date.clone();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the time when this data item was created.
     * @return the time when this data item was created
     */
    public Date getDate() {

        return (Date) date.clone();
    }

    // -------------------------------------------------------------------------------------------------------

    protected long min(final long a, final long b) {

        return a < b ? a : b;
    }

    protected double min(final double a, final double b) {

        return a < b ? a : b;
    }

    protected long max(final long a, final long b) {

        return a > b ? a : b;
    }

    protected double max(final double a, final double b) {

        return a > b ? a : b;
    }

    /**
     * Get the long value assigned to a field of the given name.
     * @param resourceName Name of a field in this instance.
     * @return The long value of the requested field.
     * @throws NoSuchFieldException This uses reflection and throws exceptions if the field does not exist.
     */
    public long getLong(final String resourceName) throws NoSuchFieldException {

        final Field field = this.getClass().getField(resourceName);

        try {
            return field.getLong(this);
        }
        catch (final IllegalArgumentException e) {
            ErrorHandling.exceptionError(e, "Should not happen because this field (" + resourceName + ") is internal to numonic and no arguments are being passed.");
        }
        catch (final IllegalAccessException e) {
            ErrorHandling.exceptionError(e, "Should not happen because this field (" + resourceName + ") is internal to numonic and the fields are accessible from this method call.");
        }

        return -1;

    }

    /**
     * Get the double value assigned to a field of the given name.
     * @param resourceName Name of a field in this instance.
     * @return The double value of the requested field.
     * @throws NoSuchFieldException This uses reflection and throws exceptions if the field does not exist.
     */
    public double getDouble(final String resourceName) throws NoSuchFieldException {

        final Field field = this.getClass().getField(resourceName);
        try {
            return field.getDouble(this);
        }
        catch (final IllegalArgumentException e) {
            ErrorHandling.exceptionError(e, "Should not happen because this field (" + resourceName + ") is internal to numonic and no arguments are being passed.");
        }
        catch (final IllegalAccessException e) {
            ErrorHandling.exceptionError(e, "Should not happen because this field (" + resourceName + ") is internal to numonic and the fields are accessible from this method call.");
        }

        return -1;
    }

    public long getLong(final ResourceType resource) throws NoSuchFieldException {

        return getLong(resource.toString());

    }

    public double getDouble(final ResourceType resource) throws NoSuchFieldException {

        return getDouble(resource.toString());

    }
}
