/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.data;

/**
 * Used to store machine utilization data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class MachineUtilisationData extends Data implements IData<MachineUtilisationData> {

    // CPU
    public double cpu_user_total; // Percentage total CPU utilisation.
    public double cpu_sys_total; // Percentage system CPU utilisation.
    public double cpu_idle_total; // Percentage CPU idle.
    public double cpu_wait_total; // Percentage CPU in I/O wait state.
    public double cpu_nice_total; // Percentage CPU running nice commands http://en.wikipedia.org/wiki/Nice_%28Unix%29.

    // Memory
    public long memory_used; // Memory used in system (in Kilobytes).
    public long memory_free; // Memory free in system (in Kilobytes).
    public long swap_used; // Swap space used in system (in Kilobytes).
    public long swap_free; // Swap space free in system (in Kilobytes).

    // TODO - How do we find out the amount of swap activity??????

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes a latency/bandwidth record.
     */
    public MachineUtilisationData() {

    }

    /**
     * Initializes a latency/bandwidth record.
     *
     * @param cpu_user_total the percentage total CPU utilization
     * @param cpu_sys_total the percentage system CPU utilization
     * @param cpu_idle_total the percentage CPU idle
     * @param cpu_wait_total the percentage CPU in I/O wait state
     * @param cpu_nice_total the percentage CPU running nice commands
     * @param memory_used the memory used in KB
     * @param memory_free the memory free in KB
     * @param swap_used the swap space used in KB
     * @param swap_free the swap space free in KB
     */
    public MachineUtilisationData(final double cpu_user_total, final double cpu_sys_total, final double cpu_idle_total, final double cpu_wait_total, final double cpu_nice_total, final long memory_used, final long memory_free, final long swap_used, final long swap_free) {

        this.cpu_user_total = cpu_user_total;
        this.cpu_sys_total = cpu_sys_total;
        this.cpu_idle_total = cpu_idle_total;
        this.cpu_wait_total = cpu_wait_total;
        this.cpu_nice_total = cpu_nice_total;
        this.memory_used = memory_used;
        this.memory_free = memory_free;
        this.swap_used = swap_used;
        this.swap_free = swap_free;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();
        builder.append("MachineUtilisationData [cpu_user_total=");
        builder.append(cpu_user_total);
        builder.append(", cpu_sys_total=");
        builder.append(cpu_sys_total);
        builder.append(", cpu_idle_total=");
        builder.append(cpu_idle_total);
        builder.append(", cpu_wait_total=");
        builder.append(cpu_wait_total);
        builder.append(", cpu_nice_total=");
        builder.append(cpu_nice_total);
        builder.append(", memory_used=");
        builder.append(memory_used);
        builder.append(", memory_free=");
        builder.append(memory_free);
        builder.append(", swap_used=");
        builder.append(swap_used);
        builder.append(", swap_free=");
        builder.append(swap_free);
        builder.append("]");
        return builder.toString();
    }

    @Override
    public MachineUtilisationData min(final MachineUtilisationData other) {

        final MachineUtilisationData min = new MachineUtilisationData(

        min(cpu_user_total, other.cpu_user_total), min(cpu_sys_total, other.cpu_sys_total), min(cpu_idle_total, other.cpu_idle_total), min(cpu_wait_total, other.cpu_wait_total), min(cpu_nice_total, other.cpu_nice_total), min(memory_used, other.memory_used), min(memory_free, other.memory_free), min(
                        swap_used, other.swap_used), min(swap_free, other.swap_free));

        return min;
    }

    @Override
    public MachineUtilisationData max(final MachineUtilisationData other) {

        final MachineUtilisationData max = new MachineUtilisationData(

        max(cpu_user_total, other.cpu_user_total), max(cpu_sys_total, other.cpu_sys_total), max(cpu_idle_total, other.cpu_idle_total), max(cpu_wait_total, other.cpu_wait_total), max(cpu_nice_total, other.cpu_nice_total), max(memory_used, other.memory_used), max(memory_free, other.memory_free), max(
                        swap_used, other.swap_used), max(swap_free, other.swap_free));

        return max;
    }

    @Override
    public MachineUtilisationData sum(final MachineUtilisationData other) {

        final MachineUtilisationData total = new MachineUtilisationData(

        cpu_user_total + other.cpu_user_total, cpu_sys_total + other.cpu_sys_total, cpu_idle_total + other.cpu_idle_total, cpu_wait_total + other.cpu_wait_total, cpu_nice_total + other.cpu_nice_total, memory_used + other.memory_used, memory_free + other.memory_free, swap_used + other.swap_used,
                        swap_free + other.swap_free);

        return total;
    }

    @Override
    public MachineUtilisationData convertToAverage(final int number_of_measurements) {

        final MachineUtilisationData average = new MachineUtilisationData(

        cpu_user_total / number_of_measurements, cpu_sys_total / number_of_measurements, cpu_idle_total / number_of_measurements, cpu_wait_total / number_of_measurements, cpu_nice_total / number_of_measurements, memory_used / number_of_measurements, memory_free / number_of_measurements, swap_used
                        / number_of_measurements, swap_free / number_of_measurements);

        return average;
    }

    @Override
    public MachineUtilisationData adjustForBase(final MachineUtilisationData other) {

        // Nothing needs to be adjusted in this class.
        return this;
    }

    @Override
    public boolean needsAdjustedForBase() {

        return false;
    }

    @Override
    public MachineUtilisationData getNewInstance() {

        return new MachineUtilisationData();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the percentage total CPU utilization.
     * @return the percentage total CPU utilization
     */
    public double getCpuUserTotal() {

        return cpu_user_total;
    }

    /**
     * Returns the percentage system CPU utilization.
     * @return the percentage system CPU utilization
     */
    public double getCpuSysTotal() {

        return cpu_sys_total;
    }

    /**
     * Returns the percentage CPU idle.
     * @return the percentage CPU idle
     */
    public double getCpuIdleTotal() {

        return cpu_idle_total;
    }

    /**
     * Returns the percentage CPU in I/O wait state.
     * @return the percentage CPU in I/O wait state
     */
    public double getCpuWaitTotal() {

        return cpu_wait_total;
    }

    /**
     * Returns the percentage CPU running nice commands.
     * @return the percentage CPU running nice commands
     */
    public double getCpuNiceTotal() {

        return cpu_nice_total;
    }

    /**
     * Returns the memory used.
     * @return the memory used in KB
     */
    public long getMemoryUsed() {

        return memory_used;
    }

    /**
     * Returns the memory free.
     * @return the memory free in KB
     */
    public long getMemoryFree() {

        return memory_free;
    }

    /**
     * Returns the swap space used.
     * @return the swap space used in KB
     */
    public long getSwapUsed() {

        return swap_used;
    }

    /**
     * Returns the swap space free.
     * @return the swap space free in KB
     */
    public long getSwapFree() {

        return swap_free;
    }

    public double getMemoryUtilization() {

        return memory_used / (memory_free + memory_used);
    }

}
