/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.data;

/**
 * Used to store system data.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class SystemInfoData {

    public final String machine_id;

    // OS Info.
    public final String os_name;
    public final String os_version;

    // Network Info.
    public final String hostname;
    public final String primary_ip;
    public final String default_gateway;

    // CPU Info.
    public final String cpu_vendor;
    public final String cpu_model;
    public final int number_of_cores;
    public final int number_of_cpus;
    public final int cpu_clock_speed;
    public final long cpu_cache_size;

    // Memory Info.
    public final long memory_total;
    public final long swap_total;

    /**
     * Initializes a system data record.
     *
     * @param machine_id the MAC address of the primary interface
     * @param os_name the name of the OS e.g. MacOSX
     * @param os_version the version of the OS e.g 10.6.4
     * @param hostname the fully qualified hostname e.g. slash.cs.st-andrews.ac.uk
     * @param primary_ip the primary IP address of the machine
     * @param default_gateway the default IP gateway used by the machine
     * @param cpu_vendor the cpu vendor e.g. Intel
     * @param cpu_model the model of the machine e.g. MacBookPro5,5
     * @param number_of_cores the number of processor cores available
     * @param number_of_cpus the number of CPUs available
     * @param cpu_clock_speed the CPU clock rate in MHz
     * @param cpu_cache_size the CPU cache size in KB
     * @param memory_total the total available memory in KB
     * @param swap_total the total available swap in KB
     */
    public SystemInfoData(final String machine_id, final String os_name, final String os_version, final String hostname, final String primary_ip, final String default_gateway, final String cpu_vendor, final String cpu_model, final int number_of_cores, final int number_of_cpus,
                    final int cpu_clock_speed, final long cpu_cache_size, final long memory_total, final long swap_total) {

        this.machine_id = machine_id;
        this.os_name = os_name;
        this.os_version = os_version;
        this.hostname = hostname;
        this.primary_ip = primary_ip;
        this.default_gateway = default_gateway;
        this.cpu_vendor = cpu_vendor;
        this.cpu_model = cpu_model;
        this.number_of_cores = number_of_cores;
        this.number_of_cpus = number_of_cpus;
        this.cpu_clock_speed = cpu_clock_speed;
        this.cpu_cache_size = cpu_cache_size;
        this.memory_total = memory_total;
        this.swap_total = swap_total;
    }

    @Override
    public String toString() {

        final StringBuilder builder = new StringBuilder();
        builder.append("SysInfoResults [os_name=");
        builder.append(os_name);
        builder.append(", os_version=");
        builder.append(os_version);
        builder.append(", hostname=");
        builder.append(hostname);
        builder.append(", primaryIP=");
        builder.append(primary_ip);
        builder.append(", default_gateway=");
        builder.append(default_gateway);
        builder.append(", cpuVendor=");
        builder.append(cpu_vendor);
        builder.append(", cpu_model=");
        builder.append(cpu_model);
        builder.append(", number_of_cores=");
        builder.append(number_of_cores);
        builder.append(", number_of_cpus=");
        builder.append(number_of_cpus);
        builder.append(", cpu_clock_speed=");
        builder.append(cpu_clock_speed);
        builder.append(", cpu_cache_size=");
        builder.append(cpu_cache_size);
        builder.append("]");
        return builder.toString();
    }

    /**
     * Returns the MAC address of the primary interface.
     * @return the MAC address of the primary interface
     */
    public String getMachineId() {

        return machine_id;
    }

    /**
     * Returns the name of the OS e.g. MacOSX.
     * @return the name of the OS e.g. MacOSX
     */
    public String getOsName() {

        return os_name;
    }

    /**
     * Returns the version of the OS e.g 10.6.4.
     * @return the version of the OS e.g 10.6.4
     */
    public String getOsVersion() {

        return os_version;
    }

    /**
     * Returns the fully qualified hostname e.g. slash.cs.st-andrews.ac.uk.
     * @return the fully qualified hostname e.g. slash.cs.st-andrews.ac.uk
     */
    public String getHostname() {

        return hostname;
    }

    /**
     * Returns the primary IP address of the machine.
     * @return the primary IP address of the machine
     */
    public String getPrimaryIp() {

        return primary_ip;
    }

    /**
     * Returns the default IP gateway used by the machine.
     * @return the default IP gateway used by the machine
     */
    public String getDefaultGateway() {

        return default_gateway;
    }

    /**
     * Returns the CPU vendor e.g. Intel.
     * @return the CPU vendor e.g. Intel
     */
    public String getCpuVendor() {

        return cpu_vendor;
    }

    /**
     * Returns the model of the machine e.g. MacBookPro5,5.
     * @return the model of the machine e.g. MacBookPro5,5
     */
    public String getCpuModel() {

        return cpu_model;
    }

    /**
     * Returns the number of processor cores available.
     * @return the number of processor cores available
     */
    public int getNumberOfCores() {

        return number_of_cores;
    }

    /**
     * Returns the number of CPUs available.
     * @return the number of CPUs available
     */
    public int getNumberOfCpus() {

        return number_of_cpus;
    }

    /**
     * Returns the CPU clock rate in MHz.
     * @return the CPU clock rate in MHz
     */
    public int getCpuClockSpeed() {

        return cpu_clock_speed;
    }

    /**
     * Returns the CPU cache size in KB.
     * @return the CPU cache size in KB
     */
    public long getCpuCacheSize() {

        return cpu_cache_size;
    }

    /**
     * Returns the total available memory in KB.
     * @return the total available memory in KB
     */
    public long getMemoryTotal() {

        return memory_total;
    }

    /**
     * Returns the total available swap in KB.
     * @return the total available swap in KB
     */
    public long getSwapTotal() {

        return swap_total;
    }
}
