/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.db;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.numonic.data.FileSystemData;
import uk.ac.standrews.cs.numonic.data.LatencyAndBandwidthData;
import uk.ac.standrews.cs.numonic.data.MachineUtilisationData;
import uk.ac.standrews.cs.numonic.data.NetworkData;
import uk.ac.standrews.cs.numonic.data.ProcessData;
import uk.ac.standrews.cs.numonic.data.SystemInfoData;
import uk.ac.standrews.cs.numonic.distribution.Distribution;
import uk.ac.standrews.cs.numonic.distribution.DistributionCollector;
import uk.ac.standrews.cs.numonic.event.Event;
import uk.ac.standrews.cs.numonic.reporting.IReporting;
import uk.ac.standrews.cs.numonic.sort.data.DistributionData;
import uk.ac.standrews.cs.numonic.summary.MultipleSummary;
import uk.ac.standrews.cs.numonic.summary.SingleSummary;

/**
 * Database-specific reporting interface.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public class DatabaseConnection implements IReporting {

    private Connection connection;

    private final QueryBuilder distribution_data_prepared_statement = new QueryBuilder();
    private final QueryBuilder event_prepared_statement = new QueryBuilder();
    private final QueryBuilder fs_prepared_statement = new QueryBuilder();
    private final QueryBuilder util_prepared_statement = new QueryBuilder();
    private final QueryBuilder net_prepared_statement = new QueryBuilder();
    private final QueryBuilder process_prepared_statement = new QueryBuilder();
    private final QueryBuilder sys_info_prepared_statement = new QueryBuilder();
    private final QueryBuilder latency_bandwith_prepared_statement = new QueryBuilder();

    /**
     * The date format used to add dates to the database.
     */
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss"); // 2010-08-04 13:51:14.539

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the connection.
     *
     * @param database_url the database URL
     * @param database_user_name the user name
     * @param database_password the password
     * @param drop_database_tables true if all tables should be dropped
     */
    public DatabaseConnection(final String database_url, final String database_user_name, final String database_password, final boolean drop_database_tables) {

        try {
            connection = DriverManager.getConnection(database_url, database_user_name, database_password);
        }
        catch (final SQLException e) {
            ErrorHandling.hardError(e, "Quitting: couldn't initialise database connection to: " + database_url);
        }

        try {
            if (drop_database_tables) {
                dropAllTables();
            }
        }
        catch (final SQLException e) {
            ErrorHandling.hardError(e, "Quitting: couldn't drop all tables from database running at: " + database_url);
        }

        // Create tables.
        try {
            executeSQLScript("sql-schema" + File.separator + "monitorSchema.sql");
            executeSQLScript("sql-schema" + File.separator + "views.sql");
        }
        catch (final Exception e) {
            ErrorHandling.hardError(e, "Quitting: couldn't configure tables in the database running at: " + database_url);
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Drops all existing tables.
     * @throws SQLException if an error occurs
     */
    public void dropAllTables() throws SQLException {

        Statement statement = null;

        try {
            statement = connection.createStatement();
            statement.executeUpdate("DROP ALL OBJECTS;");
        }
        finally {
            if (statement != null) {
                statement.close();
            }
        }
    }

    private void executeSQLScript(final String filename) throws Exception {

        InputStream resource_stream = null;
        Statement statement = null;

        try {
            // First try to read query from the jar file, in which case need absolute path.
            resource_stream = getClass().getResourceAsStream(File.separator + filename);

            // If that doesn't work, try from local file system with relative path.
            if (resource_stream == null) {
                resource_stream = new FileInputStream(new File(filename));
            }

            final StringBuilder query = new StringBuilder();

            // Read query.
            int ch;
            while ((ch = resource_stream.read()) != -1) {
                query.append((char) ch);
            }

            // Execute query.
            statement = connection.createStatement();
            statement.executeUpdate(query.toString());

        }
        finally {
            try {
                if (resource_stream != null) {
                    resource_stream.close();
                }
            }
            finally {
                if (statement != null) {
                    statement.close();
                }
            }
        }
    }

    @Override
    public void reportDistributionData(final DistributionCollector<?> machineProbability) throws SQLException {

        final String machine_id = machineProbability.getMachineId();
        final Distribution[] data = machineProbability.getProbabilities();
        final int num_measurements = machineProbability.getNumberOfMeasurements();
        final Date start_time = machineProbability.getStartTime();
        final Date end_time = machineProbability.getEndTime();

        for (final Distribution distribution : data) {
            try {
                distribution_data_prepared_statement.doInsertQuery("RESOURCE_MONITORING.DISTRIBUTION", connection, machine_id, DATE_FORMAT.format(start_time), DATE_FORMAT.format(end_time), num_measurements, distribution.getInstrumentName(), distribution.getAttributes(),
                                distribution.getProbabilityHigh(), distribution.getProbabilityMediumHigh(), distribution.getProbabilityMedium(), distribution.getProbabilityMediumLow(), distribution.getProbabilityLow());
            }
            catch (final SQLException e) {
                if (isLockException(e)) {
                    Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
                }
                else {
                    throw e;
                }
            }
        }
    }

    private boolean isLockException(final SQLException e) {

        return e.getMessage().contains("Couldn't obtain locks");
    }

    @Override
    public void reportFileSystemData(final DistributionCollector<FileSystemData> fileSystemSummary) throws SQLException {

        // TODO to implement
        throw new SQLException("method not implemented");
    }

    @Override
    public void reportFileSystemData(final MultipleSummary<FileSystemData> fileSystemSummary) throws SQLException {

        for (final SingleSummary<FileSystemData> summary : fileSystemSummary.getSummaries()) {
            final FileSystemData avg = summary.getAverage();
            final FileSystemData min = summary.getMin();
            final FileSystemData max = summary.getMax();

            try {
                fs_prepared_statement.doInsertQuery("RESOURCE_MONITORING.FILE_SYSTEM", connection, fileSystemSummary.getMachineID(), DATE_FORMAT.format(summary.getStartTime()), DATE_FORMAT.format(summary.getEndTime()), summary.getNumberOfMeasurements(), avg.getFileSystemName(),
                                avg.getFileSystemLocation(), avg.getFileSystemType(), avg.getSize(), avg.getSpaceFree(), avg.getSpaceUsed(), avg.getNumberOfFiles(), min.getNumberOfFiles(), max.getNumberOfFiles(), avg.getNumberOfDiskReads(), min.getNumberOfDiskReads(), max.getNumberOfDiskReads(),
                                avg.getBytesRead(), min.getBytesRead(), max.getBytesRead(), avg.getNumberOfDiskWrites(), min.getNumberOfDiskWrites(), max.getNumberOfDiskWrites(), avg.getBytesWritten(), min.getBytesWritten(), max.getBytesWritten());
            }
            catch (final SQLException e) {
                if (isLockException(e)) {
                    Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
                }
                else {
                    throw e;
                }
            }
        }
    }

    @Override
    public void reportMachineUtilData(final SingleSummary<MachineUtilisationData> summary) throws SQLException {

        final MachineUtilisationData avg = summary.getAverage();
        final MachineUtilisationData min = summary.getMin();
        final MachineUtilisationData max = summary.getMax();

        try {
            util_prepared_statement.doInsertQuery("RESOURCE_MONITORING.MACHINE_UTIL", connection, summary.getMachineID(), DATE_FORMAT.format(summary.getStartTime()), DATE_FORMAT.format(summary.getEndTime()), summary.getNumberOfMeasurements(), avg.getCpuUserTotal(), min.getCpuUserTotal(),
                            max.getCpuUserTotal(), avg.getCpuSysTotal(), min.getCpuSysTotal(), max.getCpuSysTotal(), avg.getCpuNiceTotal(), min.getCpuNiceTotal(), max.getCpuNiceTotal(), avg.getCpuIdleTotal(), min.getCpuIdleTotal(), max.getCpuIdleTotal(), avg.getCpuWaitTotal(),
                            min.getCpuWaitTotal(), max.getCpuWaitTotal(), avg.getMemoryUsed(), min.getMemoryUsed(), max.getMemoryUsed(), avg.getMemoryFree(), min.getMemoryFree(), max.getMemoryFree(), avg.getSwapUsed(), min.getSwapUsed(), max.getSwapUsed(), avg.getSwapFree(), min.getSwapFree(),
                            max.getSwapFree());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    @Override
    public void reportNetworkData(final SingleSummary<NetworkData> summary) throws SQLException {

        final NetworkData avg = summary.getAverage();
        final NetworkData min = summary.getMin();
        final NetworkData max = summary.getMax();

        try {
            /*
             *  machine_id  VARCHAR(40), start_ts TIMESTAMP, end_ts TIMESTAMP, measurements INT,
                device_name VARCHAR(100),
                ip_address VARCHAR(15),
                rx_bytes_avg BIGINT, rx_bytes_min BIGINT, rx_bytes_max BIGINT,
                rx_packets_avg BIGINT, rx_packets_min BIGINT, rx_packets_max BIGINT,
                rx_errors_avg BIGINT, rx_errors_min BIGINT, rx_errors_max BIGINT,
                rx_dropped_avg BIGINT, rx_dropped_min BIGINT, rx_dropped_max BIGINT,
                rx_overruns_avg BIGINT, rx_overruns_min BIGINT, rx_overruns_max BIGINT,
                tx_bytes_avg BIGINT, tx_bytes_min BIGINT, tx_bytes_max BIGINT,
                tx_packets_avg BIGINT, tx_packets_min BIGINT,tx_packets_max BIGINT,
                tx_errors_avg BIGINT, tx_errors_min BIGINT, tx_errors_max BIGINT,
                tx_dropped_avg BIGINT, tx_dropped_min BIGINT, tx_dropped_max BIGINT,
                tx_overruns_avg BIGINT, tx_overruns_min BIGINT, tx_overruns_max BIGINT,
                FOREIGN KEY (machine_id) REFERENCES SYS_INFO(machine_id));
             * */

            net_prepared_statement.doInsertQuery("RESOURCE_MONITORING.NETWORK", connection, summary.getMachineID(), DATE_FORMAT.format(summary.getStartTime()), DATE_FORMAT.format(summary.getEndTime()), summary.getNumberOfMeasurements(), avg.getDeviceName(), avg.getIPAddress(), avg.getRXBytes(),
                            min.getRXBytes(), max.getRXBytes(), avg.getRXPackets(), min.getRXPackets(), max.getRXPackets(), avg.getRXErrors(), min.getRXErrors(), max.getRXErrors(), avg.getRXDropped(), min.getRXDropped(), max.getRXDropped(), avg.getRXOverruns(), min.getRXOverruns(),
                            max.getRXOverruns(), avg.getTXBytes(), min.getTXBytes(), max.getTXBytes(), avg.getTXPackets(), min.getTXPackets(), max.getTXPackets(), avg.getTXErrors(), min.getTXErrors(), max.getTXErrors(), avg.getTXDropped(), min.getTXDropped(), max.getTXDropped(),
                            avg.getTXOverruns(), min.getTXOverruns(), max.getTXOverruns());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    @Override
    public void reportProcessData(final SingleSummary<ProcessData> summary) throws SQLException {

        final ProcessData avg = summary.getAverage();
        final ProcessData min = summary.getMin();
        final ProcessData max = summary.getMax();

        try {
            process_prepared_statement.doInsertQuery("RESOURCE_MONITORING.PROCESS", connection, summary.getMachineID(), DATE_FORMAT.format(summary.getStartTime()), DATE_FORMAT.format(summary.getEndTime()), summary.getNumberOfMeasurements(), avg.getProcessName(), avg.getProcessStartTime(),
                            avg.getProcessCpuPercent(), min.getProcessCpuPercent(), max.getProcessCpuPercent(), avg.getProcessMemSize(), min.getProcessMemSize(), max.getProcessMemSize(), avg.getProcessResident(), min.getProcessResident(), max.getProcessResident());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    @Override
    public void reportSystemInfo(final SystemInfoData data) throws SQLException {

        try {
            sys_info_prepared_statement.doInsertQuery("RESOURCE_MONITORING.SYS_INFO", connection, data.getMachineId(), DATE_FORMAT.format(new Date()), data.getHostname(), data.getPrimaryIp(), data.getCpuVendor(), data.getCpuModel(), data.getNumberOfCores(), data.getNumberOfCpus(),
                            data.getCpuClockSpeed(), data.getCpuCacheSize(), data.getOsName(), data.getOsVersion(), data.getDefaultGateway(), data.getMemoryTotal(), data.getSwapTotal());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    @Override
    public void reportLatencyAndBandwidthData(final LatencyAndBandwidthData data) throws SQLException {

        try {
            latency_bandwith_prepared_statement.doInsertQuery("RESOURCE_MONITORING.LATENCY_AND_BANDWIDTH", connection, data.getLocalMachineID(), DATE_FORMAT.format(data.getDate()), data.getLocalNetworkAdapterName(), data.getLocalGateway(), data.getRemoteMachineID(),
                            data.getRemoteNetworkAdapterName(), data.getRemoteGateway(), data.getLatency(), data.getBandwidth());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    private void insertEvent(final Event event) throws SQLException {

        // create a comma separated list of attributes
        final StringBuffer attributes = new StringBuffer();
        boolean more = false;
        if (event.getAttributes() != null) {

            for (final String attribute : event.getAttributes()) {
                if (more) {
                    attributes.append(",");
                }
                attributes.append(attribute);
                more = true;
            }
        }

        try {
            event_prepared_statement.doInsertQuery("RESOURCE_MONITORING.EVENTS", connection, event.getMachineId(), event.getLabel(), DATE_FORMAT.format(event.getTime()), attributes.toString());
        }
        catch (final SQLException e) {
            if (isLockException(e)) {
                Diagnostic.trace(DiagnosticLevel.FULL, "dropped report due to database contention");
            }
            else {
                throw e;
            }
        }
    }

    @Override
    public void reportEventData(final List<Event> events) throws SQLException {

        for (final Event event : events) {
            insertEvent(event);
        }
    }

    @Override
    public void reportAllNetworkData(final MultipleSummary<NetworkData> summaries) throws SQLException {

        for (final SingleSummary<NetworkData> summary : summaries.getSummaries()) {
            reportNetworkData(summary);
        }
    }

    @Override
    public void reportAllProcessData(final MultipleSummary<ProcessData> summaries) throws SQLException {

        for (final SingleSummary<ProcessData> summary : summaries.getSummaries()) {
            reportProcessData(summary);
        }
    }

    /**
     * Returns system information.
     * @return system information
     * @throws SQLException if an error occurs
     */
    public List<SystemInfoData> getSystemInfo() throws SQLException {

        final String query = "SELECT * FROM RESOURCE_MONITORING.SYS_INFO;";

        final List<SystemInfoData> machinesData = new LinkedList<SystemInfoData>();

        Statement stat = null;

        try {
            stat = connection.createStatement();
            final ResultSet rs = stat.executeQuery(query);

            while (rs.next()) {
                final SystemInfoData data = new SystemInfoData(rs.getString("machine_id"), rs.getString("os_name"), rs.getString("os_version"), rs.getString("hostname"), rs.getString("primary_ip"), rs.getString("default_gateway"), rs.getString("cpu_vendor"), rs.getString("cpu_model"),
                                rs.getInt("num_cores"), rs.getInt("num_cpus"), rs.getInt("cpu_mhz"), rs.getLong("cpu_cache_size"), rs.getLong("memory_total"), rs.getLong("swap_total"));

                machinesData.add(data);
            }
        }
        finally {
            stat.close();
        }

        return machinesData;
    }

    /**
     * Returns the known hosts.
     * @return the known hosts
     * @throws SQLException if an error occurs
     */
    public List<String> getKnownHosts() throws SQLException {

        final String query = "SELECT primary_ip FROM RESOURCE_MONITORING.SYS_INFO;";
        final List<String> ipNames = new LinkedList<String>();
        Statement stat = null;

        try {
            stat = connection.createStatement();
            final ResultSet rs = stat.executeQuery(query);

            while (rs.next()) {

                ipNames.add(rs.getString("primary_ip"));
            }
        }
        finally {
            stat.close();
        }

        return ipNames;
    }

    @Override
    public List<FileSystemData> getFileSystemData(final int last_n_minutes) throws SQLException {

        final String query = "SELECT " + getBasicMonitoringInfo() + getFSInfo() + ofLastNTimePeriod(last_n_minutes, "MINUTE") + groupBy("machine_id", "fs_name");
        final List<FileSystemData> fsData = new LinkedList<FileSystemData>();
        Statement stat = null;

        try {
            stat = connection.createStatement();
            final ResultSet rs = stat.executeQuery(query);

            while (rs.next()) {
                final FileSystemData data = new FileSystemData(rs.getDate("start_ts"));
                data.setMachineId(rs.getString("machine_id"));
                data.setFileSystemLocation(rs.getString("fs_location"));
                data.setFileSystemName(rs.getString("fs_name"));
                data.setFileSystemType(rs.getString("fs_type"));

                data.setSpaceFree(rs.getLong("fs_free"));
                data.setSpaceUsed(rs.getLong("fs_used"));
                data.setSize(rs.getLong("fs_size"));

                data.setBytesRead(rs.getLong("fs_disk_read_bytes_avg"));
                data.setNumberOfDiskReads(rs.getLong("fs_disk_reads_avg"));
                data.setBytesWritten(rs.getLong("fs_disk_write_bytes_avg"));
                data.setNumberOfDiskWrites(rs.getLong("fs_disk_writes_avg"));
                data.setNumberOfFiles(rs.getLong("fs_files_avg"));

                fsData.add(data);
            }
        }
        finally {
            stat.close();
        }

        return fsData;
    }

    @Override
    public Collection<Event> getEvents(final int last_n_minutes) throws SQLException {

        //        final String query = "SELECT machine_id, label, time FROM RESOURCE_MONITORING.EVENTS" + ofLastNTimePeriod(last_n_minutes, "MINUTE");
        final List<Event> data = new LinkedList<Event>();
        //        Statement stat = null;
        //
        //        try {
        //            stat = connection.createStatement();
        //            final ResultSet rs = stat.executeQuery(query);
        //
        //            while (rs.next()) {
        //                final String machine_id = rs.getString("machine_id");
        //                final String label = rs.getString("label");
        //
        //                final Event data = new Event();
        //
        //                data.add(data);
        //            }
        //        }
        //        finally {
        //            stat.close();
        //        }

        //TODO AL IS HERE

        return data;
    }

    private String groupBy(final String... columns) {

        final StringBuilder s = new StringBuilder();
        s.append(" GROUP BY ");

        for (int i = 0; i < columns.length; i++) {
            s.append(columns[i]);

            if (i < columns.length - 1) {
                s.append(", ");
            }
        }
        return s.toString();
    }

    private String getBasicMonitoringInfo() {

        final StringBuilder builder = new StringBuilder();
        builder.append("machine_id,");
        builder.append("min(start_ts) as start_ts,");
        builder.append("sum(measurements) as measurements,");
        return builder.toString();
    }

    private String getFSInfo() {

        final StringBuilder builder = new StringBuilder();
        builder.append("fs_name, fs_type,fs_location,max(fs_size) as fs_size,min(fs_free) as fs_free, max(fs_used) as fs_used, avg(fs_files_avg) as fs_files_avg, min(fs_files_min) as fs_files_min, max(fs_files_max) as fs_files_max, avg(fs_disk_reads_avg) as fs_disk_reads_avg, min(fs_disk_reads_min) as fs_disk_reads_min, max(fs_disk_reads_max) as fs_disk_reads_max, ");
        builder.append("avg(fs_disk_read_bytes_avg) as fs_disk_read_bytes_avg, min(fs_disk_read_bytes_min) as fs_disk_read_bytes_min, max(fs_disk_read_bytes_max) as fs_disk_read_bytes_max, avg(fs_disk_writes_avg) as fs_disk_writes_avg, min(fs_disk_writes_min) as fs_disk_writes_min, max(fs_disk_writes_max) as fs_disk_writes_max, ");
        builder.append("avg(fs_disk_write_bytes_avg) as fs_disk_write_bytes_avg, min(fs_disk_write_bytes_min) as fs_disk_write_bytes_min, max(fs_disk_write_bytes_max) as fs_disk_write_bytes_max FROM resource_monitoring.file_system ");
        return builder.toString();
    }

    private String getDistributionData() {

        return "MACHINE_ID, min(START_TS) as start_ts, max(END_TS) as end_ts, sum(MEASUREMENTS) as measurements, " + "INSTRUMENT_NAME, ATTRIBUTES, avg(HIGH) as HIGH, avg(MEDIUM_HIGH) as MEDIUM_HIGH, avg(MEDIUM) as MEDIUM, " + "avg(MEDIUM_LOW) as medium_low, avg(LOW) as LOW "
                        + "FROM RESOURCE_MONITORING.DISTRIBUTION";
    }

    private String ofLastNTimePeriod(final int n, final String timePeriod) {

        return " WHERE start_ts >= DATEADD('" + timePeriod + "', -" + n + ", NOW())";
    }

    @Override
    public Collection<DistributionData> getDistributionData(final int time) throws SQLException {

        final Map<String, DistributionData> newData = new HashMap<String, DistributionData>();
        final String query = "SELECT " + getDistributionData() + ofLastNTimePeriod(time, "MINUTE") + groupBy("instrument_name", "attributes", "machine_id");
        Statement stat = null;

        try {
            stat = connection.createStatement();

            final ResultSet rs = stat.executeQuery(query);

            while (rs.next()) {

                final String machine_id = rs.getString("machine_id");

                final String instrument_name = rs.getString("instrument_name");
                final String attributes = rs.getString("attributes");
                final Date date = rs.getDate("start_ts");
                final float high = rs.getFloat("HIGH");
                final float medium_high = rs.getFloat("MEDIUM_HIGH");
                final float medium = rs.getFloat("MEDIUM");
                final float medium_low = rs.getFloat("medium_low");
                final float low = rs.getFloat("LOW");

                final Distribution dist = new Distribution(instrument_name, attributes, date, high, medium_high, medium, medium_low, low);

                DistributionData dd = newData.get(machine_id);

                if (dd == null) {
                    dd = new DistributionData(machine_id);
                }

                dd.add(dist);

                newData.put(machine_id, dd);
            }
        }
        finally {
            stat.close();
        }

        return newData.values();
    }
}
