package uk.ac.standrews.cs.numonic.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Lazily builds and executes insert queries. A query can be reused for the same relation name and number of values.
 * 
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk) - "he made me do it"
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Angus Macdonald (angus@cs.st-andrews.ac.uk) - "didn't even make the coffees"
 */
public class QueryBuilder {

    private PreparedStatement prepared_statement = null;
    private String relation_name = "";
    private int number_of_values = 0;

    /**
     * Performs an insert query. A prepared statement is created if it does not already exist.
     * If a prepared statement does already exist, its relation name and number of values must
     * match those of this call.
     * 
     * @param relation_name the name of the relation into which to insert
     * @param connection the database connection
     * @param values the value list
     * @throws SQLException if the relation name or number of values do not match a previous call, or if anything goes wrong with statement execution
     */
    public void doInsertQuery(final String relation_name, final Connection connection, final Object... values) throws SQLException {

        // Construct prepared statement template if one doesn't exist
        if (prepared_statement == null) {
            initialisePreparedStatement(relation_name, connection, values);
        }
        else {
            if (!relation_name.equals(this.relation_name) || values.length != number_of_values) { throw new SQLException("query reused with different relation name or number of values"); }
        }

        // Populate the prepared statement
        int count = 1;
        for (final Object value : values) {
            if (value instanceof Long) {
                prepared_statement.setLong(count++, (Long) value);
            }
            else if (value instanceof Double) {
                prepared_statement.setDouble(count++, (Double) value);
            }
            else if (value instanceof Integer) {
                prepared_statement.setInt(count++, (Integer) value);
            }
            else if (value instanceof String) {
                prepared_statement.setString(count++, (String) value);
            }
            else if (value instanceof Float) {
                prepared_statement.setFloat(count++, (Float) value);
            }
            else {
                ErrorHandling.hardError("Unknown type encountered: " + value.getClass().getName());
            }
        }

        prepared_statement.execute();
    }

    private void initialisePreparedStatement(final String relation_name, final Connection connection, final Object... values) throws SQLException {

        final StringBuilder query = new StringBuilder();
        query.append("INSERT INTO ");
        query.append(relation_name);
        query.append(" VALUES( ");

        for (int i = 0; i < values.length; i++) {
            if (i > 0) {
                query.append(",");
            }
            query.append("?");
        }
        query.append(")");

        prepared_statement = connection.prepareStatement(query.toString());
        this.relation_name = relation_name;
        number_of_values = values.length;
    }
}
