package uk.ac.standrews.cs.numonic.monitor.bandwidthAndLatency;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;

import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.NetworkUtil;

/**
 * Server that accepts sequences of data blocks from other hosts, allowing them to estimate their latencies and
 * bandwidths relative to this host. Requires {@link BandwidthAndLatencyClient} to be running on other host.
 * 
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham.kirby@st-andrews.ac.uk)
 */
public class BandwidthAndLatencyServer extends BandwidthAndLatencyCommon {

    private final Server server;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initialises the server.
     */
    public BandwidthAndLatencyServer() {

        super();
        server = new Server();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Starts the service.
     */
    public void startServices() {

        server.start();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Runs a server.
     * @param args ignored
     */
    public static void main(final String[] args) {

        final BandwidthAndLatencyServer server = new BandwidthAndLatencyServer();
        server.startServices();
    }

    // -------------------------------------------------------------------------------------------------------

    private class Server extends Thread {

        @Override
        public synchronized void run() {

            while (true) {
                try {
                    receiveRequest();
                }
                catch (final Exception e) {
                    Diagnostic.trace(DiagnosticLevel.FULL, "Exception running listener: " + e.getMessage());
                    return;
                }
            }
        }

        /**
         * Receives and processes a request. The accepted requests are {@link #DATA} - which involves reading a specified number of data blocks,
         * each of a specified size - and {@link #CONFIG} - which involves returning this server's configuration.

         * @throws IOException if a network error occurs
         */
        void receiveRequest() throws IOException {

            ServerSocket server_socket = null;
            Socket tcp_socket = null;
            InputStream input_stream = null;
            OutputStream output_stream = null;

            try {
                server_socket = NetworkUtil.makeReusableServerSocket(DEFAULT_TCP_RECEIVE_PORT);

                tcp_socket = server_socket.accept();
                input_stream = tcp_socket.getInputStream();
                output_stream = tcp_socket.getOutputStream();

                final DataInputStream in = new DataInputStream(input_stream);
                final DataOutputStream out = new DataOutputStream(output_stream);

                final String command = readLine(in);

                if (command.equals(DATA)) {
                    doTimingRequest(in, out);
                }
                else if (command.equals(CONFIG)) {
                    doConfigRequest(out);
                }
            }
            finally {
                // Tidy up.

                closeIfNotNull(input_stream);
                closeIfNotNull(output_stream);
                closeIfNotNull(tcp_socket);
                closeIfNotNull(server_socket);
            }
        }

        private void doTimingRequest(final DataInputStream in, final DataOutputStream out) throws IOException {

            final byte[] int_buffer = new byte[SIZE_OF_INT];

            // Read the number of blocks to be received.
            in.readFully(int_buffer);
            final int number_of_repetitions = byteArrayToInt(int_buffer);

            // Read the length of each block.
            in.readFully(int_buffer);
            final int payload_length = byteArrayToInt(int_buffer);

            final byte[] payload_buffer = new byte[payload_length];

            // Read the data.
            for (int i = 0; i < number_of_repetitions; i++) {
                in.readFully(payload_buffer);
            }

            // Acknowledge successful receipt.
            out.write(ACK.getBytes());
            out.write('\n');
            out.flush();
        }

        private void doConfigRequest(final DataOutputStream out) throws IOException {

            out.write(getNetworkConfiguration().getBytes());
            out.write('\n');
            out.flush();
        }
    }
}
