/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.sort.data;

import uk.ac.standrews.cs.nds.util.ErrorHandling;
import uk.ac.standrews.cs.numonic.distribution.Distribution;

/**
 * Distribution information on the likelihood various resources will be available on a particular machine.
 *
 * <p>
 * Used to rank machines.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class DistributionData {

    private final String machine_id;

    private Distribution cpu_utilization;
    private Distribution memory_utilization;
    private Distribution disk_utilization_write;
    private Distribution disk_utilization_read;
    private Distribution network_utilization_TX;
    private Distribution network_utilization_RX;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the distribution data.
     * @param machine_id the machine ID.
     */
    public DistributionData(final String machine_id) {

        this.machine_id = machine_id;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Adds a distribution.
     * @param distribution the distribution
     */
    public void add(final Distribution distribution) {

        final String instrument = distribution.getInstrumentName();

        if (instrument.equals("cpu")) {
            cpu_utilization = distribution;
        }
        else if (instrument.equals("memory")) {
            memory_utilization = distribution;
        }
        else if (instrument.equals("net_tx")) {
            network_utilization_TX = distribution;
        }
        else if (instrument.equals("net_rx")) {
            network_utilization_RX = distribution;
        }
        else if (instrument.equals("fs_write")) {
            disk_utilization_write = distribution;
        }
        else if (instrument.equals("fs_read")) {
            disk_utilization_read = distribution;
        }
        else {
            ErrorHandling.errorNoEvent("Distribution information not recognised for instrument: " + instrument);
        }
    }

    /**
     * Returns the CPU utilization distribution.
     * @return the CPU utilization distribution
     */
    public Distribution getCpuUtilization() {

        return cpu_utilization;
    }

    /**
     * Returns the memory utilization distribution.
     * @return the memory utilization distribution
     */
    public Distribution getMemoryUtilization() {

        return memory_utilization;
    }

    /**
     * Returns the disk read utilization distribution.
     * @return the disk read utilization distribution
     */
    public Distribution getDiskUtilizationRead() {

        return disk_utilization_read;
    }

    /**
     * Returns the disk write utilization distribution.
     * @return the disk write utilization distribution
     */
    public Distribution getDiskUtilizationWrite() {

        return disk_utilization_write;
    }

    /**
     * Returns the network transmit utilization distribution.
     * @return the network transmit utilization distribution
     */
    public Distribution getNetworkUtilizationTX() {

        return network_utilization_TX;
    }

    /**
     * Returns the network read utilization distribution.
     * @return the network read utilization distribution
     */
    public Distribution getNetworkUtilizationRX() {

        return network_utilization_RX;
    }

    /**
     * Returns the machine ID.
     * @return the machine ID
     */
    public String getMachineId() {

        return machine_id;
    }
}
