/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.sort.data;

import uk.ac.standrews.cs.numonic.sort.filter.Requirements;
import uk.ac.standrews.cs.numonic.sort.ranker.Bounds;
import uk.ac.standrews.cs.numonic.sort.ranker.MachineRanker;
import uk.ac.standrews.cs.numonic.sort.ranker.MachineSortingMetric;

/**
 * Stores information on the recent availability of this machine.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class MachineInfo {

    /**
     * The name of the machine which this instance represents. This must be a unique ID - it may be a machine ID or an application-specific
     * identifier for the machine.
     */
    private final String machine_name;

    /**
     * Data indicating the probability that specific resources will be utilized. For each measurement this probability is given for various
     * levels of utilization.
     */
    private DistributionData distribution_data;

    /**
     * Data giving average utilization of various resources on this machine. This is used to perform threshold analysis and to filter
     * machines with insufficient resources available for a request.
     */
    private final SystemData system_data;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Initializes the information.
     *
     * @param machine_name the name of the machine
     * @param system_data the system data
     */
    public MachineInfo(final String machine_name, final SystemData system_data) {

        this.machine_name = machine_name;
        this.system_data = system_data;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Tests whether this machine meets a given set of requirements relating to the availability of specific resources.
     *
     * @param requirements the requirements that this machine is checked against
     * @return true iff the requirements are met
     */
    public boolean meetsRequirements(final Requirements requirements) {

        return system_data.meetsRequirements(requirements);
    }

    /**
     * Gets the value of this machine according to the sorting metric provided. The sorting metric gives weights to the availability of various
     * resources and returns a single value for this machine.
     *
     * @param metric the metric
     * @param maximums the values for the maximum cpu capacity and memory capacity among the machines being compared, used to normalize values
     * @return the value of this machine according to the metric
     */
    public double getValue(final MachineSortingMetric metric, final Bounds maximums) {

        return MachineRanker.getMachineValue(metric, system_data, distribution_data, maximums);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public int hashCode() {

        final int prime = 31;
        int result = 1;
        result = prime * result + (machine_name == null ? 0 : machine_name.hashCode());
        return result;
    }

    @Override
    public boolean equals(final Object obj) {

        if (this == obj) { return true; }
        if (obj == null) { return false; }
        if (getClass() != obj.getClass()) { return false; }
        final MachineInfo other = (MachineInfo) obj;
        if (machine_name == null) {
            if (other.machine_name != null) { return false; }
        }
        else if (!machine_name.equals(other.machine_name)) { return false; }
        return true;
    }

    @Override
    public String toString() {

        return "MachineInfo [machineName=" + machine_name + "]";
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the system data.
     * @return the system data
     */
    public SystemData getSystemData() {

        return system_data;
    }

    /**
     * Sets the distribution data.
     * @param distribution_data the distribution data
     */
    public void setDistributionData(final DistributionData distribution_data) {

        this.distribution_data = distribution_data;
    }
}
