/***************************************************************************
 *                                                                         *
 * numonic Library                                                         *
 * Copyright (C) 2009-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic                                    *
 *                                                                         *
 * This file is part of remote_management, a package providing             *
 * functionality for remotely managing a specified application.            *
 *                                                                         *
 * numonic is free software: you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * numonic is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with numonic.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                         *
 ***************************************************************************/

package uk.ac.standrews.cs.numonic.summary;

/**
 * Abstract summary.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public abstract class Summary {

    /**
     * The MAC address of the primary network interface of the machine being monitored.
     */
    private final String machine_id;

    /**
     * The number of measurements to be taken before a summary is persisted to the database.
     */
    private final int measurements_before_summary;

    /**
     * The number of measurements taken so far.
     */
    private int measurements_taken = 0;

    /**
     * Initializes a summary.
     *
     * @param measurements_before_summary the number of measurements to be accumulated before summarization
     * @param machine_id the machine ID
     */
    public Summary(final int measurements_before_summary, final String machine_id) {

        this.measurements_before_summary = measurements_before_summary + 1; // +1 because the first measurement is used as a base measurement.
        this.machine_id = machine_id;
    }

    /**
     * Returns the machine ID.
     * @return the machine ID
     */
    public String getMachineID() {

        return machine_id;
    }

    /**
     * Tests whether the summary is full.
     * @return true if the number of measurements stored so far is greater than the number of measurements required (by the constructor parameter
     * 'measurementsBeforeSummary') before the summaries can be stored
     */
    public boolean isFull() {

        return measurements_taken >= measurements_before_summary;
    }

    /**
     * Increments the number of measurements taken.
     */
    public void incrementMeasurementCount() {

        measurements_taken++;
    }

    /**
     * Returns the number of measurements taken.
     * @return the number of measurements taken
     */
    public int measurementsTaken() {

        return measurements_taken;
    }

    /**
     * Returns the number of measurements needed.
     * @return the number of measurements needed
     */
    public int getNumberOfMeasurementsNeeded() {

        return measurements_before_summary;
    }

    /**
     * Resets the measurements.
     */
    public void reset() {

        measurements_taken = 0;
    }
}
