/***************************************************************************
 *                                                                         *
 * H2O                                                                     *
 * Copyright (C) 2010-2011 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://blogs.cs.st-andrews.ac.uk/numonic/                                   *
 *                                                                         *
 * This file is part of H2O, a distributed database based on the open      *
 * source database H2 (www.h2database.com).                                *
 *                                                                         *
 * H2O is free software: you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License as          *
 * published by the Free Software Foundation, either version 3 of the      *
 * License, or (at your option) any later version.                         *
 *                                                                         *
 * H2O is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with H2O.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.numonic.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * Wrapper for java properties class, which make it easier to create, load, and save properties.
 *
 * @author Angus Macdonald (angus AT cs.st-andrews.ac.uk)
 */
public class PropertiesWrapper implements IPropertiesWrapper {

    private final Properties properties;
    private final String propertiesFileLocation;

    private String comment = "Properties File";

    private static final String TRUE = "true";

    private static Map<String, PropertiesWrapper> map = new HashMap<String, PropertiesWrapper>();

    private PropertiesWrapper(final String fileLocation) {

        propertiesFileLocation = fileLocation;
        properties = new Properties();
    }

    /**
     * Gets a properties wrapper for a given file.
     *
     * @param fileLocation the location of the file containing properties
     * @return the wrapper
     */
    public static synchronized IPropertiesWrapper getWrapper(final String fileLocation) {

        PropertiesWrapper wrapper = map.get(fileLocation);
        if (wrapper == null) {
            wrapper = new PropertiesWrapper(fileLocation);
            map.put(fileLocation, wrapper);
        }
        return wrapper;
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#loadProperties()
     */
    @Override
    public synchronized void loadProperties() throws IOException {

        if (propertiesFileLocation.startsWith("http:")) { // Parse URL, request file from webpage.

            final URL url = new URL(propertiesFileLocation);
            final InputStreamReader inputReader = new InputStreamReader(url.openStream());

            try {
                properties.load(inputReader);
            }
            finally {
                inputReader.close();
            }
        }
        else { // Try to open the file from disk.

            final File f = new File(propertiesFileLocation);

            if (!f.exists()) { throw new IOException("Properties file doesn't exist at this location (" + propertiesFileLocation + ").");
            // This check is necessary because a file will be created when FileInputStream is created.
            }

            FileInputStream inputStream = null;
            try {
                inputStream = new FileInputStream(propertiesFileLocation);
                properties.load(inputStream);
            }
            catch (final FileNotFoundException e) {
                ErrorHandling.exceptionError(e, "Unexpected file not found exception");
            }
            finally {
                if (inputStream != null) {
                    inputStream.close();
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#createNewFile()
     */
    @Override
    public synchronized void createNewFile() throws IOException {

        removePropertiesFile();
        final File f = new File(propertiesFileLocation);

        if (f.getParentFile() != null) {
            final boolean successful = f.getParentFile().mkdirs(); // create any directories specified in the path, if necessary.
            if (!successful) {
                Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Failed to create folder. It may already exist.");
            }
        }

        final boolean successful = f.createNewFile(); // create the properties file.
        if (!successful) {
            Diagnostic.traceNoEvent(DiagnosticLevel.FULL, "Failed to create file. It may already exist.");
        }
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#getProperty(java.lang.String)
     */
    @Override
    public synchronized String getProperty(final String key) {

        return properties.getProperty(key);
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#isEnabled(java.lang.String)
     */
    @Override
    public synchronized boolean isEnabled(final String key) {

        return TRUE.equals(getProperty(key));
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#setProperty(java.lang.String, java.lang.String)
     */
    @Override
    public synchronized void setProperty(final String key, final String value) {

        properties.setProperty(key, value);
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#setPropertyComment(java.lang.String)
     */
    @Override
    public synchronized void setPropertyComment(final String propertyComment) {

        comment = propertyComment;
    }

    private synchronized boolean removePropertiesFile() {

        final File f = new File(propertiesFileLocation);

        return f.delete();
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#saveAndClose()
     */
    @Override
    public synchronized void saveAndClose() throws IOException {

        FileOutputStream outputStream = null;
        try {
            outputStream = new FileOutputStream(propertiesFileLocation);
            properties.store(outputStream, comment);
        }
        catch (final FileNotFoundException e) {
            ErrorHandling.exceptionError(e, "Unexpected file not found exception");
        }
        finally {
            if (outputStream != null) {
                outputStream.close();
            }
        }
    }

    /* (non-Javadoc)
     * @see uk.ac.standrews.cs.numonic.util.IPropertiesWrapper#getKeys()
     */
    @Override
    public synchronized Set<Object> getKeys() {

        return properties.keySet();
    }
}
